Package: maaslin3
Title: "Refining and extending generalized multivariate linear models
        for meta-omic association discovery"
Year: 2025
Version: 1.4.0
Authors@R: c(
  person("William", "Nickols", email = "willnickols@g.harvard.edu", role = c("aut", "cre"), comment=c(ORCID="0000-0001-8214-9746")), 
  person("Jacob", "Nearing", email = "nearing@broadinstitute.org", role = "aut")
  )
Depends: R (>= 4.4)
Description: MaAsLin 3 refines and extends generalized multivariate
        linear models for meta-omicron association discovery. It finds
        abundance and prevalence associations between microbiome
        meta-omics features and complex metadata in population-scale
        epidemiological studies. The software includes multiple
        analysis methods (including support for multiple covariates,
        repeated measures, and ordered predictors), filtering,
        normalization, and transform options to customize analysis for
        your specific study.
License: MIT + file LICENSE
Imports: data.table, collapse, lmerTest, lme4, optparse, logging,
        mirai, multcomp, ggplot2, RColorBrewer, patchwork, scales,
        rlang, ggnewscale, survival, methods, BiocGenerics, reformulas
Suggests: knitr, testthat (>= 2.1.0), rmarkdown, markdown, kableExtra,
        dplyr, SummarizedExperiment, TreeSummarizedExperiment
VignetteBuilder: knitr
Collate: fit.R utility_scripts.R viz.R maaslin3.R
URL: http://huttenhower.sph.harvard.edu/maaslin3
biocViews: Metagenomics, Software, Microbiome, Normalization,
        MultipleComparison
BugReports: https://github.com/biobakery/maaslin3/issues
NeedsCompilation: no
Packaged: 2026-04-29 23:18:07 UTC; root
RoxygenNote: 7.3.3
Config/pak/sysreqs: cmake make
Repository: https://bioc-release.r-universe.dev
Date/Publication: 2026-04-28 13:04:13 UTC
RemoteUrl: https://github.com/bioc/maaslin3
RemoteRef: RELEASE_3_23
RemoteSha: f7a411ede14c49c6173799d252d6896bbae95450
Author: William Nickols [aut, cre] (ORCID:
    <https://orcid.org/0000-0001-8214-9746>),
  Jacob Nearing [aut]
Maintainer: William Nickols <willnickols@g.harvard.edu>
Built: R 4.6.0; ; 2026-04-30 00:35:31 UTC; unix

export(maaslin3)
export(maaslin_check_arguments)
export(maaslin_log_arguments)
export(maaslin_read_data)
export(maaslin_reorder_data)
export(maaslin_compute_formula)
export(maaslin_check_formula)
export(maaslin_filter)
export(maaslin_process_metadata)
export(maaslin_normalize)
export(maaslin_transform)
export(maaslin_fit)
export(maaslin_write_results)
export(maaslin_write_results_lefse_format)
export(maaslin_plot_results)
export(maaslin_plot_results_from_output)
export(maaslin_contrast_test)
export(preprocess_dna_mtx)
export(preprocess_taxa_mtx)

import("optparse")
importFrom("stats", "as.formula", "coef", "glm", "lm", "na.exclude",
    "p.adjust", "update", "relevel", "anova", "approx", "filter", "median",
    "model.matrix", "pbeta", "quantile", "terms",
    "update.formula", "var", "vcov", "rnorm", "cov", "setNames")
importFrom("utils", "capture.output", "read.table", "write.table", "tail")
importFrom("stats", "sd", "na.omit")
importFrom("utils","type.convert")
importFrom("rlang", ".data", "check_installed")
importFrom("methods", "is")
importFrom("survival", "strata", "coxph", "Surv")

## ----eval=FALSE, cache = FALSE------------------------------------------------
# if (!require("BiocManager", quietly = TRUE))
#     install.packages("BiocManager")
# 
# BiocManager::install("biobakery/maaslin3")

## ----eval=TRUE, cache = FALSE, echo=FALSE-------------------------------------
for (lib in c('maaslin3', 'dplyr', 'ggplot2', 'knitr')) {
    suppressPackageStartupMessages(require(lib, character.only = TRUE))
}

## ----echo = TRUE, results = 'hide', warning = FALSE, cache = FALSE------------
# Read abundance table
taxa_table_name <- system.file("extdata", "HMP2_taxonomy.tsv",
                                package = "maaslin3")
taxa_table <- read.csv(taxa_table_name, sep = '\t', row.names = 1)

# Read metadata table
metadata_name <- system.file("extdata", "HMP2_metadata.tsv",
                                package = "maaslin3")
metadata <- read.csv(metadata_name, sep = '\t', row.names = 1)

metadata$diagnosis <-
    factor(metadata$diagnosis, levels = c('nonIBD', 'UC', 'CD'))
metadata$dysbiosis_state <-
    factor(metadata$dysbiosis_state, levels = c('none', 'dysbiosis_UC',
                                                'dysbiosis_CD'))
metadata$antibiotics <-
    factor(metadata$antibiotics, levels = c('No', 'Yes'))

# Fit models
set.seed(1)
fit_out <- maaslin3(input_data = taxa_table,
                    input_metadata = metadata,
                    output = 'hmp2_output',
                    formula = '~ diagnosis + dysbiosis_state +
                    antibiotics + age + reads',
                    normalization = 'TSS',
                    transform = 'LOG',
                    augment = TRUE,
                    standardize = TRUE,
                    max_significance = 0.1,
                    median_comparison_abundance = TRUE,
                    median_comparison_prevalence = FALSE,
                    max_pngs = 20)

## ----eval=FALSE---------------------------------------------------------------
# maaslin_log_arguments(input_data = taxa_table,
#                     input_metadata = metadata,
#                     output = 'hmp2_output',
#                     formula = '~ diagnosis + dysbiosis_state +
#                     antibiotics + age + reads',
#                     normalization = 'TSS',
#                     transform = 'LOG',
#                     augment = TRUE,
#                     standardize = TRUE,
#                     max_significance = 0.1,
#                     median_comparison_abundance = TRUE,
#                     median_comparison_prevalence = FALSE,
#                     max_pngs = 20)
# 
# read_data_list <- maaslin_read_data(taxa_table,
#                                     metadata)
# read_data_list <- maaslin_reorder_data(read_data_list$data,
#                                     read_data_list$metadata)
# 
# data <- read_data_list$data
# metadata <- read_data_list$metadata
# 
# formulas <- maaslin_check_formula(data,
#                                 metadata,
#                                 '~ diagnosis + dysbiosis_state +
#                                     antibiotics + age + reads')
# 
# formula <- formulas$formula
# random_effects_formula <- formulas$random_effects_formula
# 
# normalized_data = maaslin_normalize(data,
#                                     output = 'hmp2_output',
#                                     normalization = 'TSS')
# 
# filtered_data = maaslin_filter(normalized_data,
#                             output = 'hmp2_output')
# 
# transformed_data = maaslin_transform(filtered_data,
#                                     output = 'hmp2_output',
#                                     transform = 'LOG')
# 
# standardized_metadata = maaslin_process_metadata(metadata,
#                                                 formula,
#                                                 standardize = TRUE)
# 
# maaslin_results = maaslin_fit(filtered_data,
#                             transformed_data,
#                             standardized_metadata,
#                             formula,
#                             random_effects_formula,
#                             data = data,
#                             median_comparison_abundance = TRUE,
#                             median_comparison_prevalence = FALSE)
# 
# maaslin_write_results(output = 'hmp2_output',
#                     maaslin_results$fit_data_abundance,
#                     maaslin_results$fit_data_prevalence,
#                     random_effects_formula,
#                     max_significance = 0.1)
# 
# # Invisible to avoid dumping plots
# invisible(maaslin_plot_results(output = 'hmp2_output',
#                             transformed_data,
#                             metadata,
#                             maaslin_results$fit_data_abundance,
#                             maaslin_results$fit_data_prevalence,
#                             normalization = "TSS",
#                             transform = "LOG",
#                             median_comparison_abundance = TRUE,
#                             median_comparison_prevalence = FALSE,
#                             max_significance = 0.1,
#                             max_pngs = 20))

## ----echo = TRUE, results = 'hide', warning = FALSE, cache = FALSE------------
# This section is necessary for updating
# the summary plot and the association plots

# Rename results file with clean titles
all_results <- read.csv('hmp2_output/all_results.tsv', sep='\t')
all_results <- all_results %>%
    mutate(metadata = case_when(metadata == 'age' ~ 'Age',
                                metadata == 'antibiotics' ~ 'Abx',
                                metadata == 'diagnosis' ~ 'Diagnosis',
                                metadata == 'dysbiosis_state' ~ 'Dysbiosis',
                                metadata == 'reads' ~ 'Read depth'),
        value = case_when(value == 'dysbiosis_CD' ~ 'CD',
                            value == 'dysbiosis_UC' ~ 'UC',
                            value == 'Yes' ~ 'Used', # Antibiotics
                            value == 'age' ~ 'Age',
                            value == 'reads' ~ 'Read depth',
                            TRUE ~ value),
        feature = gsub('_', ' ', feature) %>%
            gsub(pattern = 'sp ', replacement = 'sp. '))

# Write results
write.table(all_results, 'hmp2_output/all_results.tsv', sep='\t')

# Set the new heatmap and coefficient plot variables and order them
heatmap_vars = c('Dysbiosis UC', 'Diagnosis UC',
                'Abx Used', 'Age', 'Read depth')
coef_plot_vars = c('Dysbiosis CD', 'Diagnosis CD')

# This section is necessary for updating the association plots
colnames(taxa_table) <- gsub('_', ' ', colnames(taxa_table)) %>%
    gsub(pattern = 'sp ', replacement = 'sp. ')

# Rename the features in the norm transformed data file
data_transformed <-
    read.csv('hmp2_output/features/data_transformed.tsv', sep='\t')
colnames(data_transformed) <-
    gsub('_', ' ', colnames(data_transformed)) %>%
    gsub(pattern = 'sp ', replacement = 'sp. ')
write.table(data_transformed,
            'hmp2_output/features/data_transformed.tsv',
            sep='\t', row.names = FALSE)

# Rename the metadata like in the outputs table
colnames(metadata) <-
    case_when(colnames(metadata) == 'age' ~ 'Age',
            colnames(metadata) == 'antibiotics' ~ 'Abx',
            colnames(metadata) == 'diagnosis' ~ 'Diagnosis',
            colnames(metadata) == 'dysbiosis_state' ~ 'Dysbiosis',
            colnames(metadata) == 'reads' ~ 'Read depth',
            TRUE ~ colnames(metadata))
metadata <- metadata %>%
    mutate(Dysbiosis = case_when(Dysbiosis == 'dysbiosis_UC' ~ 'UC',
                                Dysbiosis == 'dysbiosis_CD' ~ 'CD',
                                Dysbiosis == 'none' ~ 'None') %>%
            factor(levels = c('None', 'UC', 'CD')),
        Abx = case_when(Abx == 'Yes' ~ 'Used',
                        Abx == 'No' ~ 'Not used') %>%
            factor(levels = c('Not used', 'Used')),
        Diagnosis = case_when(Diagnosis == 'nonIBD' ~ 'non-IBD',
                                TRUE ~ Diagnosis) %>%
            factor(levels = c('non-IBD', 'UC', 'CD')))

# Recreate the plots
scatter_plots <- maaslin_plot_results_from_output(
    output = 'hmp2_output',
    metadata,
    normalization = "TSS",
    transform = "LOG",
    median_comparison_abundance = TRUE,
    median_comparison_prevalence = FALSE,
    max_significance = 0.1,
    coef_plot_vars=coef_plot_vars,
    heatmap_vars=heatmap_vars,
    max_pngs = 20)

## ----out.width='100%', echo=FALSE, cache = FALSE, include=FALSE---------------
knitr::include_graphics("hmp2_output/figures/summary_plot.png")

## ----echo=FALSE, fig.show='hold', cache = FALSE, include=FALSE, eval=FALSE----
# prefix <- "hmp2_output/figures/association_plots"
# plot_vec <- c("/Age/linear/Age_Enterocloster clostridioformis_linear.png",
#             "/Dysbiosis/linear/Dysbiosis_Escherichia coli_linear.png",
#             "/Age/logistic/Age_Bifidobacterium longum_logistic.png",
#             paste0("/Dysbiosis/logistic/",
#                 "Dysbiosis_Faecalibacterium prausnitzii_logistic.png"
#                 ))
# knitr::include_graphics(paste0(prefix, plot_vec))

## ----echo = TRUE, results = 'hide', warning = FALSE, cache = FALSE------------
contrast_mat <- matrix(c(1, 1, 0, 0, 0, 0, 1, 1), 
    ncol = 4, nrow = 2, byrow = TRUE)
    
colnames(contrast_mat) <- c("diagnosisUC",
                            "diagnosisCD",
                            "dysbiosis_statedysbiosis_UC",
                            "dysbiosis_statedysbiosis_CD")
                            
rownames(contrast_mat) <- c("diagnosis_test", "dysbiosis_test")

contrast_mat
                            
maaslin_contrast_test(maaslin3_fit = fit_out,
                        contrast_mat = contrast_mat)

## -----------------------------------------------------------------------------
sessionInfo()

## -----------------------------------------------------------------------------
# Clean-up generated outputs
unlink('hmp2_output', recursive = TRUE)

## ----eval=FALSE, cache = FALSE------------------------------------------------
# sessionInfo()

## ----eval=FALSE, cache = FALSE------------------------------------------------
# if (!require("BiocManager", quietly = TRUE))
#     install.packages("BiocManager")
# 
# BiocManager::install("biobakery/maaslin3")

## ----eval=TRUE, cache = FALSE, echo=FALSE-------------------------------------
for (lib in c('maaslin3', 'dplyr', 'ggplot2', 'knitr', 'kableExtra')) {
    suppressPackageStartupMessages(require(lib, character.only = TRUE))
}

## ----cache = FALSE------------------------------------------------------------
# Read abundance table
taxa_table_name <- system.file("extdata", "HMP2_taxonomy.tsv",
                                package = "maaslin3")
taxa_table <- read.csv(taxa_table_name, sep = '\t', row.names = 1)

# Read metadata table
metadata_name <- system.file("extdata", "HMP2_metadata.tsv",
                            package = "maaslin3")
metadata <- read.csv(metadata_name, sep = '\t', row.names = 1)

# Factor the categorical variables to test IBD against healthy controls
metadata$diagnosis <-
    factor(metadata$diagnosis, levels = c('nonIBD', 'UC', 'CD'))
metadata$dysbiosis_state <-
    factor(metadata$dysbiosis_state, levels =
                c('none', 'dysbiosis_UC', 'dysbiosis_CD'))
metadata$antibiotics <-
    factor(metadata$antibiotics, levels = c('No', 'Yes'))

taxa_table[1:5, 1:5]
metadata[1:5, 1:5]

## ----echo = TRUE, results = 'hide', warning = FALSE, cache = FALSE------------
set.seed(1)
fit_out <- maaslin3(input_data = taxa_table,
                    input_metadata = metadata,
                    output = 'hmp2_output',
                    formula = '~ diagnosis + dysbiosis_state +
                        antibiotics + age + reads',
                    normalization = 'TSS',
                    transform = 'LOG',
                    augment = TRUE,
                    standardize = TRUE,
                    max_significance = 0.1,
                    median_comparison_abundance = TRUE,
                    median_comparison_prevalence = FALSE,
                    max_pngs = 10)

## ----echo = TRUE, results = 'hide', warning = FALSE, eval = FALSE-------------
# se <- SummarizedExperiment(
#     assays = list(taxa_table = t(taxa_table)),
#     colData = metadata
# )
# 
# fit_out <- maaslin3(input_data = se,
#                     output = 'hmp2_output',
#                     formula = '~ diagnosis + dysbiosis_state +
#                         antibiotics + age + reads',
#                     normalization = 'TSS',
#                     transform = 'LOG',
#                     augment = TRUE,
#                     standardize = TRUE,
#                     max_significance = 0.1,
#                     median_comparison_abundance = TRUE,
#                     median_comparison_prevalence = FALSE,
#                     max_pngs = 10)

## ----echo = FALSE, cache = FALSE----------------------------------------------
signif_results <- read.csv('hmp2_output/significant_results.tsv',
sep='\t')
head(signif_results, 20) %>%
    dplyr::mutate_if(is.numeric, .funs = function(x){(
        as.character(signif(x, 3)))}) %>%
    knitr::kable() %>%
    kableExtra::kable_styling("striped", position = 'center') %>%
    kableExtra::scroll_box(width = "800px", height = "400px")

## ----out.width='100%', echo=FALSE, cache = FALSE, include=FALSE, eval=FALSE----
# # Rename summary plot to avoid knitting issues later
# quiet_out <- file.rename('hmp2_output/figures/summary_plot.png',
#                         'hmp2_output/figures/summary_plot_first.png')
# 
# knitr::include_graphics("hmp2_output/figures/summary_plot_first.png")

## ----echo=FALSE, fig.show='hold',include=FALSE, eval=FALSE--------------------
# prefix <- "hmp2_output/figures/association_plots"
# plot_vec <-
#     c("/age/linear/age_Enterocloster_clostridioformis_linear.png",
#     "/dysbiosis_state/linear/dysbiosis_state_Escherichia_coli_linear.png",
#     "/age/logistic/age_Bifidobacterium_longum_logistic.png",
#     paste0("/dysbiosis_state/logistic/",
#         "dysbiosis_state_Faecalibacterium_prausnitzii_logistic.png"
#         ))
# knitr::include_graphics(paste0(prefix, plot_vec))

## ----echo = TRUE, results = 'hide', warning = FALSE, cache = FALSE------------
# This section is necessary for updating the
# summary plot and the association plots

# Rename results file with clean titles
all_results <- read.csv('hmp2_output/all_results.tsv', sep='\t')
all_results <- all_results %>%
    mutate(metadata = case_when(metadata == 'age' ~ 'Age',
                                metadata == 'antibiotics' ~ 'Abx',
                                metadata == 'diagnosis' ~ 'Diagnosis',
                                metadata == 'dysbiosis_state' ~ 'Dysbiosis',
                                metadata == 'reads' ~ 'Read depth'),
        value = case_when(value == 'dysbiosis_CD' ~ 'CD',
                            value == 'dysbiosis_UC' ~ 'UC',
                            value == 'Yes' ~ 'Used', # Antibiotics
                            value == 'age' ~ 'Age',
                            value == 'reads' ~ 'Read depth',
                            TRUE ~ value),
        feature = gsub('_', ' ', feature) %>%
            gsub(pattern = 'sp ', replacement = 'sp. '))

# Write results
write.table(all_results, 'hmp2_output/all_results.tsv', sep='\t')

# Set the new heatmap and coefficient plot variables and order them
heatmap_vars = c('Dysbiosis UC', 'Diagnosis UC',
                'Abx Used', 'Age', 'Read depth')
coef_plot_vars = c('Dysbiosis CD', 'Diagnosis CD')

# This section is necessary for updating the association plots
taxa_table_copy <- taxa_table
colnames(taxa_table_copy) <- gsub('_', ' ', colnames(taxa_table_copy)) %>%
    gsub(pattern = 'sp ', replacement = 'sp. ')

# Rename the features in the norm transformed data file
data_transformed <-
    read.csv('hmp2_output/features/data_transformed.tsv', sep='\t')
colnames(data_transformed) <-
    gsub('_', ' ', colnames(data_transformed)) %>%
    gsub(pattern = 'sp ', replacement = 'sp. ')
write.table(data_transformed,
            'hmp2_output/features/data_transformed.tsv',
            sep='\t', row.names = FALSE)

# Rename the metadata like in the outputs table
metadata_copy <- metadata
colnames(metadata_copy) <-
    case_when(colnames(metadata_copy) == 'age' ~ 'Age',
            colnames(metadata_copy) == 'antibiotics' ~ 'Abx',
            colnames(metadata_copy) == 'diagnosis' ~ 'Diagnosis',
            colnames(metadata_copy) == 'dysbiosis_state' ~ 'Dysbiosis',
            colnames(metadata_copy) == 'reads' ~ 'Read depth',
            TRUE ~ colnames(metadata_copy))
metadata_copy <- metadata_copy %>%
    mutate(Dysbiosis = case_when(Dysbiosis == 'dysbiosis_UC' ~ 'UC',
                                Dysbiosis == 'dysbiosis_CD' ~ 'CD',
                                Dysbiosis == 'none' ~ 'None') %>%
            factor(levels = c('None', 'UC', 'CD')),
        Abx = case_when(Abx == 'Yes' ~ 'Used',
                        Abx == 'No' ~ 'Not used') %>%
            factor(levels = c('Not used', 'Used')),
        Diagnosis = case_when(Diagnosis == 'nonIBD' ~ 'non-IBD',
                                TRUE ~ Diagnosis) %>%
            factor(levels = c('non-IBD', 'UC', 'CD')))

# Recreate the plots
scatter_plots <- maaslin_plot_results_from_output(
    output = 'hmp2_output',
    metadata = metadata_copy,
    normalization = "TSS",
    transform = "LOG",
    median_comparison_abundance = TRUE,
    median_comparison_prevalence = FALSE,
    max_significance = 0.1,
    max_pngs = 20)

## ----out.width='100%', echo=FALSE, cache = FALSE, include=FALSE---------------
knitr::include_graphics("hmp2_output/figures/summary_plot.png")

## ----echo=FALSE, fig.show='hold', cache = FALSE, include=FALSE, eval=FALSE----
# prefix <- "hmp2_output/figures/association_plots"
# plot_vec <- c("/Age/linear/Age_Enterocloster clostridioformis_linear.png",
#             "/Dysbiosis/linear/Dysbiosis_Escherichia coli_linear.png",
#             "/Age/logistic/Age_Bifidobacterium longum_logistic.png",
#             paste0("/Dysbiosis/logistic/",
#                 "Dysbiosis_Faecalibacterium prausnitzii_logistic.png"
#                 ))
# knitr::include_graphics(paste0(prefix, plot_vec))

## ----cache = FALSE------------------------------------------------------------
# Abundance table
taxa_table_name <- system.file("extdata", "abundance_spike_in_ex.tsv",
                            package = "maaslin3")
spike_in_taxa_table <- read.csv(taxa_table_name, sep = '\t', row.names = 1)

# Metadata table
metadata_name <- system.file("extdata", "metadata_spike_in_ex.tsv",
                            package = "maaslin3")
spike_in_metadata <- read.csv(metadata_name, sep = '\t', row.names = 1)
for (col in c('Metadata_1', 'Metadata_2', 'Metadata_5')) {
    spike_in_metadata[,col] <- factor(spike_in_metadata[,col])
}

# Spike-in table
unscaled_name <- system.file("extdata",
"scaling_factors_spike_in_ex.tsv",
                            package = "maaslin3")
spike_in_unscaled <- read.csv(unscaled_name, sep = '\t', row.names = 1)

spike_in_taxa_table[c(1:5, 101),1:5]
spike_in_metadata[1:5,]
spike_in_unscaled[1:5, , drop=FALSE]

## ----echo = TRUE, results = 'hide', warning = FALSE, cache = FALSE------------
fit_out <- maaslin3(
    input_data = spike_in_taxa_table,
    input_metadata = spike_in_metadata,
    output = 'spike_in_demo',
    formula = '~ Metadata_1 + Metadata_2 + Metadata_3 +
        Metadata_4 + Metadata_5',
    normalization = 'TSS',
    transform = 'LOG',
    median_comparison_abundance = FALSE,
    unscaled_abundance = spike_in_unscaled)

## ----cache = FALSE------------------------------------------------------------
rownames(fit_out$fit_data_abundance$results) <- NULL
head(fit_out$fit_data_abundance$results, 20) %>%
    dplyr::mutate_if(is.numeric, .funs =
                            function(x){(as.character(signif(x, 3)))}) %>%
    knitr::kable() %>%
    kableExtra::kable_styling("striped", position = 'center') %>%
    kableExtra::scroll_box(width = "800px", height = "400px")

## ----cache = FALSE------------------------------------------------------------
# Abundance table
taxa_table_name <- system.file("extdata", "abundance_total_ex.tsv",
    package = "maaslin3")
total_scaling_taxa_table <- read.csv(taxa_table_name, sep = '\t', row.names = 1)

# Metadata table
metadata_name <- system.file("extdata", "metadata_total_ex.tsv",
                                package = "maaslin3")
total_scaling_metadata <- read.csv(metadata_name, sep = '\t', row.names = 1)
for (col in c('Metadata_1', 'Metadata_3', 'Metadata_5')) {
    spike_in_metadata[,col] <- factor(spike_in_metadata[,col])
}

# Total abundance table
unscaled_name <- system.file("extdata", "scaling_factors_total_ex.tsv",
                                package = "maaslin3")
total_scaling_unscaled <- read.csv(unscaled_name, sep = '\t', row.names = 1)

total_scaling_taxa_table[1:5, 1:5]
total_scaling_metadata[1:5,]
total_scaling_unscaled[1:5, , drop=FALSE]

## ----echo = TRUE, results = 'hide', warning = FALSE, cache = FALSE------------
fit_out <- maaslin3(
    input_data = total_scaling_taxa_table,
    input_metadata = total_scaling_metadata,
    output = 'total_scaling_demo',
    formula = '~ Metadata_1 + Metadata_2 + Metadata_3 +
        Metadata_4 + Metadata_5',
    normalization = 'TSS',
    transform = 'LOG',
    median_comparison_abundance = FALSE,
    unscaled_abundance = total_scaling_unscaled)

## ----cache = FALSE------------------------------------------------------------
rownames(fit_out$fit_data_abundance$results) <- NULL
head(fit_out$fit_data_abundance$results, n = 20) %>%
    dplyr::mutate_if(is.numeric,
                    .funs = function(x){(as.character(signif(x, 3)))}) %>%
    knitr::kable() %>%
    kableExtra::kable_styling("striped", position = 'center') %>%
    kableExtra::scroll_box(width = "800px", height = "400px")

## ----echo = TRUE, results = 'hide', warning = FALSE, messages = FALSE---------
# Subset to only CD cases for time; taxa are subset automatically
fit_out <- maaslin3(
    input_data = taxa_table,
    input_metadata = metadata[metadata$diagnosis == 'CD',],
    output = 'random_effects_output',
    formula = '~ dysbiosis_state + antibiotics +
        age + reads + (1|participant_id)',
    plot_summary_plot = FALSE,
    plot_associations = FALSE)

## ----echo = FALSE, cache = FALSE----------------------------------------------
signif_results <-
read.csv('random_effects_output/significant_results.tsv',
                            sep='\t')
head(signif_results, n = 20) %>%
    dplyr::mutate_if(is.numeric, .funs =
                        function(x){(as.character(signif(x, 3)))}) %>%
    knitr::kable() %>%
    kableExtra::kable_styling("striped", position = 'center') %>%
    kableExtra::scroll_box(width = "800px", height = "400px")

## ----echo = TRUE, results = 'hide', warning = FALSE, cache = FALSE------------
metadata <- metadata %>%
    mutate(dysbiosis_general = ifelse(dysbiosis_state != 'none',
                                    'dysbiosis', 'none')) %>%
    mutate(dysbiosis_general = factor(dysbiosis_general, levels =
                                        c('none', 'dysbiosis')))

fit_out <- maaslin3(
    input_data = taxa_table,
    input_metadata = metadata,
    output = 'interaction_output',
    formula = '~ diagnosis + diagnosis:dysbiosis_general +
        antibiotics + age + reads')

## ----cache = FALSE------------------------------------------------------------
full_results <- rbind(fit_out$fit_data_abundance$results,
                    fit_out$fit_data_prevalence$results)
full_results <- full_results %>%
    dplyr::arrange(qval_joint) %>%
    dplyr::filter(metadata == "diagnosis")
rownames(full_results) <- NULL
head(full_results, n = 20) %>%
    dplyr::mutate_if(is.numeric,
                    .funs = function(x){(as.character(signif(x, 3)))}) %>%
    knitr::kable() %>%
    kableExtra::kable_styling("striped", position = 'center') %>%
    kableExtra::scroll_box(width = "800px", height = "400px")

## ----echo = TRUE, results = 'hide', warning = FALSE, cache = FALSE------------
# Put the red meat consumption responses in order
metadata <- metadata %>%
    mutate(red_meat = ifelse(
        red_meat == 'No, I did not consume these products in the last 7 days',
                            'Not in the last 7 days',
                            red_meat) %>%
            factor(levels = c('Not in the last 7 days',
                                'Within the past 4 to 7 days',
                                'Within the past 2 to 3 days',
                                'Yesterday, 1 to 2 times',
                                'Yesterday, 3 or more times'))
    )

# Create the model with only non-IBD subjects
fit_out <- maaslin3(
    input_data = taxa_table,
    input_metadata = metadata[metadata$diagnosis == 'nonIBD',],
    output = 'ordered_outputs',
    formula = '~ ordered(red_meat) + antibiotics + age + reads',
    plot_summary_plot = TRUE,
    plot_associations = TRUE,
    heatmap_vars = c('red_meat Within the past 4 to 7 days',
                    'red_meat Within the past 2 to 3 days',
                    'red_meat Yesterday, 1 to 2 times',
                    'red_meat Yesterday, 3 or more times'),
    max_pngs = 30)

## ----out.width='100%', echo=FALSE, cache = FALSE, include=FALSE---------------
knitr::include_graphics("ordered_outputs/figures/summary_plot.png")

## ----echo = FALSE, cache = FALSE----------------------------------------------
full_results <- rbind(fit_out$fit_data_abundance$results,
                    fit_out$fit_data_prevalence$results)
full_results <- full_results %>%
    dplyr::filter(metadata %in% c("red_meat") &
                    feature == 'Alistipes_shahii' &
                    model == 'logistic') %>%
    dplyr::mutate(
        value =
            factor(
                value,
                levels =
                    c('No, I did not consume these products in the last 7 days',
                                                'Within the past 4 to 7 days',
                                                'Within the past 2 to 3 days',
                                                'Yesterday, 1 to 2 times',
                                            'Yesterday, 3 or more times'))) %>%
    dplyr::arrange(value)
rownames(full_results) <- NULL
head(full_results, n = 20) %>%
    dplyr::mutate_if(is.numeric,
                    .funs = function(x){(as.character(signif(x, 3)))}) %>%
    knitr::kable() %>%
    kableExtra::kable_styling("striped", position = 'center') %>%
    kableExtra::scroll_box(width = "800px", height = "200px")

## ----echo = T, results = 'hide', warning = FALSE, cache = FALSE---------------
fit_out <- maaslin3(
    input_data = taxa_table,
    input_metadata = metadata[metadata$diagnosis == 'nonIBD',],
    output = 'group_outputs',
    formula = '~ group(red_meat) + antibiotics + age + reads',
    plot_summary_plot = TRUE,
    plot_associations = TRUE,
    heatmap_vars = c('red_meat Within the past 4 to 7 days',
                    'red_meat Within the past 2 to 3 days',
                    'red_meat Yesterday, 1 to 2 times',
                    'red_meat Yesterday, 3 or more times'),
    max_pngs = 200)

## ----echo = FALSE, cache = FALSE----------------------------------------------
full_results <- rbind(fit_out$fit_data_abundance$results,
                    fit_out$fit_data_prevalence$results)
full_results <- full_results %>%
    dplyr::filter(metadata %in% c("red_meat") &
                    feature == 'Alistipes_shahii' &
                    model == 'logistic') %>%
    dplyr::mutate(
        value =
            factor(
                value,
                levels =
                    c('No, I did not consume these products in the last 7 days',
                                                'Within the past 4 to 7 days',
                                                'Within the past 2 to 3 days',
                                                'Yesterday, 1 to 2 times',
                                            'Yesterday, 3 or more times'))) %>%
    dplyr::arrange(value)
rownames(full_results) <- NULL
head(full_results, n = 20) %>%
    dplyr::mutate_if(is.numeric,
                    .funs = function(x){(as.character(signif(x, 3)))}) %>%
    knitr::kable() %>%
    kableExtra::kable_styling("striped", position = 'center') %>%
    kableExtra::scroll_box(width = "800px", height = "100px")

## ----echo = TRUE, results = 'hide', warning = FALSE, cache = FALSE------------
fit_out <- maaslin3(input_data = taxa_table,
                    input_metadata = metadata,
                    output = 'contrast_test_output',
                    formula = '~ diagnosis + dysbiosis_state +
                        antibiotics + age + reads',
                    plot_summary_plot = FALSE,
                    plot_associations = FALSE)

## ----echo = TRUE, warning = FALSE, cache = FALSE------------------------------
contrast_mat <- matrix(c(1, -1, 0, 0, 0, 0, 1, -1), 
    ncol = 4, nrow = 2, byrow = TRUE)
    
colnames(contrast_mat) <- c("diagnosisUC",
                            "diagnosisCD",
                            "dysbiosis_statedysbiosis_UC",
                            "dysbiosis_statedysbiosis_CD")
                            
rownames(contrast_mat) <- c("diagnosis_test", "dysbiosis_test")

contrast_mat

contrast_out <- maaslin_contrast_test(maaslin3_fit = fit_out,
                        contrast_mat = contrast_mat)

head(contrast_out$fit_data_abundance$results, n = 20) %>%
    knitr::kable() %>%
    kableExtra::kable_styling("striped", position = 'center') %>%
    kableExtra::scroll_box(width = "800px", height = "400px")

## -----------------------------------------------------------------------------
sessionInfo()

## -----------------------------------------------------------------------------
# Clean-up
unlink('hmp2_output', recursive = TRUE)
unlink('spike_in_demo', recursive = TRUE)
unlink('total_scaling_demo', recursive = TRUE)
unlink('random_effects_output', recursive = TRUE)
unlink('interaction_output', recursive = TRUE)
unlink('ordered_outputs', recursive = TRUE)
unlink('group_outputs', recursive = TRUE)
unlink('contrast_test_output', recursive = TRUE)

library(testthat)
library(maaslin3)

mgx_in <- data.frame('a' = c(1, 2, 0, 4, 5), 
                    'b' = c(2, 3, 4, 5, 6),
                    'c' = c(3, 4, 5, 6, 0))
rownames(mgx_in) <- paste0("sample", c(1:5))

mtx_in <- data.frame('a' = c(1, 2, 3, 4, 5), 
                    'b' = c(2, 3, 4, 5, 0),
                    'c' = c(3, 4, 5, 6, 0))
rownames(mtx_in) <- paste0("sample", c(1:5))

data_in_tss <- data.frame(t(apply(mgx_in, MARGIN = 1, 
                                FUN = function(x){x / sum(x)})))
data_in_tss[data_in_tss == 0 & mtx_in == 0] <- NA
data_in_tss[data_in_tss == 0] <- min(data_in_tss[data_in_tss > 0], na.rm=T) / 2
data_in_tss_log <- log2(data_in_tss)

expect_that(preprocess_dna_mtx(mgx_in, mtx_in)$dna_table, 
            equals(data_in_tss_log))

data_in_tss <- data.frame(t(apply(mtx_in, MARGIN = 1, 
                            FUN = function(x){x / sum(x)})))

expect_that(preprocess_dna_mtx(mgx_in, mtx_in)$rna_table, 
                                equals(data_in_tss))

#!/usr/bin/env Rscript
utils::globalVariables(".progress")
###############################################################################
# MaAsLin3 visualizations

# Copyright (c) 2025 Harvard School of Public Health

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
###############################################################################

# MaAsLin 3 ggplot theme based on Nature journal requirements
nature_theme <- function(x_axis_labels, y_label) {
    # set default text format based on categorical and length
    angle <- NULL
    hjust <- NULL
    size <- 8
    if (max(nchar(x_axis_labels), na.rm = TRUE) > 5) {
        angle <- 45
        hjust <- 1
        size <- 6
    }
    axis_title_size <- 10
    if (nchar(y_label) > 15) {
        axis_title_size <- 8
    }
    if (nchar(y_label) > 25) {
        axis_title_size <- 6
    }
    return (
        ggplot2::theme_bw() + ggplot2::theme(
            axis.text.x = ggplot2::element_text(
                size = size,
                vjust = 1,
                hjust = hjust,
                angle = angle
            ),
            axis.text.y = ggplot2::element_text(size = 8, hjust = 1),
            axis.title = ggplot2::element_text(size = axis_title_size),
            plot.title = ggplot2::element_text(size = 7, face = 'bold'),
            legend.title = ggplot2::element_text(size = 6, face = 'bold'),
            legend.text = ggplot2::element_text(size = 6),
            axis.line = ggplot2::element_line(colour = 'black', 
                                            linewidth = .25),
            axis.line.x = ggplot2::element_line(colour = 'black', 
                                                linewidth = .25),
            axis.line.y = ggplot2::element_line(colour = 'black', 
                                                linewidth = .25),
            panel.border = ggplot2::element_blank(),
            panel.grid.major = ggplot2::element_blank(),
            panel.grid.minor = ggplot2::element_blank()
        )
    )
}

# Pre-process MaAsLin 3 results for plotting
preprocess_merged_results <- function(merged_results) {
    # Disregard abundance-induced-prevalence errors in plotting
    merged_results$error[grepl("Prevalence association possibly induced", 
                                merged_results$error)] <- NA
    # Disregard small random effect group warning
    merged_results$error[grepl("<4 average observations per random effect", 
        merged_results$error)] <- NA
    
    merged_results <-
        merged_results[is.na(merged_results$error) &
                        !is.na(merged_results$qval_individual) &
                        !is.na(merged_results$coef),]
    
    if (nrow(merged_results) == 0) {
        logging::loginfo(paste(
            "No associations were without errors. 
                No summary plot generated."
        ))
        return(NULL)
    }
    merged_results$model <-
        data.table::fifelse(merged_results$model == 'linear', 'Abundance', 'Prevalence')
    merged_results$full_metadata_name <-
        data.table::fifelse(
            merged_results$metadata == merged_results$value,
            merged_results$metadata,
            paste0(merged_results$metadata, ' ', merged_results$value)
        )
    return(merged_results)
}

# Make the coefficient plot part of the summary figure
make_coef_plot <- function(merged_results_sig, 
                        coef_plot_vars,
                        max_significance,
                        median_comparison_prevalence,
                        median_comparison_abundance,
                        median_df,
                        plot_threshold = 10) {
    coef_plot_data <-
        merged_results_sig[merged_results_sig$full_metadata_name %in% 
                            coef_plot_vars,]
    
    # Limit plotted coefficients to median +/- 10 times distance to quartiles
    get_lo = function(x, pt = plot_threshold) {
        med_x = collapse::fmedian(x)
        
        q_x = collapse::fquantile(x, .25)
        
        med_x - pt * (med_x - q_x)
    }
    
    get_hi = function(x, pt = plot_threshold) {
        med_x = collapse::fmedian(x)
        
        q_x = collapse::fquantile(x, .75)
        
        med_x + pt * (q_x - med_x)
    }
    
    quantile_df <- coef_plot_data |>
        collapse::fgroup_by("full_metadata_name") |>
        collapse::collapv(by = "full_metadata_name",
                          FUN = list(get_lo, get_hi),
                          cols = "coef") |> 
        collapse::fungroup() |> 
        collapse::fselect("full_metadata_name", 
                          "lower_q" = "get_lo.coef", 
                          "upper_q" = "get_hi.coef") |> 
        data.frame(check.names = FALSE)
        
    rownames(quantile_df) <- quantile_df$full_metadata_name
    
    # Make sure insignificant coefficients don't distort the plot
    coef_plot_data <-
        coef_plot_data[coef_plot_data$qval_individual < 
                        max_significance |
                        (coef_plot_data$coef > quantile_df[
                            coef_plot_data$full_metadata_name, 
                            'lower_q'] &
                                coef_plot_data$coef < quantile_df[
                                    coef_plot_data$full_metadata_name, 
                                    'upper_q']),]
    
    # Choose breaks for plot
    custom_break_fun <- function(n) {
        return(function(x) {
            extended_breaks <- scales::breaks_extended(n)(x)
            if (max(x) > 0) {
                extended_breaks <- extended_breaks[
                    extended_breaks <= max(x) * 0.9]
            } else {
                extended_breaks <- extended_breaks[
                    extended_breaks <= max(x) * 1.1]
            }
            if (min(x) > 0) {
                extended_breaks <- extended_breaks[
                    extended_breaks >= min(x) * 1.1]
            } else {
                extended_breaks <- extended_breaks[
                    extended_breaks >= min(x) * 0.9]
            }
            extended_breaks
        })
    }
    
    # Start plot
    p1 <-
        ggplot2::ggplot(coef_plot_data,
                        ggplot2::aes(x = .data$coef, y = .data$feature))
    
    # Show nulls the coefficients were compared against
    if (median_comparison_prevalence |
        median_comparison_abundance) {
        p1 <- p1 +
            ggplot2::guides(linetype = ggplot2::guide_legend(
                title = 'Null hypothesis', order = 1),
            ) +
            ggplot2::geom_vline(
                data = median_df[median_df$full_metadata_name %in% 
                                    coef_plot_vars,],
                ggplot2::aes(
                    xintercept = .data$median_val,
                    linetype = .data$model
                ),
                color = "darkgray"
            ) +
            ggplot2::scale_linetype_manual(values = c("Prevalence" = 
                                                        "dashed", 
                                                    "Abundance" = 
                                                        "solid"))
    } else {
        p1 <- p1 +
            ggplot2::geom_vline(
                ggplot2::aes(xintercept = 0),
                color = "darkgray",
                linetype = 'dashed'
            )
    }
    
    # Q-value color scale
    scale_fill_gradient_limits <-
        c(min(max_significance, 10 ^ floor(log10(
            min(coef_plot_data$qval_individual)
        ))), 1)
    if (min(coef_plot_data$qval_individual) < max_significance) {
        scale_fill_gradient_breaks <-
            c(10 ^ floor(log10(
                min(coef_plot_data$qval_individual)
            )), max_significance, 1)
    } else {
        scale_fill_gradient_breaks <- c(max_significance, 1)
    }
    if (min(coef_plot_data$qval_individual) < max_significance) {
        scale_fill_gradient_labels <-
            c(paste0("1e", floor(log10(
                min(coef_plot_data$qval_individual)
            ))),
            paste0(max_significance),
            "1")
    } else {
        scale_fill_gradient_labels <- c(paste0(max_significance),
                                        "1")
    }
    
    # Create the whole plot
    p1 <- p1 +
        ggplot2::geom_errorbar(
            ggplot2::aes(
                xmin = .data$coef - .data$stderr,
                xmax = .data$coef + .data$stderr
            ),
            width = 0.2
        ) +
        ggplot2::geom_point(
            data = coef_plot_data[coef_plot_data$model == 
                                    'Prevalence',],
            ggplot2::aes(
                shape = .data$model,
                fill = .data$qval_individual
            ),
            size = 4.5,
            color = "black"
        ) +
        ggplot2::scale_fill_gradient(
            low = "#008B8B",
            high = "white",
            limits = scale_fill_gradient_limits,
            breaks = scale_fill_gradient_breaks,
            labels = scale_fill_gradient_labels,
            transform = scales::pseudo_log_trans(sigma = 0.001),
            name = bquote("Prevalence" ~ P["FDR"])
        ) +
        ggnewscale::new_scale_fill() +
        ggplot2::geom_point(
            data = coef_plot_data[coef_plot_data$model == 'Abundance',],
            ggplot2::aes(
                shape = .data$model,
                fill = .data$qval_individual
            ),
            size = 4.5,
            color = "black"
        ) +
        ggplot2::scale_fill_gradient(
            low = "#8B008B",
            high = "white",
            limits = scale_fill_gradient_limits,
            breaks = scale_fill_gradient_breaks,
            labels = scale_fill_gradient_labels,
            transform = scales::pseudo_log_trans(sigma = 0.001),
            name = bquote("Abundance" ~ P["FDR"])
        ) +
        ggplot2::scale_x_continuous(
            breaks = custom_break_fun(n = 6),
            limits = c(
                min(coef_plot_data$coef) - 
                    quantile(coef_plot_data$stderr, 0.8),
                max(coef_plot_data$coef) + 
                    quantile(coef_plot_data$stderr, 0.8)
            )
        ) +
        ggplot2::scale_shape_manual(name = "Association", values =
                                        c(21, 24)) +
        ggplot2::guides(shape = ggplot2::guide_legend(order = 2), ) +
        ggplot2::labs(x = expression(paste(beta, " coefficient")),  
                    y = "Feature") +
        ggplot2::theme_bw() +
        ggplot2::theme(
            axis.title = ggplot2::element_text(size = 16),
            axis.text.y = ggplot2::element_text(size = 14),
            axis.text.x = ggplot2::element_text(size = 14),
            legend.title = ggplot2::element_text(size = 16),
            legend.text = ggplot2::element_text(size = 14, 
                                                face = "plain"),
            legend.position = "right",
            legend.background = ggplot2::element_rect(
                fill = "transparent"),
            panel.spacing = ggplot2::unit(0, "lines"),
            panel.grid.minor = ggplot2::element_blank(),
            strip.text = ggplot2::element_text(size = 14),
            strip.background = ggplot2::element_rect(
                fill = "transparent")
        ) +
        ggplot2::facet_wrap(
            ~ factor(full_metadata_name, 
                    levels = unique(coef_plot_vars)),
            scales = 'free_x',
            ncol = length(coef_plot_vars)
        )
    return(p1)
}

# Make the heatmap plot part of the summary figure
make_heatmap_plot <- function(merged_results_sig,
                            heatmap_vars,
                            max_significance,
                            median_comparison_prevalence,
                            median_comparison_abundance) {
    # Create column for significance star annotation
    merged_results_sig$sig_star <-
        cut(
            merged_results_sig$qval_individual,
            breaks = c(-Inf, max_significance / 10, max_significance, Inf),
            label = c("**", "*", "")
        )
    
    # Bin coefficients into categories
    coefficient_thresh <-
        round(max(abs(
            quantile(merged_results_sig$coef, c(0.1, 0.9))
        )) / 10, 1) * 5
    if (coefficient_thresh == 0) {
        coefficient_thresh <- 0.5
    }
    coef_breaks <-
        c(
            -coefficient_thresh,-coefficient_thresh / 2,
            0,
            coefficient_thresh / 2,
            coefficient_thresh,
            Inf
        )
    threshold_set <-
        c(
            paste0("(-Inf,", -1 * coefficient_thresh, "]"),
            paste0(
                "(",-1 * coefficient_thresh,
                ",",-1 / 2 * coefficient_thresh,
                "]"
            ),
            paste0("(", -1 / 2 * coefficient_thresh, ",0]"),
            paste0("(0,", 1 / 2 * coefficient_thresh, "]"),
            paste0(
                "(",
                1 / 2 * coefficient_thresh,
                ",",
                1 * coefficient_thresh,
                "]"
            ),
            paste0("(", 1 * coefficient_thresh, ",Inf)")
        )
    
    threshold_indices <-
        vapply(merged_results_sig$coef, function(value) {
            which(value < coef_breaks)[1]
        }, FUN.VALUE = 0)
    
    merged_results_sig <- merged_results_sig |>  
        collapse::fmutate(coef_cat = threshold_set[threshold_indices])
    
    merged_results_sig$coef_cat <-
        factor(merged_results_sig$coef_cat, levels = threshold_set)
    
    # Create color scheme
    scale_fill_values <-
        rev((RColorBrewer::brewer.pal(n = 6, name = "RdBu")))
    names(scale_fill_values) <- threshold_set
    
    # Create grid for plotting
    heatmap_data <-
        merged_results_sig[merged_results_sig$full_metadata_name %in% 
                            heatmap_vars,]
    
    grid <- expand.grid(
        feature = unique(heatmap_data$feature),
        full_metadata_name = unique(heatmap_data$full_metadata_name),
        model = unique(heatmap_data$model)
    )
    if (length(unique(grid$model)) == 2) {
        grid$model <- factor(grid$model, levels = c('Prevalence', 'Abundance'))
    }
    heatmap_data <-
        merge(
            grid,
            heatmap_data,
            by = c("feature", "full_metadata_name", "model"),
            all.x = TRUE
        )
    heatmap_data$coef[is.na(heatmap_data$coef)] <- NA
    
    # Create whole plot
    p2 <-
        ggplot2::ggplot(heatmap_data,
                        ggplot2::aes(
                            x = factor(.data$full_metadata_name, 
                                    unique(heatmap_vars)),
                            y = .data$feature
                        )) +
        ggplot2::geom_tile(
            data = heatmap_data,
            ggplot2::aes(fill = .data$coef_cat),
            colour = "white",
            linewidth = 0.2
        ) +
        ggplot2::scale_fill_manual(
            name = "Beta coefficient",
            na.value = "#EEEEEE",
            values = scale_fill_values
        ) +
        ggplot2::geom_text(
            ggplot2::aes(
                label = .data$sig_star,
                color = .data$sig_star
            ),
            size = 6,
            vjust = 0.75,
            hjust = 0.5,
            key_glyph = ggplot2::draw_key_blank
        ) +
        ggplot2::scale_color_manual(
            name = bquote("Covariates" ~ P["FDR"]),
            breaks = c("*", "**", ""),
            values = c("black", "black", "black"),
            labels = c(paste0("* < ", round(
                max_significance, 3
            )), paste0(
                "** < ", round(max_significance / 10, 5)
            ), ""),
            drop = FALSE
        ) +
        ggplot2::labs(x = '',
                    y = "Feature",
                    caption = "") +
        ggplot2::theme_bw() +
        ggplot2::theme(
            axis.title = ggplot2::element_text(size = 16),
            axis.text.x = ggplot2::element_text(
                size = 14,
                angle = 90,
                vjust = 0.5,
                hjust = 1
            ),
            legend.title = ggplot2::element_text(size = 16),
            legend.text = ggplot2::element_text(size = 14, 
                                                face = "plain"),
            legend.position = "right",
            legend.background = ggplot2::element_rect(
                fill = "transparent"),
            panel.spacing = ggplot2::unit(0, "lines"),
            panel.grid.minor = ggplot2::element_blank(),
            strip.text = ggplot2::element_text(size = 14),
            strip.background = ggplot2::element_rect(
                fill = "transparent")
        ) +
        ggplot2::guides(
            fill = ggplot2::guide_legend(order = 1),
            color = ggplot2::guide_legend(order = 2),
        ) +
        ggplot2::facet_grid(~ model, labeller = ggplot2::labeller(
            model = c("abundance" = "Abundance", 
                    "prevalence" = "Prevalence")
        ))
    return(p2)
}

# MaAsLin3 summary_plot function for overall view of associations
maaslin3_summary_plot <-
    function(merged_results,
            summary_plot_file,
            figures_folder,
            first_n = 30,
            max_significance = 0.1,
            coef_plot_vars = NULL,
            heatmap_vars = NULL,
            median_comparison_abundance = FALSE,
            median_comparison_prevalence = FALSE,
            balanced = FALSE,
            save_plots_rds = FALSE) {
        ret_plots <- list()
        if (first_n > 200) {
            logging::logerror(
                paste(
                    "At most 200 features can be plotted in the heatmap. 
                    Please choose a smaller first_n."
                )
            )
            return()
        }
        
        merged_results <- preprocess_merged_results(merged_results)
        
        if (is.null(merged_results) || nrow(merged_results) == 0) {
            return()
        }
        
        median_df <- merged_results |> 
            collapse::fselect(  c("full_metadata_name", "model", "null_hypothesis")) |> 
            collapse::fgroup_by(c("full_metadata_name", "model")) |> 
            collapse::fmedian() |> 
            collapse::frename("median_val" = "null_hypothesis")

        # Check variables can be plotted
        if (!is.null(coef_plot_vars) | !is.null(heatmap_vars)) {
            if (any(
                !c(coef_plot_vars, heatmap_vars) %in% 
                unique(merged_results$full_metadata_name)
            )) {
                logging::loginfo(
                    paste0(
                        "The following specified variables were not 
                        found in the associations: ",
                        paste0(setdiff(
                            c(coef_plot_vars, heatmap_vars),
                            unique(merged_results$full_metadata_name)
                        ), collapse = ', '),
                        collapse = ''
                    )
                )
                logging::loginfo(paste0(
                    "Available associations: ",
                    paste0(
                        unique(merged_results$full_metadata_name),
                        collapse = ', '
                    ),
                    collapse = ''
                ))
                return(NULL)
            }
            merged_results <-
                merged_results[merged_results$full_metadata_name %in%
                                c(coef_plot_vars, heatmap_vars),]
        }
        
        # Subset associations for plotting
        merged_results_joint_only <- merged_results |> 
            collapse::fselect(c('feature', 'qval_joint', 'full_metadata_name')) |> 
            collapse::funique() |> 
            collapse::roworderv("qval_joint")
        
        if (length(unique(merged_results_joint_only$feature)) < first_n) {
            first_n <- length(unique(merged_results_joint_only$feature))
            signif_taxa <-
                unique(merged_results_joint_only$feature)[seq(first_n)]
        } else {
            # If balanced is turned on but there are not
            # coefs choosen, error out
            if (balanced){
                if (is.null(coef_plot_vars)){
                    logging::logerror(
                        paste(
                            "Balanced plotting requires 
                            you set the variables you 
                            want to plot using
                            the parameter coef_plot_vars"
                        )
                    )
                    return()
                } else {
                    # grab the first N feature where 
                    # N=N/(length of coef_plot_var) to
                    # plot the coef plot
                    first_n_per <- ceiling(first_n/length(coef_plot_vars))
                    
                    signif_taxa <- merged_results_joint_only |>
                        collapse::roworderv(cols = c("full_metadata_name", 
                                                     "qval_joint")) |> 
                        collapse::fslicev("full_metadata_name",
                                          n = first_n_per) |> 
                        collapse::get_elem("feature") |> 
                        collapse::funique()
                }
            } else {
                signif_taxa <-
                unique(merged_results_joint_only$feature)[seq(first_n)]
            }
        }

        
        merged_results_sig <- merged_results |> 
            collapse::fsubset(merged_results$feature %in% signif_taxa)
        # TODO: consider collapse::`%iin%`
        
        # Order features
        ord_feature <-
            with(merged_results_sig, reorder(feature, coef))
        ord_feature <- levels(ord_feature)
        
        merged_results_sig$feature <-
            factor(merged_results_sig$feature, levels = ord_feature)
        
        # Choose variables for plotting if not set
        if (is.null(coef_plot_vars)) {
            
            mean_log_qval <- merged_results_sig |> 
                collapse::collapv(by   = "full_metadata_name",
                                  cols = "qval_joint",
                                  FUN  = \(x) collapse::fmean(log(x), na.rm = TRUE)) |> 
                collapse::fselect("full_metadata_name", 
                                  "mean_value" = "qval_joint")
                
            coef_plot_vars <-
                mean_log_qval$full_metadata_name[
                    order(mean_log_qval$mean_value)]
            
            coef_plot_vars <- setdiff(coef_plot_vars, heatmap_vars)
            
            if (length(coef_plot_vars) > 0) {
                coef_plot_vars <-
                    coef_plot_vars[seq(min(2, length(coef_plot_vars)))]
            }
        }
        
        # Choose variables for plotting if not set
        if (is.null(heatmap_vars)) {
            
            mean_log_qval <- merged_results_sig |> 
                collapse::collapv(by   = "full_metadata_name",
                                  cols = "qval_joint",
                                  FUN  = \(x) collapse::fmean(log(x), na.rm = TRUE)) |> 
                collapse::fselect("full_metadata_name", 
                                  "mean_value" = "qval_joint")
            
            heatmap_vars <-
                mean_log_qval$full_metadata_name[
                    order(mean_log_qval$mean_value)]
            
            heatmap_vars <- setdiff(heatmap_vars, coef_plot_vars)
        }
        
        # Create coefficient plot
        if (length(coef_plot_vars) > 0 &
            sum(merged_results_sig$full_metadata_name %in% 
                coef_plot_vars) >= 1) {
            if (balanced) {
                plot_thres <- 5
            } else {
                plot_thres <- 10
            }
            p1 <- make_coef_plot(merged_results_sig, 
                                coef_plot_vars,
                                max_significance,
                                median_comparison_prevalence,
                                median_comparison_abundance,
                                median_df,
                                plot_threshold = plot_thres)
            
        } else {
            p1 <- NULL
        }
        
        # Create heatmap plot
        if (length(heatmap_vars) > 0 &
            sum(merged_results_sig$full_metadata_name %in% heatmap_vars) >= 1) {
            
            p2 <- make_heatmap_plot(merged_results_sig, 
                                    heatmap_vars,
                                    max_significance,
                                    median_comparison_prevalence,
                                    median_comparison_abundance)
            
            if (!is.null(p1)) {
                p2 <- p2 + ggplot2::theme(
                    axis.text.y = ggplot2::element_blank(),
                    axis.title.y = ggplot2::element_blank(),
                    axis.ticks.y = ggplot2::element_blank(),
                )
            }
            
        } else {
            p2 <- NULL
        }
        
        # Combine plots
        if (!is.null(p1) & !is.null(p2)) {
            final_plot <- patchwork::wrap_plots(
                p1,
                p2,
                ncol = 3,
                widths = c(max(
                    0, length(coef_plot_vars) * (max(15, max(
                        nchar(as.character(coef_plot_vars))
                    ))) / 15 - 2
                ) + 2,
                max(0, length(
                    heatmap_vars
                ) / 4 - 2) + 2,
                0.5),
                guides = 'collect'
            )
        } else if (is.null(p1) & !is.null(p2)) {
            final_plot <- p2
        } else if (!is.null(p1) & is.null(p2)) {
            final_plot <- p1
        } else {
            final_plot <- NULL
        }
        ret_plots[["coefficient"]] <- p1
        ret_plots[["heat"]] <- p2
        ret_plots[["final"]] <- final_plot
        
        # Save plot
        if (!is.null(final_plot)) {
            height_out <-
                9.5 + max(first_n / 5 - 5, 0) + max(nchar(c(
                    as.character(coef_plot_vars),
                    as.character(heatmap_vars)
                ))) / 10
            width_out <-
                5 + max(nchar(merged_results$feature)) / 12 +
                (length(coef_plot_vars) * (max(20, max(
                    nchar(as.character(coef_plot_vars))
                ))) / 20) * 2.5 +
                length(heatmap_vars) * 0.25
            
            # Silence ggplot warnings
            tryCatch({
                withCallingHandlers({
                    ggplot2::ggsave(
                        summary_plot_file,
                        plot = final_plot,
                        height = height_out,
                        width = width_out
                    )
                },
                warning = function(w) {
                    invokeRestart("muffleWarning")
                })
            })
            
            png_file <- file.path(figures_folder, "summary_plot.png")
            
            tryCatch({
                withCallingHandlers({
                    ggplot2::ggsave(png_file,
                                    plot = final_plot,
                                    height = height_out,
                                    width = width_out)
                },
                warning = function(w) {
                    invokeRestart("muffleWarning")
                })
            })
            if (save_plots_rds) {
                saveRDS(final_plot,
                        file = paste(figures_folder,
                            "/" , "summary_plot_gg.RDS", sep = ""))
            }
        }
        return(ret_plots)
    }

# Create a scatterplot for abundance vs. continuous associations
make_scatterplot <- function(joined_features_metadata_abun,
                            metadata_name,
                            feature_name,
                            normalization,
                            transformation,
                            coef_val,
                            qval,
                            N_nonzero,
                            N_total,
                            results_value) {
    match.arg(normalization, c('Total sum scaling', 'Center log ratio', 'None'))
    match.arg(transformation, c('Log base 2', 'Pseudo-log base 2', 'None'))

    temp_plot <-
        ggplot2::ggplot(data = joined_features_metadata_abun,
                        ggplot2::aes(
                            as.numeric(.data$metadata),
                            as.numeric(.data$feature_abun)
                        )) +
        ggplot2::geom_point(
            fill = 'darkolivegreen4',
            color = 'black',
            alpha = .5,
            shape = 21,
            size = 1,
            stroke = 0.15
        ) +
        ggplot2::scale_x_continuous(limits = c(
            min(joined_features_metadata_abun['metadata']),
            max(joined_features_metadata_abun['metadata'])
        )) +
        ggplot2::scale_y_continuous(
            limits = c(
                min(joined_features_metadata_abun[
                    'feature_abun']),
                max(joined_features_metadata_abun[
                    'feature_abun'])
            ),
            expand = ggplot2::expansion(mult = c(0, 0.2))
        ) +
        ggplot2::stat_smooth(
            method = "glm",
            formula = 'y ~ x',
            linewidth = 0.5,
            color = 'blue',
            na.rm = TRUE
        ) +
        ggplot2::guides(alpha = 'none') +
        ggplot2::xlab(metadata_name) +
        ggplot2::ylab(
            paste0(
                feature_name,
                '\n(Normalization: ',
                normalization,
                ', Transformation: ',
                transformation,
                ')'
            )
        ) +
        nature_theme(
            joined_features_metadata_abun['metadata'],
            paste0(
                feature_name,
                '\n(Normalization: ',
                normalization,
                ', Transformation: ',
                transformation,
                ')'
            )
        ) +
        ggplot2::annotate(
            geom = "text",
            x = Inf,
            y = Inf,
            hjust = 1,
            vjust = 1,
            label = sprintf(
            "FDR: %s\nCoefficient (in full model): %sN: %s\nN (not zero): %s",
                formatC(qval, format = "e", digits = 1),
                if (is.na(coef_val)) 'NA' else formatC(
                    coef_val,
                    format = "e",
                    digits = 1
                ),
                formatC(
                    N_total,
                    format = 'f',
                    digits = 0
                ),
                formatC(
                    N_nonzero,
                    format = 'f',
                    digits = 0
                )
            ) ,
            color = "black",
            size = 2,
            fontface = "italic"
        )
    return(temp_plot)
}

# Create boxplot for abundance vs. categorical associations
make_boxplot_lm <- function(joined_features_metadata_abun,
                            metadata_name,
                            feature_name,
                            normalization,
                            transformation,
                            coef_val,
                            qval,
                            N_nonzero,
                            N_total,
                            results_value) {
    match.arg(normalization, c('Total sum scaling', 'Center log ratio', 'None'))
    match.arg(transformation, c('Log base 2', 'Pseudo-log base 2', 'None'))
    
    temp_plot <-
        ggplot2::ggplot(data = joined_features_metadata_abun,
                        ggplot2::aes(.data$metadata, 
                                    .data$feature_abun)) +
        ggplot2::geom_boxplot(
            ggplot2::aes(fill = .data$metadata),
            outlier.alpha = 0.0,
            na.rm = TRUE,
            alpha = .5,
            show.legend = FALSE
        ) +
        ggplot2::geom_point(
            ggplot2::aes(fill = .data$metadata),
            alpha = 0.75 ,
            size = 1,
            shape = 21,
            stroke = 0.15,
            color = 'black',
            position = ggplot2::position_jitterdodge()
        ) +
        ggplot2::scale_fill_brewer(palette = "Spectral") +
        ggplot2::scale_y_continuous(
            expand = ggplot2::expansion(mult = c(0, 0.2)))
    
    temp_plot <- temp_plot +
        nature_theme(
            as.character(
                joined_features_metadata_abun$metadata),
            paste0(
                feature_name,
                '\n(Normalization: ',
                normalization,
                ', Transformation: ',
                transformation,
                ')'
            )
        ) +
        ggplot2::theme(
            panel.grid.major = ggplot2::element_blank(),
            panel.grid.minor = ggplot2::element_blank(),
            panel.background = ggplot2::element_blank(),
            axis.line = ggplot2::element_line(colour = "black")
        ) +
        ggplot2::xlab(metadata_name) +
        ggplot2::ylab(
            paste0(
                feature_name,
                '\n(Normalization: ',
                normalization,
                ', Transformation: ',
                transformation,
                ')'
            )
        ) +
        ggplot2::theme(legend.position = "none") +
        ggplot2::annotate(
            geom = "text",
            x = Inf,
            y = Inf,
            hjust = 1,
            vjust = 1,
            label = sprintf(
                "Value: %s\nFDR: %s\nCoefficient (in full model): %s",
                paste0(results_value, collapse = ', '),
                paste0(
                    formatC(qval, format = "e", digits = 1),
                    collapse = ', '
                ),
                paste0(
                    vapply(coef_val, function(cv) {
                        if (is.na(cv)) 'NA' else formatC(cv, format = "e", digits = 1)
                    }, character(1)),
                    collapse = ', '
                )
            ) ,
            color = "black",
            size = 2,
            fontface = "italic"
        )
    return(temp_plot)
}

# Create plots for abundance associations
make_lm_plot <- function(this_signif_association,
                        joined_features_metadata,
                        metadata,
                        metadata_name,
                        feature_name,
                        normalization,
                        transformation,
                        feature_specific_covariate_name,
                        feature_specific_covariate) {
    match.arg(normalization, c('Total sum scaling', 'Center log ratio', 'None'))
    match.arg(transformation, c('Log base 2', 'Pseudo-log base 2', 'None'))

    coef_val <-
        this_signif_association[
            this_signif_association$model == 'linear',]$coef
    qval <-
        this_signif_association[
            this_signif_association$model == 'linear',]$qval_individual
    N_nonzero <-
        this_signif_association[
            this_signif_association$model == 'linear',]$N_not_zero
    N_total <-
        this_signif_association[
            this_signif_association$model == 'linear',]$N
    results_value <-
        this_signif_association[
            this_signif_association$model == 'linear',]$value
    
    joined_features_metadata_abun <-
        joined_features_metadata[
            !is.na(joined_features_metadata$feature_abun),]
    if (is.numeric(joined_features_metadata_abun$metadata) &
        length(unique(
            joined_features_metadata_abun$metadata)) > 1) {
        logging::loginfo(
            "Creating scatter plot for continuous 
                        data (linear), %s vs %s",
            metadata_name,
            feature_name
        )
        temp_plot <- make_scatterplot(joined_features_metadata_abun,
                                    metadata_name,
                                    feature_name,
                                    normalization,
                                    transformation,
                                    coef_val,
                                    qval,
                                    N_nonzero,
                                    N_total,
                                    results_value)
    } else {
        x_axis_label_names <- 
            unique(joined_features_metadata_abun$metadata)
        
        sorted_fixed_order <-
            order(match(results_value, levels(x_axis_label_names)))
        coef_val <- coef_val[sorted_fixed_order]
        qval <- qval[sorted_fixed_order]
        N_nonzero <- N_nonzero[sorted_fixed_order]
        N_total <- N_total[sorted_fixed_order]
        results_value <-
            results_value[sorted_fixed_order]
        
        if (!is.null(feature_specific_covariate_name)) {
            if (metadata_name == feature_specific_covariate_name) {
                renamed_levels <-
                    as.character(levels(
                        feature_specific_covariate[, feature_name]))
            } else {
                renamed_levels <- as.character(
                    levels(metadata[, metadata_name]))
            }
        } else {
            renamed_levels <- as.character(
                levels(metadata[, metadata_name]))
        }
        
        if (length(renamed_levels) == 0) {
            renamed_levels <- x_axis_label_names
        }
        for (name in x_axis_label_names) {
            total <-
                length(which(
                    joined_features_metadata_abun$metadata == name
                ))
            new_n <-
                paste(name, " (n = ", total, ")", sep = "")
            levels(joined_features_metadata_abun[, 'metadata'])[
                levels(
                    joined_features_metadata_abun[, 'metadata']) == 
                    name] <-
                new_n
            renamed_levels <-
                replace(renamed_levels,
                        renamed_levels == name,
                        new_n)
        }
        
        logging::loginfo(
            "Creating box plot for categorical data (linear), 
                        %s vs %s",
            metadata_name,
            feature_name
        )
        
        temp_plot <- make_boxplot_lm(joined_features_metadata_abun,
                                    metadata_name,
                                    feature_name,
                                    normalization,
                                    transformation,
                                    coef_val,
                                    qval,
                                    N_nonzero,
                                    N_total,
                                    results_value)
    }
    return(temp_plot)
}

# Create boxplot for prevalence vs. continuous associations
make_boxplot_logistic <- function(joined_features_metadata_prev,
                                metadata_name,
                                feature_name,
                                normalization,
                                transformation,
                                coef_val,
                                qval,
                                N_nonzero,
                                N_total,
                                results_value) {
    match.arg(normalization, c('Total sum scaling', 'Center log ratio', 'None'))
    match.arg(transformation, c('Log base 2', 'Pseudo-log base 2', 'None'))
    
    temp_plot <-
        ggplot2::ggplot(data = joined_features_metadata_prev,
                        ggplot2::aes(.data$feature_abun, 
                                    .data$metadata)) +
        ggplot2::geom_boxplot(
            ggplot2::aes(fill = .data$feature_abun),
            outlier.alpha = 0.0,
            na.rm = TRUE,
            alpha = .5,
            show.legend = FALSE
        ) +
        ggplot2::geom_point(
            ggplot2::aes(fill = .data$feature_abun),
            alpha = 0.75 ,
            size = 1,
            shape = 21,
            stroke = 0.15,
            color = 'black',
            position = ggplot2::position_jitterdodge()
        ) +
        ggplot2::scale_fill_brewer(palette = "Spectral") +
        ggplot2::scale_x_discrete(
            expand = ggplot2::expansion(mult = c(0, 0.7)))
    
    temp_plot <- temp_plot +
        nature_theme(metadata_name,
                    joined_features_metadata_prev[
                        'feature_abun']) +
        ggplot2::theme(
            panel.grid.major = ggplot2::element_blank(),
            panel.grid.minor = ggplot2::element_blank(),
            panel.background = ggplot2::element_blank(),
            axis.line = ggplot2::element_line(colour = "black")
        ) +
        ggplot2::xlab(feature_name) +
        ggplot2::ylab(metadata_name) +
        ggplot2::theme(legend.position = "none") +
        ggplot2::annotate(
            geom = "text",
            x = Inf,
            y = Inf,
            hjust = 1,
            vjust = 1,
            label = sprintf(
            "FDR: %s\nCoefficient (in full model): %s\nN: %s\nN (not zero): %s",
                formatC(qval, format = "e", digits = 1),
                if (is.na(coef_val)) 'NA' else formatC(
                    coef_val,
                    format = "e",
                    digits = 1
                ),
                formatC(
                    N_total,
                    format = 'f',
                    digits = 0
                ),
                formatC(
                    N_nonzero,
                    format = 'f',
                    digits = 0
                )
            ) ,
            color = "black",
            size = 2,
            fontface = "italic"
        ) +
        ggplot2::coord_flip()
    return(temp_plot)
}

# Create grid for prevalence vs. categorical associations
make_tile_plot <- function(joined_features_metadata_prev,
                        metadata_name,
                        feature_name,
                        normalization,
                        transformation,
                        coef_val,
                        qval,
                        N_nonzero,
                        N_total,
                        results_value) {
    match.arg(normalization, c('Total sum scaling', 'Center log ratio', 'None'))
    match.arg(transformation, c('Log base 2', 'Pseudo-log base 2', 'None'))
    
    count_df <- joined_features_metadata_prev |> 
        collapse::fcountv(c("feature_abun",
                            "metadata")) |> 
        collapse::roworderv(c("feature_abun",
                              "metadata")) |> 
        collapse::frename("count" = "N")
    
    x_vals <-
        unique(joined_features_metadata_prev$feature_abun)
    y_vals <-
        unique(joined_features_metadata_prev$metadata)
    complete_grid <-
        expand.grid(feature_abun = x_vals,
                    metadata = y_vals)
    
    table_df <- collapse::join(complete_grid,
                               count_df,
                               verbose = FALSE,
                               on = c("feature_abun", "metadata")) 
    
    table_df$count = collapse::replace_na(table_df$count, value = 0)
    
    temp_plot <-
        ggplot2::ggplot(table_df,
                        ggplot2::aes(
                            x = .data$metadata,
                            y = .data$feature_abun
                        )) +
        ggplot2::geom_tile(
            ggplot2::aes(fill = .data$count),
            color = "white",
            lwd = 1.5,
            linetype = 1
        ) +
        ggplot2::geom_text(
            ggplot2::aes(label = .data$count),
            color = "black",
            size = 32 / nrow(table_df)
        ) +
        ggplot2::coord_fixed(ratio = 0.5) +
        ggplot2::scale_fill_gradient2(
            low = "#075AFF",
            mid = "#FFFFCC",
            high = "#FF0000"
        ) +
        ggplot2::scale_y_discrete(
            expand = ggplot2::expansion(mult = c(0, 1 + nrow(
                table_df
            ) / 6))) +
        ggplot2::theme(
            panel.background = ggplot2::element_blank(),
            legend.position = "none"
        )
    
    temp_plot <- temp_plot +
        nature_theme(as.character(table_df$metadata),
                    joined_features_metadata_prev[
                        'feature_abun']) +
        ggplot2::theme(
            panel.grid.major = ggplot2::element_blank(),
            panel.grid.minor = ggplot2::element_blank(),
            panel.background = ggplot2::element_blank(),
            axis.line = ggplot2::element_line(colour = "black")
        ) +
        ggplot2::xlab(metadata_name) +
        ggplot2::ylab(feature_name) +
        ggplot2::theme(legend.position = "none") +
        ggplot2::annotate(
            geom = "text",
            x = Inf,
            y = Inf,
            hjust = 1,
            vjust = 1,
            label = sprintf(
                "Value: %s\nFDR: %s\nCoefficient (in full model): %s",
                paste0(results_value, collapse = ', '),
                paste0(
                    formatC(qval, format = "e", digits = 1),
                    collapse = ', '
                ),
                paste0(
                    vapply(coef_val, function(cv) {
                        if (is.na(cv)) 'NA' else formatC(cv, format = "e", digits = 1)
                    }, character(1)),
                    collapse = ', '
                )
            ) ,
            color = "black",
            size = 2,
            fontface = "italic"
        )
    return(temp_plot)
}

# Create plots for prevalence associations
make_logistic_plot <- function(this_signif_association,
                            joined_features_metadata,
                            metadata,
                            metadata_name,
                            feature_name,
                            normalization,
                            transformation,
                            feature_specific_covariate_name,
                            feature_specific_covariate) {
    match.arg(normalization, c('Total sum scaling', 'Center log ratio', 'None'))
    match.arg(transformation, c('Log base 2', 'Pseudo-log base 2', 'None'))
    
    coef_val <-
        this_signif_association[
            this_signif_association$model == 'logistic',]$coef
    qval <-
        this_signif_association[
            this_signif_association$model == 'logistic',
        ]$qval_individual
    N_nonzero <-
        this_signif_association[
            this_signif_association$model == 'logistic',]$N_not_zero
    N_total <-
        this_signif_association[
            this_signif_association$model == 'logistic',]$N
    results_value <-
        this_signif_association[
            this_signif_association$model == 'logistic',]$value
    
    joined_features_metadata_prev <-
        joined_features_metadata
    joined_features_metadata_prev$feature_abun <-
        data.table::fifelse(
            is.na(joined_features_metadata_prev$feature_abun),
            'Absent',
            'Present'
        )
    
    joined_features_metadata_prev$feature_abun <-
        factor(
            joined_features_metadata_prev$feature_abun,
            levels = c('Present', 'Absent')
        )
    
    if (is.numeric(joined_features_metadata_prev$metadata) &
        length(unique(
            joined_features_metadata_prev$metadata)) > 1) {
        logging::loginfo(
            "Creating boxplot for continuous data (logistic), %s vs %s",
            metadata_name,
            feature_name
        )
        temp_plot <- make_boxplot_logistic(joined_features_metadata_prev,
                                        metadata_name,
                                        feature_name,
                                        normalization,
                                        transformation,
                                        coef_val,
                                        qval,
                                        N_nonzero,
                                        N_total,
                                        results_value)
        
    } else {
        joined_features_metadata_prev$feature_abun <-
            factor(
                joined_features_metadata_prev$feature_abun,
                levels = c('Absent', 'Present')
            )
        
        x_axis_label_names <-
            unique(joined_features_metadata_prev$metadata)
        
        sorted_fixed_order <-
            order(match(results_value, levels(x_axis_label_names)))
        coef_val <- coef_val[sorted_fixed_order]
        qval <- qval[sorted_fixed_order]
        N_nonzero <- N_nonzero[sorted_fixed_order]
        N_total <- N_total[sorted_fixed_order]
        results_value <-
            results_value[sorted_fixed_order]
        
        if (!is.null(feature_specific_covariate_name)) {
            if (metadata_name == feature_specific_covariate_name) {
                renamed_levels <-
                    as.character(levels(
                        feature_specific_covariate[, feature_name]))
            } else {
                renamed_levels <- 
                    as.character(levels(metadata[, metadata_name]))
            }
        } else {
            renamed_levels <- 
                as.character(levels(metadata[, metadata_name]))
        }
        
        if (length(renamed_levels) == 0) {
            renamed_levels <- x_axis_label_names
        }
        for (name in x_axis_label_names) {
            mean_abun <- mean(
                joined_features_metadata_prev$feature_abun[
                    joined_features_metadata_prev$metadata == 
                        name] == 'Present',
                na.rm = TRUE
            )
            new_n <-
                paste(name,
                    " (p = ",
                    round(mean_abun, 2) * 100,
                    "%)",
                    sep = "")
            levels(joined_features_metadata_prev[, 'metadata'])[
                levels(joined_features_metadata_prev[, 'metadata']) 
                == name] <-
                new_n
            renamed_levels <-
                replace(renamed_levels,
                        renamed_levels == name,
                        new_n)
        }
        
        logging::loginfo(
            "Creating tile plot for categorical data (logistic), %s vs %s",
            metadata_name,
            feature_name
        )
        temp_plot <- make_tile_plot(joined_features_metadata_prev,
                                    metadata_name,
                                    feature_name,
                                    normalization,
                                    transformation,
                                    coef_val,
                                    qval,
                                    N_nonzero,
                                    N_total,
                                    results_value)
    }
    return(temp_plot)
}

# Create individual plots for significant associations
maaslin3_association_plots <-
    function(merged_results,
             metadata,
             features,
             max_significance = 0.1,
             figures_folder,
             max_pngs = 10,
             normalization,
             transform,
             feature_specific_covariate = NULL,
             feature_specific_covariate_name = NULL,
             feature_specific_covariate_record = NULL,
             save_plots_rds = FALSE) {
        
        
        match.arg(normalization, c("TSS", "CLR", "NONE"))
        
        match.arg(transform, c("LOG", "PLOG", "NONE"))
        
        # Disregard abundance-induced-prevalence errors in plotting
        merged_results$error[grepl("Prevalence association possibly induced", 
                                   merged_results$error)] <- NA
        
        # Disregard small random effect group warning
        merged_results$error[grepl("<4 average observations per random effect", 
                                   merged_results$error)] <- NA
        
        new_name_normalization <-
            c('Total sum scaling', 'Center log ratio', 'None')
        
        names(new_name_normalization) <- c("TSS", "CLR", "NONE")
        
        normalization <- new_name_normalization[normalization]
        
        new_name_transformation <-
            c('Log base 2', 'Pseudo-log base 2', 'None')
        
        names(new_name_transformation) <- c("LOG", 'PLOG', "NONE")
        
        transformation <- new_name_transformation[transform]
        
        merged_results <-
            merged_results[is.na(merged_results$error) &
                               !is.na(merged_results$qval_individual) &
                               merged_results$qval_individual < 
                               max_significance,]
        
        if (nrow(merged_results) == 0) {
            logging::loginfo(paste("All associations had errors 
                                or were insignificant."))
            return(NULL)
        }
        
        merged_results <-
            merged_results[order(merged_results$qval_individual),]
        
        logging::loginfo(
            paste(
                "Plotting associations from most",
                "to least significant,",
                "grouped by metadata"
            )
        )
        
        saved_plots <- list()
        
        features_by_metadata <-
            unique(merged_results[, c('feature', 'metadata', 'model')])
        
        names(features_by_metadata)[2] = "meta_var"
        
        assoc_num = min(nrow(features_by_metadata), max_pngs)
        
        to_map = features_by_metadata[seq_len(assoc_num),]
        
        to_map$fv = lapply(to_map$feature, \(x) {
            features[[x]]
        })
        
        stat_df = collapse::join(to_map,
                                 merged_results, 
                                 on = c("feature", 
                                        "meta_var" = "metadata",
                                        "model"),
                                 how = "left",
                                 verbose = FALSE) 
        
        stat_df = stat_df |> 
            collapse::fmutate(split_var = paste(stat_df$feature, 
                                                stat_df$meta_var,
                                                stat_df$model,
                                                sep = "_")) 
        
        stat_list = split(stat_df , 
                          stat_df$split_var) 
        # split() sorts the resultant list by f :(
        
        assoc_df = data.frame(split_var = names(stat_list),
                              assoc_stats = I(stat_list))
        
        to_map = to_map |> 
            collapse::fmutate(split_var = paste(to_map$feature, 
                                                to_map$meta_var,
                                                to_map$model,
                                                sep = "_")) |> 
            collapse::join(assoc_df, on = "split_var",
                           how = "left",
                           verbose = FALSE)
        
        collapse::get_vars(to_map, "split_var") <- NULL
        
        # Some associations have multiple rows of statistics to attach. A join +
        # tidyr::nest() would be simpler, but we don't have tidyr.
        
        # # V Attach association stats here so we don't have to pass
        # # around/re-join merged_results at each iteration.
        
        small_meta = metadata |> 
            collapse::fselect(collapse::funique(features_by_metadata$meta_var))
        
        association_plots_folder <-
            file.path(figures_folder, 'association_plots')
        
        if (!file.exists(association_plots_folder)) {
            dir.create(association_plots_folder)
        }
        
        arg_list = list(meta = small_meta,
                        feat_rn = rownames(features),
                        fscn = feature_specific_covariate_name,
                        fsc = feature_specific_covariate,
                        normalization = normalization,
                        transformation = transformation,
                        save_plots_rds = save_plots_rds,
                        ap_dir = association_plots_folder)
        
        if (mirai::daemons_set()) {
            mirai::everywhere({},
                              make_lm_plot = make_lm_plot,
                              make_logistic_plot = make_logistic_plot)
            
            plot_list <- mirai::mirai_map(to_map, 
                                   plot_one_assoc,
                                   .args = arg_list)[.progress]
        } else {
            plot_list <- mapply(plot_one_assoc,
                                to_map$feature,  #life without purrr::pmap :(
                                to_map$meta_var, 
                                to_map$model, 
                                to_map$fv, 
                                to_map$assoc_stats,
                                MoreArgs = arg_list,
                                SIMPLIFY = FALSE,
                                USE.NAMES = FALSE)
        }
        
        to_map$pl <- plot_list
        
        if (save_plots_rds) {
            # save lists for each metadata variable
             if (save_plots_rds) {
                
                split_plots = split(plot_list,
                                    to_map$meta_var)  
                
                ap_dir = association_plots_folder
                save_plot_list_fun = \(x, y) {
                    saveRDS(x, 
                            file = file.path(ap_dir, 
                                             paste0(make.names(y), 
                                                    "_gg_associations.RDS")))
                }
                
                if (mirai::daemons_set()) {
                    # I don't think parallelization helps much here, might be
                    # dependent on disk. TODO: remove?
                    plot_df = data.frame(x = I(split_plots),
                                         y = names(split_plots))
                    
                    mirai::mirai_map(.x = plot_df,
                              .f = save_plot_list_fun,
                              ap_dir = association_plots_folder)[.progress]
                } else {
                    mapply(split_plots,
                           names(split_plots),
                           FUN = save_plot_list_fun)
                }
             }
        } else {
            return(invisible())
        }
        
        names(plot_list) = paste(sep = "_",
                                 to_map$meta_var, 
                                 to_map$feature,
                                 to_map$model)
            
        return(plot_list)
    }

plot_one_assoc = function(feature, meta_var, model, fv, assoc_stats,
                          meta,
                          feat_rn,
                          fscn, # feature_specific_covariate_name
                          fsc,
                          normalization,
                          transformation,
                          save_plots_rds,
                          ap_dir) {
    
    feature_name <- feature
    
    feature_abun <- data.frame(
        sample = feat_rn,
        feature_abun = fv, # features[, feature_name], 
        check.names = FALSE)
    
    metadata_name <- meta_var # features_by_metadata[row_num, 'metadata']
    
    if (!is.null(fscn)) {
        if (metadata_name == fscn) {
            metadata_sub <-
                data.frame(
                    sample = rownames(fsc),
                    metadata = fsc[, feature_name],
                    check.names = FALSE
                )
        } else {
            metadata_sub <- data.frame(sample = rownames(meta),
                                       metadata = meta[, metadata_name], 
                                       check.names = FALSE)
        }
    } else {
        metadata_sub <- data.frame(sample = rownames(meta),
                                   metadata = meta[, metadata_name], 
                                   check.names = FALSE)
    }
    
    joined_features_metadata <-
        collapse::join(feature_abun, 
                       metadata_sub,
                       how = "inner",
                       verbose = FALSE,
                       on = c('sample'))
    
    model_name <- model # features_by_metadata[row_num, 'model']
    
    this_signif_association = assoc_stats
    
    if ('linear' == model_name) {
        temp_plot <- make_lm_plot(this_signif_association,
                                            joined_features_metadata,
                                            meta,
                                            metadata_name,
                                            feature_name,
                                            normalization,
                                            transformation,
                                            fscn,
                                            fsc)
    }
    
    if ('logistic' == model_name) {
        temp_plot <- make_logistic_plot(this_signif_association,
                                                  joined_features_metadata,
                                                  meta,
                                                  metadata_name,
                                                  feature_name,
                                                  normalization,
                                                  transformation,
                                                  fscn,
                                                  fsc)
    }
    
    out_dir = file.path(ap_dir,
                        make.names(meta_var), 
                        model)
    
    if (!dir.exists(out_dir)) dir.create(out_dir, recursive = TRUE)
    
    png_file = file.path(out_dir, 
                         paste0(make.names(meta_var),
                                "_",
                                make.names(feature_name),
                                "_",
                                model, ".png"))
    
    height <- max(960, 18 * max(nchar(unlist(strsplit(
        temp_plot$labels$y, '\n')))))
    
    tryCatch({
        withCallingHandlers({
            ggplot2::ggsave(filename = png_file, 
                            plot = temp_plot, 
                            dpi = 600, 
                            width = 960 / 300, 
                            height = height / 300)
        }, warning = function(w) { 
            invokeRestart("muffleWarning") })
    })
    
    if (save_plots_rds) {
        return(temp_plot)
    } else {
        invisible()
    }
}

library(testthat)
library(maaslin3)

taxa_in <- data.frame('tax 1' = c(1, 0, 0, 4, 5), 
                    'tax 2' = c(2, 3, 4, 5, 6), check.names = FALSE)
rownames(taxa_in) <- paste0("sample", c(1:5))

mtx_in <- data.frame('a' = c(1, 0, 0, 4, 5), 
                    'b' = c(2, 3, 0, 5, 0),
                    'c test' = c(3, 4, 5, 6, 0), check.names = FALSE)
rownames(mtx_in) <- paste0("sample", c(1:5))

rna_per_taxon <- data.frame(RNA = c('a', 'b', 'c test'),
                            taxon = c('tax 1', 'tax 1', 'tax 2'))

data_in_tss <- data.frame(t(apply(taxa_in, MARGIN = 1, 
                                FUN = function(x){x / sum(x)})), 
                        check.names = F)
data_in_tss <- data_in_tss[,c('tax 1', 'tax 1', 'tax 2')]
colnames(data_in_tss) <- c('a', 'b', 'c test')
data_in_tss[3, 1:2] <- NA # Because all RNA for taxa 1 are missing in sample 3
data_in_tss[data_in_tss == 0] <- min(data_in_tss[data_in_tss > 0], na.rm=T) / 2
data_in_tss_log <- log2(data_in_tss)

expect_that(preprocess_taxa_mtx(taxa_in, mtx_in, rna_per_taxon)$dna_table, 
            equals(data_in_tss_log))

data_in_tss <- data.frame(t(apply(mtx_in, MARGIN = 1, 
                            FUN = function(x){x / sum(x)})), check.names = FALSE)

expect_that(preprocess_taxa_mtx(taxa_in, mtx_in, rna_per_taxon)$rna_table, 
                                equals(data_in_tss))

library(testthat)
library(maaslin3)

data_in <- data.frame('a' = c(1, 2, 3, 4, 5), 
                      'b' = c(2, 3, 4, 5, 6),
                      'c' = c(3, 4, 5, 6, 7))
rownames(data_in) <- paste0("sample", c(1:5))

metadata <- data.frame('a' = c(1, 2, 3, 4, 5),
                       'b' = c(0, 1, 0, 1, 0),
                       'c' = c('a', 'b', 'c', 'a', 'b'))
rownames(metadata) <- paste0("sample", c(1:5))

data_in_tss <- data.frame(t(apply(data_in, MARGIN = 1, 
                                  FUN = function(x){x / sum(x)})))
data_in_tss_log <- log2(data_in_tss)
output_tmp <- tempfile()
expect_that(maaslin_transform(data_in_tss, output_tmp),
            equals(data_in_tss_log))

data_in <- data.frame('a' = c(0, 0, 3, 4, 5), 
                      'b' = c(2, 0, 0, 5, 6),
                      'c' = c(3, 4, 0, 0, 7))
rownames(data_in) <- paste0("sample", c(1:5))
data_in_tss <- data.frame(t(apply(data_in, MARGIN = 1, 
                                  FUN = function(x){x / sum(x)})))
data_in_tss[data_in_tss == 0] <- NA
data_in_tss_log <- log2(data_in_tss)
expect_that(maaslin_transform(data_in_tss, output_tmp),
            equals(data_in_tss_log))

data_in_tss[is.na(data_in_tss)] <- min(data_in_tss[!is.na(data_in_tss)]) / 2
data_in_tss_log <- log2(data_in_tss)
data_in_tss <- data.frame(t(apply(data_in, MARGIN = 1, 
                                  FUN = function(x){x / sum(x)})))
expect_that(maaslin_transform(data_in_tss, output_tmp, transform = 'PLOG'),
            equals(data_in_tss_log))

expect_that(maaslin_transform(data_in_tss, output_tmp, transform = 'NONE'),
            equals(data_in_tss))

unlink(output_tmp, recursive = T)

library(testthat)
library(maaslin3)

data_in <- data.frame('a' = c(1, 2, 3, 4, 5), 
                      'b' = c(2, 3, 4, 5, 6),
                      'c' = c(3, 4, 5, 6, 7))
rownames(data_in) <- paste0("sample", c(1:5))

metadata <- data.frame('a' = c(1, 2, 3, 4, 5),
                       'b' = c(0, 1, 0, 1, 0),
                       'c' = c('a', 'b', 'c', 'a', 'b'))
rownames(metadata) <- paste0("sample", c(1:5))

output_tmp <- tempfile()
expect_that(maaslin_filter(data_in, output_tmp),
            equals(data_in))

expect_that(maaslin_filter(data_in, output_tmp, 
                           min_abundance = 3, min_prevalence = 1/2),
            equals(data_in[,-1]))

expect_that(maaslin_filter(data_in, output_tmp, zero_threshold = 3,
                           min_prevalence = 1/2),
            equals(data_in[,-1]))

data_in <- data.frame('a' = c(1, 1, 1, 1, 1), 
                      'b' = c(2, 3, 4, 5, 6),
                      'c' = c(3, 4, 5, 6, 7))
rownames(data_in) <- paste0("sample", c(1:5))

expect_that(maaslin_filter(data_in, output_tmp),
            equals(data_in[,-1]))

unlink(output_tmp, recursive = T)

###############################################################################
# MaAsLin3 utility scripts

# Copyright (c) 2025 Harvard School of Public Health

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
###############################################################################

######################
## TSS Normalization #
######################

TSSnorm <- function(features, zero_threshold) {
    # Convert to Matrix from Data Frame
    features_norm <- as.matrix(features)
    X_mask <- ifelse(features > zero_threshold, 1, 0)
    dd <- colnames(features_norm)
    
    ##############
    # From vegan #
    ##############
    
    x <- as.matrix(features_norm)
    if (any(x < 0, na.rm = TRUE)) {
        k <- min(x, na.rm = TRUE)
        warning("input data contains negative entries: result may be non-sense")
    } else {
        k <- .Machine$double.eps
    }
    
    MARGIN <- 1
    
    tmp <- pmax(k, apply(x, MARGIN, sum, na.rm = TRUE))
    x <- sweep(x, MARGIN, tmp, "/")
    if (any(is.nan(x)))
        warning("result contains NaN, perhaps due to impossible mathematical\n
            operation\n")
    
    x <- ifelse(X_mask, x, NA)
    
    # Convert back to data frame
    features_TSS <- as.data.frame(x)
    
    # Rename the True Positive Features - Same Format as Before
    colnames(features_TSS) <- dd
    
    # Return
    return(features_TSS)
}

######################
## CLR Normalization #
######################

# Only CLRs on the non-zero part of the vector
CLRnorm <- function(features, zero_threshold) {
    # Convert to Matrix from Data Frame
    features_norm <- as.matrix(features)
    dd <- colnames(features_norm)
    
    #####################
    # from chemometrics #
    #####################
    
    # CLR Normalizing the Data
    X <- features_norm
    X_mask <- ifelse(X > zero_threshold, 1, 0)
    Xgeom <-
        exp(1) ^ apply(X, 1, function(x) {
            mean(log(x[x > zero_threshold]))
        })
    features_CLR <- log(X / Xgeom)
    features_CLR <- ifelse(X_mask, features_CLR, NA)
    
    # Convert back to data frame
    features_CLR <- as.data.frame(features_CLR)
    
    # Rename the True Positive Features - Same Format as Before
    colnames(features_CLR) <- dd
    
    # Return
    return(features_CLR)
}

######################
# NONE Normalization #
######################

NONEnorm <- function(features, zero_threshold) {
    X <- as.matrix(features)
    X_mask <- ifelse(X > zero_threshold, 1, 0)
    features_NONE <-
        data.frame(ifelse(X_mask, X, NA), check.names = FALSE)
    return(features_NONE)
}

##########################
# Unscaled Normalization #
##########################

UNSCALEDnorm <- function(features, abs_abundances, zero_threshold) {
    # Convert to Matrix from Data Frame
    if (colnames(abs_abundances) == 'total') {
        X_mask <- ifelse(features > zero_threshold, 1, 0)
        abs_mult_fact <- abs_abundances[rownames(features), 1]
    } else {
        abs_feature <- colnames(abs_abundances)
        X_mask <-
            ifelse(features[, colnames(features) != abs_feature] > 
                    zero_threshold, 1, 0)
        abs_mult_fact <-
            abs_abundances[rownames(features), 1] / features[, abs_feature]
        features <- features[, colnames(features) != abs_feature]
    }
    
    if (any(is.na(abs_mult_fact)) |
        any(is.infinite(abs_mult_fact))) {
        stop(
            '1+ of the unscaled abundance multipliers are NA or infinite.
        Check that the spike-in is present in all samples and the
        `unscaled_abundance` table is entirely non-NA.'
        )
    }
    
    features_norm <- as.matrix(features)
    dd <- colnames(features_norm)
    x <- as.matrix(features_norm)
    
    MARGIN <- 1
    
    x <- sweep(x, MARGIN, abs_mult_fact, "*")
    if (any(is.nan(x)))
        warning("result contains NaN, perhaps due to impossible mathematical\n
            operation\n")
    
    if (any(is.infinite(x)))
        warning("result contains Inf, perhaps due to impossible mathematical\n
            operation\n")
    
    x <- ifelse(X_mask, x, NA)
    
    # Convert back to data frame
    features_ABS <- as.data.frame(x)
    
    # Rename the features - Same Format as Before
    colnames(features_ABS) <- dd
    
    # Return
    return(features_ABS)
}

######################
# Log Transformation #
######################

LOG <- function(x) {
    if (any(x[!is.na(x)] <= 0)) {
        stop("Log transformation is only valid for values above 0")
    }
    return(log2(x))
}

#############################
# Pseudo-log Transformation #
#############################

PLOG <- function(x) {
    if (any(x[!is.na(x)] < 0)) {
        stop("Pseudo-log transformation is only valid for values >= 0")
    }
    min_observed <- min(x[!is.na(x) & x > 0])
    x[!is.na(x) & x == 0] <- min_observed / 2
    return(log2(x))
}

#################
# Write outputs #
#################

write_fits <- function(output,
                    fit_data_abundance,
                    fit_data_prevalence,
                    random_effects_formula = NULL,
                    save_models = FALSE) {
    fits_folder <- file.path(output, "fits")
    if (!file.exists(fits_folder)) {
        logging::loginfo("Creating fits folder")
        dir.create(fits_folder)
    }

    model_types <- c('linear', 'logistic')
    
    lapply(model_types, function(model_type) {
        fit_data <- if (model_type == 'linear') {
            fit_data_abundance
        } else {
            fit_data_prevalence
        }
        
        if (is.null(fit_data)) return(NULL)
        
        # Write out the raw model fits
        if (save_models) {
            model_file <- file.path(fits_folder, 
                                    paste0("models_", model_type, ".rds"))
            if (file.exists(model_file)) {
                logging::logwarn("Deleting existing model objects file: %s", 
                                    model_file)
                unlink(model_file)
            }
            logging::loginfo("Writing model objects to file %s", model_file)
            saveRDS(fit_data$fits, file = model_file)
        }
        
        # Write residuals to file
        residuals_file <- file.path(fits_folder, 
                                    paste0("residuals_", model_type, ".rds"))
        if (file.exists(residuals_file)) {
            logging::logwarn("Deleting existing residuals file: %s", 
                            residuals_file)
            unlink(residuals_file)
        }
        logging::loginfo("Writing residuals to file %s", residuals_file)
        saveRDS(fit_data$residuals, file = residuals_file)
        
        # Write fitted values to file
        fitted_file <- file.path(fits_folder, 
                                paste0("fitted_", model_type, ".rds"))
        if (file.exists(fitted_file)) {
            logging::logwarn("Deleting existing fitted file: %s", fitted_file)
            unlink(fitted_file)
        }
        logging::loginfo("Writing fitted values to file %s", fitted_file)
        saveRDS(fit_data$fitted, file = fitted_file)
        
        # Write extracted random effects to file (if specified)
        if (!is.null(random_effects_formula)) {
            ranef_file <- file.path(fits_folder, 
                                    paste0("ranef_", model_type, ".rds"))
            if (file.exists(ranef_file)) {
                logging::logwarn("Deleting existing ranef file: %s", ranef_file)
                unlink(ranef_file)
            }
            logging::loginfo("Writing extracted random effects to file %s", 
                            ranef_file)
            saveRDS(fit_data$ranef, file = ranef_file)
        }
    })
}

write_results <- function(output,
                        fit_data_abundance,
                        fit_data_prevalence,
                        max_significance = 0.1) {
    if (is.null(fit_data_abundance)) {
        fit_data <- fit_data_prevalence$results
    } else if (is.null(fit_data_prevalence)) {
        fit_data <- fit_data_abundance$results
    } else {
        fit_data <- rbind(fit_data_abundance$results,
                        fit_data_prevalence$results)
    }
    
    fit_data$model <- data.table::fcase(
            fit_data$model == 'linear', 'abundance',
            fit_data$model == 'logistic',  'prevalence'
    )
    
    small_random_effects_warning <- 
        paste0("<4 average observations per random effect group often inflates",
            " coefficients and deflates p-values: consider setting ",
            "small_random_effects=TRUE and see tutorial")
    
    fit_data <- fit_data[order(fit_data$qval_individual), ]
    # Move all that had errors to the end
    fit_data <- fit_data[order(!is.na(fit_data$error) & 
        fit_data$error != small_random_effects_warning), ] 
    
    #############################
    # Write all results to file #
    #############################
    
    results_file <- file.path(output, "all_results.tsv")
    
    logging::loginfo(
        paste(
            "Writing all the results to file (ordered 
            by increasing individual q-values): %s"
        ),
        results_file
    )
    
    write.table(
        fit_data,
        file = results_file,
        sep = "\t",
        quote = TRUE,
        row.names = FALSE
    )
    
    ###########################################
    # Write results passing threshold to file #
    ###########################################
    
    significant_results <-
        fit_data[(fit_data$qval_joint <= max_significance &
                    !is.na(fit_data$qval_joint)) | 
                    (fit_data$qval_individual <= max_significance &
                    !is.na(fit_data$qval_individual)) &
                    (is.na(fit_data$error) | 
                    fit_data$error == small_random_effects_warning),]
    significant_results_file <-
        file.path(output, "significant_results.tsv")
    
    logging::loginfo(
        paste(
            "Writing the significant results without errors",
            "(those which have joint q-values less than",
            "or equal to the threshold",
            "of %f ) to file (ordered by increasing individual q-values): %s"
        ),
        max_significance,
        significant_results_file
    )
    
    write.table(
        significant_results,
        file = significant_results_file,
        sep = "\t",
        quote = TRUE,
        row.names = FALSE
    )
}

write_results_in_lefse_format <-
    function(results, output_file_name) {
        lines_vec <- vapply(seq(nrow(results)), function(i) {
            if (is.na(results[i, ]$error) & 
                !is.na(results[i, ]$qval_individual)) {
                if (results[i, ]$qval_individual < 0.1) {
                    return(paste0(
                        c(
                            results[i, ]$feature,
                            results[i, ]$coef,
                            paste0(results[i, ]$metadata, '_', 
                                    results[i, ]$value),
                            results[i, ]$coef,
                            results[i, ]$pval_individual
                        ),
                        collapse = '\t'
                    ))
                } else {
                    return(paste0(c(results[i, ]$feature, 
                            results[i, ]$coef, "", "", "-"), collapse = '\t'))
                }
            } else {
                # return NA for rows that don't meet the condition
                return(NA_character_)
            }
        }, FUN.VALUE = character(1))
        lines_vec <- lines_vec[lines_vec != 'FALSE']
        
        writeLines(sort(lines_vec), con = output_file_name)
    }

####################
# Contrast testing #
####################

# add_qvals
# create_combined_pval
# flag_abundance_turned_prevalence
# add_joint_signif
# append_joint
# Check for highly significant likely model misfits
# Warn about prevalence associations induced by abundances changes
# Small random effects warning

maaslin_contrast_test <- function(
                        maaslin3_fit,
                        contrast_mat,
                        rhs = NULL,
                        max_significance = 0.1,
                        correction = 'BH',
                        median_comparison_abundance = TRUE,
                        median_comparison_prevalence = FALSE,
                        subtract_median = FALSE,
                        small_random_effects = FALSE,
                        bypass_small_group_warning = FALSE,
                        evaluate_only = NULL) {
    
    match.arg(correction, correction_choices)
    
    # Match variable ignoring case then set correctly as required for p.adjust
    correction <- correction_choices[match(toupper(correction),
                                        toupper(correction_choices))]
    
    # Run linear model component
    if (is.null(evaluate_only) || evaluate_only == "abundance") {
        if (is.null(maaslin3_fit$fit_data_abundance$fits)) {
            error_message <- paste0(
                "maaslin3_fit$fit_data_abundance$fits cannot be null,",
                " check if evaluate_only should be 'prevalence'")
            stop(error_message)
        }
        
        #######################
        # For non-zero models #
        #######################
        
        fit_data_abundance <- list()
        fit_data_abundance$results <- maaslin_contrast_test_runner(
            maaslin3_fit$fit_data_abundance$fits,
            contrast_mat,
            rhs = rhs,
            median_comparison = median_comparison_abundance)

        #################################################################
        # Count the total values for each feature (untransformed space) #
        #################################################################
        
        fit_data_abundance$results$N <- as.numeric(fmapvalues(
            fit_data_abundance$results$feature,
            maaslin3_fit$fit_data_abundance$results$feature,
            maaslin3_fit$fit_data_abundance$results$N))
        
        fit_data_abundance$results$N_not_zero <- as.numeric(fmapvalues(
            fit_data_abundance$results$feature,
            maaslin3_fit$fit_data_abundance$results$feature,
            maaslin3_fit$fit_data_abundance$results$N_not_zero))
    }
    
    # Run logistic model component
    if (is.null(evaluate_only) || evaluate_only == "prevalence") {
        if (is.null(maaslin3_fit$fit_data_prevalence$fits)) {
            error_message <- paste0(
                "maaslin3_fit$fit_data_prevalence$fits cannot be null,",
                " check if evaluate_only should be 'abundance'")
            stop(error_message)
        }
        
        #####################
        # For binary models #
        #####################
        
        fit_data_prevalence <- list()
        fit_data_prevalence$results <- maaslin_contrast_test_runner(
            maaslin3_fit$fit_data_prevalence$fits,
            contrast_mat,
            rhs = rhs,
            median_comparison = median_comparison_prevalence)
        
        fit_data_prevalence$results$N <- as.numeric(fmapvalues(
            fit_data_prevalence$results$feature,
            maaslin3_fit$fit_data_prevalence$results$feature,
            maaslin3_fit$fit_data_prevalence$results$N))
        
        fit_data_prevalence$results$N_not_zero <- as.numeric(fmapvalues(
            fit_data_prevalence$results$feature,
            maaslin3_fit$fit_data_prevalence$results$feature,
            maaslin3_fit$fit_data_prevalence$results$N_not_zero))
    }
    
    # Check for highly significant likely model misfits
    if (is.null(evaluate_only) || evaluate_only == "prevalence") {
        current_likely_error_subsetter <-
            !is.na(fit_data_prevalence$results$N_not_zero) &
            (
                fit_data_prevalence$results$N_not_zero < 50 &
                    fit_data_prevalence$results$N_not_zero /
                    fit_data_prevalence$results$N < 0.05
            ) &
            ((
                !is.na(fit_data_prevalence$results$coef) &
                    abs(fit_data_prevalence$results$coef) > 15
            ) |
                (
                    !is.na(fit_data_prevalence$results$pval) &
                        fit_data_prevalence$results$pval < 10 ^ -10
                )
            )
        current_errors_for_likely_issues <-
            fit_data_prevalence$results$error[current_likely_error_subsetter]
        fit_data_prevalence$results$error[current_likely_error_subsetter] <-
            data.table::fifelse(
                !is.na(current_errors_for_likely_issues),
                current_errors_for_likely_issues,
                "A large coefficient (>15 in absolute value) or small p-value (< 10^-10) was obtained from a feature present in <5% of samples. Check this is intended."
            )
        
        current_likely_error_subsetter <-
            !is.na(fit_data_prevalence$results$N_not_zero) &
            (
                fit_data_prevalence$results$N -
                    fit_data_prevalence$results$N_not_zero < 50 &
                    fit_data_prevalence$results$N_not_zero /
                    fit_data_prevalence$results$N > 0.95
            ) &
            ((
                !is.na(fit_data_prevalence$results$coef) &
                    abs(fit_data_prevalence$results$coef) > 15
            ) |
                (
                    !is.na(fit_data_prevalence$results$pval) &
                        fit_data_prevalence$results$pval < 10 ^ -10
                )
            )
        current_errors_for_likely_issues <-
            fit_data_prevalence$results$error[current_likely_error_subsetter]
        fit_data_prevalence$results$error[current_likely_error_subsetter] <-
            data.table::fifelse(
                !is.na(current_errors_for_likely_issues),
                current_errors_for_likely_issues,
                "A large coefficient (>15 in absolute value) or small p-value (< 10^-10) was obtained from a feature present in >95% of samples. Check this is intended."
            )
    } else {
        fit_data_prevalence <- NULL
    }
    
    results <- add_qvals(fit_data_abundance,fit_data_prevalence, correction)
    if (!is.null(fit_data_abundance)) {
        fit_data_abundance$results <- results[[1]]
    }
    if (!is.null(fit_data_prevalence)) {
        fit_data_prevalence$results <- results[[2]]
    }
    if (!is.null(fit_data_abundance)) {
        fit_data_abundance$results <- fit_data_abundance$results |>
            collapse::fmutate(null_hypothesis = rhs)
        
        if (subtract_median & median_comparison_abundance) {
            fit_data_abundance$results$coef <- 
                fit_data_abundance$results$coef - 
                fit_data_abundance$results$null_hypothesis
            fit_data_abundance$results$null_hypothesis <- 0
        }
    }
    if (!is.null(fit_data_prevalence)) {
        fit_data_prevalence$results <- fit_data_prevalence$results |>
            collapse::fmutate(null_hypothesis = rhs)
        
        if (subtract_median & median_comparison_prevalence) {
            fit_data_prevalence$results$coef <- 
                fit_data_prevalence$results$coef - 
                fit_data_prevalence$results$null_hypothesis
            fit_data_prevalence$results$null_hypothesis <- 0
        }
    }
    
    # Add in joint p/q-values
    if (is.null(evaluate_only)) {
        if (!is.null(fit_data_abundance)) {
            
            dtest = fit_data_abundance$results$test
            
            fit_data_abundance$results <- fit_data_abundance$results |> 
                collapse::fmutate(metadata = dtest,
                                  value    = dtest,
                                  name     = dtest)
            
        }
        if (!is.null(fit_data_prevalence)) {
            
            dtest = fit_data_prevalence$results$test
            
            fit_data_prevalence$results <- fit_data_prevalence$results |> 
                collapse::fmutate(metadata = dtest,
                                  value    = dtest,
                                  name     = dtest)
        }
        results <-
            add_joint_signif(fit_data_abundance,
                            fit_data_prevalence,
                            NULL,
                            max_significance,
                            correction)
        if (!is.null(fit_data_abundance)) {
            fit_data_abundance$results <- results[[1]]
            
            collapse::fselect(fit_data_abundance$results,
                              c("metadata", "value", "name")) <- NULL
        }
        if (!is.null(fit_data_prevalence)) {
            fit_data_prevalence$results <- results[[2]]
            collapse::fselect(fit_data_prevalence$results,
                              c("metadata", "value", "name")) <- NULL
        }
    } else if (evaluate_only == 'abundance') {
        if (!is.null(fit_data_abundance)) {
            fit_data_abundance$results$pval_joint <-
                fit_data_abundance$results$pval
            
            fit_data_abundance$results$qval_joint <-
                fit_data_abundance$results$qval
            
            collapse::setrename(fit_data_abundance$results,
                                 "pval_individual" = "pval",
                                 "qval_individual" = "qval")
        }
    } else if (evaluate_only == 'prevalence') {
        fit_data_prevalence$results$pval_joint <-
            fit_data_prevalence$results$pval
        
        fit_data_prevalence$results$qval_joint <-
            fit_data_prevalence$results$qval
        
        collapse::setrename(fit_data_prevalence$results,
                             "pval_individual" = "pval",
                             "qval_individual" = "qval")
    }
    
    if (!is.null(fit_data_prevalence)) {
        used_random_effects <- 
            any(unlist(lapply(maaslin3_fit$fit_data_prevalence$fits, 
                FUN = function(x){is(x,"glmerMod")})))
        
        if (used_random_effects & !bypass_small_group_warning) {
            if (any(grepl('<4 average observations per random effect', 
                maaslin3_fit$fit_data_prevalence$results$error))) {
                fit_data_prevalence$results$error <- 
                    data.table::fifelse(is.na(fit_data_prevalence$results$error),
                            paste0("<4 average observations per random ",
                                    "effect group often inflates ",
                                    "coefficients and deflates p-values: ",
                                    "consider setting ",
                                    "small_random_effects=TRUE and see ",
                                    "tutorial"), 
                            fit_data_prevalence$results$error)
            }
        }
    }
    
    # Set column order
    col_order <- c("feature", "test", "coef", 
                    "null_hypothesis", "stderr",
                    "pval_individual", "qval_individual", "pval_joint",
                    "qval_joint", "error", "model", "N", "N_not_zero")
    
    # Reorder outputs
    if (is.null(evaluate_only) || evaluate_only == "abundance") {
        fit_data_abundance$results <-
            fit_data_abundance$results[
                order(fit_data_abundance$results$qval_individual), ]
        # Move all that had errors to the end
        fit_data_abundance$results <-
            fit_data_abundance$results[
                order(!is.na(fit_data_abundance$results$error)), ]
        
        # Reorder columns
        fit_data_abundance$results <- fit_data_abundance$results[,col_order]
    } else {
        fit_data_abundance <- NULL
    }
    
    if (is.null(evaluate_only) || evaluate_only == "prevalence") {
        fit_data_prevalence$results <-
            fit_data_prevalence$results[
                order(fit_data_prevalence$results$qval_individual), ]
        # Move all that had errors to the end
        fit_data_prevalence$results <-
            fit_data_prevalence$results[
                order(!is.na(fit_data_prevalence$results$error)), ]
        
        # Reorder columns
        fit_data_prevalence$results <- fit_data_prevalence$results[,col_order]
    } else {
        fit_data_prevalence <- NULL
    }
    
    return(
        list(
            "fit_data_abundance" = fit_data_abundance,
            "fit_data_prevalence" = fit_data_prevalence
        )
    )
}

maaslin_contrast_test_runner <- function(fits,
                                contrast_mat,
                                rhs = NULL,
                                median_comparison = NULL) {
    
    if (is.null(fits)) {
        stop("fits must not be null")
    }
    
    if (is.null(colnames(contrast_mat))) {
        stop("contrast_mat columns must be named to match model variables")
    }
    
    # Identify model type
    if (any(unlist(lapply(fits, FUN = 
                                function(x){is(x,"lm") & !is(x,"glm")})))) {
        model <- "abundance"
        uses_random_effects <- FALSE
    } else if (any(unlist(lapply(fits, FUN = function(x){is(x,"glm")})))) {
        model <- "prevalence"
        uses_random_effects <- FALSE
    } else if (any(unlist(lapply(fits, FUN = function(x){is(x,"lmerMod")})))) {
        model <- "abundance"
        uses_random_effects <- TRUE
    } else if (any(unlist(lapply(fits, FUN = function(x){is(x,"glmerMod")})))) {
        model <- "prevalence"
        uses_random_effects <- TRUE
    } else if (any(unlist(lapply(fits, FUN = function(x){is(x,"coxph")})))) {
        model <- "prevalence"
        uses_random_effects <- FALSE
    }

    if (is.null(median_comparison)) {
        if (model == 'abundance') {
            median_comparison <- TRUE
        } else {
            median_comparison <- FALSE
        }
    }
    
    if (median_comparison & !is.null(rhs)) {
        stop("rhs must be NULL with median_comparison")
    }
    
    if (is.null(rhs)) {
        rhs <- rep(0, nrow(contrast_mat))
    }
    
    if (nrow(contrast_mat) != length(rhs)) {
        stop("Number of rows in contrast matrix must equal the length of rhs")
    }
    
    # Run contrast test on each  model
    test_out_joined <- do.call(rbind, 
        lapply(seq_along(fits), function(fit_index) {
                fit <- fits[[fit_index]]
        if (!is(fit, 'lmerMod') & !is(fit, 'glmerMod') & !is(fit, "coxph")) {
            if (all(is.na(fit))) {
                return(matrix(rep(c(NA, NA, NA, "Fit is NA"), 
                            nrow(contrast_mat)),
                            nrow = nrow(contrast_mat), byrow = TRUE))
            }
        }
        
        # Fill in contrast matrix gaps if necessary
        if (uses_random_effects) {
            included_coefs <- names(lme4::fixef(fit))
        } else {
            included_coefs <- names(coef(fit, complete = FALSE))
        }
        contrast_mat_cols <- colnames(contrast_mat)
        contrast_mat_tmp <- contrast_mat
        contrast_mat_tmp <- cbind(contrast_mat_tmp, 
                                matrix(0, 
                                    nrow = nrow(contrast_mat_tmp),
                                    ncol = length(
                                    setdiff(included_coefs,
                                    contrast_mat_cols))))
        colnames(contrast_mat_tmp) <- c(contrast_mat_cols,
                                        setdiff(included_coefs,
                                                contrast_mat_cols))
        if (any(contrast_mat_tmp[,setdiff(contrast_mat_cols, 
                                        included_coefs)] != 0)) {
            return(matrix(rep(c(NA, NA, NA, 
                "Predictors not in the model had non-zero contrast values"), 
                    nrow(contrast_mat)),
                    nrow = nrow(contrast_mat), byrow = TRUE))
        }
        
        contrast_mat_tmp <- contrast_mat_tmp[,included_coefs, drop = FALSE]
        
        # Run contrast test
        if (!uses_random_effects | model == 'prevalence') {
            test_out_joined <- vapply(X = seq(nrow(contrast_mat)), 
                FUN = function(row_num) {
                    contrast_vec <- t(matrix(contrast_mat_tmp[row_num,]))
                    
                    error_message <- NA
                    calling_env <- environment()
                    test_out <- tryCatch({
                        if (!uses_random_effects) {
                            summary_out <- summary(multcomp::glht(
                                fit,
                                linfct = contrast_vec,
                                rhs = rhs[row_num],
                                coef. = function(x) {
                                    coef(x, complete = FALSE)
                                }
                            )
                            )$test
                        } else {
                            summary_out <- summary(multcomp::glht(
                                fit,
                                linfct = contrast_vec,
                                rhs = rhs[row_num],
                            )
                            )$test
                        }
                        
                        c(summary_out$coefficients,
                            summary_out$sigma,
                            summary_out$pvalues)
                    }, warning = function(w) {
                        message(sprintf("Feature %s : %s", 
                                        names(fit), w))
                        
                        assign("error_message",
                                conditionMessage(w),
                                envir = calling_env)
                        invokeRestart("muffleWarning")
                    },
                    error = function(err) {
                        assign("error_message", err$message, 
                            envir = calling_env)
                        error_obj <- c(NA, NA, NA)
                        return(error_obj)
                    })
                    return(as.character(c(test_out, error_message)))
                }, character(4))
        } else {
            test_out_joined <- vapply(X = seq(nrow(contrast_mat)), 
                FUN = function(row_num) {
                    contrast_vec <- t(matrix(contrast_mat_tmp[row_num,]))
                    
                    error_message <- NA
                    calling_env <- environment()
                    test_out <- tryCatch({
                        pval <- lmerTest::contest(fit,
                            matrix(contrast_vec, TRUE), 
                            rhs = rhs[row_num])[['Pr(>F)']]
                        coef <- contrast_vec %*% lme4::fixef(fit)
                        sigma <- sqrt((contrast_vec %*% vcov(fit) %*% 
                                        t(contrast_vec))[1, 1])
                        c(coef, sigma, pval)
                    }, warning = function(w) {
                        message(sprintf("Feature %s : %s", names(fit), w))
                        
                        assign("error_message",
                                conditionMessage(w),
                                envir = calling_env)
                        invokeRestart("muffleWarning")
                    },
                    error = function(err) {
                        assign("error_message", err$message,
                            envir = calling_env)
                        error_obj <- c(NA, NA, NA)
                        return(error_obj)
                    })
                    return(as.character(c(test_out, error_message)))
                }, character(4))
        }
        return(t(test_out_joined))
    }))
    
    test_out_joined <- data.frame(test_out_joined, check.names = FALSE)
    colnames(test_out_joined) <- c("coefs_new", "sigmas_new", 
        "pvals_new", "errors")
    test_out_joined[, seq(3)] <- apply(test_out_joined[, seq(3)], 2, as.numeric)

    if (all(is.null(rownames(contrast_mat)))) {
        test_names <- seq(nrow(contrast_mat))
    } else {
        test_names <- rownames(contrast_mat)
    }
    
    # Prepare results for return
    paras <- data.frame(
        feature = rep(names(fits), each = nrow(contrast_mat)),
        test = rep(test_names, length(fits)),
        coef = test_out_joined$coefs_new,
        rhs = rep(rhs, length(fits)),
        stderr = test_out_joined$sigmas_new,
        pval = test_out_joined$pvals_new,
        error = test_out_joined$errors,
        model = model
    )
    
    # Perform median comparison as though these were the original fits
    if (median_comparison) {
        paras <- do.call(rbind, lapply(unique(paras$test), 
                                        function(selected_test) {
            paras_sub <- paras[paras$test == selected_test, , drop = FALSE]
            paras_sub_out <- paras_sub[is.na(paras_sub$coef) | 
                                        is.na(paras_sub$stderr), , drop = FALSE]
            paras_sub <- paras_sub[!is.na(paras_sub$coef) & 
                                    !is.na(paras_sub$stderr), , drop = FALSE]
            if (nrow(paras_sub) == 0) {
                return(paras_sub_out)
            }
            
            # Make sure other p-values are set to NA to not confuse
            # median comparison with non-comparison
            paras_sub_out$pval <- rep(NA, 
                length(paras_sub_out$pval))
            
            use_this_coef <- !is.na(paras_sub$pval) & 
                paras_sub$pval < 0.95
            
            n_coefs <- nrow(paras_sub)
            sigmas <- paras_sub$stderr
            coefs <- paras_sub$coef
            sigma_sq_med <- var(coefs[use_this_coef], na.rm=TRUE)
            
            cur_median <- median(coefs[use_this_coef])
            
            # Variance from asymptotic distribution
            sd_median <- sqrt(0.25 * 2 * base::pi * 
                            sigma_sq_med / sum(use_this_coef))
            
            # MC for covariance
            nsims <- 10000
            
            sim_results <- replicate(nsims, {
                sim_coefs <- rnorm(n_coefs, coefs, sigmas)
                sim_median <- median(sim_coefs[use_this_coef])
                c(sim_median, sim_coefs)
            })
            
            sim_medians <- sim_results[1, ]
            all_sims <- sim_results[-1, , drop = FALSE]
            cov_adjust <- apply(all_sims, 1, function(x){cov(x, sim_medians)})
            
            # Necessary offsets for contrast testing
            offsets_to_test <- abs(cur_median - coefs) * 
                sqrt((sigmas^2) / (sigmas^2+ sd_median^2 - 2 * cov_adjust)) + 
                coefs
            
            pvals_new <- vapply(seq_len(nrow(paras_sub)), function(row_index) {
                feature <- paras_sub$feature[row_index]
                fit <- fits[[feature]]
                
                # Fill in contrast matrix gaps if necessary
                if (uses_random_effects) {
                    included_coefs <- names(lme4::fixef(fit))
                } else {
                    included_coefs <- names(coef(fit, complete = FALSE))
                }
                contrast_mat_cols <- colnames(contrast_mat)
                contrast_mat_tmp <- contrast_mat
                contrast_mat_tmp <- cbind(contrast_mat_tmp, 
                                        matrix(0, 
                                        nrow = nrow(contrast_mat_tmp),
                                        ncol = length(setdiff(
                                        included_coefs, contrast_mat_cols))))
                colnames(contrast_mat_tmp) <- 
                    c(contrast_mat_cols, 
                        setdiff(included_coefs, contrast_mat_cols))
                
                contrast_mat_tmp <- 
                    contrast_mat_tmp[, included_coefs, drop = FALSE]
                
                # Run contrast test
                contrast_vec <- t(matrix(contrast_mat_tmp[selected_test,]))
                
                error_message <- NA
                calling_env <- environment()
                test_out <- tryCatch({
                    if (!uses_random_effects | model == 'prevalence') {
                        summary_out <- summary(multcomp::glht(
                            fit,
                            linfct = contrast_vec,
                            rhs = offsets_to_test[row_index],
                            coef. = function(x) { coef(x, complete = FALSE) }
                        ))$test
                    } else {
                        summary_out <- summary(multcomp::glht(
                            fit,
                            linfct = contrast_vec,
                            rhs = offsets_to_test[row_index]
                        ))$test
                    }
                    
                    c(summary_out$pvalues, summary_out$coefficients, 
                        summary_out$sigma)
                }, warning = function(w) {
                    message(sprintf("Feature %s : %s", names(fit), w))
                    
                    assign("error_message", conditionMessage(w), 
                        envir = calling_env)
                    invokeRestart("muffleWarning")
                }, error = function(err) {
                    assign("error_message", err$message, envir = calling_env)
                    error_obj <- c(NA, NA, NA)
                    return(error_obj)
                })
                
                return(test_out[1])
            }, numeric(1))
            
            paras_sub$error <- 
                data.table::fifelse(!is.na(paras_sub$pval) & 
                        is.na(pvals_new),
                    "P-value became NA during median comparison",
                    paras_sub$error
                )
            
            paras_sub$pval <- pvals_new
            paras_sub$rhs <- cur_median
            
            return(rbind(paras_sub, paras_sub_out))
        }))
    }
    
    if (!is.null(rownames(paras)) && 
        !any(is.na(as.numeric(rownames(paras))))) {
        paras <- paras[order(as.numeric(rownames(paras))),]
    }
    
    return(paras)
}

##############################
# DNA pre-processing for MTX #
##############################

# rna_table must be samples (rows) by features (cols)
preprocess_dna_mtx <- function(dna_table, rna_table) {
    samples_row_row <-
        intersect(rownames(dna_table), rownames(rna_table))
    if (length(samples_row_row) == 0) {
        samples_column_row <-
            intersect(colnames(dna_table), rownames(rna_table))
        
        if (length(samples_column_row) > 0) {
            dna_table <- as.data.frame(t(dna_table))
        } else {
            stop(paste0(
                paste0("Rows/columns do not match."),
                paste0("DNA rows: ",
                    paste(
                        rownames(dna_table), collapse = ","
                    )),
                paste0("DNA columns: ",
                    paste(
                        colnames(dna_table), collapse = ","
                    )),
                paste0("RNA rows: ",
                    paste(
                        rownames(rna_table), collapse = ","
                    )),
                paste0("RNA columns: ",
                    paste(
                        colnames(rna_table), collapse = ","
                    )),
                collapse = '\n'
            ))
        }
    }
    intersect_samples <-
        intersect(rownames(dna_table), rownames(rna_table))
    
    dna_table <- dna_table[intersect_samples, , drop = FALSE]
    rna_table <- rna_table[intersect_samples, , drop = FALSE]
    
    # At this point, DNA and RNA tables are 
    # samples x features with same samples
    
    dna_table <- TSSnorm(dna_table, 0)
    dna_table <- apply(dna_table, 2, function(x) {
        x[is.na(x)] <- 0
        return(x)
    })
    dna_table <- as.data.frame(dna_table)

    rna_table <- TSSnorm(rna_table, 0)
    rna_table <- apply(rna_table, 2, function(x) {
        x[is.na(x)] <- 0
        return(x)
    })
    rna_table <- as.data.frame(rna_table)
    
    # Transforming DNA table
    impute_val <- log2(min(dna_table[dna_table > 0]) / 2)

    intersect_features <-
        intersect(colnames(dna_table), colnames(rna_table))
    
    dna_table <- dna_table[, intersect_features, drop = FALSE]
    rna_table <- rna_table[, intersect_features, drop = FALSE]
    
    if (!all(colnames(dna_table) == colnames(rna_table)) |
        !all(rownames(dna_table) == rownames(rna_table))) {
        stop("Something went wrong in preprocessing")
    }
    
    # Transform DNA 0s as necessary
    dna_table <- log2(dna_table)
    dna_table[dna_table == -Inf & rna_table > 0] <- impute_val
    dna_table[dna_table == -Inf & rna_table == 0] <- NA
    
    return(list("dna_table" = dna_table,
                "rna_table" = rna_table))
}

###############################
# Taxa pre-processing for MTX #
###############################

# rna_table must be samples (rows) by features (cols)
preprocess_taxa_mtx <- function(taxa_table, rna_table, rna_per_taxon) {
    if (any(colnames(rna_per_taxon) != c("RNA", 'taxon'))) {
        stop("colnames of rna_per_taxon should be RNA and taxon")
    }
    
    rna_vec <- c(unlist(rna_per_taxon[,'RNA']))
    taxon_vec <- c(unlist(rna_per_taxon[,'taxon']))

    samples_row_row <-
        intersect(rownames(taxa_table), rownames(rna_table))
    if (length(samples_row_row) == 0) {
        samples_column_row <-
            intersect(colnames(taxa_table), rownames(rna_table))
        
        if (length(samples_column_row) > 0) {
            taxa_table <- as.data.frame(t(taxa_table))
        } else {
            stop(paste0(
                paste0("Rows/columns do not match."),
                paste0("Taxa rows: ",
                    paste(rownames(taxa_table), collapse = ",")),
                paste0("Taxa columns: ",
                    paste(colnames(taxa_table), collapse = ",")),
                paste0("RNA rows: ",
                    paste(rownames(rna_table), collapse = ",")),
                paste0("RNA columns: ",
                    paste(colnames(rna_table), collapse = ",")),
                collapse = '\n'
            ))
        }
    }
    intersect_samples <-
        intersect(rownames(taxa_table), rownames(rna_table))
    
    taxa_table <- taxa_table[intersect_samples, , drop = FALSE]
    rna_table <- rna_table[intersect_samples, , drop = FALSE]
    
    if (any(!colnames(rna_table) %in% rna_vec)) {
        stop("rna_per_taxon RNA must contain all columns of rna_table")
    }
    if (any(!colnames(taxa_table) %in% taxon_vec)) {
        stop("rna_per_taxon taxon must contain all columns of taxa_table")
    }
    
    # At this point, taxa and RNA tables are 
    # samples x features with same samples
    
    taxa_table <- TSSnorm(taxa_table, 0)
    taxa_table <- apply(taxa_table, 2, function(x) {
        x[is.na(x)] <- 0
        return(x)
    })
    taxa_table <- as.data.frame(taxa_table)
    
    rna_table <- TSSnorm(rna_table, 0)
    rna_table <- apply(rna_table, 2, function(x) {
        x[is.na(x)] <- 0
        return(x)
    })
    rna_table <- as.data.frame(rna_table)
    
    # Create a dna table by choosing the taxa to match the rna table
    dna_table <- taxa_table[, 
        fmapvalues(colnames(rna_table), rna_vec, taxon_vec)]
    colnames(dna_table) <- colnames(rna_table)

    # Transforming DNA table
    impute_val <- log2(min(dna_table[dna_table > 0]) / 2)
    
    if (!all(colnames(dna_table) == colnames(rna_table)) |
        !all(rownames(dna_table) == rownames(rna_table))) {
        stop("Something went wrong in preprocessing")
    }
    
    # Transform DNA 0s as necessary
    dna_table <- log2(dna_table)
    
    max_rna_table <- vapply(split(rna_per_taxon$RNA, rna_per_taxon$taxon), 
        function(rna_cols) {
            rna_cols <- rna_cols[rna_cols %in% colnames(rna_table)]
            apply(rna_table[, rna_cols, drop = FALSE], 1, max, na.rm = TRUE)
        }, FUN.VALUE = numeric(nrow(rna_table)))
    max_rna_table <- max_rna_table[, 
        fmapvalues(colnames(rna_table), rna_vec, taxon_vec)]
    colnames(max_rna_table) <- colnames(rna_table)

    dna_table[dna_table == -Inf & max_rna_table > 0] <- impute_val
    if (any(dna_table == -Inf & rna_table == 0)) {
        dna_table[dna_table == -Inf & rna_table == 0] <- NA
    }
    
    return(list("dna_table" = dna_table,
                "rna_table" = rna_table))
}

##################################
# Read from SummarizedExperiment #
##################################

maaslin_read_summarized_experiment_data <- 
    function(summarized_experiment, assay.type = 1) {
        
        rlang::check_installed("SummarizedExperiment")
        
        if (!inherits(summarized_experiment, "SummarizedExperiment")) {
            stop("Input must be a SummarizedExperiment object")
        }
        
        data <- as.data.frame(t(SummarizedExperiment::assay(
            summarized_experiment, assay.type)))
        metadata <- as.data.frame(
            SummarizedExperiment::colData(summarized_experiment))
        return(list(
            "data" = data,
            "metadata" = metadata
        ))
    }

nrw_ts = function(ts) {
    as.POSIXct(ts) |>  
        format(format = "%Y-%m-%d %H:%M:%OS2")
}

nrw_fmt = function(record) {
    # 5-6 digits of precision on the timestamp is verbose and makes it hard to
    # read. Limit it to two digits.
    
    msg <- trimws(record$msg)
    text <- paste(nrw_ts(record$timestamp), paste(record$levelname, record$logger, 
                                          msg, sep = ":"))
    return(text)
}

fmapvalues <- function(x, from, to) {
    # To avoid dependency on plyr for mapvalues()
   
    stopifnot(length(from) == length(to)) 
    
    from_matches = collapse::fmatch(x, from)
    
    mch_id = collapse::whichNA(from_matches, invert = TRUE)
    
    to_set = to[from_matches[mch_id]]
    
    return(collapse::copyv(x, mch_id, to_set)) 
}

has_auto_rownames <- function(d) {
    # tibble::has_rownames() also checks for !is.na(.row_names_info(.data,
    # 0L)[[1L]]) i.e. that the first rowname is not NA.
    
    .row_names_info(d, type = 1L) < 0L
}

ftail = function(d, i) {
    if (i < 0) i = nrow(d) + i
    
    collapse::fslice(d, how = "last", n = i)
}

library(testthat)
library(maaslin3)

data_in <- data.frame('a' = c(1, 2, 3, 4, 5), 
                      'b' = c(2, 3, 4, 5, 6),
                      'c' = c(3, 4, 5, 6, 7))
rownames(data_in) <- paste0("sample", c(1:5))

covar_data <- data.frame('a' = c(5, 4, 3, 2, 1), 
                      'b' = c(6, 5, 4, 3, 2),
                      'c' = c(7, 6, 5, 4, 3))
rownames(covar_data) <- paste0("sample", c(1:5))

unscaled <- data.frame('a' = c(5, 4, 3, 2, 1))
rownames(unscaled) <- paste0("sample", c(1:5))

metadata <- data.frame('a' = c(1, 2, 3, 4, 5),
                       'b' = c(0, 1, 0, 1, 0),
                       'c' = c('a', 'b', 'c', 'a', 'b'))
rownames(metadata) <- paste0("sample", c(1:5))

output_tmp <- tempfile()
dir.create(output_tmp)
write.table(data_in, file.path(output_tmp, 'data.tsv'), 
            row.names = T, sep = '\t', col.names=NA)
write.table(metadata, file.path(output_tmp, 'metadata.tsv'), 
            row.names = T, sep = '\t', col.names=NA)
write.table(covar_data, file.path(output_tmp, 'covar.tsv'), 
            row.names = T, sep = '\t', col.names=NA)
write.table(unscaled, file.path(output_tmp, 'unscaled.tsv'), 
            row.names = T, sep = '\t', col.names=NA)

data_in_new <- maaslin_read_data(input_data = file.path(output_tmp, 'data.tsv'), 
                input_metadata = file.path(output_tmp, 'metadata.tsv'), 
                feature_specific_covariate = file.path(output_tmp, 'covar.tsv'), 
                unscaled_abundance = file.path(output_tmp, 'unscaled.tsv'))

expect_that(data_in_new$data, equals(data_in))
expect_that(data_in_new$metadata, equals(metadata))
expect_that(data_in_new$feature_specific_covariate, equals(covar_data))
expect_that(data_in_new$unscaled_abundance, equals(unscaled))

test_that("Can read SE", {
    
    # fake data
    taxa_mat = matrix(rnorm(12), nrow = 3,
                      dimnames = list(paste0("ft_", 1:3),
                                      paste0("samp_", 1:4)))
    
    m = data.frame(grp = sample(0:1, size = 4, replace = TRUE),
                   x = rnorm(4),
                   row.names = paste0("samp_", 1:4))
    
    # construct SE here
    se = SummarizedExperiment::SummarizedExperiment(assays = list(values = taxa_mat),
                                                    colData = m)
     
    se_read = maaslin3:::maaslin_read_summarized_experiment_data(se)

    expect_equal(se_read, list(data = as.data.frame(t(taxa_mat)), metadata = m))
})

library(testthat)
library(maaslin3)

# The idea of these checks is that the contrast test should be the same as
# refactoring with categorical data
set.seed(1)

data_in <- data.frame('a' = c(0, 0, 3, 0, 5, 7, 0, 8, 4, 7), 
                      'b' = c(2, 0, 0, 5, 0, 5, 2, 6, 9, 0),
                      'c' = c(3, 4, 0, 0, 7, 2, 7, 4, 7, 0))
rownames(data_in) <- paste0("sample", c(1:10))

metadata <- data.frame('var1' = c(1, 2, 3, 4, 5, 6, 7, 8, 9, 10),
                       'var2' = c('a', 'a', 'a', 'a', 'b', 
                                  'b', 'b', 'c', 'c', 'c'))
metadata$var2 <- factor(metadata$var2)
rownames(metadata) <- paste0("sample", c(1:10))

data_in_tss <- data.frame(t(apply(data_in, MARGIN = 1, 
                                  FUN = function(x){x / sum(x)})))
data_in_tss[data_in_tss == 0] <- NA
data_in_tss_log <- log2(data_in_tss)

out_dir = tempdir()

results <- maaslin_fit(data_in_tss,
                       data_in_tss_log,
                       metadata,
                       formula = formula('expr ~ var1 + var2'),
                       random_effects_formula = NULL, 
                       min_abundance = 0, 
                       min_prevalence = 0, 
                       min_variance = 0,
                       data = data_in, 
                       save_models = TRUE)

contrast_mat <- matrix(c(0, -1, 1), 
                       ncol = 3, nrow = 1, byrow = TRUE)

colnames(contrast_mat) <- c("var1",
                            "var2b",
                            "var2c")

contrast_test_out <- maaslin_contrast_test(results, 
                      contrast_mat)

metadata$var2 <- factor(metadata$var2, levels = c('b', 'a', 'c'))

results2 <- maaslin_fit(data_in_tss,
                       data_in_tss_log,
                       metadata,
                       formula = formula('expr ~ var1 + var2'),
                       random_effects_formula = NULL, 
                       min_abundance = 0, 
                       min_prevalence = 0, 
                       min_variance = 0,
                       data = data_in, 
                       save_models = TRUE)

new_mod_results <- results2$fit_data_abundance$results[
    results2$fit_data_abundance$results$value == 'c',]
new_mod_results <- new_mod_results[order(new_mod_results$pval_individual),]

expect_that(new_mod_results$coef, equals(contrast_test_out$fit_data_abundance$results$coef))
expect_that(new_mod_results$stderr, equals(contrast_test_out$fit_data_abundance$results$stderr))
expect_equal(new_mod_results$pval_individual, 
             contrast_test_out$fit_data_abundance$results$pval_individual, tolerance = 0.01)

new_mod_results2 <- results2$fit_data_prevalence$results[
    results2$fit_data_prevalence$results$value == 'c',]
new_mod_results2 <- new_mod_results2[order(new_mod_results2$pval_individual),]

contrast_test_out$fit_data_prevalence$results <- 
    contrast_test_out$fit_data_prevalence$results[
        order(contrast_test_out$fit_data_prevalence$results$pval_individual),]

expect_that(new_mod_results2$coef, equals(contrast_test_out$fit_data_prevalence$results$coef))
expect_that(new_mod_results2$stderr, equals(contrast_test_out$fit_data_prevalence$results$stderr))
expect_equal(new_mod_results2$pval_individual, 
             contrast_test_out$fit_data_prevalence$results$pval_individual, tolerance = 0.01)

unlink(out_dir, recursive = TRUE)

#!/usr/bin/env Rscript

###############################################################################
# MaAsLin 3

# Copyright (c) 2025 Harvard School of Public Health

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
###############################################################################

###############################################################
# If running on the command line, load other Maaslin3 modules #
###############################################################

# this evaluates to true if script is being called directly as an executable
cmd_line_chk = identical(environment(), globalenv()) &&
    !length(grep("^source\\(", sys.calls()))

if (cmd_line_chk) {
    # source all R in Maaslin3 package, relative to this folder
    # same method as original maaslin
    script_options <- commandArgs(trailingOnly = FALSE)
    script_path <-
        sub("--file=", "", script_options[grep("--file=", script_options)])
    script_dir <- dirname(script_path)
    script_name <- basename(script_path)
    R_files <- c("fit.R", "utility_scripts.R", "viz.R")

    for (R_file in R_files)
    {
        if (!(R_file == script_name))
            source(file.path(script_dir, R_file))
    }
    
    # After parsing the command line arguments below, we'll use the cores
    # argument to set the daemons.
}

#### Set the default options ####

normalization_choices <- c("TSS", "CLR", "NONE")
transform_choices <- c("LOG", "PLOG", "NONE")

correction_choices <-
    c("BH", "holm", "hochberg", "hommel", "bonferroni", "BY")

# set the default run options
args <- list()
args$input_data <- NULL
args$input_metadata <- NULL
args$output <- NULL
args$min_abundance <- 0.0
args$zero_threshold <- 0.0
args$min_prevalence <- 0.0
args$max_prevalence <- 1.0
args$min_variance <- 0.0
args$max_significance <- 0.1
args$normalization <- normalization_choices[1]
args$transform <- transform_choices[1]
args$correction <- correction_choices[1]
args$formula <- NULL
args$random_effects <- NULL
args$group_effects <- NULL
args$ordered_effects <- NULL
args$strata_effects <- NULL
args$fixed_effects <- NULL
args$feature_specific_covariate <- NULL
args$feature_specific_covariate_name <- NULL
args$feature_specific_covariate_record <- NULL
args$standardize <- TRUE
args$median_comparison_abundance <- TRUE
args$median_comparison_prevalence <- FALSE
args$median_comparison_abundance_threshold <- 0
args$median_comparison_prevalence_threshold <- 0
args$subtract_median <- FALSE
args$warn_prevalence <- TRUE
args$small_random_effects <- FALSE
args$bypass_small_group_warning <- FALSE
args$augment <- TRUE
args$evaluate_only <- NULL
args$unscaled_abundance <- NULL
args$plot_summary_plot <- TRUE
args$summary_plot_first_n <- 25
args$coef_plot_vars <- NULL
args$heatmap_vars <- NULL
args$plot_associations <- TRUE
args$max_pngs <- 30
args$cores <- "1"
args$save_models <- FALSE
args$save_plots_rds <- FALSE
args$reference <- NULL
args$summary_plot_balanced <- FALSE

#### end ####

#### Add command line arguments ####

options <-
    optparse::OptionParser(usage = paste(
        "%prog",
        "<data.tsv>",
        "<metadata.tsv>",
        "<output_folder>",
        "[options]"
    ))
options <-
    optparse::add_option(
        options,
        c("--formula"),
        type = "character",
        dest = "formula",
        default = args$formula,
        help = paste("The formula for the model",
                    "[ Default: all variables fixed ]")
    )
options <-
    optparse::add_option(
        options,
        c("--fixed_effects"),
        type = "character",
        dest = "fixed_effects",
        default = args$fixed_effects,
        help = paste(
            "The fixed effects for the model,",
            "comma-delimited for multiple effects",
            "[ Default: all ]"
        )
    )
options <-
    optparse::add_option(
        options,
        c("--reference"),
        type = "character",
        dest = "reference",
        default = args$reference,
        help = paste(
            "The factor to use as a reference for",
            "a variable with more than two levels",
            "provided as a string of 'variable,reference'",
            "semi-colon delimited for multiple variables [ Default: NA ]"
        )
    )
options <-
    optparse::add_option(
        options,
        c("--random_effects"),
        type = "character",
        dest = "random_effects",
        default = args$random_effects,
        help = paste(
            "The random effects for the model,",
            "comma-delimited for multiple effects",
            "[ Default: none ]"
        )
    )
options <-
    optparse::add_option(
        options,
        c("--group_effects"),
        type = "character",
        dest = "group_effects",
        default = args$group_effects,
        help = paste(
            "The group effects for the model,",
            "comma-delimited for multiple effects",
            "[ Default: none ]"
        )
    )
options <-
    optparse::add_option(
        options,
        c("--ordered_effects"),
        type = "character",
        dest = "ordered_effects",
        default = args$ordered_effects,
        help = paste(
            "The ordered effects for the model,",
            "comma-delimited for multiple effects",
            "[ Default: none ]"
        )
    )
options <-
    optparse::add_option(
        options,
        c("--strata_effects"),
        type = "character",
        dest = "strata_effects",
        default = args$ordered_effects,
        help = paste(
            "The strata effects for the model.",
            "Only one is allowed.",
            "[ Default: none ]"
        )
    )
options <-
    optparse::add_option(
        options,
        c("--feature_specific_covariate"),
        type = "character",
        dest = "feature_specific_covariate",
        default = args$feature_specific_covariate,
        help = paste(
            "The table to use for feature-specific",
            "covariates. Row and column names should",
            "match the data input."
        )
    )
options <-
    optparse::add_option(
        options,
        c("--feature_specific_covariate_name"),
        type = "character",
        dest = "feature_specific_covariate_name",
        default = args$feature_specific_covariate,
        help = paste("The name of the feature-specific covariate")
    )
options <-
    optparse::add_option(
        options,
        c("--feature_specific_covariate_record"),
        type = "character",
        dest = "feature_specific_covariate_record",
        default = args$feature_specific_covariate_record,
        help = paste("Whether to include the feature-specific
                    covariate in the outputs")
    )
options <-
    optparse::add_option(
        options,
        c("--min_abundance"),
        type = "double",
        dest = "min_abundance",
        default = args$min_abundance,
        help = paste(
            "The minimum abundance for each feature
            (before normalization and transformation)",
            "[ Default: %default ]"
        )
    )
options <-
    optparse::add_option(
        options,
        c("--min_prevalence"),
        type = "double",
        dest = "min_prevalence",
        default = args$min_prevalence,
        help = paste(
            "The minimum proportion of samples for which",
            "a feature is detected at minimum abundance",
            "[ Default: %default ]"
        )
    )
options <-
    optparse::add_option(
        options,
        c("--max_prevalence"),
        type = "double",
        dest = "max_prevalence",
        default = args$max_prevalence,
        help = paste(
            "The maximum proportion of samples for which",
            "a feature is detected at minimum abundance",
            "[ Default: %default ]"
        )
    )
options <-
    optparse::add_option(
        options,
        c("--zero-threshold"),
        type = "double",
        dest = "zero_threshold",
        default = args$zero_threshold,
        help = paste(
            "The minimum abundance to be considered non-zero",
            "[ Default: %default ]"
        )
    )
options <-
    optparse::add_option(
        options,
        c("--min_variance"),
        type = "double",
        dest = "min_variance",
        default = args$min_variance,
        help = paste(
            "Keep features with variances",
            "greater than value",
            "[ Default: %default ]"
        )
    )
options <-
    optparse::add_option(
        options,
        c("--max_significance"),
        type = "double",
        dest = "max_significance",
        default = args$max_significance,
        help = paste("The q-value threshold for significance",
                    "[ Default: %default ]")
    )
options <-
    optparse::add_option(
        options,
        c("--normalization"),
        type = "character",
        dest = "normalization",
        default = args$normalization,
        help = paste(
            "The normalization method to apply",
            "[ Default: %default ] [ Choices:",
            toString(normalization_choices),
            "]"
        )
    )
options <-
    optparse::add_option(
        options,
        c("--transform"),
        type = "character",
        dest = "transform",
        default = args$transform,
        help = paste(
            "The transform to apply [ Default: %default ] [ Choices:",
            toString(transform_choices),
            "]"
        )
    )
options <-
    optparse::add_option(
        options,
        c("--correction"),
        type = "character",
        dest = "correction",
        default = args$correction,
        help = paste(
            "The correction method for computing",
            "the q-value [ Default: %default ]"
        )
    )
options <-
    optparse::add_option(
        options,
        c("--standardize"),
        type = "logical",
        dest = "standardize",
        default = args$standardize,
        help = paste(
            "Apply z-score so continuous metadata are on",
            "the same scale [ Default: %default ]"
        )
    )
options <-
    optparse::add_option(
        options,
        c("--unscaled_abundance"),
        type = "character",
        dest = "unscaled_abundance",
        default = args$unscaled_abundance,
        help = paste(
            "The table to use as an unscaled",
            "abundance reference (the single column",
            "name must be the same as one of the",
            "features or 'total')"
        )
    )
options <-
    optparse::add_option(
        options,
        c("--median_comparison_abundance"),
        type = "logical",
        dest = "median_comparison_abundance",
        default = args$median_comparison_abundance,
        help = paste(
            "Test abundance coefficients against the median",
            "association [ Default: %default ]"
        )
    )
options <-
    optparse::add_option(
        options,
        c("--median_comparison_prevalence"),
        type = "logical",
        dest = "median_comparison_prevalence",
        default = args$median_comparison_prevalence,
        help = paste(
            "Test prevalence coefficients against the median",
            "association [ Default: %default ]"
        )
    )
options <-
    optparse::add_option(
        options,
        c("--median_comparison_abundance_threshold"),
        type = "double",
        dest = "median_comparison_abundance_threshold",
        default = args$median_comparison_abundance_threshold,
        help = paste(
            "Radius within which the median adjustment",
            "gives a p-value of 1 [ Default: %default ]"
        )
    )
options <-
    optparse::add_option(
        options,
        c("--median_comparison_prevalence_threshold"),
        type = "double",
        dest = "median_comparison_prevalence_threshold",
        default = args$median_comparison_prevalence_threshold,
        help = paste(
            "Radius within which the median adjustment",
            "gives a p-value of 1 [ Default: %default ]"
        )
    )
options <-
    optparse::add_option(
        options,
        c("--subtract_median"),
        type = "logical",
        dest = "subtract_median",
        default = args$subtract_median,
        help = paste(
            "Subtract the median from coefficients when",
            "doing median comparisons [ Default: %default ]"
        )
    )
options <-
    optparse::add_option(
        options,
        c("--warn_prevalence"),
        type = "logical",
        dest = "warn_prevalence",
        default = args$warn_prevalence,
        help = paste(
            "Check and warn if prevalence associations are",
            "likely due to compositionality [ Default: %default ]"
        )
    )
options <-
    optparse::add_option(
        options,
        c("--small_random_effects"),
        type = "logical",
        dest = "small_random_effects",
        default = args$small_random_effects,
        help = paste(
            "Replace prevalence random effects with",
            "fixed effects because groups are small [ Default: %default ]"
        )
    )
options <-
    optparse::add_option(
        options,
        c("--bypass_small_group_warning"),
        type = "logical",
        dest = "bypass_small_group_warning",
        default = args$bypass_small_group_warning,
        help = paste(
            "Bypass the warning when there are <4 average",
            "observations per random effect group [ Default: %default ]"
        )
    )
options <-
    optparse::add_option(
        options,
        c("--augment"),
        type = "logical",
        dest = "augment",
        default = args$augment,
        help = paste(
            "Add weighted extra 0s and 1s to avoid linear",
            "separability [ Default: %default ]"
        )
    )
options <-
    optparse::add_option(
        options,
        c("--evaluate_only"),
        type = "character",
        dest = "evaluate_only",
        default = args$evaluate_only,
        help = paste(
            "Whether to evaluate just the abundnace or
            prevalence models [ Default: %default ] [ Choices:",
            toString(c("abundance", "prevalence"),),
            "]"
        )

    )

options <-
    optparse::add_option(
        options,
        c("--plot_summary_plot"),
        type = "logical",
        dest = "plot_summary_plot",
        default = args$plot_summary_plot,
        help = paste(
            "Generate a summary plot for the significant",
            "associations [ Default: %default ]"
        )
    )
options <-
    optparse::add_option(
        options,
        c("--summary_plot_first_n"),
        type = "double",
        dest = "summary_plot_first_n",
        default = args$summary_plot_first_n,
        help = paste(
            "In summary plot, plot top N features with significant",
            "associations [ Default: %default ]"
        )
    )
options <-
    optparse::add_option(
        options,
        c("--coef_plot_vars"),
        type = "character",
        dest = "coef_plot_vars",
        default = args$coef_plot_vars,
        help = paste(
            "The variables to use in the coefficient plot",
            "section of the summary plot provided as a",
            "comma-separated string. Continuous variables",
            "should match the metadata column name, and",
            "categorical variables should be of the form:",
            "[metadata] [level]. [ Default: NA ]"
        )
    )
options <-
    optparse::add_option(
        options,
        c("--heatmap_vars"),
        type = "character",
        dest = "heatmap_vars",
        default = args$coef_plot_vars,
        help = paste(
            "The variables to use in the heatmap",
            "section of the summary plot provided as a",
            "comma-separated string. Continuous variables",
            "should match the metadata column name, and",
            "categorical variables should be of the form:",
            "[metadata] [level]. [ Default: NA ]"
        )
    )
options <-
    optparse::add_option(
        options,
        c("--plot_associations"),
        type = "logical",
        dest = "plot_associations",
        default = args$plot_associations,
        help = paste(
            "Generate associations plots for the significant",
            "associations [ Default: %default ]"
        )
    )
options <-
    optparse::add_option(
        options,
        c("--max_pngs"),
        type = "double",
        dest = "max_pngs",
        default = args$max_pngs,
        help = paste(
            "The maximum number of association plots for signficant",
            "associations to save as png files [ Default: %default ]"
        )
    )
options <-
    optparse::add_option(
        options,
        c("--cores"),
        type = "character",
        dest = "cores",
        default = args$cores,
        help = paste(
            "Number of parallel workers. An integer > 1 starts mirai daemons",
            "for parallel computation. Alternatively, from the command line,",
            "a path to an R script with custom mirai::daemons() setup that",
            "will be parsed and evaluated with source().",
            "Users can also call mirai::daemons() directly before maaslin3().",
            "[default: %default]"
        )
    )
options <-
    optparse::add_option(
        options,
        c("--save_models"),
        type = "logical",
        dest = "save_models",
        default = args$save_models,
        help = paste(
            "Return the full model outputs",
            "and save to an RData file [ Default: %default ]"
        )
    )
options <-
    optparse::add_option(
        options,
        c("--save_plots_rds"),
        type = "logical",
        dest = "save_plots_rds",
        default = args$save_plots_rds,
        help = paste(
            "Save the plots to RDS files",
            "[ Default: %default ]"
        )
    )
options <-
    optparse::add_option(
        options,
        c("--summary_plot_balanced"),
        type = "logical",
        dest = "summary_plot_balanced",
        default = args$summary_plot_balanced,
        help = paste(
            "If coef_plot_vars is selected this will",
            "select balanced top features [ Default: %default ]"
        )
    )

option_not_valid_error <- function(message, valid_options) {
    logging::logerror(paste(message, ": %s"), toString(valid_options))
    stop("Option not valid", call. = FALSE)
}

#### end ####

####################################
# Check valid options are selected #
####################################

maaslin_check_arguments <-
    function(feature_specific_covariate = NULL,
            feature_specific_covariate_name = NULL,
            feature_specific_covariate_record = NULL,
            zero_threshold = 0,
            normalization = 'TSS',
            transform = 'LOG',
            correction = 'BH',
            warn_prevalence = TRUE,
            evaluate_only = NULL,
            unscaled_abundance = NULL,
            median_comparison_abundance = TRUE) {
        normalization <- toupper(normalization)
        transform <- toupper(transform)

        # Match variable ignoring case then set
        # correctly as required for p.adjust
        correction <- correction_choices[match(toupper(correction),
                                            toupper(correction_choices))]

        if (!is.null(unscaled_abundance) &&
            median_comparison_abundance) {
            stop_message <- paste0(
                "`median_comparison_abundance` usually should not be ",
                "TRUE (default) with unscaled abundances. To bypass this ",
                "check, run the maaslin steps individually, ",
                "skipping `maaslin_check_arguments`.")
            stop(stop_message)
        }

        if (!((
            is.null(feature_specific_covariate) +
            is.null(feature_specific_covariate_name) +
            is.null(feature_specific_covariate_record)
        ) %in% c(0, 3))) {
            stop_message <- paste0("`feature_specific_covariate`, ",
                            "`feature_specific_covariate_name`, ",
                            "and `feature_specific_covariate_record` ",
                            "should all be null or all be non-null")
            stop(stop_message)
        }

        # Check valid normalization option selected
        logging::loginfo("Verifying options selected are valid")
        if (!normalization %in% normalization_choices) {
            option_not_valid_error(
                paste(
                    "Please select a normalization",
                    "from the list of available options"
                ),
                toString(normalization_choices)
            )
        }

        if (!is.null(unscaled_abundance) & normalization != 'TSS') {
            stop("Normalization must be TSS if using unscaled abundance")
        }

        # check valid transform option selected
        if (!transform %in% transform_choices) {
            option_not_valid_error(
                "Please select a transform from the list of available options",
                toString(transform_choices)
            )
        }

        # check valid correction method selected
        if (!correction %in% correction_choices) {
            option_not_valid_error(
                paste(
                    "Please select a correction method",
                    "from the list of available options"
                ),
                toString(correction_choices)
            )
        }

        if (normalization == 'CLR' & transform != 'NONE') {
            stop("normalization = CLR can only be used with transform NONE")
        }

        if (transform == 'PLOG' & zero_threshold >= 0) {
            stop_message <- paste0(
                "transform set to PLOG, but zero_threshold is >= 0. ",
                "Set zero_threshold to -1 to count all ",
                "features as present and apply PLOG.")
            stop(stop_message)
        }

        if (warn_prevalence) {
            if (normalization == 'CLR') {
                stop_message <- paste0("normalization = CLR can only be used ",
                                        "with warn_prevalence FALSE")
                stop(stop_message)
            }
            if (normalization != 'TSS') {
                warn_message <- paste0("Be sure the data can be TSS normalized",
                        " when using warn_prevalence without normalization=TSS")
                warning(warn_message)
            }
            if (transform != 'LOG') {
                stop_message <- paste0(
                    "warn_prevalence has only been validated with ",
                    "transform = LOG. To bypass this ",
                    "check, run the maaslin steps individually, ",
                    "skipping `maaslin_check_arguments`.")
                stop(stop_message)
            }
            if (!is.null(evaluate_only)) {
                stop_message <-
                    "evaluate_only must be NULL when using warn_prevalence"
                stop(stop_message)
            }
        }

        if (!is.null(evaluate_only) &&
            !evaluate_only %in% c("abundance", "prevalence")) {
            stop("evaluate_only must be NULL, 'abundance', or 'prevalence'")
        }

        if (transform == 'PLOG' &&
            (is.null(evaluate_only) ||
            evaluate_only != "abundance")) {
            stop_message <- paste0("PLOG should only be used for ",
                                    "abundance-only modeling (evaluate_only)")
            stop(stop_message)
        }
    }

#####################################
# Create log file and log arguments #
#####################################

maaslin_log_arguments <- function(input_data,
                                input_metadata,
                                output,
                                formula = NULL,
                                fixed_effects = NULL,
                                reference = NULL,
                                random_effects = NULL,
                                group_effects = NULL,
                                ordered_effects = NULL,
                                strata_effects = NULL,
                                feature_specific_covariate = NULL,
                                feature_specific_covariate_name = NULL,
                                feature_specific_covariate_record = NULL,
                                min_abundance = 0,
                                min_prevalence = 0,
                                max_prevalence = 1.01,
                                zero_threshold = 0,
                                min_variance = 0,
                                max_significance = 0.1,
                                normalization = 'TSS',
                                transform = 'LOG',
                                correction = 'BH',
                                standardize = TRUE,
                                unscaled_abundance = NULL,
                                median_comparison_abundance = TRUE,
                                median_comparison_prevalence = FALSE,
                                median_comparison_abundance_threshold = 0,
                                median_comparison_prevalence_threshold = 0,
                                subtract_median = FALSE,
                                warn_prevalence = TRUE,
                                small_random_effects = FALSE,
                                bypass_small_group_warning = FALSE,
                                augment = TRUE,
                                evaluate_only = NULL,
                                plot_summary_plot = TRUE,
                                summary_plot_first_n = 25,
                                coef_plot_vars = NULL,
                                heatmap_vars = NULL,
                                plot_associations = TRUE,
                                max_pngs = 30,
                                cores = 1,
                                save_models = FALSE,
                                save_plots_rds = FALSE,
                                verbosity = 'FINEST',
                                summary_plot_balanced = FALSE) {
    # Allow for lower case variables
    normalization <- toupper(normalization)
    transform <- toupper(transform)

    # Match variable ignoring case then set correctly as required for p.adjust
    correction <- correction_choices[match(toupper(correction),
                                        toupper(correction_choices))]

    # If formula is a formula object, convert it back to a string
    if (methods::is(formula, "formula")) {
        formula <- safe_deparse(formula)
    }

    # create an output folder
    if (!file.exists(output)) {
        logging::loginfo("Creating output folder")
        dir.create(output)
    }

    # create log file (write info to stdout and debug level to log file)
    # set level to finest so all log levels are reviewed
    log_file <- file.path(output, "maaslin3.log")

    # remove log file if already exists (to avoid append)
    if (file.exists(log_file)) {
        logging::logwarn(paste(
            "Warning: Deleting existing log file:", log_file))
        unlink(log_file)
    }

    logging::logReset()
    
    logging::basicConfig(level = verbosity)
    
    logging::removeHandler("basic.stdout")
    
    logging::addHandler(logging::writeToConsole,
                        handler = "basic.stdout",
                        logger = "",
                        level = 20,
                        formatter = nrw_fmt)
    
    logging::addHandler(logging::writeToFile,
                        file = log_file, level = verbosity,
                        formatter = nrw_fmt)
    
    logging::setLevel(20, logging::getHandler('basic.stdout'))

    logging::loginfo("Writing function arguments to log file")
    logging::logdebug("Function arguments")
    if (is.character(input_data)) {
        logging::logdebug("Input data file: %s", input_data)
    }
    if (is.character(input_metadata)) {
        logging::logdebug("Input metadata file: %s", input_metadata)
    }
    logging::logdebug("Output folder: %s", output)
    logging::logdebug("Formula: %s", formula)
    logging::logdebug("Fixed effects: %s", fixed_effects)
    logging::logdebug("Reference: %s", reference)
    logging::logdebug("Random effects: %s", random_effects)
    logging::logdebug("Group effects: %s", group_effects)
    logging::logdebug("Ordered effects: %s", ordered_effects)
    logging::logdebug("Strata effects: %s", strata_effects)
    if (!is.null(feature_specific_covariate)) {
        if (is.character(feature_specific_covariate)) {
            logging::logdebug("Feature specific covariate: %s",
                                feature_specific_covariate)
        }
    }
    if (!is.null(feature_specific_covariate_name)) {
        if (is.character(feature_specific_covariate_name)) {
            logging::logdebug("Feature specific covariate name: %s",
                            feature_specific_covariate_name)
        }
    }
    if (!is.null(feature_specific_covariate_record)) {
        if (is.character(feature_specific_covariate_record)) {
            logging::logdebug(
                "Feature specific covariate include: %s",
                feature_specific_covariate_record
            )
        }
    }
    logging::logdebug("Min Abundance: %f", min_abundance)
    logging::logdebug("Min Prevalence: %f", min_prevalence)
    logging::logdebug("Max Prevalence: %f", max_prevalence)
    logging::logdebug("Zero Threshold: %f", zero_threshold)
    logging::logdebug("Min variance: %f", min_variance)
    logging::logdebug("Max significance: %f", max_significance)
    logging::logdebug("Normalization: %s", normalization)
    logging::logdebug("Transform: %s", transform)
    logging::logdebug("Correction method: %s", correction)
    logging::logdebug("Standardize: %s", standardize)
    if (is.character(unscaled_abundance)) {
        logging::logdebug("Unscaled abundance: %s", unscaled_abundance)
    }
    logging::logdebug("Abundance median comparison: %s",
                        median_comparison_abundance)
    logging::logdebug("Prevalence median comparison: %s",
                        median_comparison_prevalence)
    logging::logdebug(
        "Abundance median comparison threshold: %s",
        median_comparison_abundance_threshold
    )
    logging::logdebug(
        "Prevalence median comparison threshold: %s",
        median_comparison_prevalence_threshold
    )
    logging::logdebug(
        "Subtract median: %s",
        subtract_median
    )
    logging::logdebug(
        "Warn prevalence: %s",
        warn_prevalence
    )
    logging::logdebug(
        "Small random effects: %s",
        small_random_effects
    )
    logging::logdebug(
        "Bypass small group warning: %s",
        bypass_small_group_warning
    )
    logging::logdebug("Augment: %s", augment)
    logging::logdebug("Evaluate only: %s", evaluate_only)
    
    n_con = mirai::info()["connections"]
    
    n_daemon = if (is.null(n_con)) 0 else n_con
    
    logging::logdebug("Number of mirai daemons: %d", n_daemon)
    
    logging::logdebug("Balanced Summary plot: %s", summary_plot_balanced)


    maaslin_check_arguments(
        feature_specific_covariate,
        feature_specific_covariate_name,
        feature_specific_covariate_record,
        zero_threshold,
        normalization,
        transform,
        correction,
        warn_prevalence,
        evaluate_only,
        unscaled_abundance,
        median_comparison_abundance
    )
}

#################################
# Read in the data and metadata #
#################################

maaslin_read_data <- function(input_data,
                            input_metadata,
                            feature_specific_covariate = NULL,
                            unscaled_abundance = NULL) {
    # if a character string then this is a file name, else it
    # is a data frame
    if (is.character(input_data) && file.exists(input_data)) {
        data <- read.table(input_data,
            header = TRUE,
            row.names = 1, 
            sep = if (grepl('\\.tsv$|\\.txt$', input_data)) '\t' else ',',
            check.names = FALSE)
    } else if (is.data.frame(input_data)) {
        if (has_auto_rownames(input_data)) {
            stop("If supplying input_data as a data frame,
                it must have appropriate rownames!")
        }
        rownames_tmp <- rownames(input_data)
        data <- as.data.frame(input_data) # in case it's a tibble or something
        rownames(data) <- rownames_tmp
    } else if (inherits(input_data, 'DataFrame')) {
        data <- as.data.frame(input_data) # If it's BioC's DataFrame
    } else if (is.matrix(input_data)) {
        logging::logwarn("Input is a matrix,
                        passing through as.data.frame() .")
        data <- as.data.frame(input_data)
    } else {
        stop("input_data is neither a file nor a data frame!")
    }

    if (is.character(input_metadata) &&
        file.exists(input_metadata)) {
        metadata <- read.table(input_metadata,
            header = TRUE,
            row.names = 1, 
            sep = if (grepl('\\.tsv$|\\.txt$', input_data)) '\t' else ',',
            check.names = FALSE)
    } else if (is.data.frame(input_metadata)) {
        if (has_auto_rownames(input_metadata)) {
            stop(
                "If supplying input_metadata as a data frame,
                it must have appropriate rownames!"
            )
        }
        # in case it's a tibble or something
        rownames_tmp <- rownames(input_metadata)
        metadata <- as.data.frame(input_metadata)
        rownames(metadata) <- rownames_tmp
    } else if (inherits(input_metadata, 'DataFrame')) {
        metadata <- as.data.frame(input_metadata) # If it's BioC's DataFrame
    } else {
        stop("input_metadata is neither a file nor a data frame!")
    }

    if (is.character(unscaled_abundance) &&
        file.exists(unscaled_abundance)) {
        unscaled_abundance <-
            read.table(unscaled_abundance,
                header = TRUE,
                row.names = 1, 
                sep = if (grepl('\\.tsv$|\\.txt$', input_data)) '\t' else ',',
                check.names = FALSE)
    } else if (is.data.frame(unscaled_abundance)) {
        if (has_auto_rownames(unscaled_abundance)) {
            stop(
                "If supplying unscaled_abundance as a data frame,
                it must have appropriate rownames!"
            )
        }
        rownames_tmp <- rownames(unscaled_abundance)
        unscaled_abundance <- as.data.frame(unscaled_abundance)
        rownames(unscaled_abundance) <- rownames_tmp
    } else if (inherits(unscaled_abundance, 'DataFrame')) {
        unscaled_abundance <-
            as.data.frame(unscaled_abundance) # If it's BioC's DataFrame
    } else if (!is.null(unscaled_abundance)) {
        stop("unscaled_abundance is not a file or data frame!")
    }

    if (is.character(feature_specific_covariate) &&
        file.exists(feature_specific_covariate)) {
        feature_specific_covariate <-
            read.table(feature_specific_covariate,
                header = TRUE,
                row.names = 1, 
                sep = if (grepl('\\.tsv$|\\.txt$', input_data)) '\t' else ',',
                check.names = FALSE)
    } else if (is.data.frame(feature_specific_covariate)) {
        if (has_auto_rownames(feature_specific_covariate)) {
            stop(
                "If supplying feature_specific_covariate as a data frame,
                it must have appropriate rownames!"
            )
        }
        rownames_tmp <- rownames(feature_specific_covariate)
        feature_specific_covariate <-
            as.data.frame(feature_specific_covariate)
        rownames(feature_specific_covariate) <- rownames_tmp
    } else if (inherits(feature_specific_covariate, 'DataFrame')) {
        feature_specific_covariate <-
            as.data.frame(feature_specific_covariate) # If it's BioC's DataFrame
    } else if (!is.null(feature_specific_covariate)) {
        stop("feature_specific_covariate is not a file or data frame!")
    }

    if (any(duplicated(colnames(data)))) {
        logging::logwarn(
            "Duplicate feature names detected: %s. Making names unique.",
            toString(colnames(data)[duplicated(colnames(data))])
        )
        colnames(data) <- make.unique(colnames(data))
    }

    return(
        list(
            "data" = data,
            "metadata" = metadata,
            "feature_specific_covariate" = feature_specific_covariate,
            "unscaled_abundance" = unscaled_abundance
        )
    )
}

###############################################################
# Determine orientation of data in input and reorder to match #
###############################################################

maaslin_reorder_data <- function(data,
                                metadata,
                                feature_specific_covariate = NULL,
                                unscaled_abundance = NULL) {
    logging::loginfo("Determining format of input files")
    samples_row_row <- intersect(rownames(data), rownames(metadata))
    if (length(samples_row_row) > 0) {
        # this is the expected formatting so do not modify data frames
        logging::loginfo(paste(
            "Input format is data samples",
            "as rows and metadata samples as rows"
        ))
    } else {
        samples_column_row <- intersect(colnames(data), rownames(metadata))

        if (length(samples_column_row) == 0) {
            samples_column_row <-
                intersect(colnames(data), rownames(metadata))
        }

        if (length(samples_column_row) > 0) {
            logging::loginfo(
                paste(
                    "Input format is data samples",
                    "as columns and metadata samples as rows"
                )
            )
            # transpose data frame so samples are rows
            data <- as.data.frame(t(data))
            logging::logdebug("Transformed data so samples are rows")
        } else {
            samples_column_column <-
                intersect(colnames(data), colnames(metadata))
            if (length(samples_column_column) > 0) {
                logging::loginfo(
                    paste(
                        "Input format is data samples",
                        "as columns and metadata samples as columns"
                    )
                )
                data <- as.data.frame(BiocGenerics::t(data))
                metadata <- as.data.frame(BiocGenerics::t(metadata))
                logging::logdebug("Transformed data and metadata
                                so samples are rows")
            } else {
                samples_row_column <-
                    intersect(rownames(data), colnames(metadata))

                if (length(samples_row_column) == 0) {
                    samples_row_column <-
                        intersect(rownames(data), colnames(metadata))
                }

                if (length(samples_row_column) > 0) {
                    logging::loginfo(
                        paste(
                            "Input format is data samples",
                            "as rows and metadata samples as columns"
                        )
                    )
                    metadata <- as.data.frame(BiocGenerics::t(metadata))
                    logging::logdebug("Transformed metadata so
                                    samples are rows")
                } else {
                    logging::logerror(
                        paste(
                            "Unable to find samples in data and",
                            "metadata files.",
                            "Rows/columns do not match."
                        )
                    )
                    logging::logdebug("Data rows: %s",
                                    paste(rownames(data), collapse = ","))
                    logging::logdebug("Data columns: %s",
                                    paste(colnames(data), collapse = ","))
                    logging::logdebug("Metadata rows: %s",
                                    paste(rownames(metadata), collapse = ","))
                    logging::logdebug("Metadata columns: %s",
                                    paste(colnames(data), collapse = ","))
                    stop()
                }
            }
        }
    }

    if (!is.null(feature_specific_covariate)) {
        samples_row_row <-
            intersect(rownames(data),
                    rownames(feature_specific_covariate))
        samples_col_col <-
            intersect(colnames(data),
                    colnames(feature_specific_covariate))
        if (length(samples_row_row) > 0 &
            length(samples_col_col) > 0) {
            # this is the expected formatting so do not modify data frames
            logging::loginfo(
                paste(
                    "Input format is data samples",
                    "as rows and feature_specific_covariate samples as rows"
                )
            )
        } else {
            samples_column_row <-
                intersect(colnames(data),
                        rownames(feature_specific_covariate))
            samples_row_column <-
                intersect(rownames(data),
                        colnames(feature_specific_covariate))

            if (length(samples_column_row) == 0 |
                length(samples_row_column) == 0) {
                samples_column_row <-
                    intersect(colnames(data),
                            rownames(feature_specific_covariate))
                samples_row_column <-
                    intersect(rownames(data),
                            colnames(feature_specific_covariate))
            }

            if (length(samples_column_row) > 0 &
                length(samples_row_column) > 0) {
                logging::loginfo(
                    paste(
                        "Input format is feature_specific_covariate samples",
                        "as columns"
                    )
                )
                # transpose data frame so samples are rows
                feature_specific_covariate <-
                    as.data.frame(BiocGenerics::t(feature_specific_covariate))
                logging::logdebug("Transformed feature_specific_covariate
                                so samples are rows")
            } else {
                logging::logerror(
                    paste(
                        "Unable to find samples in feature_specific_covariate.",
                        "Rows/columns do not match."
                    )
                )
                logging::logdebug("Data rows: %s",
                                paste(rownames(data), collapse = ","))
                logging::logdebug("Data columns: %s",
                                paste(colnames(data), collapse = ","))
                logging::logdebug(
                    "Feature specific covariate rows: %s",
                    paste(
                        rownames(feature_specific_covariate),
                        collapse = ","
                    )
                )
                logging::logdebug(
                    "Feature specific covariate columns: %s",
                    paste(
                        colnames(feature_specific_covariate),
                        collapse = ","
                    )
                )
                stop()
            }
        }

    }

    # get a set of the samples with both metadata and features
    intersect_samples <-
        intersect(rownames(data), rownames(metadata))
    logging::logdebug(
        "A total of %s samples were found in both the data and metadata",
        length(intersect_samples)
    )

    if (!is.null(feature_specific_covariate)) {
        intersect_samples <-
            intersect(intersect_samples,
                    rownames(feature_specific_covariate))
        logging::logdebug(
            "A total of %s samples were found in the data, metadata,
            and feature specific covariates",
            length(intersect_samples)
        )
    }

    # check for samples without metadata
    extra_feature_samples <-
        setdiff(rownames(data), intersect_samples)
    if (length(extra_feature_samples) > 0)
        logging::loginfo(
            paste(
                "The following samples were found",
                "to have features but no metadata",
                "(or feature specific covariates if",
                "applicable).",
                "They will be removed. %s"
            ),
            paste(extra_feature_samples, collapse = ",")
        )

    # check for metadata samples without features
    extra_metadata_samples <-
        setdiff(rownames(metadata), intersect_samples)
    if (length(extra_metadata_samples) > 0)
        logging::loginfo(
            paste(
                "The following samples were found",
                "to have metadata but no features",
                "(or feature specific covariates if",
                "applicable).",
                "They will be removed. %s"
            ),
            paste(extra_metadata_samples, collapse = ",")
        )

    if (!is.null(feature_specific_covariate)) {
        extra_feature_specific_covariate_samples <-
            setdiff(rownames(feature_specific_covariate),
                    intersect_samples)
        if (length(extra_feature_specific_covariate_samples) > 0)
            logging::loginfo(
                paste(
                    "The following samples were found",
                    "to have feature specific covariates",
                    "but no features or no metadata.",
                    "They will be removed. %s"
                ),
                paste(extra_feature_specific_covariate_samples, collapse = ",")
            )
    }

    if (!is.null(unscaled_abundance)) {
        extra_unscaled_abundance_samples <-
            setdiff(rownames(unscaled_abundance), rownames(data))
        if (length(extra_unscaled_abundance_samples) > 0)
            logging::logdebug(
                paste(
                    "The following samples were found",
                    "to have unscaled abundances but no features.",
                    "They will be removed. %s"
                ),
                paste(extra_unscaled_abundance_samples, collapse = ",")
            )
    }

    if (!is.null(unscaled_abundance))  {
        if (!all(rownames(data) %in% rownames(unscaled_abundance))) {
            stop("some data samples do not have an unscaled abundance")
        } else if (length(colnames(unscaled_abundance)) > 1) {
            stop("there is more than 1 column in
                the unscaled abundance data frame")
        } else if (colnames(unscaled_abundance) %in% colnames(data)) {
            logging::logdebug("Using unscaled abundance as spike-in feature")
        } else if (colnames(unscaled_abundance) == 'total') {
            logging::logdebug("Using unscaled abundance as total abundances")
        } else {
            stop("unscaled abundance column must be a feature name or 'total'")
        }
    }

    # now order both data and metadata with the same sample ordering
    logging::logdebug("Reordering data/metadata to use same sample ordering")
    data <- data[intersect_samples, , drop = FALSE]
    metadata <- metadata[intersect_samples, , drop = FALSE]

    if (!is.null(unscaled_abundance)) {
        unscaled_abundance <-
            unscaled_abundance[intersect_samples, , drop = FALSE]
    }

    if (!is.null(feature_specific_covariate)) {
        feature_specific_covariate <-
            feature_specific_covariate[intersect_samples, , drop = FALSE]
    }

    return(
        list(
            "data" = data,
            "metadata" = metadata,
            "feature_specific_covariate" = feature_specific_covariate,
            "unscaled_abundance" = unscaled_abundance
        )
    )
}

###########################################
# Compute the formula based on user input #
###########################################

maaslin_compute_formula <- function(data,
                                    metadata,
                                    fixed_effects = NULL,
                                    random_effects = NULL,
                                    group_effects = NULL,
                                    ordered_effects = NULL,
                                    strata_effects = NULL,
                                    feature_specific_covariate_name = NULL) {
    random_effects_formula <- NULL
    # use all metadata if no fixed effects are provided
    if (is.null(fixed_effects)) {
        if (is.null(ordered_effects) & is.null(group_effects) & 
            is.null(feature_specific_covariate_name) & 
            is.null(strata_effects)) {
            fixed_effects <- colnames(metadata)
        }
    } else {
        fixed_effects <- unlist(strsplit(fixed_effects, ",", fixed = TRUE))
        # remove any fixed effects not found in metadata names
        to_remove <- setdiff(fixed_effects, colnames(metadata))
        if (length(to_remove) > 0) {
            logging::logerror(
                paste(
                    "Variable name not found in metadata",
                    "so not applied to formula as fixed effect: %s"
                ),
                paste(to_remove, collapse = " , ")
            )
            stop()
        }
    }

    if (!is.null(random_effects)) {
        random_effects <-
            unlist(strsplit(random_effects, ",", fixed = TRUE))

        common_variables <- intersect(fixed_effects, random_effects)
        if (length(common_variables) > 0) {
            logging::logwarn(
                paste(
                    "Feature name included as fixed and random effect,",
                    "check that this is intended: %s"
                ),
                paste(common_variables, collapse = " , ")
            )
        }

        # remove any random effects not found in metadata
        to_remove <- setdiff(random_effects, colnames(metadata))
        if (length(to_remove) > 0) {
            logging::logerror(
                paste(
                    "Effect name not found in metadata",
                    "so not applied to formula as random effect: %s"
                ),
                paste(to_remove, collapse = " , ")
            )
            stop()
        }

        # Create formula
        if (length(random_effects) > 0) {
            random_effects_formula_text <-
                paste("expr ~ (1 | ",
                    paste(
                        data.table::fifelse(random_effects == make.names(random_effects), 
                            random_effects, 
                            paste0('`', random_effects, '`')),
                        ")",
                        sep = '',
                        collapse = " + (1 | "
                    ),
                    sep = '')
            logging::loginfo("Formula for random effects: %s",
                            random_effects_formula_text)
            random_effects_formula <-
                tryCatch(
                    stats::as.formula(random_effects_formula_text),
                    error = function(e)
                        stop(
                            sprintf(
                                "Invalid formula for random effects: %s",
                                random_effects_formula_text
                            )
                        )
                )
        }
    }

    if (!is.null(group_effects) | !is.null(ordered_effects)) {
        multi_effects <- c()
        if (!is.null(group_effects)) {
            multi_effects <-
                c(multi_effects,
                strsplit(group_effects, ",", fixed = TRUE))
        }
        if (!is.null(ordered_effects)) {
            multi_effects <-
                c(multi_effects,
                strsplit(ordered_effects, ",", fixed = TRUE))
        }

        common_variables <- intersect(fixed_effects, multi_effects)
        if (length(common_variables) > 0) {
            logging::logerror(
                paste(
                    "Feature name included as fixed and group/ordered effect,",
                    "this is not allowed: %s"
                ),
                paste(common_variables, collapse = " , ")
            )
            stop()
        }

        # remove any random effects not found in metadata
        to_remove <- setdiff(multi_effects, colnames(metadata))
        if (length(to_remove) > 0) {
            logging::logerror(paste0(
                "Effect name not found in metadata: ",
                paste0(to_remove, collapse = ", ")
            ))
            stop()
        }
    }

    if (length(fixed_effects) == 0 &
        length(group_effects) == 0 &
        length(ordered_effects) == 0 &
        is.null(feature_specific_covariate_name)) {
        logging::logerror("No fixed/group/ordered/
                        feature-specific effects provided.")
        stop()
    }

    # reduce metadata to only include fixed/group/random effects in formula
    effects_names <-
        unique(
            c(
                fixed_effects,
                random_effects,
                group_effects,
                ordered_effects,
                strata_effects
            )
        )
    metadata <- metadata[, effects_names, drop = FALSE]

    # create the fixed effects formula text
    if (length(fixed_effects) > 0) {
        formula_effects <- data.table::fifelse(fixed_effects == make.names(fixed_effects), 
            fixed_effects, 
            paste0('`', fixed_effects, '`'))
    } else {
        formula_effects <- character(0)
    }
    if (length(group_effects) > 0) {
        formula_effects <-
            union(formula_effects, paste0("group(", 
                data.table::fifelse(group_effects == make.names(group_effects), 
                group_effects, 
                paste0('`', group_effects, '`')), ")"))
    }
    if (length(ordered_effects) > 0) {
        formula_effects <-
            union(formula_effects,
                paste0("ordered(", 
                    data.table::fifelse(ordered_effects == make.names(ordered_effects), 
                    ordered_effects, 
                    paste0('`', ordered_effects, '`')), ")"))
    }
    if (length(strata_effects) > 0) {
        formula_effects <-
            union(formula_effects,
                paste0("strata(", 
                    data.table::fifelse(strata_effects == make.names(strata_effects), 
                        strata_effects, 
                        paste0('`', strata_effects, '`')), ")"))
    }
    if (!is.null(feature_specific_covariate_name)) {
        formula_effects <-
            union(formula_effects, feature_specific_covariate_name)
    }

    formula_text <-
        paste("expr ~ ", paste(formula_effects, collapse = " + "))
    logging::loginfo("Formula for fixed effects: %s", formula_text)
    formula <-
        tryCatch(
            stats::as.formula(formula_text),
            error = function(e)
                stop(
                    sprintf(
                        "Invalid formula. Please provide
                        a different formula: %s",
                        formula_text
                    )
                )
        )

    if (!(is.null(random_effects_formula))) {
        formula <-
            paste('. ~',
                paste(formula_effects, collapse = ' + '),
                '.',
                sep = ' + ')
        formula <- update(random_effects_formula, formula)
    }

    return(list(
        "formula" = formula,
        "random_effects_formula" = random_effects_formula
    ))
}

##############################
# Check a user input formula #
##############################

maaslin_check_formula <- function(data,
                                metadata,
                                input_formula = NULL,
                                feature_specific_covariate_name = NULL) {
    if (methods::is(input_formula, "formula")) {
        input_formula <- safe_deparse(input_formula)
    }

    random_effects_formula <- NULL

    if (is.null(input_formula)) {
        logging::logerror(paste("No user formula provided"))
    }

    # Remove anything before the tilde if necessary
    input_formula <- sub(".*~\\s*", "", input_formula)

    if (!is.null(feature_specific_covariate_name)) {
        if (length(input_formula) == 0 || 
            !grepl(feature_specific_covariate_name, input_formula)) {
            if (!is.null(input_formula) && length(input_formula) > 0 && 
                nchar(input_formula) > 0) {
                input_formula <-
                    paste0("expr ~ ",
                        feature_specific_covariate_name,
                        ' + ',
                        input_formula)
            } else {
                input_formula <-
                    paste0("expr ~ ",
                           feature_specific_covariate_name)
            }
        } else {
            input_formula <- paste0("expr ~ ", input_formula)
        }
    } else {
        input_formula <- paste0("expr ~ ", input_formula)
    }

    formula <-
        tryCatch(
            stats::as.formula(input_formula),
            error = function(e)
                stop(sprintf("Invalid formula: %s",
                            input_formula))
        )

    formula_terms <- all.vars(formula)
    formula_terms <-
        formula_terms[!formula_terms %in% c("expr",
                                            feature_specific_covariate_name)]

    to_remove <- setdiff(formula_terms, colnames(metadata))
    if (length(to_remove) > 0) {
        logging::logerror(
            paste("Effect name not found in metadata: %s"),
            paste(to_remove, collapse = ", ")
        )
        stop()
    }

    term_labels <- attr(terms(formula), "term.labels")

    if (sum(!grepl("strata\\(|\\|", term_labels)) == 0 &
        is.null(feature_specific_covariate_name)) {
        logging::logerror("No fixed, group, or
                        ordered effects included in formula.")
        stop()
    }

    # Create formula
    if (sum(grepl("\\|", term_labels)) > 0) {
        random_effects_formula <- formula
    } else {
        random_effects_formula <- NULL
    }

    return(list(
        "formula" = formula,
        "random_effects_formula" = random_effects_formula
    ))
}

#################
# Normalization #
#################

maaslin_normalize <- function(data,
                            output,
                            zero_threshold = 0,
                            normalization = 'TSS',
                            unscaled_abundance = NULL) {
    features <- data

    match.arg(normalization, normalization_choices)

    normalization <- toupper(normalization)

    logging::loginfo("Running selected normalization method: %s", normalization)

    if (normalization == 'TSS') {
        features <- TSSnorm(features, zero_threshold)
    }
    if (normalization == 'CLR') {
        features <- CLRnorm(features, zero_threshold)
    }
    if (normalization == 'NONE') {
        features <- NONEnorm(features, zero_threshold)
    }

    if (!is.null(unscaled_abundance)) {
        features <-
            UNSCALEDnorm(features, unscaled_abundance, zero_threshold)
    }

    if (!is.null(output)) {
        features_folder <- file.path(output, "features")
        if (!file.exists(features_folder)) {
            logging::loginfo("Creating output feature tables folder")
            dir.create(features_folder, recursive = TRUE)
        }

        data_norm_file <- file.path(features_folder, "data_norm.tsv")
        logging::loginfo("Writing normalized data to file %s", data_norm_file)
        write.table(
            data.frame("feature" = rownames(features), 
                        features, 
                        check.names = FALSE),
            file = data_norm_file,
            sep = "\t",
            quote = FALSE,
            row.names = FALSE
        )
    }

    return(features)
}


########################################################################
# Filter data based on min abundance, min prevalence, and min variance #
########################################################################

maaslin_filter <- function(normalized_data,
                        output,
                        min_abundance = 0,
                        min_prevalence = 0,
                        max_prevalence = 1.01,
                        zero_threshold = 0,
                        min_variance = 0) {
    unfiltered_data <- normalized_data

    # require at least total samples * min prevalence values
    # and at most total samples * max prevalence values
    # for each feature to be greater than min abundance
    logging::loginfo("Filter data based on min abundance, min prevalence, and max prevalence")
    total_samples <- nrow(unfiltered_data)
    logging::loginfo("Total samples in data: %d", total_samples)
    min_samples <- total_samples * min_prevalence
    max_samples <- total_samples * max_prevalence
    logging::loginfo(
        paste(
            "Min samples required with min abundance",
            "for a feature not to be filtered: %f"
        ),
        min_samples
    )
    logging::loginfo(
        paste(
            "Max samples allowed with min abundance",
            "for a feature not to be filtered: %f"
        ),
        max_samples
    )

    # Filter by abundance
    data_zeros <- unfiltered_data
    data_zeros[is.na(data_zeros)] <-
        min(min_abundance, zero_threshold) - 1

    ##########################################
    # Apply the non-zero abundance threshold #
    ##########################################

    prevalence_mask <- ifelse(data_zeros > zero_threshold, 1, 0)
    data_zeros <- data_zeros * prevalence_mask

    filtered_data <-
        unfiltered_data[, colSums(data_zeros > min_abundance) > min_samples &
                          colSums(data_zeros > min_abundance) <= max_samples,
                        drop = FALSE]
    total_filtered_features <-
        ncol(unfiltered_data) - ncol(filtered_data)
    logging::loginfo("Total filtered features: %d", total_filtered_features)
    filtered_feature_names <-
        setdiff(names(unfiltered_data), names(filtered_data))
    logging::loginfo(
        "Filtered feature names from abundance, min prevalence, and max prevalence filtering: %s",
        toString(filtered_feature_names)
    )

    #################################
    # Filter data based on variance #
    #################################

    vars <- apply(filtered_data, 2, var, na.rm = TRUE)
    variance_filtered_data <-
        filtered_data[, which(vars > min_variance), drop = FALSE]
    variance_filtered_features <-
        ncol(filtered_data) - ncol(variance_filtered_data)
    logging::loginfo(
        "Total features filtered by non-zero variance filtering: %d",
        variance_filtered_features)
    variance_filtered_feature_names <-
        setdiff(names(filtered_data), names(variance_filtered_data))
    logging::loginfo(
        "Filtered feature names from variance filtering: %s",
        toString(variance_filtered_feature_names)
    )
    filtered_data <- variance_filtered_data

    #######################
    # Write filtered data #
    #######################

    if (!is.null(output)) {
        features_folder <- file.path(output, "features")
        if (!file.exists(features_folder)) {
            logging::loginfo("Creating output feature tables folder")
            dir.create(features_folder, recursive = TRUE)
        }

        filtered_file <- file.path(features_folder, "filtered_data.tsv")
        logging::loginfo("Writing filtered data to file %s", filtered_file)
        write.table(
            data.frame("feature" = rownames(filtered_data), 
                        filtered_data, 
                        check.names = FALSE),
            file = filtered_file,
            sep = "\t",
            quote = FALSE,
            row.names = FALSE
        )
    }

    return(filtered_data)
}

#######################################
# Filter and standardize the metadata #
#######################################

maaslin_process_metadata <- function(metadata,
                                    formula = NULL,
                                    fixed_effects = NULL,
                                    reference = NULL,
                                    feature_specific_covariate_name = NULL,
                                    standardize = TRUE) {
    if (is.null(reference)) {
        reference <- ","
    }
    split_reference <- unlist(strsplit(reference, "[,;]"))

    # Extract fixed effects from formula if null
    if (is.null(fixed_effects)) {
        term_labels <- attr(terms(formula), "term.labels")
        term_labels <-
            term_labels[!grepl("group|ordered|strata|\\|", term_labels)]
        if (length(term_labels) == 0) {
            term_labels <- '1'
        }
        tmp_formula <-
            stats::formula(paste0("~ ", paste0(term_labels, collapse = " + ")))
        formula_terms <- all.vars(tmp_formula)
        if (is.null(feature_specific_covariate_name)) {
            fixed_effects <- formula_terms
        } else {
            fixed_effects <-
                formula_terms[formula_terms != feature_specific_covariate_name]
        }
    }

    # for each fixed effect, check that a
    # reference level has been set if necessary:
    # number of levels > 2 and metadata isn't already an ordered factor
    for (i in fixed_effects) {
        # don't check for or require reference levels for numeric metadata
        if (is.numeric(metadata[, i])) {
            next
        }
        # respect ordering if a factor is
        # explicitly passed in with no reference set
        if (is.factor(metadata[, i]) && !(i %in% split_reference)) {
            logging::loginfo(
                paste(
                    "Factor detected for categorial metadata '",
                    i,
                    "'. Using as-is.",
                    sep = ""
                )
            )
            next
        }

        # set metadata as a factor (ordered alphabetically)
        metadata[, i] <- as.factor(metadata[, i])
        mlevels <- levels(metadata[, i])

        # get reference level for variable being considered,
        # returns NA if not found
        ref <- split_reference[match(i, split_reference) + 1]

        # if metadata has 2 levels, allow but don't require
        # setting reference level, otherwise require it
        if ((length(mlevels) == 2)) {
            if (!is.na(ref)) {
                metadata[, i] <- relevel(metadata[, i], ref = ref)
            }
        } else if (length(mlevels) > 2) {
            if (!is.na(ref)) {
                metadata[, i] <- relevel(metadata[, i], ref = ref)
            } else {
                stop(
                    paste(
                        "Please provide the reference for the variable '",
                        i,
                        "' which includes more than 2 levels: ",
                        paste(as.character(mlevels), collapse = ", "),
                        ". ",
                        "Alternatively, set the variable as
                        a factor beforehand.",
                        sep = ""
                    )
                )
            }
        } else {
            stop("Provided categorical metadata has
                fewer than 2 unique, non-NA values.")
        }
    }

    ################################
    # Standardize metadata, if set #
    ################################

    if (standardize) {
        logging::loginfo("Applying z-score to standardize continuous metadata")
        
        if (NCOL(collapse::num_vars(metadata)) > 0) {
            collapse::num_vars(metadata) <- collapse::fscale(collapse::num_vars(metadata))
        }
        
    } else {
        logging::loginfo("Bypass z-score application to metadata")
    }

    return(metadata)
}

##################
# Transformation #
##################

maaslin_transform <- function(filtered_data,
                            output,
                            transform = 'LOG') {
    match.arg(transform, transform_choices)

    features <- filtered_data

    logging::loginfo("Running selected transform method: %s", transform)

    if (transform == 'LOG') {
        features <- LOG(features)
    }
    if (transform == 'PLOG') {
        features <- PLOG(features)
    }

    if (!is.null(output)) {
        features_folder <- file.path(output, "features")
        if (!file.exists(features_folder)) {
            logging::loginfo("Creating output feature tables folder")
            dir.create(features_folder, recursive = TRUE)
        }

        filtered_data_norm_transformed_file <-
            file.path(features_folder, "data_transformed.tsv")
        logging::loginfo(
            "Writing normalized, filtered, transformed data to file %s",
            filtered_data_norm_transformed_file
        )
        write.table(
            data.frame("feature" = rownames(features), 
                        features, 
                        check.names = FALSE),
            file = filtered_data_norm_transformed_file,
            sep = "\t",
            quote = FALSE,
            row.names = FALSE
        )
    }

    return(features)
}

##############
# Fit models #
##############

maaslin_fit <- function(filtered_data,
                        transformed_data,
                        metadata,
                        formula,
                        random_effects_formula,
                        feature_specific_covariate = NULL,
                        feature_specific_covariate_name = NULL,
                        feature_specific_covariate_record = NULL,
                        zero_threshold = 0,
                        max_significance = 0.1,
                        correction = 'BH',
                        median_comparison_abundance = TRUE,
                        median_comparison_prevalence = FALSE,
                        median_comparison_abundance_threshold = 0,
                        median_comparison_prevalence_threshold = 0,
                        subtract_median = FALSE,
                        warn_prevalence = TRUE,
                        small_random_effects = FALSE,
                        bypass_small_group_warning = FALSE,
                        augment = TRUE,
                        evaluate_only = NULL,
                        cores = 1,
                        save_models = FALSE,
                        data = NULL,
                        min_abundance = 0,
                        min_prevalence = 0,
                        max_prevalence = 1.01,
                        min_variance = 0) {

    match.arg(correction, correction_choices)

    started_daemons <- FALSE
    if (cores > 1) {
        if (mirai::daemons_set()) {
            warning("Daemons already set; ignoring `cores` argument. ",
                    "Using existing daemons configuration.")
        } else {
            mirai::daemons(cores)
            started_daemons <- TRUE
            on.exit(mirai::daemons(0), add = TRUE)
            logging::loginfo("Started %d mirai daemons via `cores` argument", cores)
        }
    }

    if (!is.null(feature_specific_covariate)) {
        tryCatch({
            feature_specific_covariate <-
                feature_specific_covariate[rownames(filtered_data),
                                        colnames(filtered_data)]
        }, error = function(e) {
            stop(
                "feature_specific_covariate does not contain the
                features and samples of the filtered data."
            )
        })
    }

    # Match variable ignoring case then set correctly as required for p.adjust
    correction <- correction_choices[match(toupper(correction),
                                        toupper(correction_choices))]

    # Initialize fit data objects
    fit_data_abundance <- NULL
    fit_data_prevalence <- NULL

    # Run linear model component
    if (is.null(evaluate_only) || evaluate_only == "abundance") {
        logging::loginfo("Running the linear model component")
        prevalence_mask <- ifelse(!is.na(filtered_data), 1, 0)

        #######################
        # For non-zero models #
        #######################

        fit_data_abundance <-
            fit.model(
                features = transformed_data,
                metadata = metadata,
                model = 'linear',
                formula = formula,
                random_effects_formula = random_effects_formula,
                correction = correction,
                save_models = save_models,
                small_random_effects = small_random_effects,
                augment = augment,
                cores = cores,
                median_comparison = median_comparison_abundance,
                median_comparison_threshold =
                    median_comparison_abundance_threshold,
                subtract_median = subtract_median,
                feature_specific_covariate = feature_specific_covariate,
                feature_specific_covariate_name =
                    feature_specific_covariate_name,
                feature_specific_covariate_record =
                    feature_specific_covariate_record
            )

        #################################################################
        # Count the total values for each feature (untransformed space) #
        #################################################################

        logging::loginfo("Counting total values for each feature")

        fit_data_abundance$results$N <-
            apply(
                fit_data_abundance$results,
                1,
                FUN = function(x)
                    length(prevalence_mask[, x[1]])
            )
        fit_data_abundance$results$N_not_zero <-
            apply(
                fit_data_abundance$results,
                1,
                FUN = function(x)
                    length(which(prevalence_mask[, x[1]] == 1))
            )

    }

    # Run logistic model component
    if (is.null(evaluate_only) || evaluate_only == "prevalence") {
        #####################
        # For binary models #
        #####################

        logging::loginfo("Running the logistic model component")

        prevalence_mask <- ifelse(!is.na(filtered_data), 1, 0)

        fit_data_prevalence <-
            fit.model(
                features = prevalence_mask,
                metadata = metadata,
                model = 'logistic',
                formula = formula,
                random_effects_formula = random_effects_formula,
                correction = correction,
                save_models = save_models,
                small_random_effects = small_random_effects,
                augment = augment,
                cores = cores,
                median_comparison = median_comparison_prevalence,
                median_comparison_threshold =
                    median_comparison_prevalence_threshold,
                subtract_median = subtract_median,
                feature_specific_covariate = feature_specific_covariate,
                feature_specific_covariate_name =
                    feature_specific_covariate_name,
                feature_specific_covariate_record =
                    feature_specific_covariate_record
            )

        logging::loginfo("Counting total values for each feature")

        fit_data_prevalence$results$N <-
            apply(
                fit_data_prevalence$results,
                1,
                FUN = function(x)
                    length(prevalence_mask[, x[1]])
            )
        fit_data_prevalence$results$N_not_zero <-
            apply(
                fit_data_prevalence$results,
                1,
                FUN = function(x)
                    length(which(prevalence_mask[, x[1]] == 1))
            )
    }

    # Check for highly significant likely model misfits
    if (is.null(evaluate_only) || evaluate_only == "prevalence") {
        current_likely_error_subsetter <-
            !is.na(fit_data_prevalence$results$N_not_zero) &
            (
                fit_data_prevalence$results$N_not_zero < 50 &
                    fit_data_prevalence$results$N_not_zero /
                    fit_data_prevalence$results$N < 0.05
            ) &
            ((
                !is.na(fit_data_prevalence$results$coef) &
                    abs(fit_data_prevalence$results$coef) > 15
            ) |
                (
                    !is.na(fit_data_prevalence$results$pval) &
                        fit_data_prevalence$results$pval < 10 ^ -10
                )
            )
        current_errors_for_likely_issues <-
            fit_data_prevalence$results$error[current_likely_error_subsetter]
        fit_data_prevalence$results$error[current_likely_error_subsetter] <-
            data.table::fifelse(
                !is.na(current_errors_for_likely_issues),
                current_errors_for_likely_issues,
                "A large coefficient (>15 in absolute value) or small p-value (< 10^-10) was obtained from a feature present in <5% of samples. Check this is intended."
            )

        current_likely_error_subsetter <-
            !is.na(fit_data_prevalence$results$N_not_zero) &
            (
                fit_data_prevalence$results$N -
                    fit_data_prevalence$results$N_not_zero < 50 &
                    fit_data_prevalence$results$N_not_zero /
                    fit_data_prevalence$results$N > 0.95
            ) &
            ((
                !is.na(fit_data_prevalence$results$coef) &
                    abs(fit_data_prevalence$results$coef) > 15
            ) |
                (
                    !is.na(fit_data_prevalence$results$pval) &
                        fit_data_prevalence$results$pval < 10 ^ -10
                )
            )
        current_errors_for_likely_issues <-
            fit_data_prevalence$results$error[current_likely_error_subsetter]
        fit_data_prevalence$results$error[current_likely_error_subsetter] <-
            data.table::fifelse(
                !is.na(current_errors_for_likely_issues),
                current_errors_for_likely_issues,
                "A large coefficient (>15 in absolute value) or small p-value (< 10^-10) was obtained from a feature present in >95% of samples. Check this is intended."
            )
    } else {
        fit_data_prevalence <- NULL
    }

    results <- add_qvals(fit_data_abundance,
                        fit_data_prevalence,
                        correction)
    if (!is.null(fit_data_abundance)) {
        fit_data_abundance$results <- results[[1]]
    }
    if (!is.null(fit_data_prevalence)) {
        fit_data_prevalence$results <- results[[2]]
    }

    # Warn about prevalence associations induced by abundances changes
    if (warn_prevalence) {
        logging::loginfo("Re-running abundances for warn_prevalence")

        if (!is.null(evaluate_only)) {
            stop_message <- "evaluate_only must be null with warn_prevalence"
            stop(stop_message)
        }

        if (is.null(data)) {
            stop_message <- "For warn_prevalence, data must not be null"
            stop(stop_message)
        }

        new_normalized_data <- maaslin_normalize(data,
                                                NULL,
                                                zero_threshold,
                                                "TSS",
                                                NULL)

        if (is.null(min_abundance) |
            is.null(min_prevalence) |
            is.null(max_prevalence) |
            is.null(min_variance)) {
            stop_message <- paste0("For warn_prevalence, min_abundance, ",
                        "min_prevalence, max_prevalence, and min_variance must not be null")
            stop(stop_message)
        }

        new_filtered_data <- new_normalized_data[rownames(transformed_data),
                                                colnames(transformed_data)]

        new_transformed_data <- maaslin_transform(new_filtered_data,
                                                    NULL,
                                                    'LOG')

        # Run TSS LOG no median comparison model
        new_fit_data_abundance <-
            fit.model(
                features = new_transformed_data,
                metadata = metadata,
                model = 'linear',
                formula = formula,
                random_effects_formula = random_effects_formula,
                correction = correction,
                save_models = save_models,
                small_random_effects = small_random_effects,
                augment = augment,
                cores = cores,
                median_comparison = FALSE,
                median_comparison_threshold = 0,
                subtract_median = FALSE,
                feature_specific_covariate = feature_specific_covariate,
                feature_specific_covariate_name =
                    feature_specific_covariate_name,
                feature_specific_covariate_record =
                    feature_specific_covariate_record)

        results <- add_qvals(new_fit_data_abundance,
                            fit_data_prevalence,
                            correction)
        new_fit_data_abundance$results <- results[[1]]

        results <-
            add_joint_signif(fit_data_abundance,
                            fit_data_prevalence,
                            new_fit_data_abundance,
                            max_significance,
                            correction)

        if (!is.null(fit_data_abundance)) {
            fit_data_abundance$results <- results[[1]]
        }
        if (!is.null(fit_data_prevalence)) {
            fit_data_prevalence$results <- results[[2]]
        }
    } else {
        # Add in joint p/q-values
        if (is.null(evaluate_only)) {
            results <-
                add_joint_signif(fit_data_abundance,
                                fit_data_prevalence,
                                NULL,
                                max_significance,
                                correction)
            if (!is.null(fit_data_abundance)) {
                fit_data_abundance$results <- results[[1]]
            }
            if (!is.null(fit_data_prevalence)) {
                fit_data_prevalence$results <- results[[2]]
            }
        } else if (evaluate_only == 'abundance') {
            fit_data_abundance$results$pval_joint <-
                fit_data_abundance$results$pval
            
            fit_data_abundance$results$qval_joint <-
                fit_data_abundance$results$qval
            
            collapse::setrename(fit_data_abundance$results,
                                "pval_individual" = "pval",
                                "qval_individual" = "qval" )
            
        } else if (evaluate_only == 'prevalence') {
            fit_data_prevalence$results$pval_joint <-
                fit_data_prevalence$results$pval
            fit_data_prevalence$results$qval_joint <-
                fit_data_prevalence$results$qval
            
            collapse::setrename(fit_data_prevalence$results,
                                "pval_individual" = "pval",
                                "qval_individual" = "qval")
        }
    }
    
    if (!is.null(fit_data_prevalence)) {
        if (!is.null(random_effects_formula)) {
            bars <- reformulas::findbars(random_effects_formula)
            random_names <- vapply(bars, function(x) deparse(x[[3]]), 
                FUN.VALUE = character(1))
            for (random_name in random_names) {
                random_table <- table(metadata[random_name])
                random_table <- random_table[random_table > 0]
                if (mean(random_table) < 4 & !small_random_effects &
                    !bypass_small_group_warning) {
                    fit_data_prevalence$results$error <- 
                        data.table::fifelse(is.na(fit_data_prevalence$results$error),
    paste0("<4 average observations per random effect group often inflates ",
        "coefficients and deflates p-values: consider setting ",
        "small_random_effects=TRUE and see tutorial"), 
    fit_data_prevalence$results$error)
                }
            }
        }
    }

    # Set column order
    col_order <- c("feature", "metadata", "value", "name", "coef", 
                    "null_hypothesis", "stderr",
                    "pval_individual", "qval_individual", "pval_joint",
                    "qval_joint", "error", "model", "N", "N_not_zero")

    # Reorder outputs
    if (is.null(evaluate_only) || evaluate_only == "abundance") {
        fit_data_abundance$results <-
            fit_data_abundance$results[
                order(fit_data_abundance$results$qval_individual), ]
        # Move all that had errors to the end
        fit_data_abundance$results <-
            fit_data_abundance$results[
                order(!is.na(fit_data_abundance$results$error)), ]

        # Reorder columns
        fit_data_abundance$results <- fit_data_abundance$results[,col_order]
    } else {
        fit_data_abundance <- NULL
    }

    if (is.null(evaluate_only) || evaluate_only == "prevalence") {
        fit_data_prevalence$results <-
            fit_data_prevalence$results[
                order(fit_data_prevalence$results$qval_individual), ]
        # Move all that had errors to the end
        fit_data_prevalence$results <-
            fit_data_prevalence$results[
                order(!is.na(fit_data_prevalence$results$error)), ]

        # Reorder columns
        fit_data_prevalence$results <- fit_data_prevalence$results[,col_order]
        
    } else {
        fit_data_prevalence <- NULL
    }

    return(
        list(
            "fit_data_abundance" = fit_data_abundance,
            "fit_data_prevalence" = fit_data_prevalence
        )
    )
}

###########################
# Write tables of results #
###########################

maaslin_write_results <- function(output,
                                fit_data_abundance,
                                fit_data_prevalence,
                                random_effects_formula = NULL,
                                max_significance = 0.1,
                                save_models = FALSE) {
    # create an output folder if it does not exist
    if (!file.exists(output)) {
        logging::loginfo("Creating output folder")
        dir.create(output)
    }

    write_fits(
        output,
        fit_data_abundance,
        fit_data_prevalence,
        random_effects_formula,
        save_models
    )

    write_results(output,
                fit_data_abundance,
                fit_data_prevalence,
                max_significance)
}

maaslin_write_results_lefse_format <- function(output,
                                            fit_data_abundance,
                                            fit_data_prevalence) {
    # create an output folder if it does not exist
    if (!file.exists(output)) {
        logging::loginfo("Creating output folder")
        dir.create(output)
    }

    if (!is.null(fit_data_abundance$results)) {
        write_results_in_lefse_format(
            fit_data_abundance$results,
            file.path(output, 'lefse_style_results_abundance.res')
        )
    }
    if (!is.null(fit_data_prevalence$results)) {
        write_results_in_lefse_format(
            fit_data_prevalence$results,
            file.path(output, 'lefse_style_results_prevalence.res')
        )
    }
}

#######################################################
# Create visualizations for results passing threshold #
#######################################################

maaslin_plot_results <- function(output,
                                transformed_data,
                                unstandardized_metadata,
                                fit_data_abundance,
                                fit_data_prevalence,
                                normalization,
                                transform,
                                feature_specific_covariate = NULL,
                                feature_specific_covariate_name = NULL,
                                feature_specific_covariate_record = NULL,
                                median_comparison_abundance = TRUE,
                                median_comparison_prevalence = FALSE,
                                max_significance = 0.1,
                                plot_summary_plot = TRUE,
                                summary_plot_first_n = 25,
                                coef_plot_vars = NULL,
                                heatmap_vars = NULL,
                                plot_associations = TRUE,
                                max_pngs = 30,
                                balanced = FALSE,
                                save_plots_rds = FALSE) {
    ret_plots <- list()
    # create an output folder and figures folder if it does not exist
    if (!file.exists(output)) {
        logging::loginfo("Creating output folder")
        dir.create(output)
    }
    if (plot_summary_plot || plot_associations) {
        figures_folder <- file.path(output, "figures")
        if (!file.exists(figures_folder)) {
            logging::loginfo("Creating output figures folder")
            dir.create(figures_folder)
        }
    }

    if (is.null(fit_data_abundance$results)) {
        merged_results <- fit_data_prevalence$results
    } else if (is.null(fit_data_prevalence$results)) {
        merged_results <- fit_data_abundance$results
    } else {
        merged_results <- rbind(fit_data_abundance$results,
                                fit_data_prevalence$results)
    }

    # Summary plot
    if (plot_summary_plot) {
        summary_plot_file <- file.path(figures_folder, "summary_plot.pdf")
        logging::loginfo("Writing summary plot of significant
                        results to file: %s",
                        summary_plot_file)

        if (!is.null(coef_plot_vars) &
            length(coef_plot_vars) == 1) {
            coef_plot_vars <- trimws(unlist(strsplit(coef_plot_vars, ',')))
        }
        if (!is.null(heatmap_vars) & length(heatmap_vars) == 1) {
            heatmap_vars <- trimws(unlist(strsplit(heatmap_vars, ',')))
        }

        summary_plot <- maaslin3_summary_plot(
            merged_results,
            summary_plot_file,
            figures_folder,
            first_n = summary_plot_first_n,
            max_significance = max_significance,
            coef_plot_vars = coef_plot_vars,
            heatmap_vars = heatmap_vars,
            median_comparison_abundance = median_comparison_abundance,
            median_comparison_prevalence = median_comparison_prevalence,
            balanced = balanced,
            save_plots_rds = save_plots_rds
        )
        ret_plots[["summary_plot"]] <- summary_plot
    }

    # Individual association plots
    if (plot_associations) {
        logging::loginfo(
            paste(
                "Writing association plots",
                "(one for each significant association)",
                "to output folder: %s"
            ),
            figures_folder
        )

        plots_out <- tryCatch({
            withCallingHandlers({
                maaslin3_association_plots(
                    merged_results = merged_results,
                    metadata = unstandardized_metadata,
                    features = transformed_data,
                    max_significance = max_significance,
                    figures_folder = figures_folder,
                    max_pngs = max_pngs,
                    normalization = normalization,
                    transform = transform,
                    feature_specific_covariate = feature_specific_covariate,
                    feature_specific_covariate_name =
                        feature_specific_covariate_name,
                    feature_specific_covariate_record =
                        feature_specific_covariate_record,
                    save_plots_rds = save_plots_rds
                )
            },
            warning = function(w) {
                invokeRestart("muffleWarning")
            })
        })
        ret_plots[["assocation_plots"]] <- plots_out
    }
    return(ret_plots)
}

maaslin_plot_results_from_output <- function(output,
                                            metadata,
                                            normalization,
                                            transform,
                                            feature_specific_covariate =
                                                NULL,
                                            feature_specific_covariate_name =
                                                NULL,
                                            feature_specific_covariate_record =
                                                NULL,
                                            median_comparison_abundance =
                                                TRUE,
                                            median_comparison_prevalence =
                                                FALSE,
                                            max_significance = 0.1,
                                            plot_summary_plot = TRUE,
                                            summary_plot_first_n = 25,
                                            coef_plot_vars = NULL,
                                            heatmap_vars = NULL,
                                            plot_associations = TRUE,
                                            max_pngs = 30,
                                            balanced = FALSE,
                                            save_plots_rds = FALSE) {
    ret_plots <- list()

    # create an output folder and figures folder if it does not exist
    if (!file.exists(output)) {
        logging::loginfo("Creating output folder")
        dir.create(output)
    }
    if (plot_summary_plot || plot_associations) {
        figures_folder <- file.path(output, "figures")
        if (!file.exists(figures_folder)) {
            logging::loginfo("Creating output figures folder")
            dir.create(figures_folder)
        }
    }

    all_results_file <-
        paste0(gsub('/$', '', output), '/', 'all_results.tsv')
    if (!file.exists(all_results_file)) {
        stop(sprintf(
            'Please generate the results file first: %s',
            all_results_file
        ))
    }
    
    merged_results <- data.table::fread(all_results_file, 
                                        sep = '\t') |> 
        as.data.frame()
    
    merged_results$model[merged_results$model == 'abundance'] <- 'linear'
    merged_results$model[merged_results$model == 'prevalence'] <- 'logistic'

    # Summary plot
    if (plot_summary_plot) {
        summary_plot_file <- file.path(figures_folder, "summary_plot.pdf")
        logging::loginfo("Writing summary plot of
                        significant results to file: %s",
                        summary_plot_file)

        if (!is.null(coef_plot_vars) &
            length(coef_plot_vars) == 1) {
            coef_plot_vars <- trimws(unlist(strsplit(coef_plot_vars, ',')))
        }
        if (!is.null(heatmap_vars) & length(heatmap_vars) == 1) {
            heatmap_vars <- trimws(unlist(strsplit(heatmap_vars, ',')))
        }

        summary_plot <- maaslin3_summary_plot(
            merged_results,
            summary_plot_file,
            figures_folder,
            first_n = summary_plot_first_n,
            max_significance = max_significance,
            coef_plot_vars = coef_plot_vars,
            heatmap_vars = heatmap_vars,
            median_comparison_abundance = median_comparison_abundance,
            median_comparison_prevalence = median_comparison_prevalence,
            balanced = balanced,
            save_plots_rds = save_plots_rds
        )
        ret_plots[["summary_plot"]] <- summary_plot
    }

    # Individual associations
    if (plot_associations) {
        features_file <-
            paste0(gsub('/$', '', output),
                '/',
                'features/data_transformed.tsv')
        if (!file.exists(features_file)) {
            stop(sprintf(
                'Please generate the results file first: %s',
                features_file
            ))
        }
        transformed_data <-
            data.table::fread(
                features_file,
                sep = '\t',
                check.names = FALSE
            ) |> 
            as.data.frame()
        
        transformed_data = transformed_data |> 
            collapse::setRownames(transformed_data$feature)
        
        collapse::fselect(transformed_data, "feature") <- NULL

        logging::loginfo(
            paste(
                "Writing association plots",
                "(one for each significant association)",
                "to output folder: %s"
            ),
            figures_folder
        )

        # Need to redo this if not fitting the model
        if (!is.null(feature_specific_covariate)) {
            tryCatch({
                feature_specific_covariate <-
                    feature_specific_covariate[rownames(transformed_data),
                                            colnames(transformed_data)]
            }, error = function(e) {
                stop(
                    "feature_specific_covariate does not contain the features
                    and samples of the filtered data."
                )
            })
        }

        if (missing("normalization") |
            missing("transform") | missing("metadata")) {
            stop(
                "Missing normalization, transform, or metadata argument to
                maaslin_plot_results_from_output"
            )
        }

        plots_out <- tryCatch({
            withCallingHandlers({
                maaslin3_association_plots(
                    merged_results = merged_results,
                    metadata = metadata,
                    features = transformed_data,
                    max_significance = max_significance,
                    figures_folder = figures_folder,
                    max_pngs = max_pngs,
                    normalization = normalization,
                    transform = transform,
                    feature_specific_covariate = feature_specific_covariate,
                    feature_specific_covariate_name =
                        feature_specific_covariate_name,
                    feature_specific_covariate_record =
                        feature_specific_covariate_record,
                    save_plots_rds = save_plots_rds
                )
            },
            warning = function(w) {
                invokeRestart("muffleWarning")
            })
        })
        ret_plots[["assocation_plots"]] <- plots_out
    }

    if ('logging::writeToFile' %in% names(logging::getLogger()[['handlers']])) {
        logging::removeHandler('logging::writeToFile')
    }

    return(ret_plots)
}

##########################################################
# Main maaslin3 function (defaults same as command line) #
##########################################################

maaslin3 <- function(input_data,
                    input_metadata = NULL,
                    output,
                    formula = NULL,
                    fixed_effects = NULL,
                    reference = NULL,
                    random_effects = NULL,
                    group_effects = NULL,
                    ordered_effects = NULL,
                    strata_effects = NULL,
                    feature_specific_covariate = NULL,
                    feature_specific_covariate_name = NULL,
                    feature_specific_covariate_record = NULL,
                    min_abundance = 0,
                    min_prevalence = 0,
                    max_prevalence = 1.01,
                    zero_threshold = 0,
                    min_variance = 0,
                    max_significance = 0.1,
                    normalization = 'TSS',
                    transform = 'LOG',
                    correction = 'BH',
                    standardize = TRUE,
                    unscaled_abundance = NULL,
                    median_comparison_abundance = TRUE,
                    median_comparison_prevalence = FALSE,
                    median_comparison_abundance_threshold = 0,
                    median_comparison_prevalence_threshold = 0,
                    subtract_median = FALSE,
                    warn_prevalence = TRUE,
                    small_random_effects = FALSE,
                    bypass_small_group_warning = FALSE,
                    augment = TRUE,
                    evaluate_only = NULL,
                    plot_summary_plot = TRUE,
                    summary_plot_first_n = 25,
                    coef_plot_vars = NULL,
                    heatmap_vars = NULL,
                    plot_associations = TRUE,
                    max_pngs = 30,
                    cores = 1,
                    save_models = FALSE,
                    save_plots_rds = FALSE,
                    verbosity = 'FINEST',
                    summary_plot_balanced = FALSE,
                    assay.type = 1) {
    match.arg(verbosity, c("FINEST", "FINER", "FINE", "DEBUG", "INFO",
                            "WARN", "ERROR"))
    logging::logReset()
    
    started_daemons <- FALSE
    if (cores > 1) {
        if (mirai::daemons_set()) {
            warning("Daemons already set; ignoring `cores` argument. ",
                    "Using existing daemons configuration.")
        } else {
            mirai::daemons(cores)
            started_daemons <- TRUE
            on.exit(mirai::daemons(0), add = TRUE)
            logging::loginfo("Started %d mirai daemons via `cores` argument", cores)
        }
    }

    # Allow for lower case variables
    normalization <- toupper(normalization)
    transform <- toupper(transform)

    # Match variable ignoring case then set correctly as required for p.adjust
    correction <- correction_choices[match(toupper(correction),
                                        toupper(correction_choices))]

    # If formula is a formula object, convert it back to a string
    if (methods::is(formula, "formula")) {
        formula <- safe_deparse(formula)
    }

    if (inherits(input_data, "SummarizedExperiment")) {
        summarized_experiment_out <-
            maaslin_read_summarized_experiment_data(input_data, assay.type)

        input_data <- summarized_experiment_out[['data']]
        input_metadata <- summarized_experiment_out[['metadata']]
    }

    # Create log file, log arguments, and check arguments
    maaslin_log_arguments(
        input_data,
        input_metadata,
        output,
        formula,
        fixed_effects,
        reference,
        random_effects,
        group_effects,
        ordered_effects,
        strata_effects,
        feature_specific_covariate,
        feature_specific_covariate_name,
        feature_specific_covariate_record,
        min_abundance,
        min_prevalence,
        max_prevalence,
        zero_threshold,
        min_variance,
        max_significance,
        normalization,
        transform,
        correction,
        standardize,
        unscaled_abundance,
        median_comparison_abundance,
        median_comparison_prevalence,
        median_comparison_abundance_threshold,
        median_comparison_prevalence_threshold,
        subtract_median,
        warn_prevalence,
        small_random_effects,
        bypass_small_group_warning,
        augment,
        evaluate_only,
        plot_summary_plot,
        summary_plot_first_n,
        coef_plot_vars,
        heatmap_vars,
        plot_associations,
        max_pngs,
        cores,
        save_models,
        save_plots_rds,
        verbosity,
        summary_plot_balanced
    )

    # Read data in
    read_data_list <- maaslin_read_data(input_data,
                                        input_metadata,
                                        feature_specific_covariate,
                                        unscaled_abundance)

    # Set ordering of samples correctly
    read_data_list <- maaslin_reorder_data(
        read_data_list$data,
        read_data_list$metadata,
        read_data_list$feature_specific_covariate,
        read_data_list$unscaled_abundance
    )

    data <- read_data_list$data
    metadata <- read_data_list$metadata
    unscaled_abundance <- read_data_list$unscaled_abundance
    feature_specific_covariate <-
        read_data_list$feature_specific_covariate

    # Compute or check formula
    if (is.null(formula)) {
        formulas <- maaslin_compute_formula(
            data,
            metadata,
            fixed_effects,
            random_effects,
            group_effects,
            ordered_effects,
            strata_effects,
            feature_specific_covariate_name
        )
    } else {
        if (!is.null(fixed_effects) | 
            !is.null(random_effects) | !is.null(group_effects) | 
            !is.null(ordered_effects) | !is.null(strata_effects)) {
            check_null_error <- paste0("random_effects, group_effects, ",
            "ordered_effects, and strata_effects must be NULL ",
            "when formula is not NULL")
            stop(check_null_error)
        }

        formulas <- maaslin_check_formula(data,
                                        metadata,
                                        formula,
                                        feature_specific_covariate_name)
    }

    formula <- formulas$formula
    random_effects_formula <- formulas$random_effects_formula

    # Normalize data
    normalized_data <- maaslin_normalize(data,
                                        output,
                                        zero_threshold,
                                        normalization,
                                        unscaled_abundance)

    # Filter data
    filtered_data <- maaslin_filter(
        normalized_data,
        output,
        min_abundance,
        min_prevalence,
        max_prevalence,
        zero_threshold,
        min_variance
    )

    # Transform data
    transformed_data <- maaslin_transform(filtered_data,
                                        output,
                                        transform)

    # Process metadata
    standardized_metadata <- maaslin_process_metadata(
        metadata,
        formula,
        fixed_effects,
        reference,
        feature_specific_covariate_name,
        standardize
    )

    # Fit models
    maaslin_results <- maaslin_fit(
        filtered_data,
        transformed_data,
        standardized_metadata,
        formula,
        random_effects_formula,
        feature_specific_covariate,
        feature_specific_covariate_name,
        feature_specific_covariate_record,
        zero_threshold,
        max_significance,
        correction,
        median_comparison_abundance,
        median_comparison_prevalence,
        median_comparison_abundance_threshold,
        median_comparison_prevalence_threshold,
        subtract_median,
        warn_prevalence,
        small_random_effects,
        bypass_small_group_warning,
        augment,
        evaluate_only,
        cores = 1,
        save_models = TRUE,
        data,
        min_abundance,
        min_prevalence,
        max_prevalence,
        min_variance
    )

    # Write results
    maaslin_write_results(
        output,
        maaslin_results$fit_data_abundance,
        maaslin_results$fit_data_prevalence,
        random_effects_formula,
        max_significance,
        save_models
    )

    # Plot outputs
    if (plot_summary_plot | plot_associations) {
        tryCatch({
            withCallingHandlers({
                maaslin_plot_results(
                    output,
                    transformed_data,
                    metadata,
                    maaslin_results$fit_data_abundance,
                    maaslin_results$fit_data_prevalence,
                    normalization,
                    transform,
                    feature_specific_covariate,
                    feature_specific_covariate_name,
                    feature_specific_covariate_record,
                    median_comparison_abundance,
                    median_comparison_prevalence,
                    max_significance,
                    plot_summary_plot,
                    summary_plot_first_n,
                    coef_plot_vars,
                    heatmap_vars,
                    plot_associations,
                    max_pngs,
                    summary_plot_balanced,
                    save_plots_rds
                )
            },
            warning = function(w) {
                invokeRestart("muffleWarning")
            })
        })
    }

    if ('logging::writeToFile' %in% names(logging::getLogger()[['handlers']])) {
        logging::removeHandler('logging::writeToFile')
    }

    return(
        list(
            "data" = data,
            "normalized_data" = normalized_data,
            "filtered_data" = filtered_data,
            "transformed_data" = transformed_data,
            "metadata" = metadata,
            "standardized_metadata" = standardized_metadata,
            "formula" = formulas,
            "fit_data_abundance" = maaslin_results$fit_data_abundance,
            "fit_data_prevalence" = maaslin_results$fit_data_prevalence
        )
    )
}

###########################################################################
# If running on the command line, get arguments and call maaslin function #
###########################################################################

if (cmd_line_chk) {

    # get command line options and positional arguments
    parsed_arguments <- optparse::parse_args(options,
                                            positional_arguments = TRUE)
    current_args <- parsed_arguments[["options"]]
    positional_args <- parsed_arguments[["args"]]
    # check three positional arguments are provided
    if (length(positional_args) != 3) {
        optparse::print_help(options)
        stop(
            sprintf(
                "Please provide the required positional arguments
                <data.tsv> <metadata.tsv> <output_folder>"
            )
        )
    }
    
    cores_val <- tryCatch(as.integer(current_args$cores),
             warning = function(err) {
                 msg <- paste("Command line cores argument not parseable as",
                             "integer, will attempt to source()")
                 message(msg)
                 source(current_args$cores)
                 return(1L)
             })
    
    if (!is.integer(cores_val)) cores_val <- 1L
    current_args$cores <- cores_val

    # call maaslin with the command line options
    fit_data <-
        maaslin3(
            input_data = positional_args[1],
            input_metadata = positional_args[2],
            output = positional_args[3],
            min_abundance = current_args$min_abundance,
            zero_threshold = current_args$zero_threshold,
            min_prevalence = current_args$min_prevalence,
            max_prevalence = current_args$max_prevalence,
            min_variance = current_args$min_variance,
            max_significance = current_args$max_significance,
            normalization = current_args$normalization,
            transform = current_args$transform,
            random_effects = current_args$random_effects,
            fixed_effects = current_args$fixed_effects,
            group_effects = current_args$group_effects,
            ordered_effects = current_args$ordered_effects,
            strata_effects = current_args$strata_effects,
            feature_specific_covariate =
                current_args$feature_specific_covariate,
            feature_specific_covariate_name =
                current_args$feature_specific_covariate_name,
            feature_specific_covariate_record =
                current_args$feature_specific_covariate_record,
            median_comparison_abundance =
                current_args$median_comparison_abundance,
            median_comparison_prevalence =
                current_args$median_comparison_prevalence,
            median_comparison_abundance_threshold =
                current_args$median_comparison_abundance_threshold,
            median_comparison_prevalence_threshold =
                current_args$median_comparison_prevalence_threshold,
            subtract_median = current_args$subtract_median,
            warn_prevalence = current_args$warn_prevalence,
            small_random_effects = current_args$small_random_effects,
            bypass_small_group_warning = current_args$bypass_small_group_warning,
            formula = current_args$formula,
            correction = current_args$correction,
            standardize = current_args$standardize,
            cores = current_args$cores,
            plot_summary_plot = current_args$plot_summary_plot,
            summary_plot_first_n = current_args$summary_plot_first_n,
            coef_plot_vars = current_args$coef_plot_vars,
            heatmap_vars = current_args$heatmap_vars,
            plot_associations = current_args$plot_associations,
            max_pngs = current_args$max_pngs,
            save_models = current_args$save_models,
            save_plots_rds = current_args$save_plots_rds,
            augment = current_args$augment,
            evaluate_only = current_args$evaluate_only,
            reference = current_args$reference,
            unscaled_abundance = current_args$unscaled_abundance,
            summary_plot_balanced = current_args$summary_plot_balanced
        )
}

library(testthat)
library(maaslin3)

# The optimizers and optCtrlList vectors are passed to mirai daemons via
# everywhere(). Verify they exist in the package namespace.

test_that("optimizers and optCtrlList are accessible via maaslin3 namespace", {
    expect_true(is.character(maaslin3:::optimizers))
    expect_equal(length(maaslin3:::optimizers), 4)
    expect_true(is.list(maaslin3:::optCtrlList))
    expect_equal(length(maaslin3:::optCtrlList), 4)
})

test_that("logistic random effects model completes without error", {
    set.seed(42)
    n <- 60
    data_in <- data.frame(
        a = rpois(n, lambda = 5),
        b = rpois(n, lambda = 3)
    )
    rownames(data_in) <- paste0("sample", 1:n)
    
    metadata <- data.frame(
        var1 = rnorm(n),
        group = factor(rep(paste0("g", 1:6), each = 10))
    )
    rownames(metadata) <- paste0("sample", 1:n)
    
    data_in_tss <- data.frame(t(apply(data_in, MARGIN = 1,
                                      FUN = function(x) x / sum(x))))
    data_in_tss[data_in_tss == 0] <- NA
    data_in_tss_log <- log2(data_in_tss)
    
    results <- maaslin_fit(data_in_tss,
                           data_in_tss_log,
                           metadata,
                           formula = formula('expr ~ var1 + (1|group)'),
                           random_effects_formula = formula('expr ~ (1|group)'),
                           min_abundance = 0,
                           min_prevalence = 0,
                           min_variance = 0,
                           data = data_in,
                           median_comparison_abundance = FALSE,
                           median_comparison_prevalence = FALSE)
    
    expect_true(!is.null(results$fit_data_abundance))
    expect_true(!is.null(results$fit_data_prevalence))
    expect_true(nrow(results$fit_data_abundance$results) > 0)
    expect_true(nrow(results$fit_data_prevalence$results) > 0)
})

library(testthat)
library(maaslin3)

output_tmp <- tempfile()
data_in <- data.frame('a' = c(0, 0, 0, 0, 0, 0, 0, 0, 7, 5, 8, 4, 7, 8, 5, 8), 
                      'b' = c(2, 0, 0, 5, 0, 5, 3, 6, 5, 2, 6, 9, 0, 8, 3, 7),
                      'c' = c(3, 4, 0, 0, 2, 7, 0, 3, 3, 7, 7, 2, 7, 4, 7, 0))
rownames(data_in) <- paste0("sample", seq(nrow(data_in)))

metadata <- data.frame('var1' = c(rep(0, 8), rep(1, 7), 2),
                       'var2' = c(0, 1, 0, 1, 0, 1, 0, 1, 
                                  0, 1, 0, 1, 0, 1, 0, 1))
rownames(metadata) <- paste0("sample", seq(nrow(data_in)))

data_in_tss <- data.frame(t(apply(data_in, MARGIN = 1, 
                                  FUN = function(x){x / sum(x)})))
data_in_tss[data_in_tss == 0] <- NA
data_in_tss_log <- log2(data_in_tss)

results <- maaslin_fit(data_in_tss,
                       data_in_tss_log,
                       metadata,
                       formula = formula('expr ~ var1 + var2'),
                       random_effects_formula = NULL, 
                       min_abundance = 0, 
                       min_prevalence = 0, 
                       min_variance = 0,
                       data = data_in, 
                       median_comparison_abundance = FALSE)

plot_out <- maaslin_plot_results(output = output_tmp,
                     transformed_data = data_in_tss_log,
                     unstandardized_metadata = metadata,
                     fit_data_abundance = results$fit_data_abundance, 
                     fit_data_prevalence = results$fit_data_prevalence, 
                     normalization = 'TSS',
                     transform = 'LOG',
                     median_comparison_abundance = FALSE,
                     max_significance = 0.1,
                     save_plots_rds = TRUE)

expect_is(plot_out$assocation_plots[[1]], 'ggplot')

expect_equal(names(plot_out$assocation_plots)[1],
             "var1_a_logistic")

expect_equal(list.files(file.path(output_tmp, 'figures', 
                                  'association_plots', 'var1', 'logistic')),
             'var1_a_logistic.png')

expect_equal(sort(list.files(file.path(output_tmp, 'figures'))),
             sort(c("association_plots", "summary_plot_gg.RDS", "summary_plot.pdf", "summary_plot.png")))

unlink(output_tmp, recursive = T)

library(testthat)
library(maaslin3)

expect_error(
    maaslin_check_arguments(unscaled_abundance = data.frame('total' = c(1,2,3)), 
                            median_comparison_abundance = TRUE)
)

expect_error(
    maaslin_check_arguments(feature_specific_covariate = NULL,
                            feature_specific_covariate_name = "something", 
                            feature_specific_covariate_record = NULL)
)

expect_error(
    maaslin_check_arguments(normalization = 'nonsense')
)

expect_error(
    maaslin_check_arguments(normalization = 'NONE',
                            unscaled_abundance = data.frame('total' = c(1,2,3)),
                            median_comparison_abundance = FALSE)
)

expect_error(
    maaslin_check_arguments(transform = 'nonsense')
)

expect_error(
    maaslin_check_arguments(correction = 'nonsense')
)

expect_error(
    maaslin_check_arguments(normalization = 'CLR',
                            transform = 'LOG')
)

expect_error(
    maaslin_check_arguments(transform = 'PLOG',
                            zero_threshold = 0.1)
)

expect_error(
    maaslin_check_arguments(normalization = 'TSS',
                            transform = 'NONE')
)

expect_error(
    maaslin_check_arguments(evaluate_only = 'abundance')
)

expect_error(
    maaslin_check_arguments(evaluate_only = 'nonsense')
)

expect_error(
    maaslin_check_arguments(transform = 'PLOG')
)

expect_that(maaslin_check_arguments(), equals(NULL))

expect_that(maaslin_check_arguments(normalization = 'TSS', 
                                    transform = 'LOG', 
                                    zero_threshold = 0.5,
                                    median_comparison_abundance = TRUE), 
            equals(NULL))

library(testthat)
library(maaslin3)

set.seed(42)

data_in <- data.frame(
    'a' = c(0, 3, 5, 7, 0, 8, 4, 7, 2, 6, 3, 5),
    'b' = c(2, 0, 5, 0, 5, 2, 6, 9, 0, 4, 7, 1),
    'c' = c(3, 4, 0, 7, 2, 7, 4, 7, 0, 5, 8, 3)
)
rownames(data_in) <- paste0("sample", 1:12)

metadata <- data.frame(
    'var1' = rnorm(12),
    'group_var' = rep(paste0("g", 1:4), each = 3)
)
rownames(metadata) <- paste0("sample", 1:12)

data_in_tss <- data.frame(t(apply(
    data_in, MARGIN = 1,
    FUN = function(x) { x / sum(x) }
)))
data_in_tss[data_in_tss == 0] <- NA
data_in_tss_log <- log2(data_in_tss)

formula_obj <- formula('expr ~ var1 + (1 | group_var)')
random_effects_formula <- formula('expr ~ (1 | group_var)')

small_group_msg <- paste0(
    "<4 average observations per random effect group often inflates ",
    "coefficients and deflates p-values: consider setting ",
    "small_random_effects=TRUE and see tutorial")

test_that("small group warning appears when bypass is FALSE", {
    results <- maaslin_fit(
        data_in_tss,
        data_in_tss_log,
        metadata,
        formula = formula_obj,
        random_effects_formula = random_effects_formula,
        min_abundance = 0,
        min_prevalence = 0,
        min_variance = 0,
        data = data_in,
        median_comparison_abundance = FALSE,
        bypass_small_group_warning = FALSE
    )

    expect_true(
        any(grepl("<4 average observations per random effect",
                  results$fit_data_prevalence$results$error))
    )
})

test_that("small group warning is bypassed when bypass is TRUE", {
    results <- maaslin_fit(
        data_in_tss,
        data_in_tss_log,
        metadata,
        formula = formula_obj,
        random_effects_formula = random_effects_formula,
        min_abundance = 0,
        min_prevalence = 0,
        min_variance = 0,
        data = data_in,
        median_comparison_abundance = FALSE,
        bypass_small_group_warning = TRUE
    )

    has_small_group_warning <- any(grepl(
        "<4 average observations per random effect",
        results$fit_data_prevalence$results$error
    ))
    expect_false(has_small_group_warning)
})

library(testthat)
library(maaslin3)

expected_results_run1 <- read.table("expected_results_run1.tsv", header = TRUE, stringsAsFactors=FALSE, sep="\t")

taxa_table <- read.table(system.file(package="maaslin3","extdata","HMP2_taxonomy.tsv"), header = TRUE, sep="\t", row.names = 1)
metadata <- read.table(system.file(package="maaslin3","extdata","HMP2_metadata.tsv"), header = TRUE, sep="\t", row.names = 1)

metadata$diagnosis <- factor(metadata$diagnosis, levels = c('nonIBD', 'UC', 'CD'))
metadata$dysbiosis_state <- factor(metadata$dysbiosis_state, levels = c('none', 'dysbiosis_UC', 'dysbiosis_CD'))
metadata$antibiotics <- factor(metadata$antibiotics, levels = c('No', 'Yes'))

# Run MaAsLin 3
output_tmp <- tempfile()
set.seed(1)
fit_out <- maaslin3(input_data = taxa_table, 
                    input_metadata = metadata, 
                    output = output_tmp, 
                    normalization = 'TSS', 
                    transform = 'LOG', 
                    formula = '~ diagnosis + dysbiosis_state + antibiotics + age + reads', 
                    save_models = FALSE, 
                    plot_summary_plot = FALSE, 
                    plot_associations = FALSE, 
                    max_significance = 0.1, 
                    augment = TRUE, 
                    median_comparison_abundance = TRUE, 
                    median_comparison_prevalence = FALSE, 
                    cores=1, 
                    verbosity = 'ERROR')

maaslin_results = read.table(file.path(output_tmp, "significant_results.tsv"), 
                             header = TRUE, 
                             stringsAsFactors = FALSE)

expect_that(expected_results_run1$metadata[1:50],equals(maaslin_results$metadata[1:50]))
expect_that(expected_results_run1$feature[1:50],equals(maaslin_results$feature[1:50]))
expect_that(round(expected_results_run1$N[1:50],10),equals(round(maaslin_results$N[1:50],10)))
expect_that(round(as.numeric(expected_results_run1$pval_individual[1:50]),10),
            equals(round(as.numeric(maaslin_results$pval_individual[1:50]),10)))
expect_that(round(as.numeric(expected_results_run1$qval_individual[1:50]),10),
            equals(round(as.numeric(maaslin_results$qval_individual[1:50]),10)))

se <- SummarizedExperiment::SummarizedExperiment(
    assays = list(taxa_table = t(taxa_table)),
    colData = metadata
)

output_tmp <- tempfile()
fit_out <- maaslin3(input_data = se, 
                    input_metadata = metadata, 
                    output = output_tmp, 
                    normalization = 'TSS', 
                    transform = 'LOG', 
                    formula = '~ diagnosis + dysbiosis_state + antibiotics + age + reads', 
                    save_models = FALSE, 
                    plot_summary_plot = FALSE, 
                    plot_associations = FALSE, 
                    max_significance = 0.1, 
                    augment = TRUE, 
                    median_comparison_abundance = TRUE, 
                    median_comparison_prevalence = FALSE, 
                    cores=1, 
                    verbosity = 'ERROR')

se_multi <- SummarizedExperiment::SummarizedExperiment(
    assays = list(taxa_table_junk = matrix(0, nrow = ncol(taxa_table), ncol = nrow(taxa_table)),
                  another_taxa_table = t(taxa_table)),
    colData = metadata
)

metadata_df <- as(metadata, "DataFrame")
output_tmp <- tempfile()
fit_out <- maaslin3(input_data = se_multi, 
                    input_metadata = metadata_df, 
                    output = output_tmp, 
                    normalization = 'TSS', 
                    transform = 'LOG', 
                    formula = '~ diagnosis + dysbiosis_state + antibiotics + age + reads', 
                    save_models = FALSE, 
                    plot_summary_plot = FALSE, 
                    plot_associations = FALSE, 
                    max_significance = 0.1, 
                    augment = TRUE, 
                    median_comparison_abundance = TRUE, 
                    median_comparison_prevalence = FALSE, 
                    cores=1, 
                    verbosity = 'ERROR',
                    assay.type = 'another_taxa_table')

output_tmp <- tempfile()
fit_out <- maaslin3(input_data = se_multi, 
                    input_metadata = metadata, 
                    output = output_tmp, 
                    normalization = 'TSS', 
                    transform = 'LOG', 
                    formula = '~ diagnosis + dysbiosis_state + antibiotics + age + reads', 
                    save_models = FALSE, 
                    plot_summary_plot = FALSE, 
                    plot_associations = FALSE, 
                    max_significance = 0.1, 
                    augment = TRUE, 
                    median_comparison_abundance = TRUE, 
                    median_comparison_prevalence = FALSE, 
                    cores=1, 
                    verbosity = 'ERROR',
                    assay.type = 2)

output_tmp <- tempfile()
set.seed(1)
fit_out <- maaslin3(input_data = taxa_table, 
                    input_metadata = metadata, 
                    output = output_tmp, 
                    normalization = 'TSS', 
                    transform = 'LOG', 
                    formula = '~ diagnosis + dysbiosis_state + antibiotics + age + reads', 
                    save_models = FALSE, 
                    plot_summary_plot = FALSE, 
                    plot_associations = FALSE, 
                    max_significance = 0.1, 
                    augment = TRUE, 
                    median_comparison_abundance = TRUE, 
                    median_comparison_prevalence = FALSE, 
                    cores=1, 
                    verbosity = 'ERROR')

# Weird names
taxa_table <- read.table(system.file(package="maaslin3","extdata","HMP2_taxonomy.tsv"), header = TRUE, sep="\t", row.names = 1)
metadata <- read.table(system.file(package="maaslin3","extdata","HMP2_metadata.tsv"), header = TRUE, sep="\t", row.names = 1)

metadata$diagnosis <- factor(metadata$diagnosis, levels = c('nonIBD', 'UC', 'CD'))
metadata$antibiotics <- factor(metadata$antibiotics, levels = c('No', 'Yes'))

colnames(metadata)[colnames(metadata) == 'dysbiosis_state'] <- c("dysbiosis state")
colnames(metadata)[colnames(metadata) == 'red_meat'] <- c("red meat")
metadata$`red meat` <- factor(metadata$`red meat`)
rownames(metadata) <- gsub('_', ' ', rownames(metadata))
metadata$`dysbiosis state` <- gsub('_', ' ', as.character(metadata$`dysbiosis state`))
metadata$`dysbiosis state` <- factor(metadata$`dysbiosis state`, levels = c('none', 'dysbiosis UC', 'dysbiosis CD'))

rownames(taxa_table) <- gsub('_', ' ', rownames(taxa_table))
colnames(taxa_table) <- gsub('_', ' ', colnames(taxa_table))

output_tmp <- tempfile()
set.seed(1)
fit_out <- maaslin3(input_data = taxa_table, 
                    input_metadata = metadata, 
                    output = output_tmp, 
                    normalization = 'TSS', 
                    transform = 'LOG', 
                    formula = '~ ordered(`red meat`)', 
                    save_models = FALSE, 
                    plot_summary_plot = FALSE, 
                    plot_associations = FALSE, 
                    max_significance = 0.1, 
                    augment = TRUE, 
                    median_comparison_abundance = FALSE, 
                    median_comparison_prevalence = FALSE, 
                    cores=1, 
                    verbosity = 'ERROR')

output_tmp <- tempfile()
fit_out2 <- maaslin3(input_data = taxa_table, 
                    input_metadata = metadata, 
                    output = output_tmp, 
                    normalization = 'TSS', 
                    transform = 'LOG', 
                    ordered_effects = 'red meat',
                    save_models = FALSE, 
                    plot_summary_plot = FALSE, 
                    plot_associations = FALSE, 
                    max_significance = 0.1, 
                    augment = TRUE, 
                    median_comparison_abundance = FALSE, 
                    median_comparison_prevalence = FALSE, 
                    cores=1, 
                    verbosity = 'ERROR')

expect_that(fit_out$fit_data_abundance$results$coef,
            equals(fit_out2$fit_data_abundance$results$coef))


output_tmp <- tempfile()
fit_out <- maaslin3(input_data = taxa_table, 
                    input_metadata = metadata, 
                    output = output_tmp, 
                    normalization = 'TSS', 
                    transform = 'LOG', 
                    formula = '~ group(`red meat`)', 
                    save_models = FALSE, 
                    plot_summary_plot = FALSE, 
                    plot_associations = FALSE, 
                    max_significance = 0.1, 
                    augment = TRUE, 
                    median_comparison_abundance = FALSE, 
                    median_comparison_prevalence = FALSE, 
                    cores=1, 
                    verbosity = 'ERROR')

output_tmp <- tempfile()
fit_out2 <- maaslin3(input_data = taxa_table, 
                     input_metadata = metadata, 
                     output = output_tmp, 
                     normalization = 'TSS', 
                     transform = 'LOG', 
                     group_effects = 'red meat',
                     save_models = FALSE, 
                     plot_summary_plot = FALSE, 
                     plot_associations = FALSE, 
                     max_significance = 0.1, 
                     augment = TRUE, 
                     median_comparison_abundance = FALSE, 
                     median_comparison_prevalence = FALSE, 
                     cores=1, 
                     verbosity = 'ERROR')

expect_that(fit_out$fit_data_abundance$results$coef,
            equals(fit_out2$fit_data_abundance$results$coef))

output_tmp <- tempfile()
fit_out <- maaslin3(input_data = taxa_table, 
                    input_metadata = metadata, 
                    output = output_tmp, 
                    normalization = 'TSS', 
                    transform = 'LOG', 
                    formula = '~ diagnosis + `dysbiosis state` + antibiotics + age + reads', 
                    save_models = FALSE, 
                    plot_summary_plot = FALSE, 
                    plot_associations = FALSE, 
                    max_significance = 0.1, 
                    augment = TRUE, 
                    median_comparison_abundance = TRUE, 
                    median_comparison_prevalence = FALSE, 
                    cores=1, 
                    verbosity = 'ERROR')

output_tmp <- tempfile()
expect_error(maaslin3(input_data = taxa_table, 
    input_metadata = metadata, 
    output = output_tmp, 
    normalization = 'TSS', 
    transform = 'LOG', 
    formula = '~ diagnosis + `dysbiosis state` + antibiotics + age + reads', 
    fixed_effects = 'diagnosis',
    save_models = FALSE, 
    plot_summary_plot = FALSE, 
    plot_associations = FALSE, 
    max_significance = 0.1, 
    augment = TRUE, 
    median_comparison_abundance = TRUE, 
    median_comparison_prevalence = FALSE, 
    cores=1, 
    verbosity = 'ERROR'))

library(testthat)
library(maaslin3)

output_tmp <- tempfile()
mirai::daemons(0)
maaslin_log_arguments(input_data = 'something1', 
                      input_metadata = 'something2',
                      output = output_tmp, 
                      formula = 'something3', 
                      fixed_effects = 'something4', 
                      reference = 'something5', 
                      random_effects = 'something6', 
                      group_effects = 'something7',
                      ordered_effects = 'something8', 
                      strata_effects = 'something9',
                      feature_specific_covariate = 'something10',
                      feature_specific_covariate_name = 'something11',
                      feature_specific_covariate_record = 'something12',
                      min_abundance = 0,
                      min_prevalence = 1,
                      zero_threshold = 2,
                      min_variance = 3,
                      max_significance = 4,
                      normalization = 'TSS',
                      transform = 'LOG',
                      correction = 'BH',
                      standardize = TRUE,
                      unscaled_abundance = 'something16',
                      median_comparison_abundance = FALSE,
                      median_comparison_prevalence = FALSE,
                      median_comparison_abundance_threshold = 5,
                      median_comparison_prevalence_threshold = 6,
                      subtract_median = FALSE,
                      warn_prevalence = TRUE,
                      small_random_effects = FALSE,
                      bypass_small_group_warning = FALSE,
                      augment = TRUE,
                      evaluate_only = NULL,
                      plot_summary_plot = TRUE,
                      summary_plot_first_n = 7,
                      coef_plot_vars = 'something18',
                      heatmap_vars = 'something19',
                      plot_associations = TRUE,
                      max_pngs = 8,
                      cores = 1,
                      save_models = FALSE,
                      verbosity = 'FINEST',
                      summary_plot_balanced=FALSE)

lines_in <- readLines(file.path(output_tmp, 'maaslin3.log'))
lines_in <- sub('.*::', '', lines_in)

lines_to_compare <- c("Writing function arguments to log file",
                        "Function arguments",
                        "Input data file: something1",
                        "Input metadata file: something2",
                        "Output folder:",
                        "Formula: something3",
                        "Fixed effects: something4",
                        "Reference: something5",
                        "Random effects: something6",
                        "Group effects: something7",
                        "Ordered effects: something8",
                        "Strata effects: something9",
                        "Feature specific covariate: something10",
                        "Feature specific covariate name: something11",
                        "Feature specific covariate include: something12",
                        "Min Abundance: 0",
                        "Min Prevalence: 1",
                        "Max Prevalence: 1",
                        "Zero Threshold: 2",
                        "Min variance: 3",
                        "Max significance: 4",
                        "Normalization: TSS",
                        "Transform: LOG",
                        "Correction method: BH",
                        "Standardize: TRUE",
                        "Unscaled abundance: something16",
                        "Abundance median comparison: FALSE",
                        "Prevalence median comparison: FALSE",
                        "Abundance median comparison threshold: 5",
                        "Prevalence median comparison threshold: 6",
                        "Subtract median: FALSE",
                        "Warn prevalence: TRUE",
                        "Small random effects: FALSE",
                        "Bypass small group warning: FALSE",
                        "Augment: TRUE",
                        "Evaluate only:",
                        "Number of mirai daemons: 0",
                        "Balanced Summary plot: FALSE",
                        "Verifying options selected are valid")

expect_starts_with <- function(strings, prefixes) {
    expect_equal(sapply(seq_along(strings), function(i) {
        startsWith(strings[i], prefixes[i])
    }), rep(TRUE, length(strings)))
}

expect_starts_with(lines_in, lines_to_compare)

logging::logReset()
unlink(output_tmp, recursive = T)

library(testthat)
library(maaslin3)

output_tmp <- tempfile()
data_in <- data.frame('a' = c(0, 0, 0, 0, 0, 0, 0, 0, 7, 5, 8, 4, 7, 8, 5, 8), 
                      'b' = c(2, 0, 0, 5, 0, 5, 3, 6, 5, 2, 6, 9, 0, 8, 3, 7),
                      'c' = c(3, 4, 0, 0, 2, 7, 0, 3, 3, 7, 7, 2, 7, 4, 7, 0))
rownames(data_in) <- paste0("sample", seq(nrow(data_in)))

metadata <- data.frame('var1' = c(rep(0, 8), rep(1, 7), 2),
                       'var2' = c(0, 1, 0, 1, 0, 1, 0, 1, 
                                  0, 1, 0, 1, 0, 1, 0, 1))
rownames(metadata) <- paste0("sample", seq(nrow(data_in)))

data_in_tss <- data.frame(t(apply(data_in, MARGIN = 1, 
                                  FUN = function(x){x / sum(x)})))
data_in_tss[data_in_tss == 0] <- NA
data_in_tss_log <- log2(data_in_tss)

results <- maaslin_fit(data_in_tss,
                       data_in_tss_log,
                       metadata,
                       formula = formula('expr ~ var1 + var2'),
                       random_effects_formula = NULL, 
                       min_abundance = 0, 
                       min_prevalence = 0, 
                       min_variance = 0,
                       data = data_in, 
                       median_comparison_abundance = FALSE)

maaslin_write_results(output = output_tmp,
                      fit_data_abundance = results$fit_data_abundance, 
                      fit_data_prevalence = results$fit_data_prevalence, 
                      random_effects_formula = NULL, 
                      max_significance = 0.1)

results_combined <- rbind(results$fit_data_abundance$results,
                          results$fit_data_prevalence$results)
results_combined <- results_combined[order(results_combined$qval_individual, 
                                           na.last = T),]
results_combined$model <- ifelse(results_combined$model == 'linear',
                                 'abundance',
                                 'prevalence')
rownames(results_combined) <- NULL
all_results <- read.csv(file.path(output_tmp, 'all_results.tsv'), sep = '\t')
rownames(all_results) <- NULL
all_results$error <- as.character(all_results$error)
results_combined$error <- as.character(results_combined$error)

expect_equal(results_combined, all_results)

signif_results <- read.csv(file.path(output_tmp, 'significant_results.tsv'), sep = '\t')
rownames(signif_results) <- NULL
results_combined <- results_combined[results_combined$qval_joint < 0.1 | 
                                    results_combined$qval_individual < 0.1,]
rownames(results_combined) <- NULL

results_combined$error <- as.logical(results_combined$error)
expect_equal(results_combined, signif_results)

unlink(output_tmp, recursive = T)

library(testthat)
library(maaslin3)

setup_test_data <- function() {
    set.seed(123)
    n <- 30
    data_in <- data.frame(
        a = rpois(n, lambda = 5),
        b = rpois(n, lambda = 3),
        c = rpois(n, lambda = 4)
    )
    rownames(data_in) <- paste0("sample", 1:n)

    metadata <- data.frame(
        var1 = rnorm(n),
        var2 = c(rep(0, 15), rep(1, 15))
    )
    rownames(metadata) <- paste0("sample", 1:n)

    data_in_tss <- data.frame(t(apply(data_in, MARGIN = 1,
                                      FUN = function(x) x / sum(x))))
    data_in_tss[data_in_tss == 0] <- NA
    data_in_tss_log <- log2(data_in_tss)

    list(data_in = data_in, metadata = metadata,
         data_in_tss = data_in_tss, data_in_tss_log = data_in_tss_log)
}

test_that("serial execution (cores=1, no daemons) produces results", {
    if (mirai::daemons_set()) mirai::daemons(0)
    Sys.sleep(0.5)

    td <- setup_test_data()
    results <- maaslin_fit(td$data_in_tss,
                           td$data_in_tss_log,
                           td$metadata,
                           formula = formula('expr ~ var1 + var2'),
                           random_effects_formula = NULL,
                           min_abundance = 0,
                           min_prevalence = 0,
                           min_variance = 0,
                           data = td$data_in,
                           cores = 1,
                           median_comparison_abundance = FALSE,
                           median_comparison_prevalence = FALSE)

    expect_false(mirai::daemons_set())
    expect_true(!is.null(results$fit_data_abundance))
    expect_true(!is.null(results$fit_data_prevalence))
    expect_true(nrow(results$fit_data_abundance$results) > 0)
    expect_true(nrow(results$fit_data_prevalence$results) > 0)
})

test_that("cores parameter starts daemons and cleans up", {
    if (mirai::daemons_set()) mirai::daemons(0)
    Sys.sleep(0.5)

    td <- setup_test_data()
    results <- maaslin_fit(td$data_in_tss,
                           td$data_in_tss_log,
                           td$metadata,
                           formula = formula('expr ~ var1 + var2'),
                           random_effects_formula = NULL,
                           min_abundance = 0,
                           min_prevalence = 0,
                           min_variance = 0,
                           data = td$data_in,
                           cores = 2,
                           median_comparison_abundance = FALSE,
                           median_comparison_prevalence = FALSE)

    expect_false(mirai::daemons_set())
    expect_true(!is.null(results$fit_data_abundance))
    expect_true(!is.null(results$fit_data_prevalence))
    expect_true(nrow(results$fit_data_abundance$results) > 0)
    expect_true(nrow(results$fit_data_prevalence$results) > 0)
})

test_that("pre-set daemons are used and preserved", {
    if (mirai::daemons_set()) mirai::daemons(0)
    Sys.sleep(0.5)
    mirai::daemons(2)
    Sys.sleep(0.5)

    expect_true(mirai::daemons_set())

    td <- setup_test_data()
    results <- maaslin_fit(td$data_in_tss,
                           td$data_in_tss_log,
                           td$metadata,
                           formula = formula('expr ~ var1 + var2'),
                           random_effects_formula = NULL,
                           min_abundance = 0,
                           min_prevalence = 0,
                           min_variance = 0,
                           data = td$data_in,
                           cores = 1,
                           median_comparison_abundance = FALSE,
                           median_comparison_prevalence = FALSE)

    expect_true(mirai::daemons_set())
    expect_true(!is.null(results$fit_data_abundance))
    expect_true(!is.null(results$fit_data_prevalence))
    expect_true(nrow(results$fit_data_abundance$results) > 0)
    expect_true(nrow(results$fit_data_prevalence$results) > 0)

    mirai::daemons(0)
    Sys.sleep(0.5)
})

test_that("cores>1 warns when daemons already set", {
    if (mirai::daemons_set()) mirai::daemons(0)
    Sys.sleep(0.5)
    mirai::daemons(2)
    Sys.sleep(0.5)

    td <- setup_test_data()
    expect_warning(
        maaslin_fit(td$data_in_tss,
                    td$data_in_tss_log,
                    td$metadata,
                    formula = formula('expr ~ var1 + var2'),
                    random_effects_formula = NULL,
                    min_abundance = 0,
                    min_prevalence = 0,
                    min_variance = 0,
                    data = td$data_in,
                    cores = 2,
                    median_comparison_abundance = FALSE,
                    median_comparison_prevalence = FALSE),
        "Daemons already set"
    )

    expect_true(mirai::daemons_set())
    mirai::daemons(0)
    Sys.sleep(0.5)
})

test_that("maaslin3 function with cores parameter works end-to-end", {
    if (mirai::daemons_set()) mirai::daemons(0)
    Sys.sleep(0.5)

    td <- setup_test_data()
    tmp_output <- tempfile("maaslin3_parallel_test_")
    dir.create(tmp_output)

    results <- maaslin3(td$data_in,
                        td$metadata,
                        output = tmp_output,
                        formula = 'expr ~ var1 + var2',
                        normalization = 'TSS',
                        transform = 'LOG',
                        cores = 2,
                        plot_summary_plot = FALSE,
                        plot_associations = FALSE,
                        median_comparison_abundance = FALSE,
                        warn_prevalence = FALSE)

    expect_false(mirai::daemons_set())
    expect_true(!is.null(results$fit_data_abundance))
    expect_true(!is.null(results$fit_data_prevalence))
    expect_true(nrow(results$fit_data_abundance$results) > 0)
    expect_true(nrow(results$fit_data_prevalence$results) > 0)

    unlink(tmp_output, recursive = TRUE)
})

test_that("serial and parallel results are consistent", {
    if (mirai::daemons_set()) mirai::daemons(0)
    Sys.sleep(0.5)

    td <- setup_test_data()

    results_serial <- maaslin_fit(td$data_in_tss,
                                  td$data_in_tss_log,
                                  td$metadata,
                                  formula = formula('expr ~ var1 + var2'),
                                  random_effects_formula = NULL,
                                  min_abundance = 0,
                                  min_prevalence = 0,
                                  min_variance = 0,
                                  data = td$data_in,
                                  cores = 1,
                                  median_comparison_abundance = FALSE,
                                  median_comparison_prevalence = FALSE)

    results_parallel <- maaslin_fit(td$data_in_tss,
                                    td$data_in_tss_log,
                                    td$metadata,
                                    formula = formula('expr ~ var1 + var2'),
                                    random_effects_formula = NULL,
                                    min_abundance = 0,
                                    min_prevalence = 0,
                                    min_variance = 0,
                                    data = td$data_in,
                                    cores = 2,
                                    median_comparison_abundance = FALSE,
                                    median_comparison_prevalence = FALSE)

    expect_false(mirai::daemons_set())

    serial_res <- results_serial$fit_data_abundance$results
    parallel_res <- results_parallel$fit_data_abundance$results

    serial_res <- serial_res[order(serial_res$feature, serial_res$metadata, serial_res$value), ]
    parallel_res <- parallel_res[order(parallel_res$feature, parallel_res$metadata, parallel_res$value), ]

    expect_equal(nrow(serial_res), nrow(parallel_res))
    expect_equal(serial_res$feature, parallel_res$feature)
    expect_equal(serial_res$metadata, parallel_res$metadata)
    expect_equal(serial_res$coef, parallel_res$coef, tolerance = 1e-6)
    expect_equal(serial_res$pval, parallel_res$pval, tolerance = 1e-6)
})

library(testthat)
library(maaslin3)

# When evaluate_only='abundance', maaslin_fit must rename pval/qval to
# pval_individual/qval_individual on fit_data_abundance$results.
# A previous bug referenced a nonexistent variable "outputs$results" instead
# of "fit_data_abundance$results", which would crash this code path.

test_that("evaluate_only='abundance' renames pval columns correctly", {
    data_in <- data.frame('a' = c(0, 0, 3, 0, 5, 7, 0, 8, 4, 7), 
                          'b' = c(2, 0, 0, 5, 0, 5, 2, 6, 9, 0))
    rownames(data_in) <- paste0("sample", 1:10)
    
    metadata <- data.frame('var1' = 1:10,
                           'var2' = c(0, 1, 0, 1, 0, 1, 0, 1, 0, 1))
    rownames(metadata) <- paste0("sample", 1:10)
    
    data_in_tss <- data.frame(t(apply(data_in, MARGIN = 1, 
                                      FUN = function(x) x / sum(x))))
    data_in_tss[data_in_tss == 0] <- NA
    data_in_tss_log <- log2(data_in_tss)
    
    results <- maaslin_fit(data_in_tss,
                           data_in_tss_log,
                           metadata,
                           formula = formula('expr ~ var1 + var2'),
                           random_effects_formula = NULL,
                           min_abundance = 0,
                           min_prevalence = 0,
                           min_variance = 0,
                           data = data_in,
                           median_comparison_abundance = FALSE,
                           warn_prevalence = FALSE,
                           evaluate_only = 'abundance')
    
    res_cols <- colnames(results$fit_data_abundance$results)
    expect_true("pval_individual" %in% res_cols)
    expect_true("qval_individual" %in% res_cols)
    expect_false("pval" %in% res_cols)
    expect_false("qval" %in% res_cols)
    expect_true("pval_joint" %in% res_cols)
    expect_true("qval_joint" %in% res_cols)
})

library(testthat)
library(maaslin3)

metadata <- data.frame('a' = c(1, 2, 3, 4, 5),
                       'b' = c(0, 1, 0, 1, 0),
                       'c' = c('a', 'b', 'c', 'a', 'b'))
rownames(metadata) <- paste0("sample", c(1:5))

expect_error(maaslin_process_metadata(metadata = metadata, 
                         formula = formula('~ a + b + c'), 
                         standardize = T))

metadata_out <- metadata
metadata_out$a <- scale(metadata_out$a)[,1]
metadata_out$b <- scale(metadata_out$b)[,1]
metadata_out$c <- factor(metadata_out$c)
expect_equal(maaslin_process_metadata(metadata = metadata, 
                                      formula = formula('~ a + b + c'), 
                                      reference = 'c,a',
                                      standardize = T),
             metadata_out)

metadata$c <- factor(metadata$c)
expect_that(maaslin_process_metadata(metadata = metadata, 
                                     formula = formula('~ a + b + c'), 
                                     standardize = F),
            equals(metadata))

expect_that(maaslin_process_metadata(metadata = metadata, 
                             formula = formula('~ a + (1|b) + strata(c)'), 
                             standardize = F),
            equals(metadata))

library(testthat)
library(maaslin3)

data_in <- data.frame('a' = c(1, 2, 3, 4, 5), 
                    'b' = c(2, 3, 4, 5, 6),
                    'c' = c(3, 4, 5, 6, 7))
rownames(data_in) <- paste0("sample", c(1:5))

metadata <- data.frame('a' = c(1, 2, 3, 4, 5),
                       'b' = c(0, 1, 0, 1, 0),
                       'c' = c('a', 'b', 'c', 'a', 'b'))

expect_that(maaslin_check_formula(data,
                    metadata,
                    input_formula = 'a + b + c')$formula, 
            equals(formula(expr ~ a + b + c)))

expect_error(maaslin_check_formula(data,
                                  metadata,
                                  input_formula = 'strata(b)')$formula)

expect_error(maaslin_check_formula(data,
                                   metadata,
                                   input_formula = ''))

expect_error(maaslin_check_formula(data,
                                metadata,
                                input_formula = '~ a + b + d'))

expect_error(maaslin_check_formula(data,
                                   metadata,
                                   input_formula = NULL))

expect_that(maaslin_check_formula(data,
                                metadata,
                                input_formula = '~ a + b + (1|c)')$formula,
            equals(formula(expr ~ a + b + (1|c))))

expect_that(maaslin_check_formula(data,
                    metadata,
                    input_formula = '~ a + b + (1|c)')$random_effects_formula,
            equals(formula(expr ~ a + b + (1|c))))

expect_that(maaslin_check_formula(data,
                                metadata,
                                input_formula = '~ a + b + (1|c)',
                                feature_specific_covariate_name = 'd')$formula,
            equals(formula(expr ~ d + a + b + (1|c))))

expect_that(maaslin_check_formula(data,
                                  metadata,
                                  input_formula = formula('~ a + b + (1|c)'),
                                  feature_specific_covariate_name = 'd')$formula,
            equals(formula(expr ~ d + a + b + (1|c))))

context("collapse migration regressions")

# Regressions introduced by the dplyr -> collapse migration, plus a
# separately-discovered duplicate-feature-name bug in maaslin_read_data.

test_that("maaslin_process_metadata handles all-categorical metadata", {
    # collapse::num_vars(metadata) returns an empty data frame when there
    # are no numeric columns, and collapse::fscale on that empty frame
    # previously triggered:
    #   Error in `get_vars_ind<-`(x, .Call(C_vtypes, x, 1L), value) :
    #     NROW(value) must match nrow(x)
    metadata <- data.frame(
        treatment = c("A", "B", "A", "B", "A", "B"),
        group = c("X", "Y", "X", "Y", "X", "Y"),
        stringsAsFactors = FALSE,
        row.names = paste0("S", 1:6)
    )

    result <- maaslin3:::maaslin_process_metadata(
        metadata = metadata,
        fixed_effects = c("treatment", "group"),
        reference = NULL,
        feature_specific_covariate_name = NULL,
        standardize = TRUE
    )

    expect_true(is.data.frame(result))
    expect_equal(nrow(result), 6)
    expect_true(all(c("treatment", "group") %in% colnames(result)))
})

test_that("maaslin_process_metadata standardizes numeric and preserves categorical", {
    metadata <- data.frame(
        age = c(25, 30, 35, 40, 45, 50),
        treatment = c("A", "B", "A", "B", "A", "B"),
        stringsAsFactors = FALSE,
        row.names = paste0("S", 1:6)
    )

    result <- maaslin3:::maaslin_process_metadata(
        metadata = metadata,
        fixed_effects = c("age", "treatment"),
        reference = NULL,
        feature_specific_covariate_name = NULL,
        standardize = TRUE
    )

    expect_true(is.data.frame(result))
    expect_equal(nrow(result), 6)
    # numeric column should be z-scored (mean ~0, sd ~1)
    expect_equal(mean(result$age), 0, tolerance = 1e-8)
    expect_equal(sd(result$age), 1, tolerance = 1e-8)
    # categorical column unchanged
    expect_equal(as.character(result$treatment), metadata$treatment)
})

test_that("collapse::fscale on empty data frame is guarded", {
    # Document the underlying collapse behavior that motivated the guard.
    empty_df <- data.frame(row.names = paste0("S", 1:6))
    expect_equal(NCOL(empty_df), 0)
    # Without the guard, assigning fscale(empty_df) back into num_vars<-
    # errors. The maaslin_process_metadata code must therefore skip the
    # assignment when there are no numeric columns.
})

test_that("maaslin_read_data makes duplicate feature names unique", {
    tmpdir <- tempfile("maaslin_dup_")
    dir.create(tmpdir)
    on.exit(unlink(tmpdir, recursive = TRUE), add = TRUE)

    data_path <- file.path(tmpdir, "data.tsv")
    meta_path <- file.path(tmpdir, "meta.tsv")

    set.seed(1)
    n_samples <- 10
    data_mat <- matrix(
        abs(rnorm(n_samples * 4, mean = 100, sd = 30)),
        nrow = n_samples,
        ncol = 4
    )
    data_df <- as.data.frame(data_mat)
    # Two duplicate column names - exactly the condition that blew up
    # downstream subsetting with "undefined columns selected".
    colnames(data_df) <- c("Bacteroides", "Prevotella", "Bacteroides", "Firmicutes")
    rownames(data_df) <- paste0("S", 1:n_samples)

    # write.table would mangle dup names; use a manual writer to preserve them.
    header <- paste(c("", colnames(data_df)), collapse = "\t")
    body <- apply(data_df, 1, function(row) paste(row, collapse = "\t"))
    writeLines(c(header, paste(rownames(data_df), body, sep = "\t")), data_path)

    meta_df <- data.frame(
        treatment = rep(c("A", "B"), length.out = n_samples),
        row.names = paste0("S", 1:n_samples),
        stringsAsFactors = FALSE
    )
    write.table(meta_df, meta_path, sep = "\t", quote = FALSE,
                col.names = NA, row.names = TRUE)

    result <- maaslin3:::maaslin_read_data(
        input_data = data_path,
        input_metadata = meta_path,
        feature_specific_covariate = NULL,
        unscaled_abundance = NULL
    )

    expect_true(!any(duplicated(colnames(result$data))))
    expect_true("Bacteroides" %in% colnames(result$data))
    expect_true("Bacteroides.1" %in% colnames(result$data))
})

test_that("add_qvals handles prevalence/abundance fits with all-NA pvals", {
    # data.table::fifelse is strict about yes/no types. When a fit object's
    # $results$pval is all-NA (logical), fifelse(..., NA_real_, pval)
    # previously errored with
    #   'no' is of type logical but 'yes' is double
    # The fix coerces pvals to numeric first.
    fake_fit <- list(
        results = data.frame(
            pval = NA,
            error = NA,
            stringsAsFactors = FALSE
        )
    )

    result <- maaslin3:::add_qvals(
        fit_data_abundance = fake_fit,
        fit_data_prevalence = fake_fit,
        correction = "BH"
    )

    expect_true(is.list(result))
    expect_true(all(is.na(result$fit_data_abundance$results$qval_individual)))
    expect_true(all(is.na(result$fit_data_prevalence$results$qval_individual)))
})

test_that("full maaslin3 run works with only categorical metadata", {
    tmpdir <- tempfile("maaslin_cat_only_")
    dir.create(tmpdir)
    on.exit(unlink(tmpdir, recursive = TRUE), add = TRUE)

    set.seed(42)
    n_samples <- 20
    n_features <- 8
    data_mat <- matrix(
        abs(rnorm(n_samples * n_features, mean = 100, sd = 30)),
        nrow = n_samples,
        ncol = n_features
    )
    data_df <- as.data.frame(data_mat)
    colnames(data_df) <- paste0("Feature", 1:n_features)
    rownames(data_df) <- paste0("S", 1:n_samples)

    meta_df <- data.frame(
        treatment = rep(c("A", "B"), each = n_samples / 2),
        row.names = paste0("S", 1:n_samples),
        stringsAsFactors = FALSE
    )

    out_dir <- file.path(tmpdir, "out")

    expect_error(
        suppressWarnings(maaslin3::maaslin3(
            input_data = data_df,
            input_metadata = meta_df,
            output = out_dir,
            fixed_effects = c("treatment"),
            normalization = "TSS",
            transform = "LOG",
            standardize = TRUE,
            plot_summary_plot = FALSE,
            plot_associations = FALSE,
            save_models = FALSE,
            verbosity = "ERROR"
        )),
        NA
    )
})

library(testthat)
library(maaslin3)

# The inherits() check for BioC DataFrame must test input_metadata, not the
# already-converted local variable 'metadata'. The old code used
# inherits(metadata, 'DataFrame') which would always be FALSE after the
# data.frame conversion above it, causing DataFrame inputs to hit the
# stop() branch.

test_that("maaslin_read_data handles a regular data.frame for metadata", {
    data_in <- data.frame(a = c(1, 2, 3, 4, 5),
                          b = c(5, 4, 3, 2, 1))
    rownames(data_in) <- paste0("s", 1:5)

    meta <- data.frame(var1 = rnorm(5))
    rownames(meta) <- paste0("s", 1:5)

    result <- maaslin_read_data(data_in, meta, NULL, NULL)
    expect_true(is.data.frame(result$metadata))
    expect_equal(nrow(result$metadata), 5)
})

test_that("maaslin_read_data rejects non-data-frame non-file metadata", {
    data_in <- data.frame(a = c(1, 2, 3), b = c(3, 2, 1))
    rownames(data_in) <- paste0("s", 1:3)

    expect_error(
        maaslin_read_data(data_in, list(x = 1), NULL, NULL),
        "neither a file nor a data frame"
    )
})

library(testthat)
library(maaslin3)

# Association plots must use circular points (shape 21). A previous change
# switched to squares (shape 22) for rendering speed, but this was reverted
# to preserve the original visual appearance.

test_that("association plots use circular points (shape 21)", {
    output_tmp <- tempfile()
    
    data_in <- data.frame('a' = c(0, 0, 0, 0, 0, 0, 0, 0, 7, 5, 8, 4, 7, 8, 5, 8), 
                          'b' = c(2, 0, 0, 5, 0, 5, 3, 6, 5, 2, 6, 9, 0, 8, 3, 7))
    rownames(data_in) <- paste0("sample", seq(nrow(data_in)))
    
    metadata <- data.frame('var1' = c(rep(0, 8), rep(1, 7), 2))
    rownames(metadata) <- paste0("sample", seq(nrow(data_in)))
    
    data_in_tss <- data.frame(t(apply(data_in, MARGIN = 1, 
                                      FUN = function(x) x / sum(x))))
    data_in_tss[data_in_tss == 0] <- NA
    data_in_tss_log <- log2(data_in_tss)
    
    results <- maaslin_fit(data_in_tss,
                           data_in_tss_log,
                           metadata,
                           formula = formula('expr ~ var1'),
                           random_effects_formula = NULL, 
                           min_abundance = 0, 
                           min_prevalence = 0, 
                           min_variance = 0,
                           data = data_in, 
                           median_comparison_abundance = FALSE)
    
    plot_out <- maaslin_plot_results(output = output_tmp,
                                    transformed_data = data_in_tss_log,
                                    unstandardized_metadata = metadata,
                                    fit_data_abundance = results$fit_data_abundance, 
                                    fit_data_prevalence = results$fit_data_prevalence, 
                                    normalization = 'TSS',
                                    transform = 'LOG',
                                    median_comparison_abundance = FALSE,
                                    max_significance = 0.1,
                                    save_plots_rds = TRUE)
    
    gg <- plot_out$assocation_plots[[1]]
    expect_is(gg, 'ggplot')
    
    point_layers <- Filter(function(layer) {
        inherits(layer$geom, "GeomPoint")
    }, gg$layers)
    
    shapes_used <- vapply(point_layers, function(layer) {
        as.numeric(layer$aes_params$shape %||% NA_real_)
    }, numeric(1))
    
    expect_true(all(shapes_used[!is.na(shapes_used)] == 21),
                info = "All point geoms should use shape 21 (circles)")
    
    unlink(output_tmp, recursive = TRUE)
})

library(testthat)
library(maaslin3)

data_in <- data.frame('a' = c(1, 2, 3, 4, 5), 
                      'b' = c(2, 3, 4, 5, 6),
                      'c' = c(3, 4, 5, 6, 7))
rownames(data_in) <- paste0("sample", c(1:5))

metadata <- data.frame('a' = c(1, 2, 3, 4, 5),
                       'b' = c(0, 1, 0, 1, 0),
                       'c' = c('a', 'b', 'c', 'a', 'b'))
rownames(metadata) <- paste0("sample", c(1:5))

expect_that(maaslin_compute_formula(data,
              metadata,
              fixed_effects = c('a', 'b', 'c'),
              random_effects = NULL,
              group_effects = NULL,
              ordered_effects = NULL,
              strata_effects = NULL,
              feature_specific_covariate_name = NULL)$formula, 
            equals(formula(expr ~ a + b + c)))

expect_that(maaslin_compute_formula(data,
                                    metadata,
                                    fixed_effects = NULL,
                                    random_effects = NULL,
                                    group_effects = NULL,
                                    ordered_effects = NULL,
                                    strata_effects = NULL,
                                    feature_specific_covariate_name = NULL)$formula, 
            equals(formula(expr ~ a + b + c)))

expect_that(maaslin_compute_formula(data,
                                    metadata,
                                    fixed_effects = c('a', 'b'),
                                    random_effects = 'c',
                                    group_effects = NULL,
                                    ordered_effects = NULL,
                                    strata_effects = NULL,
                                    feature_specific_covariate_name = NULL)$formula, 
            equals(formula(expr ~ a + b + (1|c))))

expect_that(maaslin_compute_formula(data,
                                    metadata,
                                    fixed_effects = c('a', 'b'),
                                    random_effects = NULL,
                                    group_effects = 'c',
                                    ordered_effects = NULL,
                                    strata_effects = NULL,
                                    feature_specific_covariate_name = NULL)$formula, 
            equals(formula(expr ~ a + b + group(c))))

expect_that(maaslin_compute_formula(data,
                                    metadata,
                                    fixed_effects = c('a', 'b'),
                                    random_effects = NULL,
                                    group_effects = NULL,
                                    ordered_effects = 'c',
                                    strata_effects = NULL,
                                    feature_specific_covariate_name = NULL)$formula, 
            equals(formula(expr ~ a + b + ordered(c))))

expect_that(maaslin_compute_formula(data,
                                    metadata,
                                    fixed_effects = c('a', 'b'),
                                    random_effects = NULL,
                                    group_effects = NULL,
                                    ordered_effects = NULL,
                                    strata_effects = 'c',
                                    feature_specific_covariate_name = NULL)$formula, 
            equals(formula(expr ~ a + b + strata(c))))

expect_that(maaslin_compute_formula(data,
                                    metadata,
                                    fixed_effects = c('a,b,c'),
                                    random_effects = NULL,
                                    group_effects = NULL,
                                    ordered_effects = NULL,
                                    strata_effects = NULL,
                                    feature_specific_covariate_name = NULL)$formula, 
            equals(formula(expr ~ a + b + c)))

expect_error(maaslin_compute_formula(data,
                                    metadata,
                                    fixed_effects = c('a,b,c'),
                                    random_effects = "d",
                                    group_effects = NULL,
                                    ordered_effects = NULL,
                                    strata_effects = NULL,
                                    feature_specific_covariate_name = NULL))

expect_error(maaslin_compute_formula(data,
                                     metadata,
                                     fixed_effects = character(0),
                                     random_effects = character(0),
                                     group_effects = character(0),
                                     ordered_effects = character(0),
                                     strata_effects = character(0),
                                     feature_specific_covariate_name = NULL))

colnames(metadata) <- c("a b", "c d", "c")
expect_that(maaslin_compute_formula(data,
                                    metadata,
                                    fixed_effects = c('a b'),
                                    random_effects = NULL,
                                    group_effects = NULL,
                                    ordered_effects = NULL,
                                    strata_effects = 'c d',
                                    feature_specific_covariate_name = NULL)$formula, 
            equals(formula(expr ~ `a b` + strata(`c d`))))

library(testthat)
library(maaslin3)

data_in <- data.frame('a' = c(0, 0, 3, 0, 5, 7, 0, 8, 4, 7), 
                      'b' = c(2, 0, 0, 5, 0, 5, 2, 6, 9, 0),
                      'c' = c(3, 4, 0, 0, 7, 2, 7, 4, 7, 0))
rownames(data_in) <- paste0("sample", c(1:10))

metadata <- data.frame('var1' = c(1, 2, 3, 4, 5, 6, 7, 8, 9, 10),
                       'var2' = c(0, 1, 0, 1, 0, 1, 0, 1, 0, 1))
rownames(metadata) <- paste0("sample", c(1:10))

data_in_tss <- data.frame(t(apply(data_in, MARGIN = 1, 
                                  FUN = function(x){x / sum(x)})))
data_in_tss[data_in_tss == 0] <- NA
data_in_tss_log <- log2(data_in_tss)

results <- maaslin_fit(data_in_tss,
            data_in_tss_log,
            metadata,
            formula = formula('expr ~ var1 + var2'),
            random_effects_formula = NULL, 
            min_abundance = 0, 
            min_prevalence = 0, 
            min_variance = 0,
            data = data_in, 
            median_comparison_abundance = FALSE)

data_in_tss_log$sample <- rownames(data_in_tss_log)
metadata$sample <- rownames(metadata)
data_merged <- merge(data_in_tss_log, metadata, by = "sample", all = TRUE)
rownames(data_merged) <- data_merged$sample
data_merged$sample <- NULL

growing_df <- data.frame()
for (feature in c('a', 'b', 'c')) {
    summary_out <- summary(lm(formula(paste0(feature, '~ var1 + var2')), data_merged))$coef
    summary_out <- data.frame(summary_out, check.names = F)
    summary_out$variable <- rownames(summary_out)
    summary_out <- summary_out[-1,] # Drop intercept
    summary_out$`t value` <- NULL
    colnames(summary_out) <- c("coef", "stderr", "pval_individual", 'metadata')
    summary_out$name <- summary_out$value <- summary_out$metadata
    summary_out$feature <- feature
    summary_out$error <- NA_character_
    summary_out$model <- 'linear'
    summary_out$N <- nrow(data_merged)
    summary_out$N_not_zero <- sum(!is.na(data_merged[,feature]))
    growing_df <- rbind(growing_df, summary_out)
}

for (feature in c('a', 'b', 'c')) {
    data_subset <- data_merged[,c(feature, 'var1', 'var2')]
    data_subset[,1] <- ifelse(is.na(data_subset[,1]), 0, 1)
    orig_length <- nrow(data_subset)
    data_subset <- rbind(data_subset, data_subset, data_subset)
    data_subset[(orig_length + 1):(2 * orig_length),1] <- 1
    data_subset[(2 * orig_length + 1):(3 * orig_length),1] <- 0
    summary_out <- tryCatch({
        withCallingHandlers({
                summary(glm(formula(paste0(feature, ' ~ var1 + var2')), 
                            data_subset, family = 'binomial', 
                            weights = c(rep(1, orig_length),
                                        rep(2 / 2 / orig_length, 
                                            2 * orig_length))))$coef
            },
            warning = function(w) {
                invokeRestart("muffleWarning")
            }
        )
    })
    
    summary_out <- data.frame(summary_out, check.names = F)
    summary_out$variable <- rownames(summary_out)
    summary_out <- summary_out[-1,] # Drop intercept
    summary_out$`z value` <- NULL
    colnames(summary_out) <- c("coef", "stderr", "pval_individual", 'metadata')
    summary_out$name <- summary_out$value <- summary_out$metadata
    summary_out$feature <- feature
    summary_out$error <- NA_character_
    summary_out$model <- 'logistic'
    summary_out$N <- nrow(data_merged)
    summary_out$N_not_zero <- sum(!is.na(data_merged[,feature]))
    growing_df <- rbind(growing_df, summary_out)
}
growing_df$qval_individual <- p.adjust(growing_df$pval_individual, 'BH')

merged_df <- merge(growing_df[growing_df$model == 'linear',],
                   growing_df[growing_df$model == 'logistic',], 
                   by = c('metadata', 'value', 'name', 'feature', 'N', 
                          'N_not_zero'),
                   all = TRUE)
merged_df$pval_joint <- pbeta(pmin(merged_df$pval_individual.x, 
                                   merged_df$pval_individual.y), 1, 2)
merged_df$qval_joint <- p.adjust(merged_df$pval_joint, 'BH')
joint_lookup <- merged_df[, c('feature', 'metadata', 'value', 'name',
                               'pval_joint', 'qval_joint')]
growing_df <- merge(growing_df, joint_lookup,
                    by = c('feature', 'metadata', 'value', 'name'),
                    all.x = TRUE)
growing_df$null_hypothesis <- 0

col_order <- c("feature", "metadata", "value", "name", "coef", 
               "null_hypothesis", "stderr", "pval_individual", 
               "qval_individual", "pval_joint", 
               "qval_joint", "error", "model", "N", "N_not_zero")
growing_df <- growing_df[,col_order]
growing_df <- growing_df[order(growing_df$qval_individual),]

linear_piece <- growing_df[growing_df$model == 'linear',]
logistic_piece <- growing_df[growing_df$model == 'logistic',]
rownames(linear_piece) <- NULL
rownames(logistic_piece) <- NULL
rownames(results$fit_data_abundance$results) <- NULL
rownames(results$fit_data_prevalence$results) <- NULL

expect_equal(results$fit_data_abundance$results |> 
                 collapse::roworderv(c("feature", "metadata", "value", "name")),
             linear_piece)

expect_equal(results$fit_data_prevalence$results |> 
                 collapse::roworderv(c("feature", "metadata", "value", "name")),
            logistic_piece |> 
                collapse::roworderv(c("feature", "metadata", "value", "name")))

#!/usr/bin/env Rscript
utils::globalVariables(".progress")
###############################################################################
# MaAsLin3 fitting

# Copyright (c) 2025 Harvard School of Public Health

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
###############################################################################

# Function to augment data for logistic fitting
augment_data <- function(formula, random_effects_formula, dat_sub) {
    d2 = dat_sub 
    d3 = dat_sub 
    
    d2 = collapse::ftransform(d2, expr = 1)
    d3 = collapse::ftransform(d3, expr = 0)
    
    dat_sub_new <- collapse::rowbind(dat_sub, d2, d3)
    
    formula <- stats::formula(formula)
    
    if (is.null(random_effects_formula)) {
        # No random effects
        p <- ncol(model.matrix(formula, dat_sub)) - 1
    } else {
        # With random effects
        p <- ncol(lme4::lFormula(formula, dat_sub)$X) - 1
    }
    
    # Calculate the weights
    weight_scheme <-
        c(rep(1, nrow(dat_sub)),
        rep(p / (2 * nrow(dat_sub)), 2 * nrow(dat_sub)))
    
    return(list(
        mm_input = dat_sub_new,
        weight_scheme = weight_scheme,
        new_formula = formula
    ))
}

# Deparse formula without line breaks
safe_deparse <- function(formula) {
    paste0(trimws(deparse(formula)), collapse = " ")
}

# Extract a predictor of the form: predictor_type(variable)
extract_special_predictor <- function(formula, predictor_type) {
    match.arg(predictor_type, c("group", "ordered", "strata"))
    groups <-
        regmatches(safe_deparse(formula), gregexpr(
            paste0(predictor_type, "\\((.*?)\\)"),
            safe_deparse(formula)
        ))[[1]]
    groups <-
        gsub("\\)$", "", gsub(paste0("^", predictor_type, "\\("), "", groups))
    formula_tmp <-
        trimws(gsub(
            paste0("(\\s*\\+\\s*)?", predictor_type, "\\(.*?\\)"),
            "",
            safe_deparse(formula)
        ))
    if (substr(formula_tmp, nchar(formula_tmp), nchar(formula_tmp)) == '~') {
        formula_tmp <- paste0(formula_tmp, '1')
    }
    formula <- stats::formula(formula_tmp)
    formula <- stats::formula(gsub("~ \\+", "~", safe_deparse(formula)))
    
    if (predictor_type == "strata" & length(groups) > 1) {
        stop("Only one strata allowed. Please change the formula.")
    }
    
    return(list(formula, predictor_type = groups))
}

# Get all fixed effects, not assuming a model has or can be fit
get_fixed_effects <-
    function(formula,
            random_effects_formula,
            dat_sub,
            groups,
            ordereds,
            feature_specific_covariate_name) {
        names_to_include <- c()
        if (is.null(random_effects_formula)) {
            # Fixed and group effects only
            names_to_include <- colnames(model.matrix(stats::formula(gsub(
                "^expr ", "", safe_deparse(formula)
            )), dat_sub))
            names_to_include <-
                names_to_include[names_to_include != "(Intercept)"]
        } else {
            # Random effects
            patterns <- paste0("(", unlist(reformulas::findbars(stats::formula(
                gsub("^expr ", "", safe_deparse(formula))
            ))), ")")
            
            # Remove all the random effects from the formula
            for (pattern in patterns) {
                fixed_effects_only <- gsub(pattern, "",
                                        paste0(trimws(safe_deparse(
                                            stats::formula(gsub(
                                                "^expr ", "",
                                                safe_deparse(formula)
                                            ))
                                        )), collapse = " "),
                                        fixed = TRUE)
                fixed_effects_only <-
                    gsub("[+ ]+$", "", fixed_effects_only)
                fixed_effects_only <-
                    gsub("\\+\\s*\\++", "+", fixed_effects_only)
                formula <- stats::formula(fixed_effects_only)
            }
            
            if (!is.null(feature_specific_covariate_name)) {
                names_to_include <- colnames(model.matrix(stats::formula(gsub(
                    paste0(feature_specific_covariate_name, " \\+|^expr "), 
                    "", safe_deparse(formula)
                )), dat_sub))
                names_to_include <- c(names_to_include, 
                                        feature_specific_covariate_name)
            } else {
                names_to_include <- colnames(model.matrix(stats::formula(gsub(
                    "^expr ", "", safe_deparse(formula)
                )), dat_sub))
            }
            
            names_to_include <-
                names_to_include[names_to_include != "(Intercept)"]
        }
        
        ordered_levels <- unlist(lapply(ordereds, function(ordered) {
            ordered_no_ticks <- gsub("^`|`$", "", ordered)
            paste0(ordered, levels(dat_sub[[ordered_no_ticks]])[-1])
        }))
        
        names_to_include <-
            c(names_to_include, groups, ordered_levels)
        return(names_to_include)
    }

# Get non-baseline names for all non-numeric columns
get_character_cols <- function(dat_sub) {
    all_factors <- unlist(lapply(colnames(dat_sub), function(col) {
        if (!is.numeric(dat_sub[, col])) {
            if (is.factor(dat_sub[, col])) {
                factor_levels <- levels(dat_sub[, col])
            } else {
                factor_levels <- levels(factor(dat_sub[, col]))
            }
            # All factor levels except baseline
            return(unlist(paste0(col, unique(factor_levels[-1]))))
        }
        return(character(0))
    }))
    
    return(all_factors)
}

# Correct over prevalence and abundance to get q-values
add_qvals <- function(fit_data_abundance, fit_data_prevalence, correction) {
    match.arg(correction, 
        c("BH", "holm", "hochberg", "hommel", "bonferroni", "BY"))
    # Select out p-values and NA if errors
    if (!is.null(fit_data_abundance)) {
        abundance_pvals <- as.numeric(fit_data_abundance$results$pval)
        
        abundance_pvals <- data.table::fifelse(!is.na(fit_data_abundance$results$error),
                                               NA_real_,
                                               abundance_pvals)
    } else {
        abundance_pvals <- c()
    }
    
    if (!is.null(fit_data_prevalence)) {
        prevalence_pvals <- as.numeric(fit_data_prevalence$results$pval)
        
        prevalence_pvals <- data.table::fifelse(!is.na(fit_data_prevalence$results$error),
                                                NA_real_,
                                                prevalence_pvals)
    } else {
        prevalence_pvals <- c()
    }

    # Create and write combined q-vals
    combined_qvals <- as.numeric(p.adjust(c(abundance_pvals, prevalence_pvals), 
                                method = correction))
    
    if (!is.null(fit_data_abundance)) {
        fit_data_abundance$results$qval <- combined_qvals[
            seq_len(length(abundance_pvals))]
    }
    if (!is.null(fit_data_prevalence)) {
        fit_data_prevalence$results$qval <- combined_qvals[
            seq(length(abundance_pvals) + 1, length(abundance_pvals) + 
                    length(prevalence_pvals))]
    }
    
    return(list(
        if (!is.null(fit_data_abundance)) fit_data_abundance$results else NULL,
        if (!is.null(fit_data_prevalence)) fit_data_prevalence$results else NULL
    ))
}

# Combine abundance and prevalence p-values
create_combined_pval <- function(merged_signif, correction) {
    match.arg(correction,
            c("BH", "holm", "hochberg", "hommel", "bonferroni", "BY"))
    # Create a combined p-value
    merged_signif$pval_joint <-
        pbeta(pmin(merged_signif[, "linear"],
                merged_signif[, "logistic"]), 1, 2)
    
    # If NA or model errored, use the p-value of the non-NA
    merged_signif$pval_joint <-
        data.table::fifelse(
            is.na(merged_signif$linear) |
                !is.na(merged_signif$linear_error),
            as.numeric(merged_signif$logistic),
            merged_signif$pval_joint
        )
    # ^ as.numeric() because fifelse checks if class(yes) == class(no) and NA is
    # logical by default
    
    merged_signif$pval_joint <-
        data.table::fifelse(
            is.na(merged_signif$logistic) |
                !is.na(merged_signif$logistic_error),
            as.numeric(merged_signif$linear),
            merged_signif$pval_joint
        )
    
    merged_signif$pval_joint <-
        data.table::fifelse((
            is.na(merged_signif$logistic) |
                !is.na(merged_signif$logistic_error)
        ) & (is.na(merged_signif$linear) |
                 !is.na(merged_signif$linear_error)),
        NA_real_,
        merged_signif$pval_joint)
    
    merged_signif$qval_joint <-
        as.numeric(p.adjust(merged_signif$pval_joint, method = correction))
    
    return(merged_signif)
}

# Put note when abundance effect has likely turned into a prevalence effect
flag_abundance_turned_prevalence <- function(merged_signif,
                                            max_significance) {
    
    tst = !is.na(merged_signif$logistic_coef) &
        !is.na(merged_signif$linear_coef) &
        !is.na(merged_signif$logistic) &
        !is.na(merged_signif$linear)
    
    inner_tst = is.na(merged_signif$logistic_error) &
        is.na(merged_signif$linear_error) &
        merged_signif$linear_qval < max_significance &
        sign(merged_signif$logistic_coef) == 
        sign(merged_signif$linear_coef) &
        abs(merged_signif$linear_coef) > 
        abs(merged_signif$logistic_coef)
    
    # Join and check linear and logistic pieces
    merged_signif$logistic_error <- data.table::fifelse(
        tst,
        data.table::fifelse(
            inner_tst,
            paste0("Prevalence association possibly induced ",
                   "by stronger abundance association"),
            merged_signif$logistic_error
        ),
        merged_signif$logistic_error
    )
    
    return(merged_signif)
}

# Get joint significance for abundance and prevalence associations
add_joint_signif <-
    function(fit_data_abundance,
             fit_data_prevalence,
             new_fit_data_abundance,
             max_significance,
             correction) {
        
        match.arg(correction,
                  c("BH", "holm", "hochberg", "hommel", "bonferroni", "BY"))
        
        # Subset to shared columns
        sel_vec = c("feature", "metadata", "value",
                    "name", "coef", "null_hypothesis",
                    "pval", "qval", "error")
        
        fit_data_prevalence_signif <-
            fit_data_prevalence$results |> collapse::fselect(sel_vec)
        
        colnames(fit_data_prevalence_signif) <-
            c("feature",
            "metadata",
            "value",
            "name",
            "logistic_coef",
            "logistic_null_hypothesis",
            "logistic",
            "logistic_qval",
            "logistic_error")
        
        fit_data_abundance_signif <-
            fit_data_abundance$results |> collapse::fselect(sel_vec)
        
        colnames(fit_data_abundance_signif) <-
            c("feature",
            "metadata",
            "value",
            "name",
            "linear_coef",
            "linear_null_hypothesis",
            "linear",
            "linear_qval",
            "linear_error")
        
        # Join and check linear and logistic pieces
        merged_signif <-
            collapse::join(
                collapse::funique(fit_data_prevalence_signif),
                collapse::funique(fit_data_abundance_signif),
                how = "full",
                verbose = FALSE,
                on = c("feature", "metadata", "value", "name"),
                overid = 2 # Pretty sure 0 would be fine, but 2 to be safe
            )
        
        # Show the difference between the logistic and linear models fit
        if (nrow(merged_signif) != nrow(unique(fit_data_prevalence_signif)) |
            nrow(merged_signif) != nrow(unique(fit_data_abundance_signif))) {
            stop(
                "Merged significance tables have different associations.
                This is likely a MaAsLin 3 issue 
                due to unexpected data or models."
            )
        }
        
        if (!is.null(new_fit_data_abundance)) {
            fit_data_prevalence_signif_tmp <-
                fit_data_prevalence$results |> collapse::fselect(sel_vec)
            
            colnames(fit_data_prevalence_signif_tmp) <-
                c("feature",
                    "metadata",
                    "value",
                    "name",
                    "logistic_coef",
                    "logistic_null_hypothesis",
                    "logistic",
                    "logistic_qval",
                    "logistic_error")
            
            fit_data_abundance_signif_tmp <-
                new_fit_data_abundance$results[, 
                    c("feature", "metadata", "value",
                    "name", "coef", "null_hypothesis",
                    "pval", "qval", "error")]
            
            colnames(fit_data_abundance_signif_tmp) <-
                c("feature",
                    "metadata",
                    "value",
                    "name",
                    "linear_coef",
                    "linear_null_hypothesis",
                    "linear",
                    "linear_qval",
                    "linear_error")
            
            merged_signif_tmp <-
                collapse::join(
                    collapse::funique(fit_data_prevalence_signif_tmp),
                    collapse::funique(fit_data_abundance_signif_tmp),
                    how = "left",
                    verbose = FALSE,
                    overid = 2,
                    on = c("feature", "metadata", "value", "name")
                )
            
            merged_signif_tmp <- 
                flag_abundance_turned_prevalence(merged_signif_tmp,
                                                max_significance)
            
            overlapping_cols <- c("feature",
                                "metadata",
                                "value",
                                "name",
                                "logistic_coef",
                                "logistic_null_hypothesis",
                                "logistic",
                                "logistic_qval")
            
            merged_signif <- collapse::join(merged_signif,
                                            collapse::fselect(merged_signif_tmp,
                                                              c(overlapping_cols, 
                                                                "logistic_error")),
                                            how = "left",
                                            overid = 2,
                                            on = overlapping_cols,
                                            verbose = FALSE,
                                            suffix = c("_x", "_y"))
            
            # I don't think collapse::join can perfectly mimic the .x/.y
            # suffixing of dplyr joins.
            
            merged_signif <- collapse::fmutate(merged_signif,
                logistic_error = data.table::fcoalesce(merged_signif$logistic_error_y,
                                                       merged_signif$logistic_error_x)
                )
            
            collapse::fselect(merged_signif,
                              c("logistic_error_x",
                                "logistic_error_y")) <- NULL
            
        }
        
        merged_signif <- create_combined_pval(merged_signif,
                                            correction)
        
        return(list(
            append_joint(fit_data_abundance, merged_signif, "abundance"),
            append_joint(fit_data_prevalence, merged_signif, "prevalence")
        ))
    }

# Take logistic or linear results and add on the joint significance
append_joint <- function(outputs, merged_signif, association_type) {
    sel_vec = c("feature", "metadata", "value",
                "name", "pval_joint", "qval_joint")
    
    if (association_type == 'abundance') {
        
        merged_signif <- merged_signif |> collapse::fselect(sel_vec)
        
        collapse::setrename(outputs$results,
                            "pval_individual" = "pval",
                            "qval_individual" = "qval")
        
        # merged_signif <- collapse::join(outputs$results,
        #                                 merged_signif,
        #                                 on = sel_vec[1:4],
        #                                 how = "inner",
        #                                 overid = 2,
        #                                 verbose = FALSE) |> 
        #     collapse::roworderv(sel_vec[1:4]) |> 
        #     collapse::colorderv(sel_vec[1:4])
        
        merged_signif <- merge(outputs$results,
                               merged_signif,
                               by = c("feature", "metadata", "value", "name"))
        
        # o = base::order(merged_signif$feature,
        #                 merged_signif$metadata,
        #                 merged_signif$value,
        #                 merged_signif$name) 
        # # ^ have to use base::order to recapitulate the locale-dependent ordering of base::merge
        # 
        # merged_signif = merged_signif |> 
        #     collapse::ss(o) |> 
        #     collapse::colorderv(neworder = sel_vec[1:4])
        # Too slow. Just let the order be a little changed.
        
    } else if (association_type == 'prevalence') {
        merged_signif <- merged_signif |> 
            collapse::fselect(c(sel_vec,
                                "error" = "logistic_error"))
        
        outputs$results$error <- NULL
        
        collapse::setrename(outputs$results, 
                            "pval_individual" = "pval",
                            "qval_individual" = "qval"
        )
        
        # merged_signif <- collapse::join(outputs$results,
        #                                 merged_signif,
        #                                 on = c("feature", "metadata", 
        #                                        "value", "name"),
        #                                 how = "inner",
        #                                 overid = 2,
        #                                 verbose = FALSE) |> 
        #     collapse::roworderv(sel_vec[1:4]) |> 
        #     collapse::colorderv(sel_vec[1:4])
        
        merged_signif <- merge(outputs$results,
                               merged_signif,
                               by = c("feature", "metadata", "value", "name"))
        
    } else {
        stop("Invalid association_type")
    }
    
    merged_signif <- merged_signif |> 
        collapse::roworderv("qval_individual")
    
    return(merged_signif)
}

# All functions for export in parallelization
function_vec <-
    c(
        "augment_data",
        "safe_deparse",
        "extract_special_predictor",
        "get_fixed_effects",
        "get_character_cols",
        "add_joint_signif",
        "append_joint",
        "check_for_zero_one_obs",
        "check_missing_first_factor_level",
        "check_mffl_one",
        "fit_augmented_logistic",
        "non_augmented",
        "run_group_models",
        "run_ordered_models",
        "fitting_wrap_up"
    )

optimizers <-
    c('nloptwrap', 'nlminbwrap', 'bobyqa', 'Nelder_Mead')
optCtrlList <- list(list(maxeval = 100000),
                    list(maxit = 1500),
                    list(maxfun = 100000),
                    list(maxfun = 100000))

# Check that a formula is provided
check_formulas_valid <- function(formula, random_effects_formula) {
    if (is.null(random_effects_formula)) {
        if (is.null(formula)) {
            logging::logerror(paste("Both formula and 
                                    random_effects_formula are null"))
            stop()
        }
    }
}

# Prepare ranef, modeling, and summary functions for the linear model
choose_ranef_model_summary_funs_linear <- function(random_effects_formula) {
    if (is.null(random_effects_formula)) {
        # Fixed effects only
        ranef_function <- NULL
        
        model_function <-
            function(form,
                    data,
                    weight_scheme = NULL,
                    na.action) {
                return(stats::lm(
                    stats::formula(form),
                    data = data,
                    na.action = na.action
                ))
            }
        
        summary_function <- function(fit, names_to_include) {
            lm_summary <- summary(fit)$coefficients
            
            store_names <- rownames(lm_summary)
            
            if (!all(names_to_include %in% store_names)) {
                # If deficient rank, make sure all rownames are included
                rows_to_add <-
                    names_to_include[!(names_to_include %in% store_names)]
                lm_summary <-
                    rbind(lm_summary, matrix(rep(
                        NaN, 4 * length(rows_to_add)
                    ), nrow = length(rows_to_add)))
                rownames(lm_summary) <-
                    c(store_names, rows_to_add)
            }
            
            para <- as.data.frame(lm_summary) |> 
                ftail(-1) |> 
                collapse::fselect(-3)
            
            para$name <- rownames(lm_summary)[-1]
            
            return(para)
        }
    } else {
        # Random effects
        ranef_function <- lme4::ranef
        
        model_function <-
            function(form,
                    data,
                    weight_scheme = NULL,
                    na.action) {
                index <- 1
                
                while (index < length(optimizers)) {
                    tryCatch({
                        return(
                            lmerTest::lmer(
                                formula = form,
                                data = data,
                                na.action = na.action,
                                control = lme4::lmerControl(
                                    optimizer = optimizers[index],
                                    optCtrl = optCtrlList[[index]]
                                )
                            )
                        )
                    }, warning = function(w) {
                        'warning'
                    }, error = function(e) {
                        'error'
                    })
                    
                    # Something warned or errored if here
                    index <- index + 1
                }
                
                return(
                    lmerTest::lmer(
                        formula = form,
                        data = data,
                        na.action = na.action,
                        control = lme4::lmerControl(
                            optimizer = optimizers[index],
                            optCtrl = optCtrlList[[index]])
                    )
                )
            }
        summary_function <- function(fit, names_to_include) {
            summ = lmerTest:::summary.lmerModLmerTest # Scoped calls - this will be running on daemons

            lm_summary <- summ(fit)$coefficients # Couldn't figure out how to get a scoped call for coef.summary.merMod...

            store_names <- rownames(lm_summary)
            if (!all(names_to_include %in% store_names)) {
                # If deficient rank, make sure all rownames are included
                rows_to_add <-
                    names_to_include[!(names_to_include %in% store_names)]
                lm_summary <-
                    rbind(lm_summary, matrix(rep(
                        NaN, 5 * length(rows_to_add)
                    ), nrow = length(rows_to_add)))
                rownames(lm_summary) <-
                    c(store_names, rows_to_add)
            }
            
            para <- as.data.frame(lm_summary) |> 
                ftail(-1) |> 
                collapse::fselect(-c(3:4))
            # This yields slightly different rownames than [-1, -c(3:4)], but I
            # don't think that matters.
            
            para$name <- rownames(lm_summary)[-1]
            
            return(para)
        }
    }
    return(list("ranef_function" = ranef_function, 
                "model_function" = model_function, 
                "summary_function" = summary_function))
}

# Prepare ranef, modeling, and summary functions for the logistic model
choose_ranef_model_summary_funs_logistic <- function(random_effects_formula,
                                                    strata,
                                                    augment) {
    if (is.null(random_effects_formula)) {
        ranef_function <- NULL
        if (length(strata) > 0) {
            if (augment) {
                model_function <-
                    function(formula,
                            data,
                            weight_scheme = NULL,
                            na.action) {
                        formula <-
                            stats::formula(paste0(
                                safe_deparse(formula),
                                ' + strata(',
                                strata,
                                ')'
                            ))
                        
                        weight_sch_current <- weight_scheme
                        assign("weight_sch_current",
                            weight_scheme,
                            envir = environment(formula))
                        
                        clogit_out <- tryCatch({
                            fit1 <- survival::clogit(
                                stats::formula(formula),
                                data = data,
                                method = "breslow",
                                control = survival::coxph.control(
                                    iter.max = 1000),
                                na.action = na.action,
                                weights = weight_sch_current,
                                robust = FALSE
                            ) # Robust SE seem to be worse with weighting...
                        }, warning = function(w) {
                            'warning'
                        }, error = function(e) {
                            'error'
                        })
                        
                        if (is.character(clogit_out)) {
                            fit1 <- survival::clogit(
                                stats::formula(formula),
                                data = data,
                                method = "breslow",
                                control = survival::coxph.control(
                                    iter.max = 1000),
                                na.action = na.action,
                                weights = weight_sch_current,
                                robust = FALSE
                            ) # Robust SE seem to be worse with weighting...
                            return(fit1)
                        } else  {
                            return(clogit_out)
                        }
                    }
            } else {
                model_function <-
                    function(formula,
                            data,
                            weight_scheme = NULL,
                            na.action) {
                        clogit_out <- tryCatch({
                            fit1 <- survival::clogit(
                                stats::formula(formula),
                                data = data,
                                method = "breslow",
                                control = survival::coxph.control(
                                    iter.max = 1000),
                                na.action = na.action,
                                robust = FALSE
                            ) # Robust SE seem to be worse 
                            # with weighting...)
                        }, warning = function(w) {
                            'warning'
                        }, error = function(e) {
                            'error'
                        })
                        
                        if (is.character(clogit_out)) {
                            fit1 <- survival::clogit(
                                stats::formula(formula),
                                data = data,
                                method = "breslow",
                                control = survival::coxph.control(
                                    iter.max = 1000),
                                na.action = na.action,
                                robust = FALSE
                            ) # Robust SE seem to be 
                            # worse with weighting...)
                            return(fit1)
                        } else  {
                            return(clogit_out)
                        }
                    }
            }
            summary_function <-
                function(fit, names_to_include) {
                    summ = survival:::summary.coxph # Scoped calls - this will be running on daemons
                    
                    lm_summary <- summ(fit)$coefficients
                    
                    store_names <- rownames(lm_summary)
                    if (!all(names_to_include %in% store_names)) {
                        # If deficient rank, 
                        # make sure all rownames are included
                        rows_to_add <-
                            names_to_include[
                                !(names_to_include %in% store_names)]
                        lm_summary <-
                            rbind(lm_summary,
                                matrix(
                                    rep(
                                        NaN,
                                        ncol(lm_summary) * 
                                            length(rows_to_add)
                                    ),
                                    nrow = length(rows_to_add)
                                ))
                        rownames(lm_summary) <-
                            c(store_names, rows_to_add)
                    }
                    
                    if ('robust se' %in% colnames(lm_summary)) {
                        para <-
                            as.data.frame(lm_summary) |> 
                            collapse::fselect(-c(2, 4, 5))
                        # Don't actually use robust SE
                    } else {
                        para <- as.data.frame(lm_summary) |> 
                            collapse::fselect(-c(2, 4))
                    }
                    
                    para$name <- rownames(lm_summary)
                    return(para)
                }
        } else {
            # Fixed effects only
            if (augment) {
                model_function <-
                    function(formula,
                            data,
                            weight_scheme = NULL,
                            na.action) {
                        weight_sch_current <- weight_scheme
                        assign("weight_sch_current",
                            weight_scheme,
                            envir = environment(formula))
                        
                        glm_out <- glm(
                            formula = stats::formula(formula),
                            family = 'binomial',
                            data = data,
                            weights = weight_sch_current,
                            na.action = na.action,
                        )
                        
                        return(glm_out)
                    }
            } else {
                model_function <-
                    function(formula,
                            data,
                            weight_scheme = NULL,
                            na.action) {
                        return(
                            glm(
                                stats::formula(formula),
                                data = data,
                                family = 'binomial',
                                na.action = na.action,
                            )
                        )
                    }
            }
            summary_function <-
                function(fit, names_to_include) {
                    lm_summary <- summary(fit)$coefficients
                    store_names <- rownames(lm_summary)
                    if (!all(names_to_include %in% store_names)) {
                        # If deficient rank, 
                        # make sure all rownames are included
                        rows_to_add <-
                            names_to_include[
                                !(names_to_include %in% store_names)]
                        lm_summary <-
                            rbind(lm_summary,
                                matrix(
                                    rep(NaN, 4 * length(rows_to_add)),
                                    nrow = length(rows_to_add)
                                ))
                        rownames(lm_summary) <-
                            c(store_names, rows_to_add)
                    }
                    para <- as.data.frame(lm_summary) |> 
                        ftail(-1) |> 
                        collapse::fselect(-3)
                    
                    para$name <- rownames(lm_summary)[-1]
                    return(para)
                }
        }
    } else {
        # Random effects
        ranef_function <- lme4::ranef
        
        if (augment) {
            model_function <-
                function(formula,
                        data,
                        weight_scheme = NULL,
                        na.action) {
                    weight_sch_current <- weight_scheme
                    assign("weight_sch_current",
                        weight_scheme,
                        envir = environment(formula))
                    
                    index <- 1
                    
                    while (index < length(optimizers)) {
                        glm_out <- tryCatch({
                            withCallingHandlers({
                                # Catch non-integer # successes first
                                fit1 <- lme4::glmer(
                                    stats::formula(formula),
                                    data = data,
                                    family = 'binomial',
                                    na.action = na.action,
                                    weights = weight_sch_current,
                                    control = lme4::glmerControl(
                                        optimizer = optimizers[index],
                                        optCtrl = optCtrlList[[index]]
                                    )
                                )
                            }, warning = function(w) {
                                if (w$message == 
                                "non-integer #successes in a binomial glm!") {
                                    # Still worked
                                    invokeRestart("muffleWarning")
                                }
                            })
                        }, warning = function(w) {
                            'warning'
                        }, error = function(e) {
                            'error'
                        })
                        
                        # Something warned or errored if here
                        if (is.character(glm_out)) {
                            index <- index + 1
                        } else  {
                            break
                        }
                    }
                    
                    if (is.character(glm_out)) {
                        withCallingHandlers({
                            # Catch non-integer # successes first
                            fit1 <- lme4::glmer(
                                stats::formula(formula),
                                data = data,
                                family = 'binomial',
                                na.action = na.action,
                                weights = weight_sch_current,
                                control = lme4::glmerControl(
                                    optimizer = optimizers[index],
                                    optCtrl = optCtrlList[[index]]
                                )
                            )
                        }, warning = function(w) {
                            if (w$message == 
                                "non-integer #successes in a binomial glm!") {
                                # Still worked
                                invokeRestart("muffleWarning")
                            }
                        })
                        return(fit1)
                    } else  {
                        return(glm_out)
                    }
                }
        } else {
            model_function <-
                function(formula,
                        data,
                        weight_scheme = NULL,
                        na.action) {
                    index <- 1
                    
                    while (index < length(optimizers)) {
                        glm_out <- tryCatch({
                            lme4::glmer(
                                stats::formula(formula),
                                data = data,
                                family = 'binomial',
                                na.action = na.action,
                                control = lme4::glmerControl(
                                    optimizer = optimizers[index],
                                    optCtrl = optCtrlList[[index]]
                                )
                            )
                        }, warning = function(w) {
                            'warning'
                        }, error = function(e) {
                            'error'
                        })
                        
                        # Something warned or errored if here
                        if (is.character(glm_out)) {
                            index <- index + 1
                        } else  {
                            break
                        }
                    }
                    
                    if (is.character(glm_out)) {
                        return(
                            lme4::glmer(
                                stats::formula(formula),
                                data = data,
                                family = 'binomial',
                                na.action = na.action,
                                control = lme4::glmerControl(
                                    optimizer = optimizers[index],
                                    optCtrl = optCtrlList[[index]]
                                )
                            )
                        )
                    } else  {
                        return(glm_out)
                    }
                }
        }
        summary_function <- function(fit, names_to_include) {
            lm_summary <- summary(fit)$coefficients
            
            store_names <- rownames(lm_summary)
            if (!all(names_to_include %in% store_names)) {
                # If deficient rank, make sure all rownames are included
                rows_to_add <-
                    names_to_include[!(names_to_include %in% store_names)]
                lm_summary <-
                    rbind(lm_summary, matrix(rep(
                        NaN, 4 * length(rows_to_add)
                    ), nrow = length(rows_to_add)))
                rownames(lm_summary) <-
                    c(store_names, rows_to_add)
            }
            
            para <- as.data.frame(lm_summary) |> 
                ftail(-1) |> 
                collapse::fselect(-3)
            
            para$name <- rownames(lm_summary)[-1]
            
            return(para)
        }
    }
    return(list("ranef_function" = ranef_function, 
                "model_function" = model_function, 
                "summary_function" = summary_function))
}

# Return blank results if there are zero or one samples
check_for_zero_one_obs <- function(formula,
                                    random_effects_formula,
                                    dat_sub,
                                    groups,
                                    ordereds,
                                    # features,
                                    fn,
                                    x,
                                    model,
                                    feature_specific_covariate_name) {
    match.arg(model, c("linear", "logistic"))
    if (length(unique(dat_sub$expr)) < 2) {
        output <- list()
        
        # List fixed effects that will be included
        names_to_include <-
            get_fixed_effects(formula,
                            random_effects_formula,
                            dat_sub,
                            groups,
                            ordereds,
                            feature_specific_covariate_name)
        
        # Build outputs
        output$para <-
            as.data.frame(matrix(
                NA,
                nrow = length(names_to_include),
                ncol = 3
            ))
        output$para$name <- names_to_include
        
        output$residuals <- NA
        output$fitted <- NA
        if (!(is.null(random_effects_formula)))
            output$ranef <- NA
        output$fit <- NA
        
        output$para = collapse::setColnames(output$para, c('coef', 'stderr' , 'pval', 'name'))
        
        output$para$feature <- fn #colnames(features)[x]
        
        output$para$error <-
            if (model == "logistic") {
                "All logistic values are the same"
            } else {
                "All linear values are the same"
            }
        return(output)
    }
    return (NULL)
}

check_mffl_one = function(v, i, ex, fxf) {
    res = FALSE
    
    if (!is.factor(v)) return(FALSE)
    
    fl = collapse::ffirst(levels(v))
    
    fl_expr = ex[collapse::whichv(v, fl)]
    
    nc_fct = nchar(i)
    
    fxf_nms = substr(fxf, 1, nc_fct) |> collapse::funique()
    
    if (collapse::allNA(fl_expr) && (i %in% fxf_nms)) {
        TRUE
    } else {
        FALSE
    }
}

# If the baseline level is missing for a factor, return blank results
check_missing_first_factor_level <- function(formula,
                                            random_effects_formula,
                                            dat_sub,
                                            groups,
                                            ordereds,
                                            fn,
                                            feature_specific_covariate_name) {
    missing_first_factor_level <- FALSE
    
    fxf <- get_fixed_effects(formula,
                             random_effects_formula,
                             dat_sub,
                             groups,
                             ordereds,
                             feature_specific_covariate_name)
    
    missing_first_factor_level <- any(mapply(FUN = check_mffl_one, 
                                             as.list(dat_sub),
                                             colnames(dat_sub),
                                             MoreArgs = list(ex = dat_sub$expr,
                                                             fxf = fxf)))
    
    if (missing_first_factor_level) {
        output <- list()
        
        # List fixed effects that will be included
        names_to_include <- fxf
        
        # Build outputs
        output$para <-
            as.data.frame(matrix(
                NA,
                nrow = length(names_to_include),
                ncol = 3
            ))
        
        output$para$name <- names_to_include
        
        output$residuals <- NA
        
        output$fitted <- NA
        
        if (!(is.null(random_effects_formula))) output$ranef <- NA
        
        output$fit <- NA
        
        output$para = collapse::setColnames(output$para, c('coef', 'stderr' , 'pval', 'name')) 
        
        output$para$feature <- fn # colnames(features)[x]
        
        output$para$error <-
            "No data points have the baseline factor level"
        return(output)
    }
    
    return (NULL)
}

# Fit the augmented logistic model
fit_augmented_logistic <- function(ranef_function,
                                model_function,
                                formula,
                                random_effects_formula,
                                groups,
                                ordereds,
                                dat_sub,
                                # features,
                                fn,
                                x) {
    warning_message <- NA
    
    error_message <- NA
    
    calling_env <- environment()
    
    mm_input <- NULL
    
    weight_scheme <- NULL
    
    fit1 <- tryCatch({
        withCallingHandlers({
            withCallingHandlers({
                # Catch non-integer # successes first
                formula_new <-
                    stats::formula(paste0(
                        c(
                            safe_deparse(formula),
                            groups,
                            ordereds
                        ),
                        collapse = " + "
                    ))
                
                # Augment data
                augmented_data <-
                    augment_data(formula_new,
                                random_effects_formula,
                                dat_sub)
                mm_input <- augmented_data[["mm_input"]]
                weight_scheme <-
                    augmented_data[["weight_scheme"]]
                formula_new <-
                    augmented_data[["new_formula"]]
                
                # Fit augmented model
                fit1 <-
                    model_function(
                        formula = formula_new,
                        data = mm_input,
                        weight_scheme = weight_scheme,
                        na.action = na.exclude
                    )
                fit1
            }, warning = function(w) {
                if (w$message == 
                    "non-integer #successes in a binomial glm!") {
                    # Still worked
                    invokeRestart("muffleWarning")
                }
            })
        }, warning = function(w) {
            message(sprintf("Feature %s : %s", 
                            fn, w))
            logging::logwarn(paste(
                "Fitting problem for feature",
                x,
                "a warning was issued"
            ))
            
            assign("warning_message",
                conditionMessage(w),
                envir = calling_env)
            invokeRestart("muffleWarning")
        })
    },
    error = function(err) {
        assign("error_message", err$message, envir = calling_env)
        error_obj <-
            structure(list(message = conditionMessage(err)), 
                    class = "try-error")
        return(error_obj)
    })
    
    if (!is.na(error_message)) {
        fit_and_message <- c(list(fit1), list(error_message))
        error_message <- NA
    } else if (!is.na(warning_message)) {
        fit_and_message <- c(list(fit1), list(warning_message))
        warning_message <- NA
    } else {
        fit_and_message <- c(list(fit1), NA)
    }
    
    return(list("fit_and_message" = fit_and_message,
                "weight_scheme" = weight_scheme,
                "mm_input" = mm_input))
}

# Fit linear or non-augmented logistic models
non_augmented <- function(ranef_function,
                        model_function,
                        formula,
                        random_effects_formula,
                        groups,
                        ordereds,
                        dat_sub,
                        # features,
                        fn,
                        x) {

    warning_message <- NA
    
    error_message <- NA
    
    calling_envir <- environment()
    
    fit1 <- tryCatch({
        withCallingHandlers({
            formula_new <-
                stats::formula(paste0(
                    c(safe_deparse(formula), groups, ordereds),
                    collapse = " + "
                ))
            
            fit1 <-
                model_function(formula_new,
                            data = dat_sub,
                            na.action = na.exclude)
            fit1
        }, warning = function(w) {
            message(sprintf("Feature %s : %s", 
                            fn, w))
            logging::logwarn(paste(
                "Fitting problem for feature",
                x,
                "a warning was issued"
            ))
            
            assign("warning_message",
                conditionMessage(w),
                envir = calling_envir)
            invokeRestart("muffleWarning")
        })
    },
    error = function(err) {
        assign("error_message", err$message, envir = calling_envir)
        error_obj <-
            structure(list(message = conditionMessage(err)), 
                    class = "try-error")
        return(error_obj)
    })
    
    if (!is.na(error_message)) {
        fit_and_message <- c(list(fit1), list(error_message))
        error_message <- NA
    } else if (!is.na(warning_message)) {
        fit_and_message <- c(list(fit1), list(warning_message))
        warning_message <- NA
    } else {
        fit_and_message <- c(list(fit1), NA)
    }
    
    return(list("fit_and_message" = fit_and_message,
                "weight_scheme" = NULL,
                "mm_input" = NULL))
}

# Fit models with group effects
run_group_models <- function(ranef_function,
                            model_function,
                            groups,
                            formula,
                            random_effects_formula,
                            model,
                            fit,
                            augment,
                            weight_scheme,
                            dat_sub,
                            output,
                            mm_input) {
    match.arg(model, c("linear", "logistic"))
    
    output$para <- rbind(output$para,
                        setNames(do.call(rbind, lapply(groups, function(group) {
        as.data.frame(setNames(tryCatch({
            withCallingHandlers({
                # Catch non-integer # successes first
                if (is.null(random_effects_formula)) {
                    # Fixed effects
                    if (model == "logistic") {
                        pval_new <- tryCatch({
                            anova(fit, test = 'LRT')[group, 'Pr(>Chi)']
                        },
                        error = function(err) {
                            NA
                        })
                    } else {
                        pval_new <- tryCatch({
                            anova(fit)[group, 'Pr(>F)']
                        },
                        error = function(err) {
                            NA
                        })
                    }
                } else {
                    # Random effects
                    if (model == "logistic") {
                        if (augment) {
                            weight_sch_current <- weight_scheme
                            assign(
                                "weight_sch_current",
                                weight_scheme,
                                envir = environment(formula)
                            )
                            
                            fit_new <-
                                model_function(
                                    formula = stats::update.formula(
                                        stats::formula(fit),
                                        stats::formula(
                                            paste0('~.-', group)
                                        )
                                    ),
                                    data = mm_input,
                                    weight_scheme = 
                                        weight_sch_current,
                                    na.action = na.exclude
                                )
                        } else {
                            fit_new <-
                                model_function(
                                    stats::update.formula(
                                        stats::formula(fit),
                                        stats::formula(
                                            paste0('~.-', group)
                                        )
                                    ),
                                    data = dat_sub,
                                    na.action = na.exclude
                                )
                        }
                        
                        pval_new <-
                            tryCatch({
                                anova(fit_new, fit)[
                                    2, 'Pr(>Chisq)']
                            },
                            error = function(err) {
                                NA
                            })
                    } else {
                group_no_ticks <- gsub("^`|`$", "", group)
                contrast_mat <- matrix(0, ncol = length(
                    lme4::fixef(fit)),
                    nrow = length(levels(dat_sub[[group_no_ticks]])[-1])
                )
                if (length(which(names(lme4::fixef(fit)) %in% paste0(
                    group, levels(dat_sub[[group_no_ticks]])[-1]))) == 0) {
                    stop("group levels not found in fit")
                }
                contrast_mat[seq_along(levels(
                dat_sub[[group_no_ticks]])[-1]),
                which(names(lme4::fixef(fit)) %in% paste0(
                    group, levels(dat_sub[[group_no_ticks]])[-1]))] <-
                diag(1, nrow = length(levels(dat_sub[[group_no_ticks]])[-1]))
                        
                pval_new <- tryCatch({
                    lmerTest::contest(
                        fit, 
                        contrast_mat,
                        rhs = 
                            rep(0, length(levels(
                                dat_sub[[group_no_ticks]])[-1])))[['Pr(>F)']]
                },
                error = function(err) {
                    NA
                })
                    }
                }
                
                tmp_output <- list(NA, NA, pval_new, group)
                tmp_output
            }, warning = function(w) {
                if (w$message == "non-integer #successes in a binomial glm!") {
                    invokeRestart("muffleWarning")
                }
            })
        }, warning = function(w) {
            return(list(NA, NA, NA, group))
        }, error = function(err) {
            return(list(NA, NA, NA, group))
        }), names(output$para)), check.names = FALSE)
    })), names(output$para)))
    
    tmp_rownames <- rownames(output$para)
    
    output$para <- data.frame(lapply(output$para, function(x) {
        if (is.list(x)) unlist(x) else x
    }), check.names = FALSE)
    
    rownames(output$para) <- tmp_rownames
    
    rownames(output$para) <- c(rownames(output$para)[
        -seq((nrow(output$para) - length(groups) + 1), nrow(output$para))], 
        groups)
    return(output)
}

# Fit models with ordered effects
run_ordered_models <- function(ranef_function,
                            model_function,
                            ordereds,
                            fit_and_message,
                            formula,
                            random_effects_formula,
                            model,
                            fit,
                            augment,
                            weight_scheme,
                            dat_sub,
                            output) {
    match.arg(model, c("linear", "logistic"))
    output$para <- rbind(output$para,
                        setNames(do.call(rbind, lapply(ordereds, 
                            function(ordered) {
        ordered_no_ticks <- gsub("^`|`$", "", ordered)
        ordered_levels <- paste0(ordered,
            levels(dat_sub[[ordered_no_ticks]])[-1])
        as.data.frame(setNames(tryCatch({
            withCallingHandlers({
                # Catch non-integer # successes first
                if (is.null(random_effects_formula)) {
                    # Fixed effects
                    if (any(!ordered_levels %in% names(coef(fit)))) {
                        fit_and_message[[
                            length(fit_and_message)]] <-
                            "Error: Some ordered levels are missing"
                        stop("Some ordered levels are missing")
                    }
                    
                    contrast_mat <-
                        matrix(0, ncol = length(coef(fit, complete = FALSE)),
                        nrow = length(levels(dat_sub[[ordered_no_ticks]])[-1])
                        )
                    
                    cols_to_add_1s <-
                        which(names(coef(fit, complete = FALSE)) %in% 
                                ordered_levels)
                    contrast_mat[1, cols_to_add_1s[1]] <- 1
                    for (i in seq_along(cols_to_add_1s[-1])) {
                        contrast_mat[i + 1, cols_to_add_1s[-1][i]] <- 1
                        contrast_mat[i + 1, cols_to_add_1s[i]] <- -1
                    }
                    
                    pvals_new <- vapply(seq(nrow(contrast_mat)), 
                                        function(row_num) {
                        contrast_vec <- t(matrix(contrast_mat[row_num, ]))
                        tryCatch({
                            summary(multcomp::glht(fit, linfct = contrast_vec, 
                rhs = 0, 
                coef. = function(x) { coef(x, complete = FALSE) }))$test$pvalues
                        }, error = function(err) { NA })
                    }, numeric(1))
                    
                    coefs_new <- vapply(seq(nrow(contrast_mat)), 
                                        function(row_num) {
                        contrast_vec <- t(matrix(contrast_mat[row_num, ]))
                        tryCatch({
                            summary(multcomp::glht(fit, linfct = contrast_vec, 
                rhs = 0, 
                coef. = function(x) { 
                    coef(x, complete = FALSE) }))$test$coefficients
                        }, error = function(err) { NA })
                    }, numeric(1))
                    
                    sigmas_new <- vapply(seq(nrow(contrast_mat)), 
                                        function(row_num) {
                        contrast_vec <- t(matrix(contrast_mat[row_num, ]))
                        tryCatch({
                            summary(multcomp::glht(fit, linfct = contrast_vec, 
                rhs = 0, 
                coef. = function(x) { coef(x, complete = FALSE) }))$test$sigma
                        }, error = function(err) { NA })
                    }, numeric(1))
                } else {
                    # Random effects
                    if (any(!ordered_levels %in% names(lme4::fixef(fit)))) {
                        fit_and_message[[length(fit_and_message)]] <-
                            "Error: Some ordered levels are missing"
                        stop("Some ordered levels are missing")
                    }
                    
                    ordered_no_ticks <- gsub("^`|`$", "", ordered)
                    contrast_mat <-
                        matrix(0, ncol = length(lme4::fixef(fit)),
                        nrow = length(levels(dat_sub[[ordered_no_ticks]])[-1])
                        )
                    
                    cols_to_add_1s <-
                        which(names(lme4::fixef(fit)) %in% ordered_levels)
                    contrast_mat[1, cols_to_add_1s[1]] <- 1
                    for (i in seq_along(cols_to_add_1s[-1])) {
                        contrast_mat[i + 1, cols_to_add_1s[-1][i]] <- 1
                        contrast_mat[i + 1, cols_to_add_1s[i]] <- -1
                    }
                    
                    if (model == "logistic") {
        pvals_new <- vapply(seq(nrow(contrast_mat)), 
            function(row_num) {
                contrast_vec <- t(matrix(contrast_mat[row_num, ]))
                tryCatch({
                    summary(multcomp::glht(fit, linfct = contrast_vec, 
                                        rhs = 0))$test$pvalues
                }, error = function(err) { NA })
            }, numeric(1))
                        
        coefs_new <- vapply(seq(nrow(contrast_mat)), 
            function(row_num) {
                contrast_vec <- t(matrix(contrast_mat[row_num, ]))
                tryCatch({
                    summary(multcomp::glht(fit, linfct = contrast_vec, 
                                        rhs = 0))$test$coefficients
                }, error = function(err) { NA })
            }, numeric(1))
                        
        sigmas_new <- vapply(seq(nrow(contrast_mat)), 
            function(row_num) {
                contrast_vec <- t(matrix(contrast_mat[row_num, ]))
                tryCatch({
                    summary(multcomp::glht(fit, linfct = contrast_vec, 
                    rhs = 0))$test$sigma
                }, error = function(err) { NA })
            }, numeric(1))
                    } else {
        pvals_new <- vapply(seq(nrow(contrast_mat)), 
            function(row_num) {
                contrast_vec <- t(matrix(contrast_mat[row_num, ]))
                tryCatch({
                    lmerTest::contest(fit,
                                        matrix(
                                        contrast_vec,
                                        TRUE
                                        ), rhs = 0)[['Pr(>F)']]
                }, error = function(err) { NA })
            }, numeric(1))
        
        coefs_new <- vapply(seq(nrow(contrast_mat)), 
            function(row_num) {
                contrast_vec <- t(matrix(contrast_mat[row_num, ]))
                tryCatch({
                    contrast_vec %*% lme4::fixef(fit)
                }, error = function(err) { NA })
            }, numeric(1))
        
        sigmas_new <- vapply(seq(nrow(contrast_mat)), 
            function(row_num) {
                contrast_vec <- t(matrix(contrast_mat[row_num, ]))
                tryCatch({
                    sqrt((contrast_vec %*% vcov(fit) %*% t(contrast_vec))[1, 1])
                }, error = function(err) { NA })
            }, numeric(1))
            }
                }
                
                tmp_output <-
                    data.frame(coefs_new,
                        sigmas_new,
                        pvals_new,
                        ordered_levels
                        )
                rownames(tmp_output) <- ordered_levels
                tmp_output
            }, warning = function(w) {
                if (w$message == 
                    "non-integer #successes in a binomial glm!") {
                    # Still worked
                    invokeRestart("muffleWarning")
                }
            })
        }, warning = function(w) {
            tmp_output <- data.frame(matrix(nrow = length(ordered_levels),
                                            ncol = 4))
            tmp_output[seq(1,length(ordered_levels)), seq(3)] <- NA
            tmp_output[seq(1,length(ordered_levels)), 4] <- ordered_levels
            rownames(tmp_output)[seq(1,length(ordered_levels))] <-
                ordered_levels
            
            return(tmp_output)
        }, error = function(err) {
            tmp_output <- data.frame(matrix(nrow = length(ordered_levels),
                                            ncol = 4))
            tmp_output[seq(1,length(ordered_levels)), seq(3)] <- NA
            tmp_output[seq(1,length(ordered_levels)), 4] <- ordered_levels
            rownames(tmp_output)[seq(1,length(ordered_levels))] <-
                ordered_levels
            
            return(tmp_output)
        }), names(output$para)), check.names = FALSE)
    })), names(output$para)))
    
    return(output)
}

# Check whether the model fit properly and return modeling outputs
fitting_wrap_up <- function(fit_properly,
                            fit_and_message,
                            output,
                            fit,
                            random_effects_formula,
                            metadata,
                            median_comparison,
                            save_models,
                            formula,
                            dat_sub,
                            groups,
                            ordereds,
                            # features,
                            fn,
                            x,
                            ranef_function,
                            feature_specific_covariate_name) {
    if (fit_properly) {
        output$residuals <- stats::residuals(fit)
        output$fitted <- stats::fitted(fit)
        # ^ It would be better to unname these and put them together in a
        # co-indexed data frame. TODO:
        
        if (!(is.null(random_effects_formula))) {
            # Returns a list with a table for each random effect
            l <- ranef_function(fit)
            
            # Rename rows as random effect labels 
            # if only random intercepts
            rl_chk = length(l) == 1 &
                ncol(l[[1]]) == 1 &
                colnames(l[[1]])[1] == "(Intercept)"
            
            if (rl_chk) {
                
                d <- as.vector(unlist(l))
                
                names(d) <- row.names(l[[1]])
                
                d[setdiff(collapse::funique(metadata[, names(l)]), 
                          names(d))] <- NA
                
                d <- d[collapse::radixorder(collapse::funique(metadata[, names(l)]))]
                
                output$ranef <- d
                
            } else {
                # Otherwise return the random effects list
                output$ranef <- l
            }
        }
        
        if (median_comparison) {
            output$fit <- fit
        } else {
            if (save_models) {
                output$fit <- fit
            } else {
                output$fit <- NA
            }
        }
        
    } else {
        # Fitting issue
        logging::logwarn(paste("Fitting problem for feature",
                            x,
                            "returning NA"))
        
        names_to_include <-
            get_fixed_effects(formula,
                            random_effects_formula,
                            dat_sub,
                            groups,
                            ordereds,
                            feature_specific_covariate_name)
        
        # Store NA values for missing outputs
        output$para <-
            as.data.frame(matrix(
                NA,
                nrow = length(names_to_include),
                ncol = 3
            ))
        
        output$para$name <- names_to_include
        
        output$residuals <- NA
        
        output$fitted <- NA
        
        if (!(is.null(random_effects_formula))) output$ranef <- NA
        
        output$fit <- NA
    }
    
    if (is(output$fit, "merMod")) {
        # Delete the model frame in the object to reduce the size of the output
        # fit. It's just a redundant copy of the metadata. Same for these
        # elements of the reponse module in fit@resp.
        output$fit@frame <- data.frame()
        # For some reason this makes the models_*.rds file a lot larger TODO:
        # identify why or better yet save only what's needed.
        
        # The response slot is big too.
        # output$fit@resp$mu <- vector("numeric")
        # output$fit@resp$y <- vector("numeric")

        # TODO: check if this can/should be done for the other model types like
        # lm/glm/clogit The "y" element of clogit objects looks suspiciously
        # large. It would be best of all to just extract the
        # parameters/statistics of interest and drop everything else. If one
        # really wants access to the full fit object it would be best to save
        # individual feature fits to their own RDS files under fits/models_*/
        # from within func_to_run() and avoid returning it. That would save
        # time, memory, and avoid creating huge models_logistic.rds files.

        # The fits also contain redundant copies of the fitted / residual values
        # too. Those can/should be deleted too probably.

        # OOP - What I want is a banana. What I get is a gorilla holding a banana.
    }
    
    colnames(output$para) <-
        c('coef', 'stderr' , 'pval', 'name')
    
    output$para$feature <- fn # colnames(features)[x]
    
    output$para$error <-
        fit_and_message[[length(fit_and_message)]]
    
    output$para$error[is.na(output$para$error) &
                        is.na(output$para$pval)] <-
        'Fitting error (NA p-value returned from fitting procedure)'
    
    return(output)
}

# Run the median comparison procedure for ordered predictors
run_median_comparison_ordered <- function(paras_sub,
                                        fits,
                                        ordereds,
                                        metadata,
                                        random_effects_formula,
                                        median_comparison_threshold,
                                        metadata_variable,
                                        pvals_new,
                                        cur_median,
                                        model) {
    match.arg(model, c("linear", "logistic"))
    
    ordered <- ordereds[which(startsWith(
        metadata_variable, ordereds))]
    
    use_this_coef <- !is.na(paras_sub$pval) & paras_sub$pval < 0.95
    
    n_coefs <- nrow(paras_sub)
    
    sigmas <- paras_sub$stderr
    
    coefs <- paras_sub$coef
    
    sigma_sq_med <- var(coefs[use_this_coef], na.rm=TRUE)
    
    # Variance from asymptotic distribution
    sd_median <- sqrt(0.25 * 2 * base::pi * sigma_sq_med / sum(use_this_coef))
    
    # MC for covariance
    nsims <- 10000
    
    sim_results <- replicate(nsims, {
        sim_coefs <- rnorm(n_coefs, coefs, sigmas)
        sim_median <- median(sim_coefs[use_this_coef])
        c(sim_median, sim_coefs)
    })
    
    sim_medians <- sim_results[1, ]
    
    all_sims <- sim_results[-1, , drop = FALSE]
    
    cov_adjust <- apply(all_sims, 1, function(x){cov(x, sim_medians)})
    
    # Necessary offsets for contrast testing
    offsets_to_test <- abs(cur_median - coefs) * 
        sqrt((sigmas^2) / (sigmas^2+ sd_median^2 - 2 * cov_adjust)) + coefs
    
    pvals_new <- c(pvals_new,
        vapply(seq(nrow(paras_sub)), function(row_index) {
            feature <- paras_sub$feature[row_index]
            if (is.null(random_effects_formula)) {
                # Fixed effects
                cur_fit <- fits[[feature]]
                
                if (!metadata_variable %in% names(coef(cur_fit))) {
                    return(NA)
                }
                
                if (any(!paste0(ordered, 
                                levels(
                                    metadata[[ordered]])[-1]) %in% 
                        names(coef(cur_fit)))) {
                    return(NA)
                }
                
                mm_variable <-
                    model.matrix(cur_fit)[, metadata_variable]
                if (any(!unique(
                    mm_variable[!is.na(mm_variable)]) %in% c(0, 1))) {
                    median_comparison_threshold_updated <-
                        median_comparison_threshold / 
                        sd(mm_variable)
                } else {
                    median_comparison_threshold_updated <- 
                        median_comparison_threshold
                }
                
                if (is.na(coef(cur_fit, complete = FALSE)[
                    which(names(coef(
                        cur_fit, complete = FALSE
                    )) == metadata_variable)])) {
                    pval_new_current <- NA
                } else if (abs(coef(cur_fit, complete = FALSE)[
                    which(names(coef(
                        cur_fit, complete = FALSE
                    )) == metadata_variable)] -
                    cur_median) < median_comparison_threshold_updated) {
                    pval_new_current <- 1
                } else {
                    contrast_mat <-
                        matrix(0, ncol = length(
                            coef(cur_fit, complete = FALSE)),
                            nrow = length(levels(metadata[[ordered]])[-1])
                        )
                    
                    cols_to_add_1s <-
                        which(names(coef(cur_fit, complete = FALSE)) %in% 
                            paste0(ordered, levels(metadata[[ordered]])[-1]))
                    contrast_mat[1, cols_to_add_1s[1]] <- 1
                    for (i in seq_along(cols_to_add_1s[-1])) {
                        contrast_mat[i + 1, cols_to_add_1s[-1][i]] <- 1
                        contrast_mat[i + 1, cols_to_add_1s[i]] <- -1
                    }
                    
                    contrast_vec <-
                        t(matrix(contrast_mat[which(paste0(
                            ordered, levels(
                                metadata[[ordered]]
                            )[-1]) == metadata_variable),]))
                    pval_new_current <-
                        tryCatch({
                            summary(
                                multcomp::glht(
                                    cur_fit,
                                    linfct = contrast_vec,
                                    rhs = offsets_to_test[row_index],
                                    coef. = function(x) {
                                        coef(x, complete = FALSE)
                                    }
                                )
                            )$test$pvalues
                        },
                        error = function(err) {
                            NA
                        })
                }
                
                return(pval_new_current)
            } else {
                # Random effects
                cur_fit <- fits[[feature]]
                
                if (!metadata_variable %in% 
                    names(lme4::fixef(cur_fit))) {
                    NA
                }
                
                if (any(!paste0(ordered, 
                                levels(metadata[[
                                    ordered]])[-1]) %in% 
                        names(lme4::fixef(cur_fit)))) {
                    NA
                }
                
                mm_variable <-
                    model.matrix(cur_fit)[, metadata_variable]
                if (any(
                    !unique(mm_variable[!is.na(mm_variable)]) %in% 
                    c(0, 1))) {
                    median_comparison_threshold_updated <-
                        median_comparison_threshold / 
                        sd(mm_variable)
                } else {
                    median_comparison_threshold_updated <- 
                        median_comparison_threshold
                }
                
                contrast_mat <-
                    matrix(
                        0,
                        ncol = length(lme4::fixef(cur_fit)),
                        nrow = length(levels(
                            metadata[[ordered]])[-1])
                    )
                cols_to_add_1s <-
                    which(names(lme4::fixef(cur_fit)) %in% 
                        paste0(ordered, levels(
                        metadata[[ordered]])[-1]))
                contrast_mat[1, cols_to_add_1s[1]] <- 1
                for (i in seq_along(cols_to_add_1s[-1])) {
                    contrast_mat[i + 1, cols_to_add_1s[-1][i]] <- 1
                    contrast_mat[i + 1, cols_to_add_1s[i]] <- -1
                }
                contrast_vec <-
                    t(matrix(
                        contrast_mat[
                            which(paste0(
                                ordered, levels(metadata[[
                                    ordered]])[-1]) == 
                                    metadata_variable),]))
                
                if (model == "logistic") {
                    if (is.na(lme4::fixef(cur_fit)[
                        which(names(lme4::fixef(
                            cur_fit
                        )) == metadata_variable)])) {
                        pval_new_current <- NA
                    } else if (abs(lme4::fixef(cur_fit)[
                        which(names(lme4::fixef(
                            cur_fit
                        )) == metadata_variable)] -
                        cur_median) < 
                        median_comparison_threshold_updated) {
                        pval_new_current <- 1
                    } else {
                        pval_new_current <-
                            tryCatch({
                                summary(
                                    multcomp::glht(
                                        cur_fit,
                                        linfct = contrast_vec,
                                        rhs = offsets_to_test[row_index]
                                    )
                                )$test$pvalues
                            },
                            error = function(err) {
                                NA
                            })
                    } 
                    return(pval_new_current)
                } else {
                    if (is.na(lme4::fixef(cur_fit)[
                        which(names(lme4::fixef(
                            cur_fit
                        )) == metadata_variable)])) {
                        pval_new_current <- NA
                    } else if (abs(lme4::fixef(cur_fit)[
                        which(names(lme4::fixef(
                            cur_fit
                        )) == metadata_variable)] -
                        cur_median) < 
                        median_comparison_threshold_updated) {
                        pval_new_current <- 1
                    } else {
                        pval_new_current <-
                            tryCatch({
                                lmerTest::contest(cur_fit,
                                            matrix(contrast_vec, TRUE),
                                            rhs = offsets_to_test[row_index])[[
                                                'Pr(>F)']]
                            },
                            error = function(err) {
                                NA
                            })
                    }
                    
                    return(pval_new_current)
                }
            }
        }, numeric(1)))
    
    return(pvals_new)
}

# Run the median comparison procedure for non-ordered predictors
run_median_comparison_general <- function(paras_sub,
                                        fits,
                                        metadata,
                                        random_effects_formula,
                                        median_comparison_threshold,
                                        metadata_variable,
                                        pvals_new,
                                        cur_median,
                                        model) {
    match.arg(model, c("linear", "logistic"))
    use_this_coef <- !is.na(paras_sub$pval) & paras_sub$pval < 0.95
    
    n_coefs <- nrow(paras_sub)
    sigmas <- paras_sub$stderr
    coefs <- paras_sub$coef
    sigma_sq_med <- var(coefs[use_this_coef], na.rm=TRUE)
    
    # Variance from asymptotic distribution
    sd_median <- sqrt(0.25 * 2 * base::pi * sigma_sq_med / sum(use_this_coef))

    # MC for covariance
    nsims <- 10000
    sim_results <- replicate(nsims, {
        sim_coefs <- rnorm(n_coefs, coefs, sigmas)
        sim_median <- median(sim_coefs[use_this_coef])
        c(sim_median, sim_coefs)
    })
    
    sim_medians <- sim_results[1, ]
    all_sims <- sim_results[-1, , drop = FALSE]
    cov_adjust <- apply(all_sims, 1, function(x){cov(x, sim_medians)})
    
    # Necessary offsets for contrast testing
    offsets_to_test <- abs(cur_median - coefs) * 
        sqrt((sigmas^2) / (sigmas^2+ sd_median^2 - 2 * cov_adjust)) + coefs
    
    pvals_new <- c(pvals_new,
        vapply(seq(nrow(paras_sub)), function(row_index) {
            feature <- paras_sub$feature[row_index]
            if (is.null(random_effects_formula)) {
                # Fixed effects
                cur_fit <- fits[[feature]]
                
                if (!metadata_variable %in% names(coef(cur_fit))) {
                    return(NA)
                }
                
                mm_variable <-
                    model.matrix(cur_fit)[, metadata_variable]
                if (any(
                    !unique(mm_variable[!is.na(mm_variable)]) %in% 
                    c(0, 1))) {
                    median_comparison_threshold_updated <-
                        median_comparison_threshold / 
                        sd(mm_variable)
                } else {
                    median_comparison_threshold_updated <- 
                        median_comparison_threshold
                }
                
                contrast_vec <-
                    rep(0, length(coef(
                        cur_fit, complete = FALSE
                    )))
                contrast_vec[which(names(coef(
                    cur_fit, complete = FALSE
                )) == metadata_variable)] <- 1
                
                if (is.na(coef(cur_fit, complete = FALSE)[
                    which(names(coef(
                        cur_fit, complete = FALSE
                    )) == metadata_variable)])) {
                    pval_new_current <- NA
                } else if (abs(coef(cur_fit, complete = FALSE)[
                    which(names(coef(
                        cur_fit, complete = FALSE
                    )) == metadata_variable)] -
                    cur_median) < median_comparison_threshold_updated) {
                    pval_new_current <- 1
                } else {
                    pval_new_current <-
                        tryCatch({
                            summary(
                                multcomp::glht(
                                    cur_fit,
                                    linfct = matrix(contrast_vec, 
                                                    TRUE),
                                    rhs = offsets_to_test[row_index],
                                    coef. = function(x) {
                                        coef(x, complete = FALSE)
                                    }
                                )
                            )$test$pvalues[1]
                        },
                        error = function(err) {
                            NA
                        })
                }
                
                return(pval_new_current)
            } else {
                # Random effects
                cur_fit <- fits[[feature]]
                
                if (!metadata_variable %in% 
                    names(lme4::fixef(cur_fit))) {
                    return(NA)
                }
                
                mm_variable <-
                    model.matrix(cur_fit)[, metadata_variable]
                if (any(!unique(
                    mm_variable[!is.na(mm_variable)]) %in% 
                    c(0, 1))) {
                    median_comparison_threshold_updated <-
                        median_comparison_threshold / 
                        sd(mm_variable)
                } else {
                    median_comparison_threshold_updated <- 
                        median_comparison_threshold
                }
                
                contrast_vec <- rep(0, length(lme4::fixef(cur_fit)))
                contrast_vec[which(names(lme4::fixef(cur_fit)) == 
                            metadata_variable)] <- 1
                
                if (model == "logistic") {
                    if (is.na(lme4::fixef(cur_fit)[
                        which(names(lme4::fixef(
                            cur_fit
                        )) == metadata_variable)])) {
                        pval_new_current <- NA
                    } else if (abs(lme4::fixef(cur_fit)[
                        which(names(lme4::fixef(
                            cur_fit
                        )) == metadata_variable)] -
                        cur_median) < 
                        median_comparison_threshold_updated) {
                        pval_new_current <- 1
                    } else {
                        pval_new_current <-
                            tryCatch({
                                summary(
                                    multcomp::glht(
                                        cur_fit,
                                        linfct = matrix(
                                            contrast_vec, TRUE),
                                        rhs = offsets_to_test[row_index]
                                    )
                                )$test$pvalues[1]
                            },
                            error = function(err) {
                                NA
                            })
                    }
                    
                    return(pval_new_current)
                } else {
                    if (is.na(lme4::fixef(cur_fit)[
                        which(names(lme4::fixef(
                            cur_fit
                        )) == metadata_variable)])) {
                        pval_new_current <- NA
                    } else if (abs(lme4::fixef(cur_fit)[
                        which(names(lme4::fixef(
                            cur_fit
                        )) == metadata_variable)] -
                        cur_median) < 
                        median_comparison_threshold_updated) {
                        pval_new_current <- 1
                    } else {
                        pval_new_current <-
                            tryCatch({
                                lmerTest::contest(
                                    cur_fit,
                                    matrix(contrast_vec, TRUE),
                                    rhs = 
                                    offsets_to_test[row_index])[['Pr(>F)']]
                            },
                            error = function(err) {
                                NA
                            })
                    }
                    
                    return(pval_new_current)
                }
            }
        }, numeric(1)))
    return(pvals_new)
}

# Run the median comparison for all variables
run_median_comparison <- function(paras,
                                fits,
                                ordereds,
                                groups,
                                metadata,
                                random_effects_formula,
                                median_comparison_threshold,
                                subtract_median,
                                model) {
    match.arg(model, c("linear", "logistic"))
    logging::loginfo("Performing tests against medians")
    
    
    if (length(ordereds) > 0) {
        ordered_levels <- unlist(lapply(ordereds, function(ordered) {
            unlist(paste0(ordered, levels(metadata[[ordered]])[-1]))
        }))
    } else {
        ordered_levels <- c()
    }
    
    final_paras <-
        paras[!is.na(paras$error) | paras$name %in% groups,]
    final_paras$null_hypothesis <- rep(NA, nrow(final_paras))
    paras <-
        paras[is.na(paras$error) & !paras$name %in% groups,]
    
    process_metadata_variable <- function(metadata_variable,
                                        paras,
                                        fits,
                                        ordered_levels,
                                        ordereds,
                                        metadata,
                                        random_effects_formula, 
                                        median_comparison_threshold, 
                                        subtract_median, 
                                        model) {
        
        paras_sub <- paras[paras$name == metadata_variable, ]
        cur_median <- median(paras_sub$coef[!is.na(paras_sub$pval) & 
                                        paras_sub$pval < 0.95], na.rm = TRUE)
        
        if (is.na(cur_median)) {
            pvals_new <- rep(NA, nrow(paras_sub))
            coefs_new <- paras_sub$coef
        } else {
            pvals_new <- vector()
            if (subtract_median) {
                coefs_new <- paras_sub$coef - cur_median
            } else {
                coefs_new <- paras_sub$coef
            }
            
            if (metadata_variable %in% ordered_levels) {
                pvals_new <- run_median_comparison_ordered(
                    paras_sub, 
                    fits, 
                    ordereds, 
                    metadata, 
                    random_effects_formula, 
                    median_comparison_threshold, 
                    metadata_variable, 
                    pvals_new, 
                    cur_median, 
                    model
                )
            } else {
                pvals_new <- run_median_comparison_general(
                    paras_sub, 
                    fits, 
                    metadata, 
                    random_effects_formula, 
                    median_comparison_threshold, 
                    metadata_variable, 
                    pvals_new, 
                    cur_median, 
                    model
                )
            }
        }
        
        paras_sub$error <- data.table::fifelse(
            is.na(pvals_new) & !is.na(paras_sub$pval), 
            "P-value became NA in median comparison, try rerunning without the median comparison", 
            paras_sub$error
        )
        
        paras_sub$pval <- pvals_new
        paras_sub$coef <- coefs_new
        paras_sub$null_hypothesis <- cur_median
        
        return(paras_sub)
    }
    
    final_paras_list <- lapply(unique(paras$name), function(metadata_variable) {
        process_metadata_variable(metadata_variable,
                                paras,
                                fits,
                                ordered_levels,
                                ordereds,
                                metadata,
                                random_effects_formula,
                                median_comparison_threshold,
                                subtract_median,
                                model)
    })
    
    final_paras <- collapse::rowbind(final_paras, 
                                     collapse::rowbind(final_paras_list))

    paras <- final_paras
    return(paras)
}

# fit the data using the model selected and applying the correction
fit.model <- function(features,
                    metadata,
                    model,
                    formula = NULL,
                    random_effects_formula = NULL,
                    correction = "BH",
                    save_models = FALSE,
                    small_random_effects = FALSE,
                    augment = FALSE,
                    cores = 1,
                    median_comparison = FALSE,
                    median_comparison_threshold = 0,
                    subtract_median = FALSE,
                    feature_specific_covariate = NULL,
                    feature_specific_covariate_name = NULL,
                    feature_specific_covariate_record = NULL) {
    match.arg(model, c("linear", "logistic"))
    match.arg(correction,
            c("BH", "holm", "hochberg", "hommel", "bonferroni", "BY"))
    check_formulas_valid(formula, random_effects_formula)
    formula <- stats::as.formula(formula, env = baseenv())

    extract_out <- extract_special_predictor(formula, 'group')
    formula <- extract_out[[1]]
    groups <- extract_out[[2]]

    extract_out <- extract_special_predictor(formula, 'ordered')
    formula <- extract_out[[1]]
    ordereds <- extract_out[[2]]

    extract_out <- extract_special_predictor(formula, 'strata')
    formula <- extract_out[[1]]
    strata <- extract_out[[2]]
    
    if (small_random_effects & model == 'logistic') {
        fixed_part <- lme4::nobars(formula)
        random_terms <- reformulas::findbars(formula)
        if (length(random_terms) > 0) {
            grouping <- unique(
                vapply(random_terms, function(x) deparse(x[[3]]), 
                    FUN.VALUE = character(length(random_terms))))
            grouping <- data.table::fifelse(grouping == make.names(grouping),
                grouping, paste0('`', grouping, '`'))
            new_formula <- as.formula(
                paste(safe_deparse(fixed_part), "+", 
                    paste(grouping, collapse = " + ")),
                env = environment(formula)
            )
            environment(new_formula) <- environment(formula)
            formula <- new_formula
        } else {
            stop("small_random_effects=TRUE but no random effects specified")
        }
        random_effects_formula <- NULL
    }
    
    if (length(strata) > 0 & !is.null(random_effects_formula)){
        stop(
            "Strata and random effects cannot be combined. Please only use
            random effects if you have multiple grouping categories."
        )
    }
    
    if (length(strata) > 0 & model == 'linear') {
        formula <-
            stats::formula(paste0(safe_deparse(formula), ' + (1 | ', strata, ')'))
        random_effects_formula <- formula
    }
    
    ################
    # Linear Model #
    ################
    
    if (model == "linear")
        fun_list <- choose_ranef_model_summary_funs_linear(
            random_effects_formula)
    
    ##################
    # Logistic Model #
    ##################
    
    if (model == "logistic") {
        fun_list <- choose_ranef_model_summary_funs_logistic(
            random_effects_formula,
            strata = strata,
            augment = augment)
    }
    
    ranef_function <- fun_list$ranef_function
    model_function <- fun_list$model_function
    summary_function <- fun_list$summary_function
    
    #######################################
    # Init cluster for parallel computing #
    #######################################
    
    # cluster <- NULL
    # if (cores > 1) {
    #     logging::loginfo("Creating cluster of %s R processes", cores)
    #     cluster <- parallel::makeCluster(cores)
        # parallel::clusterExport(cluster, c(ls(), function_vec),
                                # envir = environment())
    # }
    
    ##############################
    # Apply per-feature modeling #
    ##############################
   
    fit_vars = all.vars(formula) |> tail(-1)
    
    small_meta = metadata |>
        collapse::fselect(fit_vars)
    
    if (mirai::daemons_set()) {
        mirai::everywhere({}, 
                          metadata = small_meta,
                          random_effects_formula = random_effects_formula,
                          groups = groups,
                          ordereds = ordereds,
                          model = model,
                          feature_specific_covariate = feature_specific_covariate,
                          feature_specific_covariate_name = feature_specific_covariate_name,
                          formula = formula,
                          summary_function = summary_function,
                          ranef_function = ranef_function,
                          model_function = model_function,
                          augment = augment,
                          median_comparison = median_comparison,
                          get_fixed_effects = get_fixed_effects,
                          safe_deparse = safe_deparse,
                          save_models = save_models,
                          ftail = ftail,
                          check_for_zero_one_obs = check_for_zero_one_obs,
                          check_missing_first_factor_level = check_missing_first_factor_level,
                          fit_augmented_logistic = fit_augmented_logistic,
                          non_augmented = non_augmented,
                          run_group_models = run_group_models,
                          run_ordered_models = run_ordered_models,
                          get_character_cols = get_character_cols,
                          fitting_wrap_up = fitting_wrap_up,
                          optimizers = optimizers,
                          optCtrlList = optCtrlList)
        
        mirai::everywhere({
            requireNamespace("stats")
            requireNamespace("methods")
        })
    }
    
    func_to_run <- function(fv, fn, fi) {
        # Extract Features One by One
        featuresVector <- fv 
        
        logging::loginfo("Fitting model to feature number %d, %s",
                         fi,
                         fn)
        
        # Make fitting matrix of features and metadata
        if (!is.null(feature_specific_covariate)) {
            covariateVector <- feature_specific_covariate[, fi]
            
            dat_sub <-
                data.frame(
                    expr = as.numeric(featuresVector),
                    feature_specific_covariate = covariateVector,
                    metadata, 
                    check.names = FALSE
                )
            
            colnames(dat_sub)[colnames(dat_sub) == 
                            'feature_specific_covariate'] <-
                feature_specific_covariate_name
        } else {
            dat_sub <- data.frame(expr = 
                as.numeric(featuresVector), metadata, 
                check.names = FALSE)
        }
        
        # 0 or 1 observations
        zero_one_out <- check_for_zero_one_obs(formula,
            random_effects_formula,
            dat_sub,
            groups,
            ordereds,
            # features, # TODO: avoid handing the full matrix to all daemons... Will need to fix functions like this one
            fn = fn,
            fi,
            model,
            feature_specific_covariate_name)
        
        if (!is.null(zero_one_out)) {
            return(zero_one_out)
        }
        
        # Missing first factor level
        check_out <- check_missing_first_factor_level(formula,
            random_effects_formula,
            dat_sub,
            groups,
            ordereds,
            fn = fn,
            feature_specific_covariate_name)
        
        if (!is.null(check_out)) {
            return(check_out)
        }
        
        # Augment logistic fitting
        if (augment &
            model == "logistic" &
            length(unique(featuresVector)) >= 2) {
            fitting_out <- fit_augmented_logistic(
                ranef_function,
                model_function,
                formula,
                random_effects_formula,
                groups,
                ordereds,
                dat_sub,
                # features,
                fn = fn,
                fi)
        } else { # linear or non-augmented logistic
            fitting_out <- non_augmented(
                ranef_function,
                model_function,
                formula,
                random_effects_formula,
                groups,
                ordereds,
                dat_sub,
                # features,
                fn = fn,
                fi)
        }
        
        fit_and_message <- fitting_out[["fit_and_message"]]
        weight_scheme <- fitting_out[["weight_scheme"]]
        mm_input <- fitting_out[["mm_input"]]
        
        fit <- fit_and_message[[1]]
        
        # Gather Output
        output <- list()
        
        # Check for fitting errors and add on special predictors
        low_n_error <- FALSE
        if (all(!inherits(fit, "try-error"))) {
            names_to_include <-
                get_fixed_effects(formula,
                                             random_effects_formula,
                                             dat_sub,
                                             character(0),
                                             character(0),
                                             feature_specific_covariate_name)
            # Suppress warnings about variance-covariance matrix calculation
            fit_properly <- FALSE
            
            withCallingHandlers({
                tryCatch({
                    output$para <-
                        summary_function(fit, c('(Intercept)', 
                                            names_to_include))
                    fit_properly <- TRUE
                }, error = function(e) {
                    return()
                })
            }, warning = function(w) {
                invokeRestart("muffleWarning")
            })
            
            if (fit_properly) {
                output$para <-
                    output$para[names_to_include, , drop = FALSE]
                
                n_uni_cols <- nrow(output$para)
                
                if (length(groups) > 0) {
                    output <- run_group_models(ranef_function,
                                            model_function,
                                            groups,
                                            formula,
                                            random_effects_formula,
                                            model,
                                            fit,
                                            augment,
                                            weight_scheme,
                                            dat_sub,
                                            output,
                                            mm_input)
                }
                
                if (length(ordereds) > 0) {
                    output <- run_ordered_models(ranef_function,
                                                model_function,
                                                ordereds,
                                                fit_and_message,
                                                formula,
                                                random_effects_formula,
                                                model,
                                                fit,
                                                augment,
                                                weight_scheme,
                                                dat_sub,
                                                output)
                }
                
                # Check whether summaries are correct
                names_to_include <-
                    get_fixed_effects(formula,
                                    random_effects_formula,
                                    dat_sub,
                                    groups,
                                    ordereds,
                                    feature_specific_covariate_name)
                if (any(!(names_to_include %in% rownames(output$para)))) {
                    # Don't worry about dropped factor levels
                    missing_names <- names_to_include[
                        !(names_to_include %in% rownames(output$para))]
                    
                    character_cols <- get_character_cols(dat_sub)
                    
                    if (!all(missing_names %in% character_cols)) {
                        fit_properly <- FALSE
                        
                        fit_and_message[[length(fit_and_message)]] <-
                            "Metadata dropped during fitting (rank deficient)"
                    } else {
                        fit_properly <- TRUE
                    }
                } else {
                    # No errors, summaries are correct
                    fit_properly <- TRUE
                }
                
                # Prevents individual group levels from ending up in results
                output$para <-
                    output$para[rownames(output$para) %in% names_to_include,]
            }
        } else {
            # Fit issue occurred
            fit_properly <- FALSE
        }
        
        output <- fitting_wrap_up(fit_properly,
                                fit_and_message,
                                output,
                                fit,
                                random_effects_formula,
                                metadata,
                                median_comparison,
                                save_models,
                                formula,
                                dat_sub,
                                groups,
                                ordereds,
                                # features,
                                fn = fn,
                                fi,
                                ranef_function,
                                feature_specific_covariate_name)
       
        
        return(output)
    }
    
    env_objects <- ls(environment(func_to_run))
    for (obj in env_objects) {
        size <- utils::object.size(get(obj, envir = environment(func_to_run)))
        logging::logdebug(paste0("Object: ", obj, ", Size: ", size))
    }
    size <- utils::object.size(func_to_run)
    logging::logdebug(paste0("Object: ", "func_to_run", ", Size: ", size))

    feat_list = as.list(as.data.frame(features))
    feat_nm = colnames(features)
    feat_i = seq_len(ncol(features))
    
    # purrr::list_transpose() would be a better way to do this V.
    map_input = data.frame(fv = I(feat_list), # feature value, name, index
                           fn = feat_nm, 
                           fi = feat_i)
    
    if (mirai::daemons_set()) {
        outputs <- mirai::mirai_map(map_input, func_to_run)[.progress]
        # ^ TODO, figure out how to ensure this writes to the log file
    } else {
        outputs <- mapply(func_to_run, 
                          map_input$fv,
                          map_input$fn,
                          map_input$fi,
                          SIMPLIFY = FALSE)
        # mirai_map() deliberately doesn't fall back to serial without daemons set: https://github.com/r-lib/mirai/issues/397
        
        # To get progress here:
        # lightest: error out and tell user to set daemons with mirai::require_daemons()
        # alternative: utils::txtProgressBar + a loop
        # alternative 2: reintroduce pbapply dependency (it's pretty light)
    }
   
    # Check to run once on a daemon: 
    # print(mirai::mirai(func_to_run(map_input$fv[[1]],
    #                   map_input$fn[1],
    #                   map_input$fi[1]),
    #                   func_to_run = func_to_run,
    #                   map_input = map_input)[])
    
    # cli::cli_alert("First result: ")
    # print(outputs[[1]])
    # outputs <-
    #     pbapply::pblapply(seq_len(ncol(features)), cl = cluster, func_to_run)
    
    # # stop the cluster
    # if (!is.null(cluster))
    #     parallel::stopCluster(cluster)
    
    # bind the results for each feature
    paras <- collapse::rowbind(lapply(outputs, function(x) {
        return(x$para)
    }))
   
    residuals <-
        do.call(rbind, lapply(outputs, function(x) {
            return(x$residuals)
        }))
   
    row.names(residuals) <- colnames(features)
    
    
    fitted <-
        do.call(rbind, lapply(outputs, function(x) {
            return(x$fitted)
        }))
    
    row.names(fitted) <- colnames(features)
    
    fits <-
        lapply(outputs, function(x) {
            return(x$fit)
        })
    names(fits) <- colnames(features)
    
    # Adjust the significance levels based on tests against the median
    if (median_comparison) {
        paras <- run_median_comparison(paras,
                                    fits,
                                    ordereds,
                                    groups,
                                    metadata,
                                    random_effects_formula,
                                    median_comparison_threshold,
                                    subtract_median,
                                    model)
    } else {
        paras$null_hypothesis <- 0
    }
    
    # Return NULL rather than empty object if fits aren't saved
    if (all(is.na(fits)) | !save_models) {
        fits <- NULL
    }
    
    if (!(is.null(random_effects_formula))) {
        
        l <- lapply(outputs, function(x) {
            return(x$ranef)
        })
        
        ranef <- do.call(rbind, l) |> 
            collapse::setRownames(colnames(features))
            
    }
    
    #####################################################
    # Determine the metadata names from the model names #
    #####################################################
    
    metadata_names <- colnames(metadata)
    if (!is.null(feature_specific_covariate)) {
        metadata_names <-
            union(metadata_names, feature_specific_covariate_name)
    }
    # order the metadata names by decreasing length
    metadata_names_ordered <-
        metadata_names[order(nchar(metadata_names), decreasing = TRUE)]
    # find the metadata name based on the match
    # to the beginning of the string
    extract_metadata_name <- function(name) {
        tmp_val <- metadata_names_ordered[mapply(startsWith,
                                                name,
                                                metadata_names_ordered)][1]
        if (is.na(tmp_val)) {
            metadata_names_ordered[mapply(grepl,
                                        metadata_names_ordered,
                                        name)][1]
        } else {
            return(tmp_val)
        }
    }
    paras$metadata <-
        unlist(lapply(paras$name, extract_metadata_name))
    # compute the value as the model contrast minus metadata
    paras$value <-
        mapply(function(x, y) {
            if (x == y)
                x
            else
                gsub("^\\:", "", sub(x, "", y))
        }, paras$metadata, paras$name)
    # Trim ticks if metadata had ticks
    paras$value <- gsub('^``', '', paras$value)
    
    if (!is.null(feature_specific_covariate_record)) {
        if (!feature_specific_covariate_record) {
            paras <-
                paras[!grepl(feature_specific_covariate_name, paras$name),]
        }
    }
    
    if (small_random_effects & model == 'logistic') {
        paras <- paras[!paras$metadata %in% trimws(grouping, whitespace = '`'),]
    }
    
    ##############################
    # Sort by decreasing q-value #
    ##############################
    
    paras <- paras[order(paras$pval, decreasing = FALSE), ]
    
    paras <- collapse::colorderv(paras, c('feature', 'metadata', 'value'))
    paras$model <- model
    rownames(paras) <- NULL
    
    if (!(is.null(random_effects_formula))) {
        return(
            list(
                "results" = paras,
                "residuals" = residuals,
                "fitted" = fitted,
                "ranef" = ranef,
                "fits" = fits
            )
        )
    } else {
        return(
            list(
                "results" = paras,
                "residuals" = residuals,
                "fitted" = fitted,
                "ranef" = NULL,
                "fits" = fits
            )
        )
    }
}

n = 30

dat_sub = data.frame(expr = rnorm(n),
               g = sample(letters[1:3], size = n, replace = TRUE),
               y = factor(sample(LETTERS[1:3], size = n, replace = TRUE),
                          levels = LETTERS[1:3]),
               x = rnorm(n)
)

dat_sub$expr[dat_sub$y == "A"] = NA

formula = ~ x + y + (1|g)
random_effects_formula = expr ~ (1|g)
groups = NULL
ordereds = NULL

features = matrix(rnorm(n*3), nrow = n)
colnames(features) = tail(LETTERS, 3)
x = 1
fn = colnames(features)[x]
feature_specific_covariate_name = NULL

# copied from the original function
expect_true(any(c(vapply(colnames(dat_sub), 
                         function(col) {
                             if (is.factor(dat_sub[, col])) {
                                 if (all(is.na(dat_sub$expr[dat_sub[, col] == 
                                                            levels(dat_sub[, col])[1]]))) {
                                     fixed_effects <-
                                         maaslin3:::get_fixed_effects(formula,
                                                           random_effects_formula,
                                                           dat_sub,
                                                           groups,
                                                           ordereds,
                                                           feature_specific_covariate_name)
                                     if (col %in% substr(fixed_effects, 1, nchar(col))) {
                                         return(TRUE)
                                     }
                                 }
                             }
                             return(FALSE)
                         }, logical(1)))))
fxf <-
        maaslin3:::get_fixed_effects(formula,
                          random_effects_formula,
                          dat_sub,
                          groups,
                          ordereds,
                          feature_specific_covariate_name)

# Check my simplified function:
expect_equal(any(c(vapply(colnames(dat_sub), 
                          function(col) {
                              if (is.factor(dat_sub[, col])) {
                                  if (all(is.na(dat_sub$expr[dat_sub[, col] == 
                                                             levels(dat_sub[, col])[1]]))) {
                                      fixed_effects <-
                                          maaslin3:::get_fixed_effects(formula,
                                                            random_effects_formula,
                                                            dat_sub,
                                                            groups,
                                                            ordereds,
                                                            feature_specific_covariate_name)
                                      if (col %in% substr(fixed_effects, 1, nchar(col))) {
                                          return(TRUE)
                                      }
                                  }
                              }
                              return(FALSE)
                          }, logical(1)))),
             any(mapply(FUN = maaslin3:::check_mffl_one, 
                        as.list(dat_sub),
                        colnames(dat_sub),
                        MoreArgs = list(ex = dat_sub$expr,
                                        fxf = fxf))))
             
goal = list(para = structure(list(coef = c(NA, NA, NA), 
                                  stderr = c(NA, NA, NA),
                                  pval = c(NA, NA, NA),
                                  name = c("x", "yB", "yC"), 
                                  feature = rep(fn, 3),
                                  error = c("No data points have the baseline factor level", 
                                            "No data points have the baseline factor level",
                                            "No data points have the baseline factor level")), 
                             row.names = c(NA, -3L), 
                             class = "data.frame"), 
            residuals = NA, 
            fitted = NA, 
            ranef = NA, 
            fit = NA)

expect_equal(goal,
             maaslin3:::check_missing_first_factor_level(formula, random_effects_formula,
                                              dat_sub, groups, ordereds, fn,
                                              feature_specific_covariate_name))

library(testthat)
library(maaslin3)

data_in <- data.frame('a' = c(1, 2, 3, 4, 5), 
                      'b' = c(2, 3, 4, 5, 6),
                      'c' = c(3, 4, 5, 6, 7))
rownames(data_in) <- paste0("sample", c(1:5))

metadata <- data.frame('a' = c(1, 2, 3, 4, 5),
                       'b' = c(0, 1, 0, 1, 0),
                       'c' = c('a', 'b', 'c', 'a', 'b'))
rownames(metadata) <- paste0("sample", c(1:5))

data_in_tss <- data.frame(t(apply(data_in, MARGIN = 1, 
                                  FUN = function(x){x / sum(x)})))

output_tmp <- tempfile()
expect_that(maaslin_normalize(data_in, output_tmp),
            equals(data_in_tss))

expect_that(maaslin_normalize(data_in, output_tmp, normalization = 'NONE'),
            equals(data_in))

unscaled <- data.frame('total' = c(5, 4, 3, 2, 1))
rownames(unscaled) <- paste0("sample", c(1:5))
scaled_data <- data.frame(apply(data_in_tss, MARGIN = 2, 
                    FUN = function(x){x * unscaled$total}))
expect_that(maaslin_normalize(data_in, output_tmp, 
                              normalization = 'TSS', 
                              unscaled_abundance = unscaled),
            equals(scaled_data))

unscaled <- data.frame('a' = c(5, 4, 3, 2, 1))
rownames(unscaled) <- paste0("sample", c(1:5))
scaled_data <- data.frame(apply(data_in_tss, MARGIN = 2, 
                        FUN = function(x){x * unscaled$a / data_in_tss$a}))
scaled_data$a <- NULL
expect_that(maaslin_normalize(data_in, output_tmp, 
                              normalization = 'TSS', 
                              unscaled_abundance = unscaled),
            equals(scaled_data))

data_in <- data.frame('a' = c(0, 0, 3, 4, 5), 
                      'b' = c(2, 0, 0, 5, 6),
                      'c' = c(3, 4, 0, 0, 7))
rownames(data_in) <- paste0("sample", c(1:5))
data_in_tss <- data.frame(t(apply(data_in, MARGIN = 1, 
                                  FUN = function(x){x / sum(x)})))
data_in_tss[data_in_tss == 0] <- NA
expect_that(maaslin_normalize(data_in, output_tmp),
            equals(data_in_tss))

unlink(output_tmp, recursive = T)

library(testthat)
library(maaslin3)

data_in <- data.frame('a' = c(1, 2, 3, 4, 5), 
                      'b' = c(2, 3, 4, 5, 6),
                      'c' = c(3, 4, 5, 6, 7))
rownames(data_in) <- paste0("sample", c(1:5))

covar_data <- data.frame('a' = c(5, 4, 3, 2, 1), 
                         'b' = c(6, 5, 4, 3, 2),
                         'c' = c(7, 6, 5, 4, 3))
rownames(covar_data) <- paste0("sample", c(1:5))

unscaled <- data.frame('total' = c(5, 4, 3, 2, 1))
rownames(unscaled) <- paste0("sample", c(1:5))

metadata <- data.frame('var1' = c(1, 2, 3, 4, 5),
                       'var2' = c(0, 1, 0, 1, 0),
                       'var3' = c('a', 'b', 'c', 'a', 'b'))
rownames(metadata) <- paste0("sample", c(1:5))

data_in_new <- maaslin_reorder_data(data_in, metadata, covar_data, unscaled)

expect_that(data_in_new$data, equals(data_in))
expect_that(data_in_new$metadata, equals(metadata))
expect_that(data_in_new$feature_specific_covariate, equals(covar_data))
expect_that(data_in_new$unscaled_abundance, equals(unscaled))

data_in_new <- maaslin_reorder_data(t(data_in), metadata, 
                                    t(covar_data), unscaled)

expect_that(data_in_new$data, equals(data_in))
expect_that(data_in_new$metadata, equals(metadata))
expect_that(data_in_new$feature_specific_covariate, equals(covar_data))
expect_that(data_in_new$unscaled_abundance, equals(unscaled))

data_in_new <- maaslin_reorder_data(data_in[-1,], metadata, 
                                    t(covar_data), unscaled)

expect_that(data_in_new$data, equals(data_in[-1,]))
expect_that(data_in_new$metadata, equals(metadata[-1,]))
expect_that(data_in_new$feature_specific_covariate, equals(covar_data[-1,]))
expect_that(data_in_new$unscaled_abundance, equals(unscaled[-1, ,drop=F]))

library(testthat)
library(maaslin3)

# After migration to data.table::fifelse and if/else, the only remaining
# base::ifelse calls should be on matrices (where fifelse doesn't work).
# This test checks the installed package's decompiled code for stray ifelse.

get_function_body_text <- function(fn) {
    paste(deparse(body(fn)), collapse = "\n")
}

test_that("fit.model does not use base ifelse", {
    fn_body <- get_function_body_text(maaslin3:::fit.model)
    ifelse_matches <- gregexpr("\\bifelse\\(", fn_body)[[1]]
    n_matches <- if (ifelse_matches[1] == -1) 0 else length(ifelse_matches)
    expect_equal(n_matches, 0,
                 info = "fit.model should not use base ifelse()")
})

test_that("key viz functions do not use base ifelse", {
    for (fn_name in c("maaslin_plot_results_from_output", 
                       "maaslin_plot_results")) {
        fn <- getFromNamespace(fn_name, "maaslin3")
        fn_body <- get_function_body_text(fn)
        ifelse_matches <- gregexpr("\\bifelse\\(", fn_body)[[1]]
        n_matches <- if (ifelse_matches[1] == -1) 0 else length(ifelse_matches)
        expect_equal(n_matches, 0,
                     info = paste(fn_name, "should not use base ifelse()"))
    }
})

library(testthat)
library(maaslin3)

test_that("test files do not use dplyr in active code", {
    test_dir <- system.file("tests", "testthat", package = "maaslin3")

    if (nchar(test_dir) == 0) {
        test_dir <- file.path(getwd(), "tests", "testthat")
    }

    test_files <- list.files(test_dir, pattern = "\\.R$", full.names = TRUE)
    test_files <- test_files[basename(test_files) != "test_no_dplyr_in_tests.R"]

    for (tf in test_files) {
        lines <- readLines(tf, warn = FALSE)
        code_lines <- lines[!grepl("^\\s*#", lines)]
        pattern <- paste0("dplyr", "::", "|library\\(dplyr\\)")
        dplyr_calls <- grep(pattern, code_lines, value = TRUE)
        expect_equal(length(dplyr_calls), 0,
                     info = paste("File", basename(tf),
                                  "should not reference dplyr in code:",
                                  paste(dplyr_calls, collapse = "; ")))
    }
})

library(testthat)
library(maaslin3)

# Verify NAMESPACE declares all imports that R CMD check requires.

test_that("tail is imported from utils", {
    ns_file <- system.file("NAMESPACE", package = "maaslin3")
    ns_lines <- readLines(ns_file)
    tail_import <- grep("importFrom.*utils.*tail", ns_lines, value = TRUE)
    expect_true(length(tail_import) > 0,
                info = "NAMESPACE should importFrom('utils', 'tail')")
})

test_that("dplyr is declared in Suggests for vignette use", {
    desc <- packageDescription("maaslin3")
    expect_true(grepl("\\bdplyr\\b", desc$Suggests),
                info = "dplyr should be in Suggests (used in vignettes)")
})

test_that("internal functions passed to daemons are accessible", {
    fns <- c("check_for_zero_one_obs", "check_missing_first_factor_level",
             "fit_augmented_logistic", "non_augmented", "run_group_models",
             "run_ordered_models", "get_character_cols", "fitting_wrap_up",
             "get_fixed_effects", "make_lm_plot", "make_logistic_plot",
             "optimizers", "optCtrlList")
    
    for (fn_name in fns) {
        obj <- get(fn_name, envir = asNamespace("maaslin3"))
        expect_true(!is.null(obj),
                    info = paste(fn_name, "should exist in maaslin3 namespace"))
    }
})

library(testthat)
library(maaslin3)

# R CMD check warns when a package uses ::: to access its own objects.
# All internal functions should be called directly or passed to daemons
# via mirai::everywhere().

test_that("no maaslin3::: self-calls in R source files", {
    r_dir <- system.file("R", package = "maaslin3")
    r_files <- list.files(r_dir, pattern = "\\.R$", full.names = TRUE)
    
    all_violations <- character(0)
    for (f in r_files) {
        lines <- readLines(f)
        hits <- grep("maaslin3:::", lines, value = TRUE)
        hits <- hits[!grepl("^\\s*#", hits)]
        if (length(hits) > 0) {
            all_violations <- c(all_violations,
                                paste0(basename(f), ": ", trimws(hits)))
        }
    }
    expect_equal(length(all_violations), 0,
                 info = paste("Found maaslin3::: self-calls:",
                              paste(all_violations, collapse = "\n")))
})

test_that("no library() or require() calls in R source files", {
    r_dir <- system.file("R", package = "maaslin3")
    r_files <- list.files(r_dir, pattern = "\\.R$", full.names = TRUE)
    
    all_violations <- character(0)
    for (f in r_files) {
        lines <- readLines(f)
        hits <- grep("\\b(library|require)\\s*\\(", lines, value = TRUE)
        hits <- hits[!grepl("^\\s*#", hits)]
        hits <- hits[!grepl("requireNamespace", hits)]
        if (length(hits) > 0) {
            all_violations <- c(all_violations,
                                paste0(basename(f), ": ", trimws(hits)))
        }
    }
    expect_equal(length(all_violations), 0,
                 info = paste("Found library/require calls:",
                              paste(all_violations, collapse = "\n")))
})

library(testthat)
library(maaslin3)

# The NAMESPACE previously imported dplyr::%>% and plyr::mapvalues even though
# neither is used. The DESCRIPTION previously listed pbapply and parallel in
# Imports even though neither is used in active code paths.

test_that("dplyr pipe is not imported", {
    ns_file <- system.file("NAMESPACE", package = "maaslin3")
    ns_lines <- readLines(ns_file)
    dplyr_pipe_lines <- grep('importFrom.*dplyr.*%>%', ns_lines, value = TRUE)
    expect_equal(length(dplyr_pipe_lines), 0,
                 info = "NAMESPACE should not import dplyr pipe")
})

test_that("plyr mapvalues is not imported", {
    ns_file <- system.file("NAMESPACE", package = "maaslin3")
    ns_lines <- readLines(ns_file)
    plyr_lines <- grep('importFrom.*plyr.*mapvalues', ns_lines, value = TRUE)
    expect_equal(length(plyr_lines), 0,
                 info = "NAMESPACE should not import plyr::mapvalues")
})

test_that("pbapply is not in Imports", {
    desc <- packageDescription("maaslin3")
    expect_false(grepl("pbapply", desc$Imports),
                 info = "pbapply should not be in Imports")
})

test_that("parallel is not in Imports", {
    desc <- packageDescription("maaslin3")
    expect_false(grepl("\\bparallel\\b", desc$Imports),
                 info = "parallel should not be in Imports")
})

---
title: "MaAsLin 3 Tutorial"
author:
- name: William Nickols
  email: willnickols@g.harvard.edu
- name: Jacob Nearing
  email: nearing@broadinstitute.org
- name: Sagun Maharjan
  email: smaharjan@hsph.harvard.edu
output: html_document
vignette: >
    %\VignetteIndexEntry{Tutorial}
    %\VignetteEngine{knitr::rmarkdown}
    %\VignetteEncoding{UTF-8}
---

## Introduction ##

[**MaAsLin 3**](http://huttenhower.sph.harvard.edu/MaAsLin3) is the next
generation of **M**a**A**s**L**in (**M**icrobiome **M**ultivariable
**A**ssociations with **L**inear Models). This comprehensive R package
efficiently determines multivariable associations between clinical
metadata and microbial meta-omics features. Relative to MaAsLin 2,
MaAsLin 3 introduces the ability to quantify and test for **both
abundance and prevalence associations** while **accounting for
compositionality**. By incorporating generalized linear models, MaAsLin
3 accommodates most modern epidemiological study designs including
cross-sectional and longitudinal studies.

If you use the MaAsLin 3 software, please cite our manuscript:

> William A. Nickols, Thomas Kuntz, Jiaxian Shen, Sagun Maharjan, 
Himel Mallick, Eric A. Franzosa, Kelsey N. Thompson, Jacob T. Nearing, 
Curtis Huttenhower. MaAsLin 3: Refining and extending generalized 
multivariable linear models for meta-omic association discovery. 
bioRxiv 2024.12.13.628459; doi: https://doi.org/10.1101/2024.12.13.628459

### Support ###
The user manual can be found
[here](https://github.com/biobakery/maaslin3). If using vignettes,
users should start with the `maaslin3_tutorial.Rmd` vignette and then
refer to the `maaslin3_manual.Rmd` vignette as necessary. If you have questions,
please direct them to the [MaAsLin
Forum](https://forum.biobakery.org/c/Downstream-analysis-and-statistics/maaslin)

## Contents
* [1. Installing R](#installing-r)
* [2. Installing MaAsLin 3](#installing-maaslin-3)
* [3. Microbiome association detection with MaAsLin
3](#microbiome-association-detection-with-maaslin-3)
    * [3.1 MaAsLin 3 input](#maaslin-3-input)
    * [3.2 Running MaAsLin 3](#running-maaslin-3)
    * [3.3 MaAsLin 3 output](#maaslin-3-output)
* [4. Advanced Topics](#advanced-topics)
    * [4.1 Absolute abundance](#absolute-abundance)
    * [4.2 Random Effects (repeated sampling)](#random-effects)
    * [4.3 Interactions (differences in differences)](#interactions)
    * [4.4 Level contrasts](#level-contrasts)
    * [4.5 Group-wise differences](#group-wise-differences)
    * [4.6 Contrast tests](#contrast-tests)
* [5. Command line](#command-line)

## 1. Installing R {#installing-r}

[R](https://www.r-project.org/) is a programming language specializing
in statistical computing and graphics. You can use R just the same as
any other programming languages, but it is most useful for statistical
analyses, with well-established packages for common tasks such as
[linear
modeling](https://cran.r-project.org/web/packages/lme4/index.html),
['omic data analysis](https://bioconductor.org/), [machine
learning](http://topepo.github.io/caret/index.html), and
[visualization](https://ggplot2.tidyverse.org/).

#### Installing R for the first time

You can download and install the free R software environment
[here](https://cloud.r-project.org/). Note that you should download the
latest release - this will ensure the R version you have is compatible
with MaAsLin 3.

#### (Optional) the RStudio IDE

[RStudio](https://rstudio.com/products/rstudio/) is a freely available
IDE (integrated development environment) for R. It is a "wrapper" around
R with some additional functionalities that make programming in R a bit
easier. Feel free to download RStudio and explore its interface - but it
is not required for this tutorial.

#### Important: the correct R version

If you already have R installed, then it is possible that the R version
you have does not support MaAsLin 3. The easiest way to check this is to
launch R/RStudio, and in the console ("Console" pane in RStudio), type
in the following command (without the `>` symbol):

```{R, eval=FALSE, cache = FALSE}
sessionInfo()
```

The first line of output message should indicate your current R version.
For example:

```
> sessionInfo()
R version 4.4.1 (2024-06-14)
```

For MaAsLin 3 to install, you will need R >= 4.4. If your version is
older than that, please refer to section [Installing R for the first
time](#installing-r-for-the-first-time) to download the latest R. Note
that RStudio users will need to have RStudio "switch" to this later
version once it is installed. This should happen automatically for
Windows and Mac OS users when you relaunch RStudio. For Linux users you
might need to bind the correct R executable.
Either way, once you have the correct version installed, launch the
software and use `sessionInfo()` to make sure that you indeed have R >=
4.4.

## 2. Installing MaAsLin 3 {#installing-maaslin-3}

The latest version of MaAsLin 3 can be installed from BiocManager.

```{r, eval=FALSE, cache = FALSE}
if (!require("BiocManager", quietly = TRUE))
    install.packages("BiocManager")

BiocManager::install("biobakery/maaslin3")
```

To compile vignettes yourself, specify `dependencies = TRUE`.

```{r, eval=TRUE, cache = FALSE, echo=FALSE}
for (lib in c('maaslin3', 'dplyr', 'ggplot2', 'knitr', 'kableExtra')) {
    suppressPackageStartupMessages(require(lib, character.only = TRUE))
}
```

## 3. Microbiome association detection with MaAsLin 3

To run MaAsLin 3, the user must supply a table of per-sample feature
abundances (with zeros still included), a table of per-sample metadata,
and a model specifying how the metadata should relate to the feature
prevalence (how likely the feature is to be present or absent) and
abundance (how much of the feature is there if it's there). MaAsLin 3
will return a table of associations including an effect size and p-value
for each feature-metadatum association and a folder of visuals including
a summary plot and diagnostic plots for significant associations.

### 3.1 MaAsLin 3 input {#maaslin-3-input}

MaAsLin3 requires two input files.

1. Feature abundance data frame
    * Formatted with features as columns and samples as rows.
    * The transpose of this format is also okay.
    * Possible features include taxonomy or genes. These can be relative
    abundances or counts.
    * This can be a filepath to a tab-delimited file.
2. Metadata data frame
    * Formatted with variables as columns and samples as rows.
    * The transpose of this format is also okay.
    * Possible metadata include gender or age.
    * This can be a filepath to a tab-delimited file.
    
The data file can contain samples not included in the metadata file
(along with the reverse case). For both cases, those samples not
included in both files will be removed from the analysis. Also, the
samples do not need to be in the same order in the two files.

Samples with NA values for the metadata will be excluded when fitting the 
models. It is recommended to drop these samples in advance under a 
missing-at-random assumption or use multiple imputation to run 
the complete dataset.

Example input files can be found in the ``inst/extdata`` folder of the
MaAsLin 3 source. The files provided were generated from the Human
Microbiome Project 2 (HMP2) data which can be downloaded from
https://ibdmdb.org/.

* ``HMP2_taxonomy.tsv``: a tab-delimited file with samples as rows and
species as columns. It is a subset of the full HMP2 taxonomy that
includes just some of the the species abundances.
* ``HMP2_metadata.tsv``: a tab-delimited file with samples as rows and
metadata as columns. It is a subset of the full HMP2 metadata that
includes just some of the fields.

```{R, cache = FALSE}  
# Read abundance table
taxa_table_name <- system.file("extdata", "HMP2_taxonomy.tsv",
                                package = "maaslin3")
taxa_table <- read.csv(taxa_table_name, sep = '\t', row.names = 1)

# Read metadata table
metadata_name <- system.file("extdata", "HMP2_metadata.tsv",
                            package = "maaslin3")
metadata <- read.csv(metadata_name, sep = '\t', row.names = 1)

# Factor the categorical variables to test IBD against healthy controls
metadata$diagnosis <-
    factor(metadata$diagnosis, levels = c('nonIBD', 'UC', 'CD'))
metadata$dysbiosis_state <-
    factor(metadata$dysbiosis_state, levels =
                c('none', 'dysbiosis_UC', 'dysbiosis_CD'))
metadata$antibiotics <-
    factor(metadata$antibiotics, levels = c('No', 'Yes'))

taxa_table[1:5, 1:5]
metadata[1:5, 1:5]
```

### 3.2 Running MaAsLin 3 {#running-maaslin-3}

MaAsLin 3 is run by specifying the
abundance table (`input_data`), the metadata table (`input_metadata`),
the output directory (`output`), and a model. The model can come from a
formula or vectors of terms. In either case, variable names should not
have spaces or unusual characters.

* *Formula*: The `formula` parameter should be set to any formula that
satisfies the `lme4` specifications: fixed effects, random effects
([section 4.2](#random-effects)), interaction terms ([section
4.3](#interactions)), polynomial terms, and more can all be included. If
categorical variables are included as fixed effects, each level will be
tested against the first factor level. In addition, ordered predictors
([section 4.4](#level-contrasts)) and group predictors ([section
4.5](#group-wise-differences)) can be included by including
`group(variable_name)` and `ordered(variable_name)` in the formula. When
the data involve paired or matched case-control samples, conditional
logistic regression can be run by specifying the grouping variable with
`strata(variable_name)`.
* *Vectors*: Alternatively, a vector of variable names can be supplied
to the parameters `fixed_effects`, `random_effects`, `group_effects`,
`ordered_effects`, and `strata_effects`. Categorical variables 
should either be ordered as
factors beforehand, or `reference` should be provided as a string of
'variable,reference' semi-colon delimited for multiple variables (e.g.,
`variable_1,reference_1;variable_2,reference_2`). NOTE: adding a space
between the variable and level might result in the wrong reference level
being used.

**Because MaAsLin 3 identifies prevalence (presence/absence)
associations, sample read depth (number of reads) should be included as
a covariate if available. Deeper sequencing will likely increase feature
detection in a way that could spuriously correlate with metadata of
interest when read depth is not included in the model.**

#### Running MaAsLin 3 on HMP2 data

The following command runs MaAsLin 3 on the HMP2 data, running a
multivariable
regression model to test for the association between microbial species
abundance and prevalence
versus IBD diagnosis and [dysbiosis
scores](https://www.nature.com/articles/s41586-019-1237-9) while
controlling for antibiotic usage, age, and the sampling depth (reads).
The abundance associations come from fitting multivariate linear models
to the log base 2 relative abundances of features in samples in which
the feature is present. The prevalence associations come from fitting
multivariate logistic models to the presence/absence data for a feature
across all samples. Outputs are directed to a folder called
`hmp2_output` under the current working directory (`output =
"hmp2_output"`).

To show one of each type of plot (see below), the maximum number of
significant associations to plot has been increased with `max_pngs=100`
(default `30`). All other parameters beyond the first four are the
default parameters but are still included for clarity:

* Total sum scaling (`normalization = 'TSS'`) with a log transformation
(`'transform = 'LOG'`, base 2) afterwards will be used. These are almost
always the recommended choices (and all MaAsLin 3 evaluations were
performed with these options), but other normalizations and
transformations are allowed (see `normalization` and `transform` in
`?maaslin3`).
* A data augmentation procedure is used to avoid linear separability in
the logistic regressions (`augment = TRUE`). This is almost always
recommended, though it can be turned off (see `augment` in the manual).
* Continuous metadatum variables are z-score standardized (subtract the
mean, divide by the standard deviation) with `standardize = TRUE` so
that the coefficients will be on the same scale (improving
visualization).
* A nominal FDR level of 0.1 (`max_significance = 0.1`) determines which
associations are significant, but this can also be changed (see
`max_significance` in the manual).
* For the abundance coefficients, to account for compositionality in
relative abundance data, significance is determined by comparing against
the median coefficient for a metadatum across the features
(`median_comparison_abundance`). Since prevalence coefficients do not
have the same compositional properties, they are instead compared
against 0 (`median_comparison_prevalence`). See below for a discussion of
these parameters.
* One core is used by default. You can enable parallelization by creating
background [daemons](https://mirai.r-lib.org/reference/daemons.html) using the
[mirai](https://mirai.r-lib.org/index.html) package e.g. call
`mirai::daemons(2)` before running the `maaslin3()` command to use two cores.

The abundance, prevalence, and variance filtering parameters are not
included, and they are 0 by default. **Different from other differential
abundance tools, low prevalence features need not be filtered out**
since the prevalence modeling in MaAsLin 3 already accounts for high
proportions of zeros. However, filtering low prevalence features might improve
power.

```{R, echo = TRUE, results = 'hide', warning = FALSE, cache = FALSE}
set.seed(1)
fit_out <- maaslin3(input_data = taxa_table,
                    input_metadata = metadata,
                    output = 'hmp2_output',
                    formula = '~ diagnosis + dysbiosis_state +
                        antibiotics + age + reads',
                    normalization = 'TSS',
                    transform = 'LOG',
                    augment = TRUE,
                    standardize = TRUE,
                    max_significance = 0.1,
                    median_comparison_abundance = TRUE,
                    median_comparison_prevalence = FALSE,
                    max_pngs = 10)
```

By default, many lines of information are printed while the models fit, 
but the verbosity can be changed with the `verbosity` argument (e.g., 
`verbosity = 'WARN'`).

MaAsLin 3 can also take as input a `SummarizedExperiment` or 
`TreeSummarizedExperiment` object from BioConductor:

```{R, echo = TRUE, results = 'hide', warning = FALSE, eval = FALSE}
se <- SummarizedExperiment(
    assays = list(taxa_table = t(taxa_table)),
    colData = metadata
)

fit_out <- maaslin3(input_data = se,
                    output = 'hmp2_output',
                    formula = '~ diagnosis + dysbiosis_state +
                        antibiotics + age + reads',
                    normalization = 'TSS',
                    transform = 'LOG',
                    augment = TRUE,
                    standardize = TRUE,
                    max_significance = 0.1,
                    median_comparison_abundance = TRUE,
                    median_comparison_prevalence = FALSE,
                    max_pngs = 10)
```

#### Median comparisons

When `median_comparison_abundance` or `median_comparison_prevalence` are
on, the coefficients for a metadatum will be tested against the median
coefficient for that metadatum (median across the features). Otherwise,
the coefficients will be tested against 0. For abundance associations,
this is designed to account for compositionality, the idea that
**testing against zero on the relative scale is not equivalent to
testing against zero on the absolute scale**. The user manual contains a
more extensive discussion of when to use `median_comparison_abundance`
and `median_comparison_prevalence`, but in summary:

* `median_comparison_abundance` is `TRUE` by default and should be used
to make inference on the absolute scale when using relative abundance
data. When `median_comparison_abundance` is `TRUE`, only the p-values
and q-values change; the coefficients returned are still the relative
abundance coefficients unless `subtract_median` is `TRUE` in which case
the median will be subtracted from the coefficients.
* `median_comparison_abundance` should be `FALSE` when (1) testing for
significant relative associations, (2) testing for absolute abundance
associations under the assumption that the total absolute abundance is
not changing, or (3) testing for significant absolute associations when
supplying spike-in or total abundances with `unscaled_abundance`.
* `median_comparison_prevalence` is `FALSE` by default.

### 3.3 MaAsLin 3 output {#maaslin-3-output}

#### Significant associations

The main output from MaAsLin 3 is the list of significant associations
in `significant_results.tsv`. This lists all associations with joint 
or individual q-values 
that pass MaAsLin 3's significance
threshold, ordered by increasing individual q-value. The format is:

* `feature` and `metadata` are the feature and metadata names.
* `value` and `name` are the value of the metadata and variable name
from the model.
* `coef` and `stderr` are the fit coefficient and standard error from
the model. In abundance models, a one-unit change in the metadatum
variable corresponds to a $2^{\textrm{coef}}$ fold change in the
relative abundance of the feature. In prevalence models, a one-unit
change in the metadatum variable corresponds to a $\textrm{coef}$ change
in the log-odds of a feature being present.
* `null_hypothesis` is the value of the null hypothesis against which 
the coefficient is tested (zero or the per-metadatum median).
* `pval_individual` and `qval_individual` are the p-value and false
discovery rate (FDR) corrected q-value of the individual association.
FDR correction is performed over all p-values without errors in the abundance
and prevalence modeling together.
* `pval_joint` and `qval_joint` are the p-value and q-value of the joint
prevalence and abundance association. The p-value comes from plugging in
the minimum of the association's abundance and prevalence p-values into
the Beta(1,2) CDF. It is interpreted as the probability that either the
abundance or prevalence association would be as extreme as observed if
there was neither an abundance nor prevalence association between the
feature and metadatum.
* `model` specifies whether the association is abundance or prevalence.
* `N` and `N_not_zero` are the total number of data points and the total
number of non-zero data points for the feature.


```{r, echo = FALSE, cache = FALSE}
signif_results <- read.csv('hmp2_output/significant_results.tsv',
sep='\t')
head(signif_results, 20) %>%
    dplyr::mutate_if(is.numeric, .funs = function(x){(
        as.character(signif(x, 3)))}) %>%
    knitr::kable() %>%
    kableExtra::kable_styling("striped", position = 'center') %>%
    kableExtra::scroll_box(width = "800px", height = "400px")
```
#### Full output file structure

MaAsLin 3 generates two types of output files explained below: data and
visualizations. In addition, the object returned from `maaslin3`
(`fit_out` above) contains all the data and results (see
`?maaslin_fit`).

1. Data output files
    * ``significant_results.tsv``
        * Described above.
    * ``all_results.tsv``
        * Same format as above but for all associations and with an extra column
        listing any errors from the model fitting.
    * ``features``
        * This folder includes the filtered, normalized, and transformed
        versions of the input feature table.
        * These steps are performed sequentially in the above order.
        * If an option is set such that a step does not change the data, the
        resulting table will still be output.
    * ``models_linear.rds`` and ``models_logistic.rds``
        * These files contain a list with every model fit object (`linear` for
        linear models, `logistic` for logistic models).
        * It will only be generated if `save_models` is set to TRUE.
    * ``residuals_linear.rds`` and ``residuals_logstic.rds``
        * These files contain a data frame with residuals for each feature.
    * ``fitted_linear.rds`` and ``fitted_logistic.rds``
        * These files contain a data frame with fitted values for each feature.
    * ``ranef_linear.rds`` and ``ranef_logistic.rds``
        * These files contain a data frame with extracted random effects for
        each feature (when random effects are specified).
    * ``maaslin3.log``
        * This file contains all log information for the run.
        * It includes all settings, warnings, errors, and steps run.
2. Visualization output files
    * ``summary_plot.pdf``
        * This file contain a combined coefficient plot and heatmap of the most
        significant associations. In the heatmap, one star indicates the
        individual q-value is below the parameter `max_significance`, and two
        stars indicate the individual q-value is below `max_significance`
        divided by 10. (Make sure to set `include=T` to view the plots when
        knitting the Rmd.)
```{r, out.width='100%', echo=FALSE, cache = FALSE, include=FALSE, eval=FALSE}
# Rename summary plot to avoid knitting issues later
quiet_out <- file.rename('hmp2_output/figures/summary_plot.png',
                        'hmp2_output/figures/summary_plot_first.png')

knitr::include_graphics("hmp2_output/figures/summary_plot_first.png")
```
    * ``association_plots/[metadatum]/[association]/
[metadatum]_[feature]_[association].png``
        * A plot is generated for each significant association up to `max_pngs`.
        * Scatter plots are used for continuous metadata abundance associations.
        * Box plots are used for categorical data abundance associations.
        * Box plots are used for continuous data prevalence associations.
        * Grids are used for categorical data prevalence associations.
        * Data points plotted are after filtering, normalization, and
        transformation so that the scale in the plot is the scale that was used
        in fitting.
        * **To see the association plots, set `max_pngs = 250` above and set 
        `eval=TRUE` and `include=TRUE` in this chunk**
```{r, echo=FALSE, fig.show='hold',include=FALSE, eval=FALSE}
prefix <- "hmp2_output/figures/association_plots"
plot_vec <- 
    c("/age/linear/age_Enterocloster_clostridioformis_linear.png",
    "/dysbiosis_state/linear/dysbiosis_state_Escherichia_coli_linear.png",
    "/age/logistic/age_Bifidobacterium_longum_logistic.png",
    paste0("/dysbiosis_state/logistic/",
        "dysbiosis_state_Faecalibacterium_prausnitzii_logistic.png"
        ))
knitr::include_graphics(paste0(prefix, plot_vec))
```

At the top right of each association plot is the name of the significant
association in the results file, the FDR corrected q-value for the
individual association, the number of samples in the dataset, and the
number of samples with non-zero abundances for the feature. In the plots
with categorical metadata variables, the reference category is on the
left, and the significant q-values and coefficients in the top right are
in the order of the values specified above. Because the displayed
coefficients correspond to the full fit model with (possibly) scaled
metadata variables, the marginal association plotted might not match the
coefficient displayed. However, the plots are intended to provide an
interpretable visual while usually agreeing with the full model.

#### Diagnostics

There are a few common issues to check for in the results:

1. When warnings or errors are thrown during the fitting process, they
are recorded in the `error` column of `all_results.tsv`. Often, these
indicate model fitting failures or poor fits that should not be trusted,
but sometimes the warnings can be benign, and the model fit might still
be reasonable. Users should check associations of interest if they
produce errors.
    * With `warn_prevalence=TRUE`, a note will be produced in the error column
    indicating when prevalence associations are likely induced by abundance 
    associations. These should be checked visually with the diagnosic visuals.
2. Despite the error checking, significant results could still result
from poor model fits. These can usually be diagnosed with the visuals in
the `association_plots` directory.
    * Any significant abundance associations with a categorical variable
    should usually have **at least 10 observations in each category**.
    * Significant prevalence associations with categorical variables should
    also have **at least 10 samples in which the feature was present and at
    least 10 samples in which it was absent for each significant category**.
    * Significant abundance associations with continuous metadata should be
    checked visually for influential outliers.
3. The q-values are FDR corrected over all abundance and prevalence
relationships, so it may be preferable to FDR correct just
the p-values from the variables of interest. This can reduce false
positives when there are many significant but uninteresting associations
(e.g., many read depth associations).
4. There are also a few rules of thumb to keep in mind.
    * Models should ideally have about 10 times as many samples (all samples
    for logistic fits, non-zero samples for linear fits) as covariate terms
    (all continuous variables plus all categorical variable levels).
    * Coefficients (effect sizes) larger than about 15 in absolute value are
    usually suspect unless very small unstandardized predictors are being
    included. (A coefficient of 15 corresponds to a fold change >30000!). If
    you encounter such coefficients, check that (1) no error was thrown, (2)
    the diagnostics look reasonable, (3) a sufficient number of samples were
    used in fitting, (4) the q-value is significant, (5) the metadata are
    not highly collinear, and (6) the random effects are plausible.
    
#### Replotting

Once `maaslin3` has been run once, `maaslin_plot_results_from_output`
can be run to
(re-)create the plots. This allows the user to plot the associations
even without having the R object returned by `maaslin_fit` or `maaslin3`
(e.g., if fitting models through the command line). It is recommended to
fit models with simple variables names that are robust to formula
plotting and then convert these into proper names for plotting.
Likewise, `heatmap_vars` and `coef_plot_vars` can be specified at any point, 
but it is usually easier to see how the names
come out first and then choose which metadata variables will go in the
coefficient plot and which will go in the heatmap afterwards with
`maaslin_plot_results_from_output`.

```{R, echo = TRUE, results = 'hide', warning = FALSE, cache = FALSE}
# This section is necessary for updating the
# summary plot and the association plots

# Rename results file with clean titles
all_results <- read.csv('hmp2_output/all_results.tsv', sep='\t')
all_results <- all_results %>%
    mutate(metadata = case_when(metadata == 'age' ~ 'Age',
                                metadata == 'antibiotics' ~ 'Abx',
                                metadata == 'diagnosis' ~ 'Diagnosis',
                                metadata == 'dysbiosis_state' ~ 'Dysbiosis',
                                metadata == 'reads' ~ 'Read depth'),
        value = case_when(value == 'dysbiosis_CD' ~ 'CD',
                            value == 'dysbiosis_UC' ~ 'UC',
                            value == 'Yes' ~ 'Used', # Antibiotics
                            value == 'age' ~ 'Age',
                            value == 'reads' ~ 'Read depth',
                            TRUE ~ value),
        feature = gsub('_', ' ', feature) %>%
            gsub(pattern = 'sp ', replacement = 'sp. '))

# Write results
write.table(all_results, 'hmp2_output/all_results.tsv', sep='\t')

# Set the new heatmap and coefficient plot variables and order them
heatmap_vars = c('Dysbiosis UC', 'Diagnosis UC',
                'Abx Used', 'Age', 'Read depth')
coef_plot_vars = c('Dysbiosis CD', 'Diagnosis CD')

# This section is necessary for updating the association plots
taxa_table_copy <- taxa_table
colnames(taxa_table_copy) <- gsub('_', ' ', colnames(taxa_table_copy)) %>%
    gsub(pattern = 'sp ', replacement = 'sp. ')

# Rename the features in the norm transformed data file
data_transformed <-
    read.csv('hmp2_output/features/data_transformed.tsv', sep='\t')
colnames(data_transformed) <-
    gsub('_', ' ', colnames(data_transformed)) %>%
    gsub(pattern = 'sp ', replacement = 'sp. ')
write.table(data_transformed,
            'hmp2_output/features/data_transformed.tsv',
            sep='\t', row.names = FALSE)

# Rename the metadata like in the outputs table
metadata_copy <- metadata
colnames(metadata_copy) <-
    case_when(colnames(metadata_copy) == 'age' ~ 'Age',
            colnames(metadata_copy) == 'antibiotics' ~ 'Abx',
            colnames(metadata_copy) == 'diagnosis' ~ 'Diagnosis',
            colnames(metadata_copy) == 'dysbiosis_state' ~ 'Dysbiosis',
            colnames(metadata_copy) == 'reads' ~ 'Read depth',
            TRUE ~ colnames(metadata_copy))
metadata_copy <- metadata_copy %>%
    mutate(Dysbiosis = case_when(Dysbiosis == 'dysbiosis_UC' ~ 'UC',
                                Dysbiosis == 'dysbiosis_CD' ~ 'CD',
                                Dysbiosis == 'none' ~ 'None') %>%
            factor(levels = c('None', 'UC', 'CD')),
        Abx = case_when(Abx == 'Yes' ~ 'Used',
                        Abx == 'No' ~ 'Not used') %>%
            factor(levels = c('Not used', 'Used')),
        Diagnosis = case_when(Diagnosis == 'nonIBD' ~ 'non-IBD',
                                TRUE ~ Diagnosis) %>%
            factor(levels = c('non-IBD', 'UC', 'CD')))

# Recreate the plots
scatter_plots <- maaslin_plot_results_from_output(
    output = 'hmp2_output',
    metadata = metadata_copy,
    normalization = "TSS",
    transform = "LOG",
    median_comparison_abundance = TRUE,
    median_comparison_prevalence = FALSE,
    max_significance = 0.1,
    max_pngs = 20)
```

In the new summary plot below, we can see that the feature names are
cleaned up, the metadata names are cleaned up, the set of metadata
variables used in the coefficient plot is different, and the metadata
used in the heatmap is reordered.
```{r, out.width='100%', echo=FALSE, cache = FALSE, include=FALSE}
knitr::include_graphics("hmp2_output/figures/summary_plot.png")
```

In the association plots, the taxa and metadata have been renamed to be
consistent with the results file from earlier.
```{r, echo=FALSE, fig.show='hold', cache = FALSE, include=FALSE, eval=FALSE}
prefix <- "hmp2_output/figures/association_plots"
plot_vec <- c("/Age/linear/Age_Enterocloster clostridioformis_linear.png",
            "/Dysbiosis/linear/Dysbiosis_Escherichia coli_linear.png",
            "/Age/logistic/Age_Bifidobacterium longum_logistic.png",
            paste0("/Dysbiosis/logistic/",
                "Dysbiosis_Faecalibacterium prausnitzii_logistic.png"
                ))
knitr::include_graphics(paste0(prefix, plot_vec))
```

## 4. Advanced Topics {#advanced-topics}

### 4.1 Absolute abundance {#absolute-abundance}

Most microbiome methodologies have historically focused on relative
abundances (proportions out of 1). However, some experimental protocols
can enable estimation of absolute abundances (cell count/concentration).
MaAsLin 3 can be used with two types of absolute abundance estimation:
spike-ins and total abundance scaling. In a spike-in procedure, a small,
known quantity of a microbe that otherwise would not be present in the
sample is added, and the sequencing procedure is carried out as usual.
Then, the absolute abundance of a microbe already in the community is
estimated as:
$$\textrm{Absolute abundance other microbe}=\frac{\textrm{Relative
abundance other microbe}}{\textrm{Relative abundance spike-in
microbe}}\cdot (\textrm{Absolute abundance spike-in microbe})$$
Alternatively, the total microbial abundance of a sample can be
determined (e.g., with qPCR of a marker gene or by cell counting). Then,
the absolute abundance of a microbe in the community is estimated as:
$$\textrm{Absolute abundance microbe}=(\textrm{Total absolute
abundance})\cdot(\textrm{Relative abundance microbe})$$

#### Spike-in

The following section shows a synthetic spike-in procedure with
simulated data from SparseDOSSA2. The abundance table is like any other
abundance table input to MaAsLin 3 except that 'Feature101' is the
spike-in (which must be present in every sample). The scaling factor
data frame has 'Feature101' as its only column name with the absolute
abundance of the spike-in for each sample (rownames). If the same
quantity of the spike-in was added to all samples, this column could be
entirely 1 (or any other arbitrary value). When using a spike-in
procedure, the scaling factor data frame should have a single column
named the same as one of the features in the abundance table, which will
be used as the spike-in.

```{R, cache = FALSE}
# Abundance table
taxa_table_name <- system.file("extdata", "abundance_spike_in_ex.tsv",
                            package = "maaslin3")
spike_in_taxa_table <- read.csv(taxa_table_name, sep = '\t', row.names = 1)

# Metadata table
metadata_name <- system.file("extdata", "metadata_spike_in_ex.tsv",
                            package = "maaslin3")
spike_in_metadata <- read.csv(metadata_name, sep = '\t', row.names = 1)
for (col in c('Metadata_1', 'Metadata_2', 'Metadata_5')) {
    spike_in_metadata[,col] <- factor(spike_in_metadata[,col])
}

# Spike-in table
unscaled_name <- system.file("extdata",
"scaling_factors_spike_in_ex.tsv",
                            package = "maaslin3")
spike_in_unscaled <- read.csv(unscaled_name, sep = '\t', row.names = 1)

spike_in_taxa_table[c(1:5, 101),1:5]
spike_in_metadata[1:5,]
spike_in_unscaled[1:5, , drop=FALSE]
```

The following code fits the absolute abundance model. Note that the
normalization must be set to `TSS` since the procedure involves scaling
the relative abundance ratios. Also, `median_comparison_abundance` is
now set to `FALSE` since we want to test the absolute coefficients
against zero. The data frame of absolute abundances is included as the
`unscaled_abundance` parameter, and the spike-in strategy will
automatically be detected based on the column name.

```{R, echo = TRUE, results = 'hide', warning = FALSE, cache = FALSE}
fit_out <- maaslin3(
    input_data = spike_in_taxa_table,
    input_metadata = spike_in_metadata,
    output = 'spike_in_demo',
    formula = '~ Metadata_1 + Metadata_2 + Metadata_3 +
        Metadata_4 + Metadata_5',
    normalization = 'TSS',
    transform = 'LOG',
    median_comparison_abundance = FALSE,
    unscaled_abundance = spike_in_unscaled)
```

The absolute abundance coefficients can be inspected:

```{R, cache = FALSE}
rownames(fit_out$fit_data_abundance$results) <- NULL
head(fit_out$fit_data_abundance$results, 20) %>%
    dplyr::mutate_if(is.numeric, .funs =
                            function(x){(as.character(signif(x, 3)))}) %>%
    knitr::kable() %>%
    kableExtra::kable_styling("striped", position = 'center') %>%
    kableExtra::scroll_box(width = "800px", height = "400px")
```

#### Total abundance scaling

The following section shows a synthetic total abundance scaling
procedure with simulated data from SparseDOSSA2. The abundance table is
like any other abundance table input to MaAsLin 3 without any extra
features. The scaling factor data frame has 'total' as its only column
name with the total absolute abundance for each sample (rownames).

```{r, cache = FALSE}
# Abundance table
taxa_table_name <- system.file("extdata", "abundance_total_ex.tsv",
    package = "maaslin3")
total_scaling_taxa_table <- read.csv(taxa_table_name, sep = '\t', row.names = 1)

# Metadata table
metadata_name <- system.file("extdata", "metadata_total_ex.tsv",
                                package = "maaslin3")
total_scaling_metadata <- read.csv(metadata_name, sep = '\t', row.names = 1)
for (col in c('Metadata_1', 'Metadata_3', 'Metadata_5')) {
    spike_in_metadata[,col] <- factor(spike_in_metadata[,col])
}

# Total abundance table
unscaled_name <- system.file("extdata", "scaling_factors_total_ex.tsv",
                                package = "maaslin3")
total_scaling_unscaled <- read.csv(unscaled_name, sep = '\t', row.names = 1)

total_scaling_taxa_table[1:5, 1:5]
total_scaling_metadata[1:5,]
total_scaling_unscaled[1:5, , drop=FALSE]
```

The following code fits the absolute abundance model. As before, `TSS`
must be set to `TRUE`, `median_comparison_abundance` is set to `FALSE`,
and the data frame of absolute abundances is included as the
`unscaled_abundance` parameter. The total abundance scaling procedure
will be detected based on the column name.

```{R, echo = TRUE, results = 'hide', warning = FALSE, cache = FALSE}
fit_out <- maaslin3(
    input_data = total_scaling_taxa_table,
    input_metadata = total_scaling_metadata,
    output = 'total_scaling_demo',
    formula = '~ Metadata_1 + Metadata_2 + Metadata_3 +
        Metadata_4 + Metadata_5',
    normalization = 'TSS',
    transform = 'LOG',
    median_comparison_abundance = FALSE,
    unscaled_abundance = total_scaling_unscaled)
```

The absolute abundance coefficients can be inspected:

```{R, cache = FALSE}
rownames(fit_out$fit_data_abundance$results) <- NULL
head(fit_out$fit_data_abundance$results, n = 20) %>%
    dplyr::mutate_if(is.numeric,
                    .funs = function(x){(as.character(signif(x, 3)))}) %>%
    knitr::kable() %>%
    kableExtra::kable_styling("striped", position = 'center') %>%
    kableExtra::scroll_box(width = "800px", height = "400px")
```

#### Computationally estimated absolute abundance

When using a purely computational approach to estimate absolute abundance
(e.g., Nishijima et al. Cell 2024 with the function `MLP`), the abundance 
table can be estimated
and then run with the `normalization = 'NONE'` to avoid scaling the results.

### 4.2 Random effects (clustered samples) {#random-effects}

Certain studies have a natural "grouping" of sample observations, such
as by subject in longitudinal designs or by family in family designs. It
is  important for statistic analysis to address the non-independence
between samples belonging to the same group. MaAsLin 3 provides a simple
interface for this with random effects, where the user can specify the
grouping variable to run a [mixed effect
model](https://cran.r-project.org/web/packages/lme4/index.html). This
grouping variable can be provided with the `random_effects` parameter or
specified in the model with `(1|grouping_variable)`. This will add in a
"random intercept," a per-group adjustment to the model intercept. More
complicated random effects can be specified in the formula in accordance
with `lme4` formula parsing. For example, HMP2 has a longitudinal design
where the same subject (column `participant_id`) can have multiple
samples. We thus use `participant_id` as a random effect grouping
variable in the model.

```{R, echo = TRUE, results = 'hide', warning = FALSE, messages = FALSE}
# Subset to only CD cases for time; taxa are subset automatically
fit_out <- maaslin3(
    input_data = taxa_table,
    input_metadata = metadata[metadata$diagnosis == 'CD',],
    output = 'random_effects_output',
    formula = '~ dysbiosis_state + antibiotics +
        age + reads + (1|participant_id)',
    plot_summary_plot = FALSE,
    plot_associations = FALSE)
```

```{r, echo = FALSE, cache = FALSE}
signif_results <-
read.csv('random_effects_output/significant_results.tsv',
                            sep='\t')
head(signif_results, n = 20) %>%
    dplyr::mutate_if(is.numeric, .funs =
                        function(x){(as.character(signif(x, 3)))}) %>%
    knitr::kable() %>%
    kableExtra::kable_styling("striped", position = 'center') %>%
    kableExtra::scroll_box(width = "800px", height = "400px")
```

Note that no `participant_id` terms are included in the outputs; the
random intercepts are just used to control for grouping. If you are
interested in testing the effect of time in a longitudinal study, the
time point variable should be included as a fixed effect in your MaAsLin
3 formula.

#### Alternatives to random effects

Random effects are not the solution to every correlated data problem. In 
particular, **logistic random intercept models may produce poor model fits 
when there are fewer than 4 average observations per random effect group.** The
fit coefficients can be large in magnitude and very significant while
modeling the data poorly. By default, a warning will be added to the `error`
column of the outputs in these scenarios. This warning can be disregarded
if the user is confident the model is fit correctly, but recommendations
are provided in the table below:

| Scenario |$\geq4$ average observations per group|$<4$ average observations\
per group, metadata not collinear with the groups |$<4$ average observations\
per group, metadata collinear with the groups|
|----|----|-----|------|
| Example   | Subjects sampled $\geq4$ times | Subjects sampled once before \
and once after treatment | Subjects have fixed disease status \
and are sampled twice (disease \
status is constant per-subject, so collinear with the grouping) |
| Action   | Use random effects with  `small_random_effects=FALSE` \
(default) | Use random effects for abundance but fixed effects for \
prevalence by setting `small_random_effects=TRUE` | Subset or average data to \
one observation per group |
| Reason | Random effects are robust and \
increase power  | Prevalence random effects are not robust but fixed effects \
are robust and control for grouping | Prevalence random effects are \
not robust and fixed effects cannot be used when groups are \
collinear with the metadata |

### 4.3 Interactions (differences in differences) {#interactions}

One of the benefits of MaAsLin 3's formula mode is the ability to
include interaction terms. Mathematically, an interaction term
corresponds to the product of two columns in the design matrix. When a
continuous variable is interacted with a categorical term, the
interaction term corresponds to the change in the continuous variable's
slope between the categories. For two categorical variables interacted,
see below; it is better explained through an example. In the formula,
interactions can be specified with the `:` symbol to include only the
interaction term(s) or the `*` symbol to include both the interaction
term and the non-interacted terms (i.e., `a*b` adds the terms `a`, `b`,
and `a:b`).

```{R, echo = TRUE, results = 'hide', warning = FALSE, cache = FALSE}
metadata <- metadata %>%
    mutate(dysbiosis_general = ifelse(dysbiosis_state != 'none',
                                    'dysbiosis', 'none')) %>%
    mutate(dysbiosis_general = factor(dysbiosis_general, levels =
                                        c('none', 'dysbiosis')))

fit_out <- maaslin3(
    input_data = taxa_table,
    input_metadata = metadata,
    output = 'interaction_output',
    formula = '~ diagnosis + diagnosis:dysbiosis_general +
        antibiotics + age + reads')
```

```{R, cache = FALSE}
full_results <- rbind(fit_out$fit_data_abundance$results,
                    fit_out$fit_data_prevalence$results)
full_results <- full_results %>%
    dplyr::arrange(qval_joint) %>%
    dplyr::filter(metadata == "diagnosis")
rownames(full_results) <- NULL
head(full_results, n = 20) %>%
    dplyr::mutate_if(is.numeric,
                    .funs = function(x){(as.character(signif(x, 3)))}) %>%
    knitr::kable() %>%
    kableExtra::kable_styling("striped", position = 'center') %>%
    kableExtra::scroll_box(width = "800px", height = "400px")
```
In the model above, we have converted `dysbiosis_CD` and `dysbiosis_UC`
into `dysbiosis` in the variable `dysbiosis_general` and then specified
an interaction between `diagnosis` and `dysbiosis_general` with
`diagnosis:dysbiosis_general`. Since `diagnosis` has 3 levels itself
(`nonIBD`, `UC`, `CD`), this will produce five terms (`name` column) for
each feature:

* `diagnosisCD` is the difference between non-dysbiotic CD and non-IBD
* `diagnosisUC` is the difference between non-dysbiotic UC and non-IBD
* `diagnosisCD:dysbiosis_generaldysbiosis` (the interaction of
`diagnosisCD` and `dysbiosis_generaldysbiosis`) is the difference
between non-dysbiotic CD and dysbiotic CD
* `diagnosisnonIBD:dysbiosis_generaldysbiosis` (the interaction of
`diagnosisnonIBD` and `dysbiosis_generaldysbiosis`) is the difference
between dysbiotic and non-dysbiotic non-IBD. Note that these
coefficients are all NA since no subjects labeled non-IBD were marked as
being dysbiotic.
* `diagnosisUC:dysbiosis_generaldysbiosis` (the interaction of
`diagnosisUC` and `dysbiosis_generaldysbiosis`) is the difference
between non-dysbiotic UC and dysbiotic UC

Since we have effectively fit the original model in a different way,
this output table is equivalent to the original output table but with
new names in the columns `metadata`, `value`, and `name` based on the
interaction terms.

### 4.4 Level contrasts {#level-contrasts}

Another feature of MaAsLin 3 is the ability to test for
level-versus-level differences in ordered predictors with contrast
tests. Ordered predictors are categorical variables with a natural
ordering such as cancer stage, consumption frequency of a dietary
factor, or dosage group. Here, we assess how microbial abundances and
prevalences are associated with eating red meat by including
`ordered(red_meat)` in the formula. This will perform a contrast test
for whether there is a difference between each pair of subsequent levels
(e.g., "Yesterday, 3 or more times" versus "Yesterday, 1 to 2 times")
rather than whether there are differences between the levels and the
baseline (e.g., "Yesterday, 3 or more times" versus "Not in the last 7
days"). The coefficient, standard error, and p-value all correspond to
the difference between the level in `value` and the previous level.
Ordered predictors should only be included as fixed effects (i.e., no
ordered predictors as random effects etc.).

```{R, echo = TRUE, results = 'hide', warning = FALSE, cache = FALSE}
# Put the red meat consumption responses in order
metadata <- metadata %>%
    mutate(red_meat = ifelse(
        red_meat == 'No, I did not consume these products in the last 7 days',
                            'Not in the last 7 days',
                            red_meat) %>%
            factor(levels = c('Not in the last 7 days',
                                'Within the past 4 to 7 days',
                                'Within the past 2 to 3 days',
                                'Yesterday, 1 to 2 times',
                                'Yesterday, 3 or more times'))
    )

# Create the model with only non-IBD subjects
fit_out <- maaslin3(
    input_data = taxa_table,
    input_metadata = metadata[metadata$diagnosis == 'nonIBD',],
    output = 'ordered_outputs',
    formula = '~ ordered(red_meat) + antibiotics + age + reads',
    plot_summary_plot = TRUE,
    plot_associations = TRUE,
    heatmap_vars = c('red_meat Within the past 4 to 7 days',
                    'red_meat Within the past 2 to 3 days',
                    'red_meat Yesterday, 1 to 2 times',
                    'red_meat Yesterday, 3 or more times'),
    max_pngs = 30)
```

```{r, out.width='100%', echo=FALSE, cache = FALSE, include=FALSE}
knitr::include_graphics("ordered_outputs/figures/summary_plot.png")
```

If we look at the resulting heatmap, we can identify the *Alistipes
shahii* prevalence association as potentially interesting since it
increases in prevalence with more meat consumption in all but one
level-versus-level comparison.

```{r, echo = FALSE, cache = FALSE}
full_results <- rbind(fit_out$fit_data_abundance$results,
                    fit_out$fit_data_prevalence$results)
full_results <- full_results %>%
    dplyr::filter(metadata %in% c("red_meat") &
                    feature == 'Alistipes_shahii' &
                    model == 'logistic') %>%
    dplyr::mutate(
        value =
            factor(
                value,
                levels =
                    c('No, I did not consume these products in the last 7 days',
                                                'Within the past 4 to 7 days',
                                                'Within the past 2 to 3 days',
                                                'Yesterday, 1 to 2 times',
                                            'Yesterday, 3 or more times'))) %>%
    dplyr::arrange(value)
rownames(full_results) <- NULL
head(full_results, n = 20) %>%
    dplyr::mutate_if(is.numeric,
                    .funs = function(x){(as.character(signif(x, 3)))}) %>%
    knitr::kable() %>%
    kableExtra::kable_styling("striped", position = 'center') %>%
    kableExtra::scroll_box(width = "800px", height = "200px")
```

### 4.5 Group-wise differences {#group-wise-differences}

MaAsLin 3 can also test
for group-wise differences in categorical predictors using an ANOVA or
ANOVA-like procedure. Group-wise predictors are categorical variables
with or without an ordering such as race, country, or consumption
frequency of a dietary factor. When performing a group-wise difference
test, we are not comparing any two particular levels; rather, we are
investigating whether abundances and prevalences are the same for all
levels. Therefore, no coefficient is returned, only a p-value. Here, we
test whether there are differences in microbial abundances and
prevalences across people with different levels of red meat consumption
by including `group(red_meat)`. Group-wise predictors should only be
included as fixed effects (i.e., no group-wise predictors as random
effects etc.).

```{R, echo = T, results = 'hide', warning = FALSE, cache = FALSE}
fit_out <- maaslin3(
    input_data = taxa_table,
    input_metadata = metadata[metadata$diagnosis == 'nonIBD',],
    output = 'group_outputs',
    formula = '~ group(red_meat) + antibiotics + age + reads',
    plot_summary_plot = TRUE,
    plot_associations = TRUE,
    heatmap_vars = c('red_meat Within the past 4 to 7 days',
                    'red_meat Within the past 2 to 3 days',
                    'red_meat Yesterday, 1 to 2 times',
                    'red_meat Yesterday, 3 or more times'),
    max_pngs = 200)
```

If we look at the same *Alistipes shahii* association from before, we
see that no coefficient or standard error is returned for a group
predictor, but the q-value is very low. This corroborates the
observation earlier that there are differences in *Alistipes shahii*
prevalence with red meat consumption.

```{r, echo = FALSE, cache = FALSE}
full_results <- rbind(fit_out$fit_data_abundance$results,
                    fit_out$fit_data_prevalence$results)
full_results <- full_results %>%
    dplyr::filter(metadata %in% c("red_meat") &
                    feature == 'Alistipes_shahii' &
                    model == 'logistic') %>%
    dplyr::mutate(
        value =
            factor(
                value,
                levels =
                    c('No, I did not consume these products in the last 7 days',
                                                'Within the past 4 to 7 days',
                                                'Within the past 2 to 3 days',
                                                'Yesterday, 1 to 2 times',
                                            'Yesterday, 3 or more times'))) %>%
    dplyr::arrange(value)
rownames(full_results) <- NULL
head(full_results, n = 20) %>%
    dplyr::mutate_if(is.numeric,
                    .funs = function(x){(as.character(signif(x, 3)))}) %>%
    knitr::kable() %>%
    kableExtra::kable_styling("striped", position = 'center') %>%
    kableExtra::scroll_box(width = "800px", height = "100px")
```

### 4.6 Contrast tests {#contrast-tests}

We can also perform contrast testing in MaAsLin 3 to check whether a linear
combination of the coefficients is significantly different from some constant 
on the right hand size (RHS). Specify a contrast matrix
with column names matching the coefficient names and rows with the contrasts
of interest. For each row $C^T$, the hypothesis $H_0:C^T\vec{\beta}=RHS$ 
will be tested. If no RHS is specified, a median test will be performed,
and the median test RHS will be returned.

Below, we test whether the UC and CD diagnosis coefficients are the same
and whether the UC and CD dysbiosis coefficients are the same. The contrast
matrix can easily be extended to test any groups pairwise.

```{R, echo = TRUE, results = 'hide', warning = FALSE, cache = FALSE}
fit_out <- maaslin3(input_data = taxa_table,
                    input_metadata = metadata,
                    output = 'contrast_test_output',
                    formula = '~ diagnosis + dysbiosis_state +
                        antibiotics + age + reads',
                    plot_summary_plot = FALSE,
                    plot_associations = FALSE)
```

```{R, echo = TRUE, warning = FALSE, cache = FALSE}
contrast_mat <- matrix(c(1, -1, 0, 0, 0, 0, 1, -1), 
    ncol = 4, nrow = 2, byrow = TRUE)
    
colnames(contrast_mat) <- c("diagnosisUC",
                            "diagnosisCD",
                            "dysbiosis_statedysbiosis_UC",
                            "dysbiosis_statedysbiosis_CD")
                            
rownames(contrast_mat) <- c("diagnosis_test", "dysbiosis_test")

contrast_mat

contrast_out <- maaslin_contrast_test(maaslin3_fit = fit_out,
                        contrast_mat = contrast_mat)

head(contrast_out$fit_data_abundance$results, n = 20) %>%
    knitr::kable() %>%
    kableExtra::kable_styling("striped", position = 'center') %>%
    kableExtra::scroll_box(width = "800px", height = "400px")
```


## 5. Command line #### {#command-line}

MaAsLin 3 can also be run with a command line interface. For example,
the first HMP2 analysis can be performed with the following command (the
slashes may need to be removed):

```{r, engine = 'bash', eval = FALSE, cache = FALSE}
./R/maaslin3.R \
    inst/extdata/HMP2_taxonomy.tsv \
    inst/extdata/HMP2_metadata.tsv \
    command_line_output \
    --formula='~ diagnosis + dysbiosis_state + antibiotics + age + reads' \
    --reference='diagnosis,nonIBD;dysbiosis_state,none;antibiotics,No'
```

* Make sure to provide the full path to the MaAsLin3 executable (i.e.
`./R/maaslin3.R`).
* In the demo command:
    * ``inst/extdata/HMP2_taxonomy.tsv`` is the path to your data (or
    features) file
    * ``inst/extdata/HMP2_metadata.tsv`` is the path to your metadata file
    * ``command_line_output`` is the path to the folder to write the output

## Tool comparison ##

| Tool | Models differential abundance | \
Models differential prevalence | Models allowed | Accounts for\
compositionality | Can use experimental absolute abundance information |
|:---:|:---:|:---:|:---:|:---:|:---:|
| MaAsLin 3 | Yes | Yes, for any model type | Any lme4 model\
(general mixed models), level-vs-level differences, group-wise differences, \
feature-specific covariates, contrast tests | Yes | Yes |
| MaAsLin 2 | Yes | Only insofar as it affects abundance | Fixed and random \
effects (subset of general mixed models) | No | No |
| ANCOM-BC2 | Yes | Only when completely absent in a group | Any lme4 model \
(general mixed models), level-vs-level differences, group-wise differences, \
contrast tests | Yes | No |
| ALDEx2 | Yes | Only insofar as it affects abundance | Fixed \
effects | Yes | No |
| LinDA | Yes | Only insofar as it affects abundance | Any lme4 model \
(general mixed models), group-wise differences, contrast tests | Yes | No |


```{R}
sessionInfo()
```

```{R}
# Clean-up
unlink('hmp2_output', recursive = TRUE)
unlink('spike_in_demo', recursive = TRUE)
unlink('total_scaling_demo', recursive = TRUE)
unlink('random_effects_output', recursive = TRUE)
unlink('interaction_output', recursive = TRUE)
unlink('ordered_outputs', recursive = TRUE)
unlink('group_outputs', recursive = TRUE)
unlink('contrast_test_output', recursive = TRUE)
```

---
title: "MaAsLin 3 User Manual"
author:
- name: William Nickols
  email: willnickols@g.harvard.edu
- name: Jacob Nearing
  email: nearing@broadinstitute.org
- name: Sagun Maharjan
  email: smaharjan@hsph.harvard.edu
output:
  html_document: default
  pdf_document: default
vignette: >
    %\VignetteIndexEntry{Manual}
    %\VignetteEngine{knitr::rmarkdown}
    %\VignetteEncoding{UTF-8}
---

## Introduction ##

[**MaAsLin 3**](http://huttenhower.sph.harvard.edu/MaAsLin3) is the next
generation of **M**a**A**s**L**in (**M**icrobiome **M**ultivariable
**A**ssociations with **L**inear Models). This comprehensive R package
efficiently determines multivariable associations between clinical
metadata and microbial meta-omics features. Relative to MaAsLin 2,
MaAsLin 3 introduces the ability to quantify and test for **both
abundance and prevalence associations** while **accounting for
compositionality**. By incorporating generalized linear models, MaAsLin
3 accommodates most modern epidemiological study designs including
cross-sectional and longitudinal studies.

If you use the MaAsLin 3 software, please cite our manuscript:

> William A. Nickols, Thomas Kuntz, Jiaxian Shen, Sagun Maharjan, 
Himel Mallick, Eric A. Franzosa, Kelsey N. Thompson, Jacob T. Nearing, 
Curtis Huttenhower. MaAsLin 3: Refining and extending generalized 
multivariable linear models for meta-omic association discovery. 
bioRxiv 2024.12.13.628459; doi: https://doi.org/10.1101/2024.12.13.628459

### Support ###
Check out the [MaAsLin 3
tutorial](https://github.com/biobakery/biobakery/wiki/MaAsLin3) for an
overview of analysis options and some example runs. If using vignettes,
users should start with the `maaslin3_tutorial.Rmd` vignette and then
refer to the `maaslin3_manual.Rmd` vignette as necessary.

If you have questions, please direct them to the [MaAsLin 3
Forum](https://forum.biobakery.org/c/Downstream-analysis-and-statistics/maaslin)

--------------------------------------------

## Contents ##
- [Introduction](#introduction)
- [Support](#support)
- [Contents](#contents)
- [Requirements](#requirements)
- [Installation](#installation)
- [Running MaAsLin 3](#running-maaslin-3)
- [Input data](#input-data)
- [Output files](#output-files)
- [Run a demo](#run-a-demo)
    - [In R](#in-r)
    - [Plot revisions](#plot-revisions)
    - [Command line](#command-line)
    - [Contrast testing](#contrast-testing)
- [Options](#options)
    - [Required parameters](#required-parameters)
    - [Model formula](#model-formula)
    - [Analysis options](#analysis-options)
    - [Compositionality corrections](#compositionality-corrections)
    - [Plotting parameters](#plotting-parameters)
    - [Technical parameters](#technical-parameters)
- [Troubleshooting](#troubleshooting)

## Requirements ##

MaAsLin3 is an R package that can be run on the command line or as an R
function.

## Installation ##

#### Install using GitHub and devtools

The latest version of MaAsLin 3 can be installed from BiocManager.

```{r, eval=FALSE, cache = FALSE}
if (!require("BiocManager", quietly = TRUE))
    install.packages("BiocManager")

BiocManager::install("biobakery/maaslin3")
```

To compile vignettes yourself, specify `dependencies = TRUE`.

```{r, eval=TRUE, cache = FALSE, echo=FALSE}
for (lib in c('maaslin3', 'dplyr', 'ggplot2', 'knitr')) {
    suppressPackageStartupMessages(require(lib, character.only = TRUE))
}
```

## Running MaAsLin 3 ##

MaAsLin3 can be run from the command line or as an R function. Both
methods require the same arguments, have the same options, and use the
same default settings. To run MaAsLin 3, the user must supply a table of
per-sample feature abundances (with zeros still included), a table of
per-sample metadata, and a model specifying how the metadata should
relate to the feature prevalence (how likely the feature is to be
present or absent) and abundance (how much of the feature is there if
it's there). MaAsLin 3 will return a table of associations including an
effect size and p-value for each feature-metadatum association and a
folder of visuals including a summary plot and diagnostic plots for
significant associations.

### Input data ###

MaAsLin3 requires two input files.

1. Feature abundance data frame
    * Formatted with features as columns and samples as rows.
    * The transpose of this format is also okay.
    * Possible features include taxonomy or genes. These can be relative
    abundances or counts.
    * This can be a filepath to a tab-delimited file.
2. Metadata data frame
    * Formatted with variables as columns and samples as rows.
    * The transpose of this format is also okay.
    * Possible metadata include gender or age.
    * This can be a filepath to a tab-delimited file.

The data file can contain samples not included in the metadata file
(along with the reverse case). For both cases, those samples not
included in both files will be removed from the analysis. Also, the
samples do not need to be in the same order in the two files.

To run MaAsLin 3, it is also necessary to specify a model. The model can
come from a formula or vectors of terms. In either case, variable names
should not have spaces or unusual characters.

* *Formula*: The `formula` parameter should be set to any formula that
satisfies the `lme4` specifications: fixed effects, random effects,
interaction terms, polynomial terms, and more can all be included. If
categorical variables are included as fixed effects, each level will be
tested against the first factor level. In addition, ordered predictors, group
predictors, and strata variables can be included by including
`group(variable_name)`, `ordered(variable_name)`, and `strata(variable_name)`
respectively in the formula. Ordered and group
predictors should stand alone in the formula (i.e., no group predictors
in random effects). Only one strata variable can be included.
* *Vectors*: Alternatively, a vector of variable names can be supplied
to the parameters `fixed_effects`, `random_effects`, `group_effects`,
`ordered_effects`, and `strata_effects`. Categorical variables should
either be ordered as factors beforehand, or `reference` should be
provided as a string of 'variable,reference' semi-colon delimited for
multiple variables (e.g.,
`variable_1,reference_1;variable_2,reference_2`). NOTE: adding a space
between the variable and level might result in the wrong reference level
being used.

**Because MaAsLin 3 identifies prevalence (presence/absence)
associations, sample read depth (number of reads) should be included as
a covariate if available. Deeper sequencing will likely increase feature
detection in a way that could spuriously correlate with metadata of
interest when read depth is not included in the model.**

### Output files ###

MaAsLin 3 generates two types of output files explained below: data and
visualizations. In addition, the object returned from `maaslin3`
contains all the data and results (see `?maaslin_fit`).

1. Data output files
    * ``all_results.tsv``
        * `feature` and `metadata` are the feature and metadata names.
        * `value` and `name` are the value of the metadata and variable name
        from the model.
        * `coef` and `stderr` are the fit coefficient and standard error from
        the model. In abundance models, a one-unit change in the metadatum
        variable corresponds to a $2^{\textrm{coef}}$ fold change in the
        relative abundance of the feature. In prevalence models, a one-unit
        change in the metadatum variable corresponds to a $\textrm{coef}$ change
        in the log-odds of a feature being present.
        * `null_hypothesis` is the value of the null hypothesis against which 
        the coefficient is tested (zero or the per-metadatum median).
        * `pval_individual` is the p-value of the individual association.
        * `qval_individual` is the false discovery rate (FDR) corrected q-value
        of the individual association. FDR correction is performed over all
        p-values without errors in the abundance and prevalence modeling 
        together.
        * `pval_joint` and `qval_joint` are the p-value and q-value of the joint
        prevalence and abundance association. The p-value comes from plugging in
        the minimum of the association's abundance and prevalence p-values into
        the Beta(1,2) CDF. It is interpreted as the probability that either the
        abundance or prevalence association would be as extreme as observed if
        there was neither an abundance nor prevalence association between the
        feature and metadatum.
        * `error` lists any errors from the model fitting.
        * `model` specifies whether the association is abundance or prevalence.
        * `N` and `N_not_zero` are the total number of data points and the total
        number of non-zero data points for the feature.
    * ``significant_results.tsv``
        * This file is a subset of the results in the first file.
        * It only includes associations with joint or individual q-values less 
than or equal to the threshold.
    * ``features``
        * This folder includes the filtered, normalized, and transformed
        versions of the input feature table.
        * These steps are performed sequentially in the above order.
        * If an option is set such that a step does not change the data, the
        resulting table will still be output.
    * ``models_linear.rds`` and ``models_logistic.rds``
        * These files contain a list with every model fit object (`linear` for
        linear models, `logistic` for logistic models).
        * It will only be generated if `save_models` is set to TRUE.
    * ``residuals_linear.rds`` and ``residuals_logstic.rds``
        * These files contain a data frame with residuals for each feature.
    * ``fitted_linear.rds`` and ``fitted_logistic.rds``
        * These files contain a data frame with fitted values for each feature.
    * ``ranef_linear.rds`` and ``ranef_logistic.rds``
        * These files contain a data frame with extracted random effects for
        each feature (when random effects are specified).
    * ``maaslin3.log``
        * This file contains all log information for the run.
        * It includes all settings, warnings, errors, and steps run.
2. Visualization output files
    * ``summary_plot.pdf``
        * This file contain a combined coefficient plot and heatmap of the most
        significant associations. In the heatmap, one star indicates the
        individual q-value is below the parameter `max_significance`, and two
        stars indicate the individual q-value is below `max_significance`
        divided by 10.
    * ``association_plots/[metadatum]/[association]/
[metadatum]_[feature]_[association].png``
        * A plot is generated for each significant association up to `max_pngs`.
        * Scatter plots are used for continuous metadata abundance associations.
        * Box plots are used for categorical data abundance associations.
        * Box plots are used for continuous data prevalence associations.
        * Grids are used for categorical data prevalence associations.
        * Data points plotted are after filtering, normalization, and
        transformation so that the scale in the plot is the scale that was used
        in fitting.

At the top right of each association plot is the name of the significant
association in the results file, the FDR corrected q-value for the
individual association, the number of samples in the dataset, and the
number of samples with non-zero abundances for the feature. In the plots
with categorical metadata variables, the reference category is on the
left, and the significant q-values and coefficients in the top right are
in the order of the values specified above. Because the displayed
coefficients correspond to the full fit model with (possibly) scaled
metadata variables, the marginal association plotted might not match the
coefficient displayed. However, the plots are intended to provide an
interpretable visual while usually agreeing with the full model.

#### Diagnostics

1. When warnings or errors are thrown during the fitting process, they are 
recorded in the `error` column of `all_results.tsv`. Often, these indicate 
model fitting failures or poor fits that should not be trusted, but sometimes 
the warnings can be benign, and the model fit might still be reasonable. Users 
should check associations of interest if they produce errors.
2. Despite the error checking, significant results could still result from poor 
model fits. These can usually be diagnosed with the visuals in the 
`association_plots` directory. 
    * Any significant abundance associations with a categorical variable should 
    usually have **at least 10 observations in each category**.
    * Significant prevalence associations with categorical variables should 
    also have **at least 10 samples in which the feature was present and at 
    least 10 samples in which it was absent for each significant category**. 
    * Significant abundance associations with continuous metadata should be 
    checked visually for influential outliers.
3. The q-values are FDR corrected over all abundance or prevalence 
relationships (separately), so it may be preferable to FDR correct just the 
p-values from the variables of interest. This can reduce false positives when 
there are many significant but uninteresting associations (e.g., many read 
depth associations).
4. There are also a few rules of thumb to keep in mind.
    * Models should ideally have about 10 times as many samples (all samples 
    for logistic fits, non-zero samples for linear fits) as covariate terms 
    (all continuous variables plus all categorical variable levels).
    * Coefficients (effect sizes) larger than about 15 in absolute value are 
    usually suspect unless very small unstandardized predictors are being 
    included. (A coefficient of 15 corresponds to a fold change >30000!). If 
    you encounter such coefficients, check that (1) no error was thrown, (2) 
    the plots look reasonable, (3) a sufficient number of samples were 
    used in fitting, (4) the q-value is significant, (5) the metadata are not 
    highly collinear, and (6) the random effects are plausible.

### Run a demo ###

Example input files can be found in the ``inst/extdata`` folder of the
MaAsLin 3 source. The files provided were generated from the Human
Microbiome Project 2 (HMP2) data which can be downloaded from
https://ibdmdb.org/.

* ``HMP2_taxonomy.tsv``: a tab-delimited file with samples as rows and
species as columns. It is a subset of the full HMP2 taxonomy that
includes just some of the the species abundances.
* ``HMP2_metadata.tsv``: a tab-delimited file with samples as rows and
metadata as columns. It is a subset of the full HMP2 metadata that
includes just some of the fields.

#### In R ####

The following code identifies associations between patient metadata and
microbial species in the HMP2 cohort.

```{R, echo = TRUE, results = 'hide', warning = FALSE, cache = FALSE}
# Read abundance table
taxa_table_name <- system.file("extdata", "HMP2_taxonomy.tsv",
                                package = "maaslin3")
taxa_table <- read.csv(taxa_table_name, sep = '\t', row.names = 1)

# Read metadata table
metadata_name <- system.file("extdata", "HMP2_metadata.tsv",
                                package = "maaslin3")
metadata <- read.csv(metadata_name, sep = '\t', row.names = 1)

metadata$diagnosis <-
    factor(metadata$diagnosis, levels = c('nonIBD', 'UC', 'CD'))
metadata$dysbiosis_state <-
    factor(metadata$dysbiosis_state, levels = c('none', 'dysbiosis_UC',
                                                'dysbiosis_CD'))
metadata$antibiotics <-
    factor(metadata$antibiotics, levels = c('No', 'Yes'))

# Fit models
set.seed(1)
fit_out <- maaslin3(input_data = taxa_table,
                    input_metadata = metadata,
                    output = 'hmp2_output',
                    formula = '~ diagnosis + dysbiosis_state +
                    antibiotics + age + reads',
                    normalization = 'TSS',
                    transform = 'LOG',
                    augment = TRUE,
                    standardize = TRUE,
                    max_significance = 0.1,
                    median_comparison_abundance = TRUE,
                    median_comparison_prevalence = FALSE,
                    max_pngs = 20)
```

The `maaslin3` wrapper function runs all steps at once, but equivalent
results would be produced from running one step at a time:

```{r, eval=FALSE}
maaslin_log_arguments(input_data = taxa_table,
                    input_metadata = metadata,
                    output = 'hmp2_output',
                    formula = '~ diagnosis + dysbiosis_state +
                    antibiotics + age + reads',
                    normalization = 'TSS',
                    transform = 'LOG',
                    augment = TRUE,
                    standardize = TRUE,
                    max_significance = 0.1,
                    median_comparison_abundance = TRUE,
                    median_comparison_prevalence = FALSE,
                    max_pngs = 20)

read_data_list <- maaslin_read_data(taxa_table,
                                    metadata)
read_data_list <- maaslin_reorder_data(read_data_list$data,
                                    read_data_list$metadata)

data <- read_data_list$data
metadata <- read_data_list$metadata

formulas <- maaslin_check_formula(data,
                                metadata,
                                '~ diagnosis + dysbiosis_state +
                                    antibiotics + age + reads')

formula <- formulas$formula
random_effects_formula <- formulas$random_effects_formula

normalized_data = maaslin_normalize(data,
                                    output = 'hmp2_output',
                                    normalization = 'TSS')

filtered_data = maaslin_filter(normalized_data,
                            output = 'hmp2_output')

transformed_data = maaslin_transform(filtered_data,
                                    output = 'hmp2_output',
                                    transform = 'LOG')

standardized_metadata = maaslin_process_metadata(metadata,
                                                formula,
                                                standardize = TRUE)

maaslin_results = maaslin_fit(filtered_data,
                            transformed_data,
                            standardized_metadata,
                            formula,
                            random_effects_formula,
                            data = data,
                            median_comparison_abundance = TRUE,
                            median_comparison_prevalence = FALSE)

maaslin_write_results(output = 'hmp2_output',
                    maaslin_results$fit_data_abundance,
                    maaslin_results$fit_data_prevalence,
                    random_effects_formula,
                    max_significance = 0.1)

# Invisible to avoid dumping plots
invisible(maaslin_plot_results(output = 'hmp2_output',
                            transformed_data,
                            metadata,
                            maaslin_results$fit_data_abundance,
                            maaslin_results$fit_data_prevalence,
                            normalization = "TSS",
                            transform = "LOG",
                            median_comparison_abundance = TRUE,
                            median_comparison_prevalence = FALSE,
                            max_significance = 0.1,
                            max_pngs = 20))
```

#### Replotting

Once `maaslin3` has been run once, `maaslin_plot_results_from_output`
can be run to
(re-)create the plots. This allows the user to plot the associations
even without having the R object returned by `maaslin_fit` or `maaslin3`
(e.g., if fitting models through the command line). It is recommended to
fit models with simple variables names that are robust to formula
plotting and then convert these into proper names for plotting.
Likewise, `heatmap_vars` and `coef_plot_vars` can be specified at 
any point, but it is usually easier to see how the names
come out first and then choose which metadata variables will go in the
coefficient plot and which will go in the heatmap afterwards with
`maaslin_plot_results_from_output`.

```{R, echo = TRUE, results = 'hide', warning = FALSE, cache = FALSE}
# This section is necessary for updating
# the summary plot and the association plots

# Rename results file with clean titles
all_results <- read.csv('hmp2_output/all_results.tsv', sep='\t')
all_results <- all_results %>%
    mutate(metadata = case_when(metadata == 'age' ~ 'Age',
                                metadata == 'antibiotics' ~ 'Abx',
                                metadata == 'diagnosis' ~ 'Diagnosis',
                                metadata == 'dysbiosis_state' ~ 'Dysbiosis',
                                metadata == 'reads' ~ 'Read depth'),
        value = case_when(value == 'dysbiosis_CD' ~ 'CD',
                            value == 'dysbiosis_UC' ~ 'UC',
                            value == 'Yes' ~ 'Used', # Antibiotics
                            value == 'age' ~ 'Age',
                            value == 'reads' ~ 'Read depth',
                            TRUE ~ value),
        feature = gsub('_', ' ', feature) %>%
            gsub(pattern = 'sp ', replacement = 'sp. '))

# Write results
write.table(all_results, 'hmp2_output/all_results.tsv', sep='\t')

# Set the new heatmap and coefficient plot variables and order them
heatmap_vars = c('Dysbiosis UC', 'Diagnosis UC',
                'Abx Used', 'Age', 'Read depth')
coef_plot_vars = c('Dysbiosis CD', 'Diagnosis CD')

# This section is necessary for updating the association plots
colnames(taxa_table) <- gsub('_', ' ', colnames(taxa_table)) %>%
    gsub(pattern = 'sp ', replacement = 'sp. ')

# Rename the features in the norm transformed data file
data_transformed <-
    read.csv('hmp2_output/features/data_transformed.tsv', sep='\t')
colnames(data_transformed) <-
    gsub('_', ' ', colnames(data_transformed)) %>%
    gsub(pattern = 'sp ', replacement = 'sp. ')
write.table(data_transformed,
            'hmp2_output/features/data_transformed.tsv',
            sep='\t', row.names = FALSE)

# Rename the metadata like in the outputs table
colnames(metadata) <-
    case_when(colnames(metadata) == 'age' ~ 'Age',
            colnames(metadata) == 'antibiotics' ~ 'Abx',
            colnames(metadata) == 'diagnosis' ~ 'Diagnosis',
            colnames(metadata) == 'dysbiosis_state' ~ 'Dysbiosis',
            colnames(metadata) == 'reads' ~ 'Read depth',
            TRUE ~ colnames(metadata))
metadata <- metadata %>%
    mutate(Dysbiosis = case_when(Dysbiosis == 'dysbiosis_UC' ~ 'UC',
                                Dysbiosis == 'dysbiosis_CD' ~ 'CD',
                                Dysbiosis == 'none' ~ 'None') %>%
            factor(levels = c('None', 'UC', 'CD')),
        Abx = case_when(Abx == 'Yes' ~ 'Used',
                        Abx == 'No' ~ 'Not used') %>%
            factor(levels = c('Not used', 'Used')),
        Diagnosis = case_when(Diagnosis == 'nonIBD' ~ 'non-IBD',
                                TRUE ~ Diagnosis) %>%
            factor(levels = c('non-IBD', 'UC', 'CD')))

# Recreate the plots
scatter_plots <- maaslin_plot_results_from_output(
    output = 'hmp2_output',
    metadata,
    normalization = "TSS",
    transform = "LOG",
    median_comparison_abundance = TRUE,
    median_comparison_prevalence = FALSE,
    max_significance = 0.1,
    coef_plot_vars=coef_plot_vars,
    heatmap_vars=heatmap_vars,
    max_pngs = 20)
```

In the new summary plot below, we can see that the feature names are
cleaned up, the metadata names are cleaned up, the set of metadata
variables used in the coefficient plot is different, and the metadata
used in the heatmap is reordered. (Make sure to set `include=T` to view
the plots when knitting the Rmd.)
```{r, out.width='100%', echo=FALSE, cache = FALSE, include=FALSE}
knitr::include_graphics("hmp2_output/figures/summary_plot.png")
```

In the association plots below, the taxa and metadata have been renamed to be
consistent with the results file from earlier. **To see the plots, set 
`max_pngs = 250` above and set `eval=TRUE` in this chunk.**
```{r, echo=FALSE, fig.show='hold', cache = FALSE, include=FALSE, eval=FALSE}
prefix <- "hmp2_output/figures/association_plots"
plot_vec <- c("/Age/linear/Age_Enterocloster clostridioformis_linear.png",
            "/Dysbiosis/linear/Dysbiosis_Escherichia coli_linear.png",
            "/Age/logistic/Age_Bifidobacterium longum_logistic.png",
            paste0("/Dysbiosis/logistic/",
                "Dysbiosis_Faecalibacterium prausnitzii_logistic.png"
                ))
knitr::include_graphics(paste0(prefix, plot_vec))
```

#### Contrast testing

We can also perform contrast testing in MaAsLin 3 to check whether a linear
combination of the coefficients is significantly different from some constant. 
Specify a contrast matrix
with column names matching the coefficient names and rows with the contrasts
of interest. A separate contrast test will be performed for each row.

```{R, echo = TRUE, results = 'hide', warning = FALSE, cache = FALSE}
contrast_mat <- matrix(c(1, 1, 0, 0, 0, 0, 1, 1), 
    ncol = 4, nrow = 2, byrow = TRUE)
    
colnames(contrast_mat) <- c("diagnosisUC",
                            "diagnosisCD",
                            "dysbiosis_statedysbiosis_UC",
                            "dysbiosis_statedysbiosis_CD")
                            
rownames(contrast_mat) <- c("diagnosis_test", "dysbiosis_test")

contrast_mat
                            
maaslin_contrast_test(maaslin3_fit = fit_out,
                        contrast_mat = contrast_mat)
```


#### Command line ####

MaAsLin 3 can also be run with a command line interface. For example,
the first HMP2 analysis can be performed with the following command (the
slashes may need to be removed):

```{r, engine = 'bash', eval = FALSE, cache = FALSE}
./R/maaslin3.R \
    inst/extdata/HMP2_taxonomy.tsv \
    inst/extdata/HMP2_metadata.tsv \
    command_line_output \
    --formula='~ diagnosis + dysbiosis_state + antibiotics + age + reads' \
    --reference='diagnosis,nonIBD;dysbiosis_state,none;antibiotics,No'
```

* Make sure to provide the full path to the MaAsLin3 executable (i.e.
`./R/maaslin3.R`).
* In the demo command:
    * ``inst/extdata/HMP2_taxonomy.tsv`` is the path to your data (or
    features) file
    * ``inst/extdata/HMP2_metadata.tsv`` is the path to your metadata file
    * ``command_line_output`` is the path to the folder to write the output

### Options ###

From the command line, the following command will print the list of
MaAsLin 3 options and default settings:

```$ ./R/maaslin3.R --help```

When running MaAsLin 3 in R, the manual page for each function (e.g.,
`?maaslin3`) will show the available options and default settings. For
both, the options and settings are as follows:

#### Required parameters ####

* `input_data`: A data frame of feature abundances or read counts or a
filepath to a tab-delimited file with abundances. It should be formatted
with features as columns and samples as rows (or the transpose). The
column and row names should be the feature names and sample names
respectively.
* `input_metadata`: A data frame of per-sample metadata or a filepath to
a tab-delimited file with metadata. It should be formatted with
variables as columns and samples as rows (or the transpose). The column
and row names should be the variable names and sample names
respectively.
* `output`: The output folder to write results.

#### Model formula ####

* `formula`: A formula in `lme4` format. Random effects, interactions,
and functions of the metadata can be included (note that these functions
will be applied after standardization if `standardize = TRUE`). Group,
ordered, and strata variables can be specified as:
`group(grouping_variable)`, `ordered(ordered_variable)`, and
`strata(strata_variable)`. The other variable options below will not be
considered if a formula is set.
* `fixed_effects`: A vector of variable names to be included as fixed
effects.
    * Fixed effects models are fit with `lm` (linear) or `glm` (logistic).
* `reference`: For a variable with more than two levels supplied with
`fixed_effects`, the factor to use as a reference provided as a string
of 'variable,reference' semi-colon delimited for multiple variables.
* `random_effects`: A vector of variable names to be included as random
intercepts. **Random intercept models may produce poor model fits when
there are fewer than 5 observations per group.** In these scenarios,
per-group fixed effects should be used and subsequently filtered out.
(See `strata_effects` as well.)
    * Random effects models are fit with `lmer` (linear) and `glmer`
(logistic), and the significance tests come from `lmerTest` and `glmer`
respectively.
* `group_effects`: A factored categorical variable to be included for
group testing. An ANOVA-style test will be performed to assess whether
any of the variable's levels are significant, and no coefficients or
individual p-values will be returned.
    * Tests are performed with the `anova` function's `LRT` option (logistic
fixed and mixed effects), the `anova` function's F test (linear fixed
effects), or `lmerTest::contest` (linear mixed effects).
* `ordered_effects`: A factored categorical variable to be included.
Consecutive levels will be tested for significance against each other
with contrast tests, and the resulting associations will correspond to
effect sizes, standard errors, and significances of each level versus
the previous.
    * Contrast tests are performed with `multcomp::glht` (fixed effects and
logistic mixed effects) and `lmerTest::contest` (linear mixed effects).
* `strata_effects`: A single grouping variable to be included in matched
case-control studies. If a strata variable is included, no random
effects can be included. When a strata variable is included, a
conditional logistic regression will be run to account for the strata.
The abundance model will be run with a random intercept in place of the
strata. Strata can include more than two observations per group. Only
variables that differ within the groups can be tested. In general, strata
effects are not recommended except for advanced users. Fixed or random
intercepts are recommended instead.

#### Feature specific covariates ####
Particularly for use in metatranscriptomics workflows, a table of
feature-specific covariates can be included. A feature's covariates will
be included like a fixed effect metadatum when fitting the model for
that feature. The covariate's name does not need to be included in the
formula.

* `feature_specific_covariate`: A table of feature-specific covariates
or a filepath to a tab-delimited file with feature-specific covariates.
It should be formatted with features as columns and samples as rows (or
the transpose). The row names and column names should be the same as
those of the `input_data`: the column and row names should be the
feature names and sample names respectively. Typically, this table should be
generated by `preprocess_dna_mtx` or `preprocess_taxa_mtx` first.
* `feature_specific_covariate_name`: The name for the feature-specific
covariates when fitting the models.
* `feature_specific_covariate_record`: Whether to keep the
feature-specific covariates in the outputs when calculating p-values,
writing results, and displaying plots.

#### Analysis options ####

* `min_abundance` (default `0`): Features with abundances of more than
`min_abundance` in more than `min_prevalence` of the samples will be 
included for analysis. The threshold is applied after normalization and before
transformation.
* `min_prevalence` (default `0`): See above.
* `max_prevalence` (default `1.01`): Features with abundances more than
`min_abundance` in fewer than `max_prevalence` of the samples will be 
included for analysis. The threshold is applied after normalization and
before transformation.
* `zero_threshold` (default `0`): Abundances less than or equal to
`zero_threshold` will be treated as zeros. This is primarily to be used
when the abundance table has likely low-abundance false positives.
* `min_variance` (default `0`): Features with abundance variances less
than or equal to `min_variance` will be dropped. This is primarily used
for dropping features that are entirely zero.
* `max_significance` (default `0.1`): The FDR corrected q-value
threshold for significance used in selecting which associations to write
as significant and to plot.
* `normalization` (default `TSS`): The normalization to apply to the
features before transformation and analysis. The option `TSS` (total-sum
scaling) is recommended, but `CLR` (centered log ratio) and `NONE` can
also be used.
* `transform` (default `LOG`, base 2): The transformation to apply to
the features after normalization and before analysis. The option `LOG`
is recommended, but `PLOG` (pseudo-log with a pseudo-count of half the 
dataset minimum non-zero abundance replacing zeros, particularly for
metabolomics data) and `NONE` can also be used.
* `correction` (default `BH`): The correction to obtain FDR-corrected
q-values from raw p-values. Any valid options for `p.adjust` can be
used.
* `standardize` (default `TRUE`): Whether to apply z-scores to
continuous metadata variables so they are on the same scale. This is
recommended in order to compare coefficients across metadata variables,
but note that functions of the metadata specified in the `formula` will
apply after standardization.
* `warn_prevalence` (default `TRUE`): Warn when prevalence associations 
are likely induced by abundance associations. This requires re-fitting the 
linear models on the TSS log-transformed data. A prevalence coefficient is
flagged if its corresponding abundance coefficient is significantly different
from 0, of the same sign, and larger in absolute value.
* `small_random_effects` (default `FALSE`): Automatically replace random 
effects with fixed effects in the logistic prevalence model to handle low 
numbers of observations per group.
* `bypass_small_group_warning` (default `FALSE`): When `TRUE`, bypass the 
warning that is normally added to results when there are fewer than 4 average 
observations per random effect group.
* `augment` (default `TRUE`): To avoid linear separability in the
logistic regression, at each input data point, add an extra 0 and an
extra 1 observation weighted as the number of predictors divided by two
times the number of data points. This is almost always recommended to
avoid linear separability while having a minor effect on fit
coefficients otherwise.
* `evaluate_only` (default `NULL`): To fit only the abundance or only the
prevalence models, `evaluate_only` can be set to `abundance` or `prevalence`.

#### Compositionality corrections ####

##### Absolute abundance

Most microbiome methodologies have historically focused on relative
abundances (proportions out of 1). However, some experimental protocols
can enable estimation of absolute abundances (cell count/concentration).
MaAsLin 3 can be used with two types of absolute abundance estimation:
spike-ins and total abundance scaling. In a spike-in procedure, a small,
known quantity of a microbe that otherwise would not be present in the
sample is added, and the sequencing procedure is carried out as usual.
Then, the absolute abundance of a microbe already in the community is
estimated as:
$$\textrm{Absolute abundance other microbe}=\frac{\textrm{Relative
abundance other microbe}}{\textrm{Relative abundance spike-in
microbe}}\cdot (\textrm{Absolute abundance spike-in microbe})$$
Alternatively, the total microbial abundance of a sample can be
determined (e.g., with qPCR of a marker gene or by cell counting). Then,
the absolute abundance of a microbe in the community is estimated as:
$$\textrm{Absolute abundance microbe}=(\textrm{Total absolute
abundance})\cdot(\textrm{Relative abundance microbe})$$

##### Compositionality corrections continued

* `unscaled_abundance`: A data frame with a single column of absolute
abundances or a filepath to such a tab-delimited file. The row names
should match the names of the samples in `input_data` and
`input_metadata`. When using spike-ins, the single column should have
the same name as one of the features in `input_data`, and the values
should correspond to the absolute quantity of the spike-in. For example,
if the same spike-in quantity is used in each sample, the entire column
can be set to 1. When using total abundance scaling, the single column
should have the name 'total', and the values should correspond to the
total abundance of each sample. In both cases,
`median_comparison_abundance` should be set to `FALSE` since the
spike-in or total abundance normalization accounts for compositionality.

Alternatively, if the `input_data` abundances have already been scaled
to be absolute abundances, the user should set `normalization = NONE`
and `median_comparison_abundance = FALSE` and not include anything for
`unscaled_abundance`. Then, the absolute abundances will be log
transformed, and models will be fit on those values directly.

##### Median comparisons

When `median_comparison_abundance` or `median_comparison_prevalence` are
on, the coefficients for a metadatum will be tested against the median
coefficient for that metadatum (median across the features). Otherwise,
the coefficients will be tested against 0. For abundance associations,
this is designed to account for compositionality, the idea that if only
one feature has a positive association with a metadatum on the absolute
scale (cell count), the other features will have apparent negative
associations with that metadatum on the relative scale (proportion of
the community) because relative abundances must sum to 1. More
generally, associations on the relative scale are not necessarily the
same as the associations on the absolute scale in magnitude or sign, so
**testing against zero on the relative scale is not equivalent to
testing against zero on the absolute scale**. When testing associations
on the relative scale, the coefficients should be tested against 0
(median comparison off). However, since these tests do not correspond to
tests for associations on the absolute scale, we instead use a test
against the median, which can enable some inference on the absolute
scale. There are two interpretations of this test for absolute abundance
associations:

1. In linear models without sparsity (or with sparsity under some assumptions), 
if two features' associations with a particular
metadatum on the log *absolute* scale differ by some value $d$, the
features' associations with that metadatum on the log *relative* scale
(total-sum scaling) will also differ by $d$. That is, the absolute and
relative coefficients for a particular feature-metadatum association are
different, but **the ordering of the relative coefficients is the same
as the ordering of the absolute coefficients for a metadatum, and the
difference between two coefficients on the relative scale is the same as
the difference between the corresponding coefficients on the absolute
scale**. Therefore, the test against the relative coefficient median can
always be interpreted as "a test of whether a particular association is
significantly different from the typical (median) association on the
absolute scale."
2. Under the assumption that at least half the features are not changing
on the absolute scale, the median true absolute coefficient is 0, so
this can be interpreted as a test of whether the feature has a non-zero
association on the absolute scale.

By contrast, sparsity should be less affected by compositionality since
a feature should still be present even if another increases or decreases
in abundance. (Note that, because the read depth is finite, this might
not always be true in practice.) Therefore,
`median_comparison_prevalence` is off by default, but it can be turned
on if the user is interested in testing whether a particular prevalence
association is significantly different from the typical prevalence
association.

In both cases, if the tested coefficient is within
`median_comparison_[abundance/prevalence]_threshold` of the median, it
will automatically receive a p-value of 1. This is based on the idea
that the association might be statistically significantly different but
not substantially different from the median and therefore is likely
still a result of compositionality.

To conclude:

* `median_comparison_abundance` is `TRUE` by default and should be used
to make inference on the absolute scale when using relative abundance
data. When `median_comparison_abundance` is `TRUE`, only the p-values
and q-values change. The coefficients returned are still the relative
abundance coefficients unless `subtract_median` is set to `TRUE` in which
case the median will be subtracted.
* `median_comparison_abundance` should be `FALSE` when (1) testing for
significant relative associations, (2) testing for absolute abundance
associations under the assumption that the total absolute abundance is
not changing, or (3) testing for significant absolute associations when
supplying spike-in or total abundances with `unscaled_abundance`.
* `median_comparison_prevalence` is `FALSE` by default.

##### Compositionality corrections continued

* `median_comparison_abundance` (default `TRUE`): Test abundance
coefficients against a null value corresponding to the median
coefficient for a metadata variable across the features. Otherwise, test
against 0. This is recommended for relative abundance data but should
not be used for absolute abundance data.
* `median_comparison_prevalence` (default `FALSE`): Test prevalence
coefficients against a null value corresponding to the median
coefficient for a metadata variable across the features. Otherwise, test
against 0. This is only recommended if the analyst is interested in how
feature prevalence associations compare to each other or if there is
likely strong compositionality-induced sparsity.
* `median_comparison_abundance_threshold` (default `0`): Coefficients
within `median_comparison_abundance_threshold` of the median association
will automatically be counted as insignificant (p-value set to 1) since
they likely represent compositionality-induced associations. This
threshold will be divided by the metadata variable's standard deviation
if the metadatum is continuous to ensure the threshold applies to the
right scale.
* `median_comparison_prevalence_threshold` (default `0`): Same as
`median_comparison_abundance_threshold` but applied to the prevalence
associations.
* `subtract_median` (default `FALSE`): Subtract the median from the 
coefficients.

#### Plotting parameters ####

* `plot_summary_plot` (default `TRUE`): Generate a summary plot of
significant associations.
* `summary_plot_first_n` (default `25`): Include the top
`summary_plot_first_n` features with significant associations.
* `coef_plot_vars`: Vector of variable names to be used in the
coefficient plot section of the summary plot. Continuous variables
should match the metadata column name, and categorical variables should
be of the form `"[variable] [level]"`.

By default, the (up to) two metadata variables with the most significant
associations will be plotted in the coefficient plot, and the rest will
be plotted in the heatmap. Because predicting the output variable names
can be tricky, it is recommended to first run `maaslin3` without setting
`coef_plot_vars` or `heatmap_vars`, look at the names of the variables
in the summary plot, and then rerun with
`maaslin_plot_results_from_output` after updating `coef_plot_vars` and
`heatmap_vars` with the desired variables.
* `heatmap_vars`: Vector of variable names to be used in the heatmap
section of the summary plot. Continuous variables should match the
metadata column name, and categorical variables should be of the form
`"[variable] [level]"`.
* `plot_associations` (default `TRUE`): Whether to generate plots for
significant associations.
* `max_pngs` (default `30`): The top `max_pngs` significant associations
will be plotted.

#### Technical/miscellaneous parameters ####

* `save_models` (default `FALSE`): Whether to return the fit models and
save them to an RData file.
* `verbosity` (default `'FINEST'`): The level of verbosity for the 
`logging` package.
* `save_plots_rds` (default `FALSE`): Whether to save the plots as RDS files.
* `assay.type` (default `1`): A string or index to select the assay when using
a `SummarizedExperiment` object.
* Enable parallelization by creating background 
[daemons](https://mirai.r-lib.org/reference/daemons.html) using 
[mirai](https://mirai.r-lib.org/articles/mirai.html) e.g. call 
`mirai::daemons(2)` before running `maaslin3()` to use two cores.

## Tool comparison ##

| Tool | Models differential abundance | \
Models differential prevalence | Models allowed | Accounts for\
compositionality | Can use experimental absolute abundance information |
|:---:|:---:|:---:|:---:|:---:|:---:|
| MaAsLin 3 | Yes | Yes, for any model type | Any lme4 model\
(general mixed models), level-vs-level differences, group-wise differences, \
feature-specific covariates, contrast tests | Yes | Yes |
| MaAsLin 2 | Yes | Only insofar as it affects abundance | Fixed and random \
effects (subset of general mixed models) | No | No |
| ANCOM-BC2 | Yes | Only when completely absent in a group | Any lme4 model \
(general mixed models), level-vs-level differences, group-wise differences, \
contrast tests | Yes | No |
| ALDEx2 | Yes | Only insofar as it affects abundance | Fixed \
effects | Yes | No |
| LinDA | Yes | Only insofar as it affects abundance | Any lme4 model \
(general mixed models), group-wise differences, contrast tests | Yes | No |

## Troubleshooting ##

1. Question: When I run from the command line I see the error
``maaslin3.R: command not found``. How do I fix this?
    * Answer: Provide the full path to the executable when running
    maaslin3.R
2. Question: When I run as a function I see the error ``Error in
library(maaslin3): there is no package called 'maaslin3'``. How do I fix
this?
    * Answer: Install the R package and then try loading the library again.

```{R}
sessionInfo()
```

```{R}
# Clean-up generated outputs
unlink('hmp2_output', recursive = TRUE)
```

Example input files can be found in the ``inst/extdata`` folder of the
MaAsLin 3 source.

* ``abundance_spike_in_ex.tsv`` is the abundance table generated from a 
modified version of SparseDOSSA 2 as documented here:
https://github.com/willnickols/maaslin3_benchmark. SparseDOSSA 2 uses
a zero-inflated log-linear abundance model to generate microbiome profiles
based on user-defined metadata and effect sizes. This procedure was modified
to include an extra feature with a recorded absolute abundance before
transforming the feature table to relative abundances.
* ``abundance_total_ex.tsv`` is the abundance table generated from a 
modified version of SparseDOSSA 2 as documented here:
https://github.com/willnickols/maaslin3_benchmark. SparseDOSSA 2 uses
a zero-inflated log-linear abundance model to generate microbiome profiles
based on user-defined metadata and effect sizes. This procedure was modified
to record the total absolute abundance before
transforming the feature table to relative abundances.
* ``HMP2_metadata.tsv``: a tab-delimited file with samples as rows and
metadata as columns. It is a subset of the full HMP2 metadata available at
https://www.ibdmdb.org/results that includes a subset of fields that
are of interest.
* ``HMP2_taxonomy.tsv``: a tab-delimited file with samples as rows and
species as columns. It is a subset of the full HMP2 taxonomic profiles table
generated by MetaPhlAn 4 (https://huttenhower.sph.harvard.edu/metaphlan/) from
sequencing data available at https://www.ibdmdb.org/results. The table 
includes a subset of taxa that are of interest.
* ``metadata_spike_in_ex.tsv`` is the metadata table generated from a 
modified version of SparseDOSSA 2 as documented here:
https://github.com/willnickols/maaslin3_benchmark. SparseDOSSA 2 uses
a zero-inflated log-linear abundance model to generate microbiome profiles
based on user-defined metadata and effect sizes. This procedure was modified
to include an extra feature with a recorded absolute abundance before
transforming the feature table to relative abundances.
* ``metadata_total_ex.tsv`` is the metadata table generated from a 
modified version of SparseDOSSA 2 as documented here:
https://github.com/willnickols/maaslin3_benchmark. SparseDOSSA 2 uses
a zero-inflated log-linear abundance model to generate microbiome profiles
based on user-defined metadata and effect sizes. This procedure was modified
to record the total absolute abundance before
transforming the feature table to relative abundances.
* ``scaling_factors_spike_in_ex.tsv`` is the spike-in table generated from a 
modified version of SparseDOSSA 2 as documented here:
https://github.com/willnickols/maaslin3_benchmark. SparseDOSSA 2 uses
a zero-inflated log-linear abundance model to generate microbiome profiles
based on user-defined metadata and effect sizes. This procedure was modified
to include an extra feature with a recorded absolute abundance before
transforming the feature table to relative abundances.
* ``scaling_factors_total_ex.tsv`` is the total per-sample abundance table 
generated from a modified version of SparseDOSSA 2 as documented here:
https://github.com/willnickols/maaslin3_benchmark. SparseDOSSA 2 uses
a zero-inflated log-linear abundance model to generate microbiome profiles
based on user-defined metadata and effect sizes. This procedure was modified
to record the total absolute abundance before
transforming the feature table to relative abundances.

No licenses are required or provided.

\name{maaslin_plot_results_from_output}
\alias{maaslin_plot_results_from_output}
\title{
Plot the results from a MaAsLin 3 run.
}
\description{
Two types of plots are generated. First, the summary plot contains
sorted per-feature coefficients plotted with their standard errors for
key variables and a heatmap summarizing the remaining variables. Second,
for significant features, association plots (scatterplots, boxplots, or
tables depending on the association) are generated to visualize and
verify the model fits. The data are shown with their transformed values
in the association plots since this is the scale on which the models are
fit. In comparison to \code{maaslin_plot_results} that needs the entire
\code{maaslin_fit} list, only the parameter list and an outputs
directory containing a completed run are needed for
\code{maaslin_plot_results_from_output}.
}
\usage{
maaslin_plot_results_from_output(output,
                                metadata,
                                normalization,
                                transform,
                                feature_specific_covariate = NULL,
                                feature_specific_covariate_name = NULL,
                                feature_specific_covariate_record = NULL,
                                median_comparison_abundance = TRUE,
                                median_comparison_prevalence = FALSE,
                                max_significance = 0.1,
                                plot_summary_plot = TRUE,
                                summary_plot_first_n = 25,
                                coef_plot_vars = NULL,
                                heatmap_vars = NULL,
                                plot_associations = TRUE,
                                max_pngs = 30,
                                balanced = FALSE,
                                save_plots_rds = FALSE)
}
\arguments{
    \item{output}{The output folder to write results.}
    \item{metadata}{A data frame of per-sample metadata. It should be
    formatted with variables as columns and samples as rows. The column and
    row names should be the variable names and sample names respectively.}
    \item{normalization}{The normalization to apply to the features before
    transformation and analysis. The option \code{TSS} (total sum scaling)
    is recommended, but \code{CLR} (centered log ratio) and \code{NONE} can
    also be used.}
    \item{transform}{The transformation to apply to the features after
    normalization and before analysis. The option \code{LOG} (base 2) is
    recommended, but \code{PLOG} (pseudo-log) and \code{NONE} can also be
    used.}
    \item{feature_specific_covariate}{A table of feature-specific covariates
    or
    a filepath to a tab-delimited file with feature-specific covariates. It
    should be formatted with features as columns and samples as rows (or the
    transpose). The row names and column names should be the same as those
    of the \code{input_data}: the column and row names should be the feature
    names and sample names respectively. Typically, this table should be
    generated by `preprocess_dna_mtx` or `preprocess_taxa_mtx` first.}
    \item{feature_specific_covariate_name}{The name for the feature-specific
    covariates when fitting the models.}
    \item{feature_specific_covariate_record}{Whether to keep the
    feature-specific covariates in the outputs when calculating p-values,
    writing results, and displaying plots.}
    \item{median_comparison_abundance}{Test abundance coefficients against a
    null value corresponding to the median coefficient for a metadata
    variable across the features. This is recommended for relative abundance
    data but should not be used for absolute abundance data.}
    \item{median_comparison_prevalence}{Test prevalence coefficients against
    a null value corresponding to the median coefficient for a metadata
    variable across the features. This is only recommended if the analyst is
    interested in how feature prevalence associations compare to each other
    or if there is likely strong compositionality-induced sparsity.}
    \item{max_significance}{The FDR corrected q-value threshold for
    significance used in selecting which associations to write as
    significant and to plot.}
    \item{plot_summary_plot}{Generate a summary plot of significant
    associations.}
    \item{summary_plot_first_n}{Include the top \code{summary_plot_first_n}
    features with significant associations.}
    \item{coef_plot_vars}{Vector of variable names to be used in the
    coefficient plot section of the summary plot. Continuous variables
    should match the metadata column name, and categorical variables should
    be of the form \code{"[variable] [level]"}.}
    \item{heatmap_vars}{Vector of variable names to be used in the heatmap
    section of the summary plot. Continuous variables should match the
    metadata column name, and categorical variables should be of the form
    \code{"[variable] [level]"}.}
    \item{plot_associations}{Whether to generate plots for significant
    associations.}
    \item{max_pngs}{The top \code{max_pngs} significant associations will be
    plotted.}
    \item{balanced}{If set to TRUE the summary plot will 
    show the top N features of each variable included in 
    \code{coef_plot_vars} and \code{heatmap_vars} where N is equal to: 
    \code{ceiling(summary_plot_first_n/length(coef_plot_vars))}. Will error
    if \code{coef_plot_vars} = \code{NULL}}
    \item{save_plots_rds}{Whether to return the plots to an RDS file.}
}

\value{
Results will be written to the \code{figures} folder within the folder
\code{output}. The list of individual association plots is returned if
\code{plot_associations=TRUE}. In the heatmap of the summary plot, one
star corresponds to the user-set \code{max_significance} and two stars
corresponds to the user-set \code{max_signifiance} divided by 10.
}
\author{
    William Nickols<willnickols@g.harvard.edu>,\cr
    Jacob Nearing<nearing@broadinstitute.org>,\cr
    Maintainers: Lauren McIver<lauren.j.mciver@gmail.com>,\cr
}
\examples{
    # Read features table
    taxa_table_name <- system.file("extdata", "HMP2_taxonomy.tsv", package =
    "maaslin3")
    taxa_table <- read.csv(taxa_table_name, sep = '\t', row.names = 1)
    
    # Read metadata table
    metadata_name <- system.file("extdata", "HMP2_metadata.tsv", package =
    "maaslin3")
    metadata <- read.csv(metadata_name, sep = '\t', row.names = 1)
    
    metadata$diagnosis <-
        factor(metadata$diagnosis, levels = c('nonIBD', 'UC', 'CD'))
    metadata$dysbiosis_state <-
        factor(metadata$dysbiosis_state, levels = c('none', 'dysbiosis_UC',
        'dysbiosis_CD'))
    metadata$antibiotics <-
        factor(metadata$antibiotics, levels = c('No', 'Yes'))
    
    #Run MaAsLin3
    maaslin3::maaslin_log_arguments(
        input_data = taxa_table,
        input_metadata = metadata,
        output = 'output',
        formula = '~ diagnosis + dysbiosis_state + antibiotics +
        age + reads')
    
    read_data_list <- maaslin3::maaslin_read_data(
        taxa_table,
        metadata)
    read_data_list <- maaslin3::maaslin_reorder_data(
        read_data_list$data,
        read_data_list$metadata)
    
    data <- read_data_list$data
    metadata <- read_data_list$metadata
    
    formulas <- maaslin3::maaslin_check_formula(
        data,
        metadata,
        input_formula = '~ diagnosis + dysbiosis_state + antibiotics +
        age + reads')
    
    formula <- formulas$formula
    random_effects_formula <- formulas$random_effects_formula
    
    normalized_data = maaslin3::maaslin_normalize(data,
                                output = 'output')
                                
    filtered_data = maaslin3::maaslin_filter(normalized_data,
                                output = 'output')
                                
    transformed_data = maaslin3::maaslin_transform(filtered_data,
                                output = 'output')
                                
    standardized_metadata = maaslin3::maaslin_process_metadata(
        metadata,
        formula = formula)
    
    maaslin_results = maaslin3::maaslin_fit(
        filtered_data,
        transformed_data,
        standardized_metadata,
        formula,
        random_effects_formula,
        warn_prevalence = FALSE)
    
    maaslin3::maaslin_write_results(
        output = 'output',
        maaslin_results$fit_data_abundance,
        maaslin_results$fit_data_prevalence,
        random_effects_formula)
    
    maaslin3::maaslin_plot_results_from_output(
        output = 'output',
        metadata,
        normalization = "TSS",
        transform = "LOG")
    
    unlink('output', recursive=TRUE)
    logging::logReset()
}

\name{maaslin_read_data}
\alias{maaslin_read_data}
\title{
Read in the abundance data and metadata.
}
\description{
Read in the abundance data and metadata from files if necessary.
}
\usage{
maaslin_read_data(input_data,
                input_metadata,
                feature_specific_covariate = NULL,
                unscaled_abundance = NULL)
}
\arguments{
    \item{input_data}{A data frame of feature abundances or read counts or a
    filepath to a tab-delimited file with abundances. It should be formatted
    with features as columns and samples as rows (or the transpose). The
    column and row names should be the feature names and sample names
    respectively.}
    \item{input_metadata}{A data frame of per-sample metadata or a filepath
    to a tab-delimited file with metadata. It should be formatted with
    variables as columns and samples as rows (or the transpose). The column
    and row names should be the variable names and sample names
    respectively.}
    \item{feature_specific_covariate}{A table of feature-specific covariates
    or
    a filepath to a tab-delimited file with feature-specific covariates. It
    should be formatted with features as columns and samples as rows (or the
    transpose). The row names and column names should be the same as those
    of the \code{input_data}: the column and row names should be the feature
    names and sample names respectively. Typically, this table should be
    generated by `preprocess_dna_mtx` or `preprocess_taxa_mtx` first.}
    \item{unscaled_abundance}{A data frame with a single column of absolute
    abundances or a filepath to such a tab-delimited file. The row names
    should match the names of the samples in \code{input_data} and
    \code{input_metadata}. When using spike-ins, the single column should
    have the same name as one of the features in \code{input_data}, and the
    \code{unscaled_abundance} should correspond to the absolute quantity of
    the spike-in. When using total abundance scaling, the single column
    should have the name 'total', and the \code{unscaled_abundance} should
    correspond to the total abundance of each sample.}
}
\value{
    A list containing the following items:
    \describe{
    \item{(1)}{\code{data}: A data frame of feature abundances.}
    \item{(2)}{\code{metadata}: A data frame of metadata.}
    \item{(3)}{\code{feature_specific_covariate}: A data frame of feature
    specific covariates.}
    \item{(4)}{\code{unscaled_abundance}: A data frame of unscaled
    abundances.}

    }
}
\author{
    William Nickols<willnickols@g.harvard.edu>,\cr
    Jacob Nearing<nearing@broadinstitute.org>,\cr
    Maintainers: Lauren McIver<lauren.j.mciver@gmail.com>,\cr
}
\examples{
    # Read features table
    taxa_table_name <- system.file("extdata", "HMP2_taxonomy.tsv", package =
    "maaslin3")
    taxa_table <- read.csv(taxa_table_name, sep = '\t', row.names = 1)
    
    # Read metadata table
    metadata_name <- system.file("extdata", "HMP2_metadata.tsv", package =
    "maaslin3")
    metadata <- read.csv(metadata_name, sep = '\t', row.names = 1)
    
    metadata$diagnosis <-
        factor(metadata$diagnosis, levels = c('nonIBD', 'UC', 'CD'))
    metadata$dysbiosis_state <-
        factor(metadata$dysbiosis_state, levels = c('none', 'dysbiosis_UC',
        'dysbiosis_CD'))
    metadata$antibiotics <-
        factor(metadata$antibiotics, levels = c('No', 'Yes'))
    
    #Run MaAsLin3
    maaslin3::maaslin_log_arguments(
        input_data = taxa_table,
        input_metadata = metadata,
        output = 'output',
        formula = '~ diagnosis + dysbiosis_state + antibiotics +
        age + reads',
        plot_summary_plot = FALSE,
        plot_associations = FALSE)
    read_data_list <- maaslin3::maaslin_read_data(
        taxa_table,
        metadata)
    
    unlink('output', recursive=TRUE)
    logging::logReset()
}

\name{maaslin_process_metadata}
\alias{maaslin_process_metadata}
\title{
Process metadata before MaAsLin 3 model fitting.
}
\description{
Check that references are set properly if the metadata variables are
categorical and provided through \code{fixed_effects}. Standardize the
continuous metadata variables as a z-score (subtract the mean, divide by
the standard deviation) if \code{standardize} is set.
}
\usage{
maaslin_process_metadata(metadata,
                        formula = NULL,
                        fixed_effects = NULL,
                        reference = NULL,
                        feature_specific_covariate_name = NULL,
                        standardize = TRUE)
}
\arguments{
    \item{metadata}{A data frame of per-sample metadata. It should be
    formatted with variables as columns and samples as rows. The column and
    row names should be the variable names and sample names respectively.}
    \item{formula}{A formula in \code{lme4} format. Random effects,
    interactions, and functions of the metadata can be included (note that
    these functions will be applied after standardization if
    \code{standardize=TRUE}). Group, ordered, and strata variables can be
    specified as: \code{group(grouping_variable)},
    \code{ordered(ordered_variable)} and \code{strata(strata_variable)}. The
    other variable options below will not be considered if a formula is
    set.}
    \item{fixed_effects}{A vector of variable names to be included as fixed
    effects.}
    \item{reference}{For a variable with more than two levels supplied with
    \code{fixed_effects}, the factor to use as a reference provided as a
    string of 'variable,reference' semi-colon delimited for multiple
    variables.}
    \item{feature_specific_covariate_name}{The name for the feature-specific
    covariates when fitting the models.}
    \item{standardize}{Whether to apply z-scores to continuous metadata
    variables so they are on the same scale. This is recommended in order to
    compare coefficients across metadata variables, but note that functions
    of the metadata specified in the \code{formula} will apply after
    standardization.}
}
\value{
The processed metadata.
}
\author{
    William Nickols<willnickols@g.harvard.edu>,\cr
    Jacob Nearing<nearing@broadinstitute.org>,\cr
    Maintainers: Lauren McIver<lauren.j.mciver@gmail.com>,\cr
}
\examples{
    # Read features table
    taxa_table_name <- system.file("extdata", "HMP2_taxonomy.tsv", package =
    "maaslin3")
    taxa_table <- read.csv(taxa_table_name, sep = '\t', row.names = 1)
    
    # Read metadata table
    metadata_name <- system.file("extdata", "HMP2_metadata.tsv", package =
    "maaslin3")
    metadata <- read.csv(metadata_name, sep = '\t', row.names = 1)
    
    metadata$diagnosis <-
        factor(metadata$diagnosis, levels = c('nonIBD', 'UC', 'CD'))
    metadata$dysbiosis_state <-
        factor(metadata$dysbiosis_state, levels = c('none', 'dysbiosis_UC',
        'dysbiosis_CD'))
    metadata$antibiotics <-
        factor(metadata$antibiotics, levels = c('No', 'Yes'))
    
    #Run MaAsLin3
    maaslin3::maaslin_log_arguments(
        input_data = taxa_table,
        input_metadata = metadata,
        output = 'output',
        formula = '~ diagnosis + dysbiosis_state + antibiotics +
        age + reads',
        plot_summary_plot = FALSE,
        plot_associations = FALSE)
    
    read_data_list <- maaslin3::maaslin_read_data(
        taxa_table,
        metadata)
        read_data_list <- maaslin3::maaslin_reorder_data(
        read_data_list$data,
        read_data_list$metadata)
    
    data <- read_data_list$data
    metadata <- read_data_list$metadata
    
    formulas <- maaslin3::maaslin_check_formula(
        data,
        metadata,
        input_formula = '~ diagnosis + dysbiosis_state + antibiotics +
        age + reads')
    
    formula <- formulas$formula
    random_effects_formula <- formulas$random_effects_formula
    
    normalized_data = maaslin3::maaslin_normalize(data,
                                    output = 'output')
                                    
    filtered_data = maaslin3::maaslin_filter(normalized_data,
                                    output = 'output')
                                    
    standardized_metadata = maaslin3::maaslin_process_metadata(
        metadata,
        formula = formula)
    
    unlink('output', recursive=TRUE)
    logging::logReset()
}

\name{maaslin3}
\alias{maaslin3}
\title{
MaAsLin 3: A multivariable statistical framework for finding abundance
and
prevalence associations between metadata and high-dimensional microbial
multi-omics data.
}
\description{
This wrapper for all MaAsLin 3 steps finds abundance and prevalence
    associations between microbiome meta-omics features and
    complex metadata in population-scale epidemiological studies.
    The software includes multiple analysis methods (including
    support for multiple covariates, repeated measures, and ordered
    predictors),
    filtering, normalization, and transform options to customize analysis
    for
    your specific study.
}
\usage{
maaslin3(input_data,
    input_metadata = NULL,
    output,
    formula = NULL,
    fixed_effects = NULL,
    reference = NULL,
    random_effects = NULL,
    group_effects = NULL,
    ordered_effects = NULL,
    strata_effects = NULL,
    feature_specific_covariate = NULL,
    feature_specific_covariate_name = NULL,
    feature_specific_covariate_record = NULL,
    min_abundance = 0,
    min_prevalence = 0,
    max_prevalence = 1.01,
    zero_threshold = 0,
    min_variance = 0,
    max_significance = 0.1,
    normalization = 'TSS',
    transform = 'LOG',
    correction = 'BH',
    standardize = TRUE,
    unscaled_abundance = NULL,
    median_comparison_abundance = TRUE,
    median_comparison_prevalence = FALSE,
    median_comparison_abundance_threshold = 0,
    median_comparison_prevalence_threshold = 0,
    subtract_median = FALSE,
    warn_prevalence = TRUE,
    small_random_effects = FALSE,
    bypass_small_group_warning = FALSE,
    augment = TRUE,
    evaluate_only = NULL,
    plot_summary_plot = TRUE,
    summary_plot_first_n = 25,
    coef_plot_vars = NULL,
    heatmap_vars = NULL,
    plot_associations = TRUE,
    max_pngs = 30,
    cores = 1,
    save_models = FALSE,
    save_plots_rds = FALSE,
    verbosity = 'FINEST',
    summary_plot_balanced = FALSE,
    assay.type = 1)
}
\arguments{
    \item{input_data}{A data frame of feature abundances or read counts, a
    filepath to a tab-delimited file with abundances, or a SummarizedExperiment
    or TreeSummarizedExperiment object with the taxa table in `assays` and 
    metadata in `colData`. 
    If a data frame or a filepath is supplied, the table should be formatted 
    with features as columns and samples as rows (or the transpose). The
    column and row names should be the feature names and sample names
    respectively.}
    \item{input_metadata}{A data frame of per-sample metadata or a filepath
    to
    a tab-delimited file with metadata. It should be formatted with
    variables
    as columns and samples as rows (or the transpose). The column and row
    names
    should be the variable names and sample names respectively.}
    \item{output}{The output folder to write results.}
    \item{formula}{A formula in \code{lme4} format. Random effects,
    interactions, and functions of the metadata can be included (note that
    these functions will be applied after standardization if
    \code{standardize=TRUE}). Group, ordered, and strata variables can be
    specified as: \code{group(grouping_variable)},
    \code{ordered(ordered_variable)} and \code{strata(strata_variable)}. The
    other variable options below will not be considered if a formula is
    set.}
    \item{fixed_effects}{A vector of variable names to be included as fixed
    effects.}
    \item{reference}{For a variable with more than two levels supplied with
    \code{fixed_effects}, the factor to use as a reference provided as a
    string
    of 'variable,reference' semi-colon delimited for multiple variables.}
    \item{random_effects}{A vector of variable names to be included as
    random
    intercepts.}
    \item{group_effects}{A factored categorical variable to be included for
    group testing. An ANOVA-style test will be performed to assess whether
    any
    of the variable's levels are significant, and no coefficients or
    individual
    p-values will be returned.}
    \item{ordered_effects}{A factored categorical variable to be included.
    Consecutive levels will be tested for significance against each other,
    and
    the resulting associations will correspond to effect sizes, standard
    errors, and significances of each level versus the previous.}
    \item{strata_effects}{A vector with one variable name to be included as
    the
    strata variable in case-control studies. Strata cannot be combined with
    random effects.}
    \item{feature_specific_covariate}{A table of feature-specific covariates
    or
    a filepath to a tab-delimited file with feature-specific covariates. It
    should be formatted with features as columns and samples as rows (or the
    transpose). The row names and column names should be the same as those
    of the \code{input_data}: the column and row names should be the feature
    names and sample names respectively. Typically, this table should be
    generated by `preprocess_dna_mtx` or `preprocess_taxa_mtx` first.}
    \item{feature_specific_covariate_name}{The name for the feature-specific
    covariates when fitting the models. This string must be parse-able
    in a formula (e.g., no spaces).}
    \item{feature_specific_covariate_record}{Whether to keep the
    feature-specific covariates in the outputs when calculating p-values,
    writing results, and displaying plots.}
    \item{min_abundance}{Features with abundances more than
    \code{min_abundance} in more than \code{min_prevalence} of the samples will 
    be included for analysis. The threshold is applied after normalization and
    before transformation.}
    \item{min_prevalence}{See \code{min_abundance}.}
    \item{max_prevalence}{Features with abundances more than
    \code{min_abundance} in fewer than \code{max_prevalence} of the samples will 
    be included for analysis. The threshold is applied after normalization and
    before transformation.}
    \item{zero_threshold}{Abundances less than or equal to
    \code{zero_threshold} will be treated as zeros. This is primarily to be
    used when the abundance table has likely low-abundance false positives.}
    \item{min_variance}{Features with abundance variances less than or equal
    to
    \code{min_variance} will be dropped. This is primarily used for dropping
    features that are entirely zero.}
    \item{max_significance}{The FDR corrected q-value threshold for
    significance used in selecting which associations to write as
    significant and to plot.}
    \item{normalization}{The normalization to apply to the features before
    transformation and analysis. The option \code{TSS} (total sum scaling)
    is
    recommended, but \code{CLR} (centered log ratio) and \code{NONE} can
    also
    be used.}
    \item{transform}{The transformation to apply to the features after
    normalization and before analysis. The option \code{LOG} (base 2) is
    recommended, but \code{PLOG} (pseudo-log) and \code{NONE} can also be
    used.}
    \item{correction}{The correction to obtain FDR-corrected q-values from
    raw
    p-values. Any valid options for \code{p.adjust} can be used.}
    \item{standardize}{Whether to apply z-scores to continuous metadata
    variables so they are on the same scale. This is recommended in order to
    compare coefficients across metadata variables, but note that functions
    of the metadata specified in the \code{formula} will apply after
    standardization.}
    \item{unscaled_abundance}{A data frame with a single column of absolute
    abundances or a filepath to such a tab-delimited file. The row names
    should match the names of the samples in \code{input_data} and
    \code{input_metadata}. When using spike-ins, the single column should
    have the same name as one of the features in \code{input_data}, and the
    \code{unscaled_abundance} should correspond to the absolute quantity of
    the spike-in. When using total abundance scaling, the single column
    should have the name 'total', and the \code{unscaled_abundance} should
    correspond to the total abundance of each sample.}
    \item{median_comparison_abundance}{Test abundance coefficients against a
    null value corresponding to the median coefficient for a metadata
    variable across the features. This is recommended for relative abundance
    data but should not be used for absolute abundance data.}
    \item{median_comparison_prevalence}{Test prevalence coefficients against
    a null value corresponding to the median coefficient for a metadata
    variable across the features. This is only recommended if the analyst is
    interested in how feature prevalence associations compare to each other
    or if there is likely strong compositionality-induced sparsity.}
    \item{median_comparison_abundance_threshold}{Coefficients within
    \code{median_comparison_abundance_threshold} of the median association
    will automatically be counted as insignificant (p-value set to 1) since
    they likely represent compositionality-induced associations. This
    threshold will be divided by the metadata variable's standard deviation
    if the metadatum is continuous to ensure the threshold applies to the
    right scale.}
    \item{median_comparison_prevalence_threshold}{Same as
    \code{median_comparison_abundance_threshold} but applied to the
    prevalence associations.}
    \item{subtract_median}{Subtract the median from the coefficients.}
    \item{warn_prevalence}{Warn when prevalence associations are likely
    induced by abundance associations. This requires re-fitting the linear
    models on the TSS log-transformed data.}
    \item{small_random_effects}{Automatically replace random effects with 
    fixed effects in the logistic prevalence model to handle low numbers
    of observations per group.}
    \item{bypass_small_group_warning}{When \code{TRUE}, bypass the warning
    that is normally added to results when there are fewer than 4 average
    observations per random effect group.}
    \item{augment}{Add extra lowly-weighted 0s and 1s to avoid linear
    separability.}
    \item{evaluate_only}{Whether to evaluate just the abundnace
    ("abundance") or prevalence ("prevalence") models}
    \item{plot_summary_plot}{Generate a summary plot of significant
    associations.}
    \item{summary_plot_first_n}{Include the top \code{summary_plot_first_n}
    features with significant associations.}
    \item{coef_plot_vars}{Vector of variable names to be used in the
    coefficient plot section of the summary plot. Continuous variables
    should match the metadata column name, and categorical variables should
    be of the form \code{"[variable] [level]"}.}
    \item{heatmap_vars}{Vector of variable names to be used in the heatmap
    section of the summary plot. Continuous variables should match the
    metadata column name, and categorical variables should be of the form
    \code{"[variable] [level]"}.}
    \item{plot_associations}{Whether to generate plots for significant
    associations.}
    \item{max_pngs}{The top \code{max_pngs} significant associations will be
    plotted.}
    \item{cores}{Deprecated. Use \link[=mirai]{daemons} to enable parallelization.}
    \item{save_models}{Whether to return the fit models and save them to an
    RData file.}
    \item{save_plots_rds}{Whether to return the fit models and save them to an
    RData file.}
    \item{verbosity}{The level of verbosity for the \code{logging} package.}
    \item{summary_plot_balanced}{If set to TRUE the summary plot will 
    show the top N features of each variable included in 
    \code{coef_plot_vars} where N is equal to: 
    \code{ceiling(summary_plot_first_n/length(coef_plot_vars))}. Will error
    if \code{coef_plot_vars} = \code{NULL}}
    \item{assay.type}{A string or index to select the assay when using
    a \code{SummarizedExperiment} object}
}
\value{
    A list containing the following items:
    \describe{
    \item{(1)}{\code{data}: A dataframe of feature abundances with the
    retained samples for fitting.}
    \item{(2)}{\code{normalized_data}: A dataframe of normalized feature
    abundances.}
    \item{(3)}{\code{filtered_data}: A dataframe of feature abundances on
    the original scale after normalization and filtering.}
    \item{(4)}{\code{transformed_data}: A dataframe of feature abundances
    after filtering, normalization, and transformation.}
    \item{(5)}{\code{metadata}: A dataframe of metadata with the retained
    samples for fitting.}
    \item{(6)}{\code{standardized_metadata}: A dataframe of metadata after
    scaling (if selected).}
    \item{(7)}{\code{formula}: Checked or constructed formula(s) specifying
    the model to be fit.}
    \item{(8)}{\code{fit_data_abundance}: The results from the fit abundance
    models (see \code{\link{maaslin_fit}}).}
    \item{(9)}{\code{fit_data_prevalence}: The results from the fit
    prevalence models (see \code{\link{maaslin_fit}}).}
    }
}
\author{
    William Nickols<willnickols@g.harvard.edu>,\cr
    Jacob Nearing<nearing@broadinstitute.org>,\cr
    Maintainers: Lauren McIver<lauren.j.mciver@gmail.com>,\cr
}
\examples{
    # Read features table
    taxa_table_name <- system.file("extdata", "HMP2_taxonomy.tsv", package =
    "maaslin3")
    taxa_table <- read.csv(taxa_table_name, sep = '\t', row.names = 1)
    
    # Read metadata table
    metadata_name <- system.file("extdata", "HMP2_metadata.tsv", package =
    "maaslin3")
    metadata <- read.csv(metadata_name, sep = '\t', row.names = 1)
    
    metadata$diagnosis <-
        factor(metadata$diagnosis, levels = c('nonIBD', 'UC', 'CD'))
    metadata$dysbiosis_state <-
        factor(metadata$dysbiosis_state, levels = c('none', 'dysbiosis_UC',
        'dysbiosis_CD'))
    metadata$antibiotics <-
        factor(metadata$antibiotics, levels = c('No', 'Yes'))
    
    #Run MaAsLin3
    fit_out <- maaslin3::maaslin3(input_data = taxa_table,
                                input_metadata = metadata,
                                output = 'output',
                                formula = '~ diagnosis + dysbiosis_state +
                                antibiotics + age + reads',
                                plot_summary_plot = FALSE,
                                plot_associations = FALSE)
                                
    unlink('output', recursive=TRUE)
    logging::logReset()
}

\name{maaslin_contrast_test}
\alias{maaslin_contrast_test}
\title{
Perform a contrast from a fit MaAsLin 3 model.
}
\description{
Perform a contrast test (\code{lmerTest::contest} for mixed effects linear; 
\code{multcomp::glht} for all others) using a named contrast matrix and 
right hand side. One contrast test is applied per row of the matrix.
}
\usage{
maaslin_contrast_test(maaslin3_fit,
                        contrast_mat,
                        rhs = NULL,
                        max_significance = 0.1,
                        correction = 'BH',
                        median_comparison_abundance = TRUE,
                        median_comparison_prevalence = FALSE,
                        subtract_median = FALSE,
                        small_random_effects = FALSE,
                        bypass_small_group_warning = FALSE,
                        evaluate_only = NULL)
}
\arguments{
    \item{maaslin3_fit}{The output of \code{\link{maaslin_fit}} with 
    \code{save_models = TRUE}.}
    \item{contrast_mat}{A matrix with one row per contrast test to run. The
    columns will be matched to the coefficients of the model by name. 
    Contrast vector coefficients need not be specified if they would be zero. 
    If row names are provided, they will be used to label the test in the
    results.}
    \item{rhs}{The right hand size of the contrast test. The length should
    be the same as the number of rows in the \code{contrast_mat}. This will
    default to 0 or the median comparison if \code{median_comparison=TRUE}.}
    \item{max_significance}{The FDR corrected q-value threshold for
    significance used in selecting which associations to write as
    significant and to plot.}
    \item{correction}{The correction to obtain FDR-corrected q-values from
    raw p-values. Any valid options for \code{p.adjust} can be used.}
    \item{median_comparison_abundance}{Test abundance coefficients against a
    null value corresponding to the median coefficient for a metadata
    variable across the features. This is recommended for relative abundance
    data but should not be used for absolute abundance data.}
    \item{median_comparison_prevalence}{Test prevalence coefficients against
    a null value corresponding to the median coefficient for a metadata
    variable across the features. This is only recommended if the analyst is
    interested in how feature prevalence associations compare to each other
    or if there is likely strong compositionality-induced sparsity.}
    \item{subtract_median}{Subtract the median from the coefficients.}
    \item{small_random_effects}{Automatically replace random effects with 
    fixed effects in the logistic prevalence model to handle low numbers
    of observations per group.}
    \item{bypass_small_group_warning}{When \code{TRUE}, bypass the warning
    that is normally added to results when there are fewer than 4 average
    observations per random effect group.}
    \item{evaluate_only}{Whether to evaluate just the abundnace
    ("abundance") or prevalence ("prevalence") models}
}
\value{
A dataframe with the following columns:
\describe{
    \item{(1)}{\code{feature}: The feature involved in the association.}
    \item{(2)}{\code{test}: The contrast test name.}
    \item{(3)}{\code{coef}: The coefficient of the association: the slope
    coefficient in the abundance model and the change in log odds in the
    prevalence model.}
    \item{(4)}{\code{null_hypothesis}: The value of the null hypothesis
    against which the coefficients are tested (zero or the per-metadatum 
    median).}
    \item{(5)}{\code{stderr}: The standard error of the coefficient.}
    \item{(6)}{\code{pval_individual}: The (uncorrected) p-value of the
    association.}
    \item{(7)}{\code{qval_individual}: The FDR corrected q-value of the
    association. FDR correction is performed over all associations in the
    abundance and prevalence modeling without errors together.}
    \item{(8)}{\code{pval_joint}: The p-value of the overall association
    (combining abundance and prevalence) by taking the minimum of the
    abundance and logistic p-values and applying the Beta(1,2) CDF. These
    will be the same in the abundance and prevalence results for an
    association.}
    \item{(9)}{\code{qval_joint}: The FDR corrected q-value of the
    association. FDR correction is performed over all joint p-values
    without errors.}
    \item{(10)}{\code{error}: Any error produced by the model during fitting.
    NA otherwise.}
    \item{(11)}{\code{model}: \code{linear} for the abundance models and
    \code{logistic} for the prevalence models.}
    \item{(12)}{\code{N}: The number of data points for the association's
    feature.}
    \item{(13)}{\code{N_not_zero}: The number of non-zero data points for
    the association's feature.}
    }
}
\author{
    William Nickols<willnickols@g.harvard.edu>,\cr
    Jacob Nearing<nearing@broadinstitute.org>,\cr
    Maintainers: Lauren McIver<lauren.j.mciver@gmail.com>,\cr
}
\examples{
    # Read features table
    taxa_table_name <- system.file("extdata", "HMP2_taxonomy.tsv", package =
    "maaslin3")
    taxa_table <- read.csv(taxa_table_name, sep = '\t', row.names = 1)
    
    # Read metadata table
    metadata_name <- system.file("extdata", "HMP2_metadata.tsv", package =
    "maaslin3")
    metadata <- read.csv(metadata_name, sep = '\t', row.names = 1)
    
    metadata$diagnosis <-
        factor(metadata$diagnosis, levels = c('nonIBD', 'UC', 'CD'))
    metadata$dysbiosis_state <-
        factor(metadata$dysbiosis_state, levels = c('none', 'dysbiosis_UC',
        'dysbiosis_CD'))
    metadata$antibiotics <-
        factor(metadata$antibiotics, levels = c('No', 'Yes'))
    
    #Run MaAsLin3
    fit_out <- maaslin3::maaslin3(input_data = taxa_table,
                                input_metadata = metadata,
                                output = 'output',
                                formula = '~ diagnosis + dysbiosis_state +
                                antibiotics + age + reads',
                                plot_summary_plot = FALSE,
                                plot_associations = FALSE)
                                
    contrast_mat <- matrix(c(1, 1, 0, 0, 0, 0, 1, 1), 
        ncol = 4, nrow = 2, byrow = TRUE)
        
    colnames(contrast_mat) <- c("diagnosisUC",
                                "diagnosisCD",
                                "dysbiosis_statedysbiosis_UC",
                                "dysbiosis_statedysbiosis_CD")
                                
    rownames(contrast_mat) <- c("diagnosis_test", "dysbiosis_test")
                                
    maaslin_contrast_test(maaslin3_fit = fit_out,
                            contrast_mat = contrast_mat)
                                
    unlink('output', recursive=TRUE)
    logging::logReset()
}

\name{maaslin_fit}
\alias{maaslin_fit}
\title{
Fit MaAsLin 3 models.
}
\description{
Fit the abundance data with abundance and prevalence models to discover
feature-metadata associations.
}
\usage{
maaslin_fit(filtered_data,
            transformed_data,
            metadata,
            formula,
            random_effects_formula,
            feature_specific_covariate = NULL,
            feature_specific_covariate_name = NULL,
            feature_specific_covariate_record = NULL,
            zero_threshold = 0,
            max_significance = 0.1,
            correction = 'BH',
            median_comparison_abundance = TRUE,
            median_comparison_prevalence = FALSE,
            median_comparison_abundance_threshold = 0,
            median_comparison_prevalence_threshold = 0,
            subtract_median = FALSE,
            warn_prevalence = TRUE,
            small_random_effects = FALSE,
            bypass_small_group_warning = FALSE,
            augment = TRUE,
            evaluate_only = NULL,
            cores = 1,
            save_models = FALSE,
            data = NULL,
            min_abundance = 0,
            min_prevalence = 0,
            max_prevalence = 1.01,
            min_variance = 0)
}
\arguments{
    \item{filtered_data}{A data frame of filtered feature abundances. It
    should be formatted with features as columns and samples as rows. The
    column and row names should be the feature names and sample names
    respectively.}
    \item{transformed_data}{A data frame of transformed feature abundances.
    It should be formatted with features as columns and samples as rows. The
    column and row names should be the feature names and sample names
    respectively.}
    \item{metadata}{A data frame of per-sample metadata. It should be
    formatted with variables as columns and samples as rows. The column and
    row names should be the variable names and sample names respectively.}
    \item{formula}{A formula in \code{lme4} format as from
    \code{\link{maaslin_check_formula}}.}
    \item{random_effects_formula}{A formula in \code{lme4} format as from
    \code{\link{maaslin_check_formula}}.}
    \item{feature_specific_covariate}{A table of feature-specific covariates
    or
    a filepath to a tab-delimited file with feature-specific covariates. It
    should be formatted with features as columns and samples as rows (or the
    transpose). The row names and column names should be the same as those
    of the \code{input_data}: the column and row names should be the feature
    names and sample names respectively. Typically, this table should be
    generated by `preprocess_dna_mtx` or `preprocess_taxa_mtx` first.}
    \item{feature_specific_covariate_name}{The name for the feature-specific
    covariates when fitting the models.}
    \item{feature_specific_covariate_record}{Whether to keep the
    feature-specific covariates in the outputs when calculating p-values,
    writing results, and displaying plots.}
    \item{zero_threshold}{Abundances less than or equal to
    \code{zero_threshold} will be treated as zeros. This is primarily to be
    used when the abundance table has likely low-abundance false positives.}
    \item{max_significance}{The FDR corrected q-value threshold for
    significance used in selecting which associations to write as
    significant and to plot.}
    \item{correction}{The correction to obtain FDR-corrected q-values from
    raw p-values. Any valid options for \code{p.adjust} can be used.}
    \item{median_comparison_abundance}{Test abundance coefficients against a
    null value corresponding to the median coefficient for a metadata
    variable across the features. This is recommended for relative abundance
    data but should not be used for absolute abundance data.}
    \item{median_comparison_prevalence}{Test prevalence coefficients against
    a null value corresponding to the median coefficient for a metadata
    variable across the features. This is only recommended if the analyst is
    interested in how feature prevalence associations compare to each other
    or if there is likely strong compositionality-induced sparsity.}
    \item{median_comparison_abundance_threshold}{Coefficients within
    \code{median_comparison_abundance_threshold} of the median association
    will automatically be counted as insignificant (p-value set to 1) since
    they likely represent compositionality-induced associations. This
    threshold will be divided by the metadata variable's standard deviation
    if the metadatum is continuous to ensure the threshold applies to the
    right scale.}
    \item{median_comparison_prevalence_threshold}{Same as
    \code{median_comparison_abundance_threshold} but applied to the
    prevalence associations.}
    \item{subtract_median}{Subtract the median from the coefficients.}
    \item{warn_prevalence}{Warn when prevalence associations are likely
    induced by abundance associations. This requires re-fitting the linear
    models on the TSS log-transformed data.}
    \item{small_random_effects}{Automatically replace random effects with 
    fixed effects in the logistic prevalence model to handle low numbers
    of observations per group.}
    \item{bypass_small_group_warning}{When \code{TRUE}, bypass the warning
    that is normally added to results when there are fewer than 4 average
    observations per random effect group.}
    \item{augment}{Add extra lowly-weighted 0s and 1s to avoid linear
    separability.}
    \item{evaluate_only}{Whether to evaluate just the abundnace
    ("abundance") or prevalence ("prevalence") models}
    \item{cores}{Deprecated. Use \link[=mirai]{daemons} to enable parallelization.}
    \item{save_models}{Whether to return the fit models and save them to an
    RData file.}
    \item{data}{The original data (only necessary if warn_prevalence is
    TRUE).}
    \item{min_abundance}{The original min_abundance parameter (only necessary 
    if warn_prevalence is TRUE).}
    \item{min_prevalence}{The original min_prevalence parameter (only necessary 
    if warn_prevalence is TRUE).}
    \item{max_prevalence}{The original max_prevalence parameter (only necessary 
    if warn_prevalence is TRUE).}
    \item{min_variance}{The original min_variance parameter (only necessary if 
    min_variance is TRUE).}
}
\value{
A list containing the following named items:
    \describe{
    \item{(1)}{\code{fit_data_abundance}: The results from the fit abundance
    models.}
    \item{(2)}{\code{fit_data_prevalence}: The results from the fit
    prevalence models.}
    }
    
The \code{fit_data_abundance} and \code{fit_data_prevalence} items have
the same structure. They are both lists with the following named items:
\describe{
    \item{(1)}{\code{results}: A results table with the modeled associations
    (see below).}
    \item{(2)}{\code{residuals}: A features (rows) by samples (columns)
    dataframe of residuals from the models.}
    \item{(3)}{\code{fitted}: A features (rows) by samples (columns)
    dataframe of fitted values from the models.}
    \item{(4)}{\code{ranef}: A features (rows) by random effect (columns)
    dataframe of random effects from the models. If multiple random effects
    are specified, this is a dataframe of dataframes.}
    \item{(5)}{\code{fits}: If \code{save_models=TRUE}, this is a list of
    the fit models.}
    }
    
The \code{results} tables contain the following columns for each
association (row):
\describe{
    \item{(1)}{\code{feature}: The feature involved in the association.}
    \item{(2)}{\code{metadata}: The metadata variable involved in the
    association.}
    \item{(3)}{\code{value}: The value of the metadata variable: the
    metadata variable itself if continuous or the level if categorical.}
    \item{(4)}{\code{name}: The name of the model component involved in the
    association: the metadata variable itself if continuous or a
    concatenated version of the metadata variable and level if categorical.}
    \item{(5)}{\code{coef}: The coefficient of the association: the slope
    coefficient in the abundance model and the change in log odds in the
    prevalence model.}
    \item{(6)}{\code{null_hypothesis}: The value of the null hypothesis
    against which the coefficients are tested (zero or the per-metadatum 
    median).}
    \item{(7)}{\code{stderr}: The standard error of the coefficient.}
    \item{(8)}{\code{pval_individual}: The (uncorrected) p-value of the
    association.}
    \item{(9)}{\code{qval_individual}: The FDR corrected q-value of the
    association. FDR correction is performed over all associations in the
    abundance and prevalence modeling without errors together.}
    \item{(10)}{\code{pval_joint}: The p-value of the overall association
    (combining abundance and prevalence) by taking the minimum of the
    abundance and logistic p-values and applying the Beta(1,2) CDF. These
    will be the same in the abundance and prevalence results for an
    association.}
    \item{(11)}{\code{qval_joint}: The FDR corrected q-value of the
    association. FDR correction is performed over all joint p-values
    without errors.}
    \item{(12)}{\code{error}: Any error produced by the model during fitting.
    NA otherwise.}
    \item{(13)}{\code{model}: \code{linear} for the abundance models and
    \code{logistic} for the prevalence models.}
    \item{(14)}{\code{N}: The number of data points for the association's
    feature.}
    \item{(15)}{\code{N_not_zero}: The number of non-zero data points for
    the association's feature.}
    }
}
\author{
    William Nickols<willnickols@g.harvard.edu>,\cr
    Jacob Nearing<nearing@broadinstitute.org>,\cr
    Maintainers: Lauren McIver<lauren.j.mciver@gmail.com>,\cr
}
\examples{
    # Read features table
    taxa_table_name <- system.file("extdata", "HMP2_taxonomy.tsv", package =
    "maaslin3")
    taxa_table <- read.csv(taxa_table_name, sep = '\t', row.names = 1)
    
    # Read metadata table
    metadata_name <- system.file("extdata", "HMP2_metadata.tsv", package =
    "maaslin3")
    metadata <- read.csv(metadata_name, sep = '\t', row.names = 1)
    
    metadata$diagnosis <-
        factor(metadata$diagnosis, levels = c('nonIBD', 'UC', 'CD'))
    metadata$dysbiosis_state <-
        factor(metadata$dysbiosis_state, levels = c('none', 'dysbiosis_UC',
        'dysbiosis_CD'))
    metadata$antibiotics <-
        factor(metadata$antibiotics, levels = c('No', 'Yes'))
    
    #Run MaAsLin3
    maaslin3::maaslin_log_arguments(
        input_data = taxa_table,
        input_metadata = metadata,
        output = 'output',
        formula = '~ diagnosis + dysbiosis_state + antibiotics +
        age + reads',
        plot_summary_plot = FALSE,
        plot_associations = FALSE)
    
    read_data_list <- maaslin3::maaslin_read_data(
        taxa_table,
        metadata)
    read_data_list <- maaslin3::maaslin_reorder_data(
        read_data_list$data,
        read_data_list$metadata)
    
    data <- read_data_list$data
    metadata <- read_data_list$metadata
    
    formulas <- maaslin3::maaslin_check_formula(
        data,
        metadata,
        input_formula = '~ diagnosis + dysbiosis_state + antibiotics +
        age + reads')
    
    formula <- formulas$formula
    random_effects_formula <- formulas$random_effects_formula
    
    normalized_data = maaslin3::maaslin_normalize(data,
                                    output = 'output')
                                    
    filtered_data = maaslin3::maaslin_filter(normalized_data,
                                    output = 'output')
                                    
    transformed_data = maaslin3::maaslin_transform(filtered_data,
                                    output = 'output')
                                    
    standardized_metadata = maaslin3::maaslin_process_metadata(
        metadata,
        formula = formula)
    
    maaslin_results = maaslin3::maaslin_fit(
        filtered_data,
        transformed_data,
        standardized_metadata,
        formula,
        random_effects_formula,
        warn_prevalence = FALSE)
    
    unlink('output', recursive=TRUE)
    logging::logReset()
}

\name{preprocess_taxa_mtx}
\alias{preprocess_taxa_mtx}
\title{
Pre-process the taxa covariates for metatranscriptomics
}
\description{
Pre-process the taxa covariates for metatranscriptomics by
total-sum-scaling the taxa, matching the taxa to the RNAs coming from those
taxa, and then, for each sample in each feature:
\enumerate{
    \item{Log 2 transforming the taxon abundance if the taxon abundance is >=0}
    \item{Setting the taxon abundance to \code{log2([minimum non-zero relative
    abundance in the dataset] / 2)} if any of the corresponding RNA abundances
    are non-zero but the taxon abundance is zero}
    \item{Setting the taxon abundance to NA if both are zero}
}
When the taxon is present, the RNA data can be modeled as usual in MaAsLin
3 with
\code{log2(taxon)} as a covariate. When the taxon is not present, if any of
its RNA is present, we assume the taxon was missed due to finite read depth, 
so the taxon abundance is imputed with a small pseudo-count. When neither the
taxon nor RNA is
present, we assume the gene/microbe was not in the sample and therefore no
information about the transcription level can be obtained. Setting the taxon
covariate to NA has the effect of dropping the sample when fitting the relevant
feature's model in MaAsLin 3. Unlike most MaAsLin functions that will infer the
samples from the row names and column names, \bold{the \code{rna_table} must be
formated as samples (rows) by features (columns)}.
}
\usage{
preprocess_taxa_mtx(taxa_table, rna_table, rna_per_taxon)
}
\arguments{
\item{taxa_table}{The samples (rows) by features (columns) data frame of
taxon abundances to preprocess. These can be relative abundances or
counts.}
\item{rna_table}{The samples (rows) by features (columns) data frame of
RNA to preprocess. These can be relative abundances or counts.}
\item{rna_per_taxon}{A dataframe with the columns `RNA` and `taxon` with
one row per `RNA` column found in `rna_table` giving both the `RNA` column 
and which `taxon` column it corresponds to in `taxa_table`.}
}
\value{
A list containing the following named items:
    \enumerate{
    \item{\code{dna_table}: The table of log2 transformed taxon relative
    abundances with NAs for any feature-sample pairs for which both the taxon
    and RNA abundances were 0.}
    \item{\code{rna_table}: The table of total sum scaled RNA abundances.
    These are not log2 transformed.}
    }
}
\author{
    William Nickols<willnickols@g.harvard.edu>,\cr
    Jacob Nearing<nearing@broadinstitute.org>,\cr
    Maintainers: Lauren McIver<lauren.j.mciver@gmail.com>,\cr
}
\examples{
    taxa_in <- data.frame('tax1' = c(1, 2, 0, 4, 5), 
                    'tax2' = c(2, 3, 4, 5, 6), check.names = FALSE)
    rownames(taxa_in) <- paste0("sample", c(1:5))
    
    mtx_in <- data.frame('a' = c(1, 2, 3, 4, 5), 
                        'b' = c(2, 3, 4, 5, 0),
                        'c' = c(3, 4, 5, 6, 0), check.names = FALSE)
    rownames(mtx_in) <- paste0("sample", c(1:5))
    
    rna_per_taxon <- data.frame(RNA = c('a', 'b', 'c'),
                                taxon = c('tax1', 'tax1', 'tax2'))
    
    preprocess_out <- preprocess_taxa_mtx(taxa_in, mtx_in, rna_per_taxon)
}

\name{preprocess_dna_mtx}
\alias{preprocess_dna_mtx}
\title{
Pre-process the DNA covariates for metatranscriptomics
}
\description{
Pre-process the DNA covariates for metatranscriptomics by
total-sum-scaling DNA
abundances per sample and then, for each sample in each feature:
\enumerate{
    \item{Log 2 transforming the DNA abundance if the DNA abundance is >=0}
    \item{Setting the DNA abundance to \code{log2([minimum non-zero relative
    abundance in the dataset] / 2)} if the corresponding RNA abundance is
    non-zero but the DNA abundance is zero}
    \item{Setting the DNA abundance to NA if both are zero}
}
When the DNA is present, the RNA data can be modeled as usual in MaAsLin
3 with
\code{log2(DNA)} as a covariate. When the DNA is not present, if the RNA
is
present, we assume the DNA was missed due to finite read depth, so the
DNA
abundance is imputed with a small pseudo-count. When neither the DNA nor
RNA is
present, we assume the gene/microbe was not in the sample and therefore
no
information about the transcription level can be obtained. Setting the
DNA
covariate to NA has the effect of dropping the sample when fitting the
relevant
feature's model in MaAsLin 3. Unlike most MaAsLin functions that will
infer the
samples from the row names and column names, \bold{the \code{rna_table}
must be
formated as samples (rows) by features (columns)}.
}
\usage{
preprocess_dna_mtx(dna_table, rna_table)
}
\arguments{
\item{dna_table}{The samples (rows) by features (columns) data frame of
DNA abundances to preprocess. These can be relative abundances or
counts.}
\item{rna_table}{The samples (rows) by features (columns) data frame of
RNA to
preprocess. These can be relative abundances or counts.}
}
\value{
A list containing the following named items:
    \enumerate{
    \item{\code{dna_table}: The table of log2 transformed DNA relative
    abundances with NAs for any feature-sample pairs for which both the DNA
    and
    RNA abundances were 0.}
    \item{\code{rna_table}: The table of total sum scaled RNA abundances.
    These
    are not log2 transformed.}
    }
}
\author{
    William Nickols<willnickols@g.harvard.edu>,\cr
    Jacob Nearing<nearing@broadinstitute.org>,\cr
    Maintainers: Lauren McIver<lauren.j.mciver@gmail.com>,\cr
}
\examples{
    mgx_in <- data.frame('a' = c(1, 2, 0, 4, 5), 
                        'b' = c(2, 3, 4, 5, 6),
                        'c' = c(3, 4, 5, 6, 0))
    rownames(mgx_in) <- paste0("sample", c(1:5))
    
    mtx_in <- data.frame('a' = c(1, 2, 3, 4, 5), 
                        'b' = c(2, 3, 4, 5, 0),
                        'c' = c(3, 4, 5, 6, 0))
    rownames(mtx_in) <- paste0("sample", c(1:5))
    
    preprocess_out <- preprocess_dna_mtx(mgx_in, mtx_in)
}

\name{maaslin_check_arguments}
\alias{maaslin_check_arguments}
\title{
Check parameter arguments to ensure a successful MaAsLin 3 run.
}
\description{
Check the arguments provided are valid for further MaAsLin 3 use.
}
\usage{
maaslin_check_arguments(feature_specific_covariate = NULL,
                        feature_specific_covariate_name = NULL,
                        feature_specific_covariate_record = NULL,
                        zero_threshold = 0,
                        normalization = 'TSS',
                        transform = 'LOG',
                        correction = 'BH',
                        warn_prevalence = TRUE,
                        evaluate_only = NULL,
                        unscaled_abundance = NULL,
                        median_comparison_abundance = TRUE)
}
\arguments{
    \item{feature_specific_covariate}{A table of feature-specific covariates
    or
    a filepath to a tab-delimited file with feature-specific covariates. It
    should be formatted with features as columns and samples as rows (or the
    transpose). The row names and column names should be the same as those
    of the \code{input_data}: the column and row names should be the feature
    names and sample names respectively. Typically, this table should be
    generated by `preprocess_dna_mtx` or `preprocess_taxa_mtx` first.}
    \item{feature_specific_covariate_name}{The name for the feature-specific
    covariates when fitting the models.}
    \item{feature_specific_covariate_record}{Whether to keep the
    feature-specific covariates in the outputs when calculating p-values,
    writing results, and displaying plots.}
    \item{zero_threshold}{Abundances less than or equal to
    \code{zero_threshold} will be treated as zeros. This is primarily to be
    used when the abundance table has likely low-abundance false positives.}
    \item{normalization}{The normalization to apply to the features before
    transformation and analysis. The option \code{TSS} (total sum scaling)
    is recommended, but \code{CLR} (centered log ratio) and \code{NONE} can
    also be used.}
    \item{transform}{The transformation to apply to the features after
    normalization and before analysis. The option \code{LOG} (base 2) is
    recommended, but \code{PLOG} (pseudo-log) and \code{NONE} can also be
    used.}
    \item{correction}{The correction to obtain FDR-corrected q-values from
    raw p-values. Any valid options for \code{p.adjust} can be used.}
    \item{warn_prevalence}{Warn when prevalence associations are likely
    induced by abundance associations. This requires re-fitting the linear
    models on the TSS log-transformed data.}
    \item{evaluate_only}{Whether to evaluate just the abundnace
    ("abundance") or prevalence ("prevalence") models}
    \item{unscaled_abundance}{A data frame with a single column of absolute
    abundances or a filepath to such a tab-delimited file. The row names
    should match the names of the samples in \code{input_data} and
    \code{input_metadata}. When using spike-ins, the single column should
    have the same name as one of the features in \code{input_data}, and the
    \code{unscaled_abundance} should correspond to the absolute quantity of
    the spike-in. When using total abundance scaling, the single column
    should have the name 'total', and the \code{unscaled_abundance} should
    correspond to the total abundance of each sample.}
    \item{median_comparison_abundance}{Test abundance coefficients against a
    null value corresponding to the median coefficient for a metadata
    variable across the features. This is recommended for relative abundance
    data but should not be used for absolute abundance data.}
}
\value{
    No value is returned, but incompatibile arguments will produce an error.
}
\author{
    William Nickols<willnickols@g.harvard.edu>,\cr
    Jacob Nearing<nearing@broadinstitute.org>,\cr
    Maintainers: Lauren McIver<lauren.j.mciver@gmail.com>,\cr
}
\examples{
    # Read features table
    taxa_table_name <- system.file("extdata", "HMP2_taxonomy.tsv", package =
    "maaslin3")
    taxa_table <- read.csv(taxa_table_name, sep = '\t', row.names = 1)
    
    # Read metadata table
    metadata_name <- system.file("extdata", "HMP2_metadata.tsv", package =
    "maaslin3")
    metadata <- read.csv(metadata_name, sep = '\t', row.names = 1)
    
    metadata$diagnosis <-
        factor(metadata$diagnosis, levels = c('nonIBD', 'UC', 'CD'))
    metadata$dysbiosis_state <-
        factor(metadata$dysbiosis_state, levels = c('none', 'dysbiosis_UC',
        'dysbiosis_CD'))
    metadata$antibiotics <-
        factor(metadata$antibiotics, levels = c('No', 'Yes'))
    
    # Prepare parameter lists
    maaslin3::maaslin_check_arguments(zero_threshold = 0,
                                    normalization = 'TSS',
                                    transform = 'LOG',
                                    correction = 'BH',
                                    median_comparison_abundance = TRUE)
                                    
    unlink('output', recursive=TRUE)
    logging::logReset()
}

\name{maaslin_transform}
\alias{maaslin_transform}
\title{
Transform abundance data for MaAsLin 3 modeling.
}
\description{
Transform the abundance data according to the \code{transform}
parameter.
}
\usage{
maaslin_transform(filtered_data,
                output,
                transform = 'LOG')
}
\arguments{
    \item{filtered_data}{A data frame of filtered feature abundances. It
    should be formatted with features as columns and samples as rows. The
    column and row names should be the feature names and sample names
    respectively.}
    \item{output}{The output folder to write results.}
    \item{transform}{The transformation to apply to the features after
    normalization and before analysis. The option \code{LOG} (base 2) is
    recommended, but \code{PLOG} (pseudo-log) and \code{NONE} can also be
    used.}
}
\value{
A dataframe of transformed features (features are columns; samples are
rows).
}
\author{
    William Nickols<willnickols@g.harvard.edu>,\cr
    Jacob Nearing<nearing@broadinstitute.org>,\cr
    Maintainers: Lauren McIver<lauren.j.mciver@gmail.com>,\cr
}
\examples{
    # Read features table
    taxa_table_name <- system.file("extdata", "HMP2_taxonomy.tsv", package =
    "maaslin3")
    taxa_table <- read.csv(taxa_table_name, sep = '\t', row.names = 1)
    
    # Read metadata table
    metadata_name <- system.file("extdata", "HMP2_metadata.tsv", package =
    "maaslin3")
    metadata <- read.csv(metadata_name, sep = '\t', row.names = 1)
    
    metadata$diagnosis <-
        factor(metadata$diagnosis, levels = c('nonIBD', 'UC', 'CD'))
    metadata$dysbiosis_state <-
        factor(metadata$dysbiosis_state, levels = c('none', 'dysbiosis_UC',
        'dysbiosis_CD'))
    metadata$antibiotics <-
        factor(metadata$antibiotics, levels = c('No', 'Yes'))
    
    #Run MaAsLin3
    maaslin3::maaslin_log_arguments(
        input_data = taxa_table,
        input_metadata = metadata,
        output = 'output',
        formula = '~ diagnosis + dysbiosis_state + antibiotics +
        age + reads',
        plot_summary_plot = FALSE,
        plot_associations = FALSE)
    
    read_data_list <- maaslin3::maaslin_read_data(
        taxa_table,
        metadata)
    read_data_list <- maaslin3::maaslin_reorder_data(
        read_data_list$data,
        read_data_list$metadata)
    
    data <- read_data_list$data
    metadata <- read_data_list$metadata
    
    formulas <- maaslin3::maaslin_check_formula(
        data,
        metadata,
        input_formula = '~ diagnosis + dysbiosis_state + antibiotics +
        age + reads')
    
    normalized_data = maaslin3::maaslin_normalize(data,
                                    output = 'output')
                                    
    filtered_data = maaslin3::maaslin_filter(normalized_data,
                                    output = 'output')
                                    
    transformed_data = maaslin3::maaslin_transform(filtered_data,
                                    output = 'output')
    
    unlink('output', recursive=TRUE)
    logging::logReset()
}

\name{maaslin_normalize}
\alias{maaslin_normalize}
\title{
Normalize abundance data for MaAsLin 3 model fitting.
}
\description{
Normalize the abundance data according to the \code{normalization}
parameter. If \code{unscaled_abundance} is specified, compute the
absolute abundances.
}
\usage{
maaslin_normalize(data,
                output,
                zero_threshold = 0,
                normalization = 'TSS',
                unscaled_abundance = NULL)
}
\arguments{
    \item{data}{A data frame of feature abundances. It should be formatted
    with features as columns and samples as rows. The column and row names
    should be the feature names and sample names respectively.}
    \item{output}{The output folder to write results.}
    \item{zero_threshold}{Abundances less than or equal to
    \code{zero_threshold} will be treated as zeros. This is primarily to be
    used when the abundance table has likely low-abundance false positives.}
    \item{normalization}{The normalization to apply to the features before
    transformation and analysis. The option \code{TSS} (total sum scaling)
    is recommended, but \code{CLR} (centered log ratio) and \code{NONE} can
    also be used.}
    \item{unscaled_abundance}{A data frame with a single column of absolute
    abundances. The row names should match the names of the samples in
    \code{input_data} and \code{input_metadata}. When using spike-ins, the
    single column should have the same name as one of the features in
    \code{input_data}, and the \code{unscaled_abundance} should correspond
    to the absolute quantity of the spike-in. When using total abundance
    scaling, the single column should have the name 'total', and the
    \code{unscaled_abundance} should correspond to the total abundance of
    each sample.}
}
\value{
A dataframe of normalized features (features are columns; samples are
rows).
}
\author{
    William Nickols<willnickols@g.harvard.edu>,\cr
    Jacob Nearing<nearing@broadinstitute.org>,\cr
    Maintainers: Lauren McIver<lauren.j.mciver@gmail.com>,\cr
}
\examples{
    # Read features table
    taxa_table_name <- system.file("extdata", "HMP2_taxonomy.tsv", package =
    "maaslin3")
    taxa_table <- read.csv(taxa_table_name, sep = '\t', row.names = 1)
    
    # Read metadata table
    metadata_name <- system.file("extdata", "HMP2_metadata.tsv", package =
    "maaslin3")
    metadata <- read.csv(metadata_name, sep = '\t', row.names = 1)
    
    metadata$diagnosis <-
        factor(metadata$diagnosis, levels = c('nonIBD', 'UC', 'CD'))
    metadata$dysbiosis_state <-
        factor(metadata$dysbiosis_state, levels = c('none', 'dysbiosis_UC',
        'dysbiosis_CD'))
    metadata$antibiotics <-
        factor(metadata$antibiotics, levels = c('No', 'Yes'))
    
    #Run MaAsLin3
    maaslin3::maaslin_log_arguments(
        input_data = taxa_table,
        input_metadata = metadata,
        output = 'output',
        formula = '~ diagnosis + dysbiosis_state + antibiotics +
        age + reads',
        plot_summary_plot = FALSE,
        plot_associations = FALSE)
    
    read_data_list <- maaslin3::maaslin_read_data(
        taxa_table,
        metadata)
        read_data_list <- maaslin3::maaslin_reorder_data(
        read_data_list$data,
        read_data_list$metadata)
    
    data <- read_data_list$data
    metadata <- read_data_list$metadata
    
    formulas <- maaslin3::maaslin_check_formula(
        data,
        metadata,
        input_formula = '~ diagnosis + dysbiosis_state + antibiotics +
        age + reads')
    
    normalized_data = maaslin3::maaslin_normalize(data,
                                    output = 'output')
    
    unlink('output', recursive=TRUE)
    logging::logReset()
}

\name{maaslin_check_formula}
\alias{maaslin_check_formula}
\title{
Check a MaAsLin 3 formula to ensure a proper MaAsLin 3 run.
}
\description{
Ensure that the formula provided is valid. Only one of
\code{maaslin_compute_formula} or \code{maaslin_check_formula} should be
used.
}
\usage{
maaslin_check_formula(data,
                    metadata,
                    input_formula = NULL,
                    feature_specific_covariate_name = NULL)
}
\arguments{
    \item{data}{A data frame of feature abundances. It should be formatted
    with features as columns and samples as rows. The column and row names
    should be the feature names and sample names respectively.}
    \item{metadata}{A data frame of per-sample metadata. It should be
    formatted with variables as columns and samples as rows. The column and
    row names should be the variable names and sample names respectively.}
    \item{input_formula}{A formula in \code{lme4} format. Random effects,
    interactions, and functions of the metadata can be included (note that
    these functions will be applied after standardization if
    \code{standardize=TRUE}). Group, ordered, and strata variables can be
    specified as: \code{group(grouping_variable)},
    \code{ordered(ordered_variable)} and \code{strata(strata_variable)}. The
    other variable options below will not be considered if a formula is
    set.}
    \item{feature_specific_covariate_name}{The name for the feature-specific
    covariates when fitting the models.}
}
\value{
A list containing the following named items:
    \describe{
    \item{(1)}{\code{formula}: The constructed formula.}
    \item{(2)}{\code{random_effects_formula}: A formula for the random
    effects.}
    }
}
\author{
    William Nickols<willnickols@g.harvard.edu>,\cr
    Jacob Nearing<nearing@broadinstitute.org>,\cr
    Maintainers: Lauren McIver<lauren.j.mciver@gmail.com>,\cr
}
\examples{
    # Read features table
    taxa_table_name <- system.file("extdata", "HMP2_taxonomy.tsv", package =
    "maaslin3")
    taxa_table <- read.csv(taxa_table_name, sep = '\t', row.names = 1)
    
    # Read metadata table
    metadata_name <- system.file("extdata", "HMP2_metadata.tsv", package =
    "maaslin3")
    metadata <- read.csv(metadata_name, sep = '\t', row.names = 1)
    
    metadata$diagnosis <-
        factor(metadata$diagnosis, levels = c('nonIBD', 'UC', 'CD'))
    metadata$dysbiosis_state <-
        factor(metadata$dysbiosis_state, levels = c('none', 'dysbiosis_UC',
        'dysbiosis_CD'))
    metadata$antibiotics <-
        factor(metadata$antibiotics, levels = c('No', 'Yes'))
    
    #Run MaAsLin3
    maaslin3::maaslin_log_arguments(
        input_data = taxa_table,
        input_metadata = metadata,
        output = 'output',
        formula = '~ diagnosis + dysbiosis_state + antibiotics +
        age + reads',
        plot_summary_plot = FALSE,
        plot_associations = FALSE)
        read_data_list <- maaslin3::maaslin_read_data(
        taxa_table,
        metadata)
        read_data_list <- maaslin3::maaslin_reorder_data(
        read_data_list$data,
        read_data_list$metadata)
    
    data <- read_data_list$data
    metadata <- read_data_list$metadata
    
    formulas <- maaslin3::maaslin_check_formula(
        data,
        metadata,
        input_formula = '~ diagnosis + dysbiosis_state + antibiotics +
        age + reads')
    
    unlink('output', recursive=TRUE)
    logging::logReset()
}

\name{maaslin_compute_formula}
\alias{maaslin_compute_formula}
\title{
Compute a formula for a MaAsLin 3 run based on the specified effects.
}
\description{
Compute a formula using variables provided through \code{fixed_effects},
\code{random_effects}, \code{group_effects}, \code{ordered_effects}, and
\code{strata_effects}. Only one of \code{maaslin_compute_formula} or
\code{maaslin_check_formula} should be used.
}
\usage{
maaslin_compute_formula(data,
                        metadata,
                        fixed_effects = NULL,
                        random_effects = NULL,
                        group_effects = NULL,
                        ordered_effects = NULL,
                        strata_effects = NULL,
                        feature_specific_covariate_name = NULL)
}
\arguments{
    \item{data}{A data frame of feature abundances. It should be formatted
    with features as columns and samples as rows. The column and row names
    should be the feature names and sample names respectively.}
    \item{metadata}{A data frame of per-sample metadata. It should be
    formatted with variables as columns and samples as rows. The column and
    row names should be the variable names and sample names respectively.}
    \item{fixed_effects}{A vector of variable names to be included as fixed
    effects.}
    \item{random_effects}{A vector of variable names to be included as
    random intercepts.}
    \item{group_effects}{A factored categorical variable to be included for
    group testing. An ANOVA-style test will be performed to assess whether
    any of the variable's levels are significant, and no coefficients or
    individual p-values will be returned.}
    \item{ordered_effects}{A factored categorical variable to be included.
    Consecutive levels will be tested for significance against each other,
    and the resulting associations will correspond to effect sizes, standard
    errors, and significances of each level versus the previous.}
    \item{strata_effects}{A vector with one variable name to be included as
    the strata variable in case-control studies. Strata cannot be combined
    with random effects.}
    \item{feature_specific_covariate_name}{The name for the feature-specific
    covariates when fitting the models.}
}
\value{
A list containing the following named items:
    \describe{
    \item{(1)}{\code{formula}: The constructed formula.}
    \item{(2)}{\code{random_effects_formula}: A formula for the random
    effects.}
    }
}
\author{
    William Nickols<willnickols@g.harvard.edu>,\cr
    Jacob Nearing<nearing@broadinstitute.org>,\cr
    Maintainers: Lauren McIver<lauren.j.mciver@gmail.com>,\cr
}
\examples{
    # Read features table
    taxa_table_name <- system.file("extdata", "HMP2_taxonomy.tsv", package =
    "maaslin3")
    taxa_table <- read.csv(taxa_table_name, sep = '\t', row.names = 1)
    
    # Read metadata table
    metadata_name <- system.file("extdata", "HMP2_metadata.tsv", package =
    "maaslin3")
    metadata <- read.csv(metadata_name, sep = '\t', row.names = 1)
    
    metadata$diagnosis <-
        factor(metadata$diagnosis, levels = c('nonIBD', 'UC', 'CD'))
    metadata$dysbiosis_state <-
        factor(metadata$dysbiosis_state, levels = c('none', 'dysbiosis_UC',
        'dysbiosis_CD'))
    metadata$antibiotics <-
        factor(metadata$antibiotics, levels = c('No', 'Yes'))
    
    #Run MaAsLin3
    maaslin3::maaslin_log_arguments(
        input_data = taxa_table,
        input_metadata = metadata,
        output = 'output',
        fixed_effects = c('diagnosis', 'dysbiosis_state', 'antibiotics',
                        'age', 'reads'),
        random_effects = c('participant_id'),
        plot_summary_plot = FALSE,
        plot_associations = FALSE)
    read_data_list <- maaslin3::maaslin_read_data(
        taxa_table,
        metadata)
    read_data_list <- maaslin3::maaslin_reorder_data(
        read_data_list$data,
        read_data_list$metadata)
        
    data <- read_data_list$data
    metadata <- read_data_list$metadata
    
    formulas <- maaslin3::maaslin_compute_formula(
        data,
        metadata,
        fixed_effects = c('diagnosis', 'dysbiosis_state', 'antibiotics',
                        'age', 'reads'),
        random_effects = c('participant_id'))
    
    unlink('output', recursive=TRUE)
    logging::logReset()
}

\name{maaslin_write_results_lefse_format}
\alias{maaslin_write_results_lefse_format}
\title{
Write the results from a MaAsLin 3 run in LEfSe format.
}
\description{
Write the results from a MaAsLin 3 run to the output folder in LEfSe
format.
}
\usage{
maaslin_write_results_lefse_format(output,
                                fit_data_abundance,
                                fit_data_prevalence)
}
\arguments{
    \item{output}{The output folder to write results.}
    \item{fit_data_abundance}{The abundance outputs of
    \code{\link{maaslin_fit}}.}
    \item{fit_data_prevalence}{The prevalence outputs of
    \code{\link{maaslin_fit}}.}
}
\value{
Results will be written to the \code{lefse_style_results_abundance.res}
file in the folder \code{output}. No value is returned.
}
\author{
    William Nickols<willnickols@g.harvard.edu>,\cr
    Jacob Nearing<nearing@broadinstitute.org>,\cr
    Maintainers: Lauren McIver<lauren.j.mciver@gmail.com>,\cr
}
\examples{
    # Read features table
    taxa_table_name <- system.file("extdata", "HMP2_taxonomy.tsv", package =
    "maaslin3")
    taxa_table <- read.csv(taxa_table_name, sep = '\t', row.names = 1)
    
    # Read metadata table
    metadata_name <- system.file("extdata", "HMP2_metadata.tsv", package =
    "maaslin3")
    metadata <- read.csv(metadata_name, sep = '\t', row.names = 1)
    
    metadata$diagnosis <-
        factor(metadata$diagnosis, levels = c('nonIBD', 'UC', 'CD'))
    metadata$dysbiosis_state <-
        factor(metadata$dysbiosis_state, levels = c('none', 'dysbiosis_UC',
        'dysbiosis_CD'))
    metadata$antibiotics <-
        factor(metadata$antibiotics, levels = c('No', 'Yes'))
    
    #Run MaAsLin3
    maaslin3::maaslin_log_arguments(
        input_data = taxa_table,
        input_metadata = metadata,
        output = 'output',
        formula = '~ diagnosis + dysbiosis_state + antibiotics +
        age + reads',
        plot_summary_plot = FALSE,
        plot_associations = FALSE)
    
    read_data_list <- maaslin3::maaslin_read_data(
        taxa_table,
        metadata)
        read_data_list <- maaslin3::maaslin_reorder_data(
        read_data_list$data,
        read_data_list$metadata)
    
    data <- read_data_list$data
    metadata <- read_data_list$metadata
    
    formulas <- maaslin3::maaslin_check_formula(
        data,
        metadata,
        input_formula = '~ diagnosis + dysbiosis_state + antibiotics +
        age + reads')
    
    formula <- formulas$formula
    random_effects_formula <- formulas$random_effects_formula
    
    normalized_data = maaslin3::maaslin_normalize(data,
                                    output = 'output')
                                    
    filtered_data = maaslin3::maaslin_filter(normalized_data,
                                    output = 'output')
                                    
    transformed_data = maaslin3::maaslin_transform(filtered_data,
                                    output = 'output')
                                    
    standardized_metadata = maaslin3::maaslin_process_metadata(
        metadata,
        formula = formula)
    
    maaslin_results = maaslin3::maaslin_fit(
        filtered_data,
        transformed_data,
        standardized_metadata,
        formula,
        random_effects_formula,
        warn_prevalence = FALSE)
    
    maaslin3::maaslin_write_results_lefse_format(
        output = 'output',
        maaslin_results$fit_data_abundance,
        maaslin_results$fit_data_prevalence)
    
    unlink('output', recursive=TRUE)
    logging::logReset()
}

\name{maaslin_filter}
\alias{maaslin_filter}
\title{
Filter abundance data before MaAsLin 3 model fitting.
}
\description{
Set abundances below \code{zero_threshold} to zero, remove features
without abundances more than \code{min_abundance} in
\code{min_prevalence} of the samples, remove features with abundances 
more than \code{min_abundance} in more than \code{max_prevalence} of 
the samples, and remove features with variances less than or equal to 
\code{min_variance}.
}
\usage{
maaslin_filter(normalized_data,
            output,
            min_abundance = 0,
            min_prevalence = 0,
            max_prevalence = 1.01,
            zero_threshold = 0,
            min_variance = 0)
}
\arguments{
    \item{normalized_data}{A data frame of normalized feature abundances. It
    should be formatted with features as columns and samples as rows. The
    column and row names should be the feature names and sample names
    respectively.}
    \item{output}{The output folder to write results.}
    \item{min_abundance}{Features with abundances more than
    \code{min_abundance} in more than \code{min_prevalence} of the samples will 
    be included for analysis. The threshold is applied after normalization and
    before transformation.}
    \item{min_prevalence}{See \code{min_abundance}.}
    \item{max_prevalence}{Features with abundances more than
    \code{min_abundance} in fewer than \code{max_prevalence} of the samples will 
    be included for analysis. The threshold is applied after normalization and
    before transformation.}
    \item{zero_threshold}{Abundances less than or equal to
    \code{zero_threshold} will be treated as zeros. This is primarily to be
    used when the abundance table has likely low-abundance false positives.}
    \item{min_variance}{Features with abundance variances less than or equal
    to \code{min_variance} will be dropped. This is primarily used for
    dropping features that are entirely zero.}
}
\value{
A dataframe of filtered features (features are columns; samples are
rows).
}
\author{
    William Nickols<willnickols@g.harvard.edu>,\cr
    Jacob Nearing<nearing@broadinstitute.org>,\cr
    Maintainers: Lauren McIver<lauren.j.mciver@gmail.com>,\cr
}
\examples{
    # Read features table
    taxa_table_name <- system.file("extdata", "HMP2_taxonomy.tsv", package =
    "maaslin3")
    taxa_table <- read.csv(taxa_table_name, sep = '\t', row.names = 1)
    
    # Read metadata table
    metadata_name <- system.file("extdata", "HMP2_metadata.tsv", package =
    "maaslin3")
    metadata <- read.csv(metadata_name, sep = '\t', row.names = 1)
    
    metadata$diagnosis <-
        factor(metadata$diagnosis, levels = c('nonIBD', 'UC', 'CD'))
    metadata$dysbiosis_state <-
        factor(metadata$dysbiosis_state, levels = c('none', 'dysbiosis_UC',
        'dysbiosis_CD'))
    metadata$antibiotics <-
        factor(metadata$antibiotics, levels = c('No', 'Yes'))
    
    #Run MaAsLin3
    maaslin3::maaslin_log_arguments(
        input_data = taxa_table,
        input_metadata = metadata,
        output = 'output',
        formula = '~ diagnosis + dysbiosis_state + antibiotics +
        age + reads',
        plot_summary_plot = FALSE,
        plot_associations = FALSE)
    
    read_data_list <- maaslin3::maaslin_read_data(
        taxa_table,
        metadata)
    read_data_list <- maaslin3::maaslin_reorder_data(
        read_data_list$data,
        read_data_list$metadata)
    
    data <- read_data_list$data
    metadata <- read_data_list$metadata
    
    formulas <- maaslin3::maaslin_check_formula(
        data,
        metadata,
        input_formula = '~ diagnosis + dysbiosis_state + antibiotics +
        age + reads')
    
    normalized_data = maaslin3::maaslin_normalize(data,
                                    output = 'output')
                                    
    filtered_data = maaslin3::maaslin_filter(normalized_data,
                                    output = 'output')
    
    unlink('output', recursive=TRUE)
    logging::logReset()
}
